@@ -32,10 +32,19 @@ def test_dimension_selection():
   assert ts.d[1, 2, 3] == ts.d[1, [2, 3]]
   assert ts.d[1, 2, 3] == ts.d[1, ts.d[2, 3]]
 
-  with pytest.raises(TypeError, match=".* is not iterable"):
+  with pytest.raises(TypeError):
     iter(ts.d)
 
 
+def test_no_operations():
+  x = ts.IndexTransform(input_rank=3)
+  expr = ts.d[0, 1]
+  with pytest.raises(
+      IndexError,
+      match="Must specify at least one operation in dimension expression"):
+    x[expr]  # pylint: disable=pointless-statement
+
+
 def test_translate_by_vector():
   x = ts.IndexTransform(input_shape=[2, 3], input_labels=["x", "y"])
 
@@ -85,6 +94,46 @@ def test_translate_to():
   )
 
 
+def test_stride_vector():
+  x = ts.IndexTransform(
+      input_inclusive_min=[0, 2, 1],
+      input_inclusive_max=[6, 5, 8],
+      input_labels=["x", "y", "z"])
+
+  expr = ts.d["x", "z"].stride[-2, 3]
+  assert repr(expr) == "d['x','z'].stride[-2,3]"
+  assert x[expr] == ts.IndexTransform(
+      input_inclusive_min=[-3, 2, 1],
+      input_inclusive_max=[0, 5, 2],
+      input_labels=["x", "y", "z"],
+      output=[
+          ts.OutputIndexMap(stride=-2, input_dimension=0),
+          ts.OutputIndexMap(stride=1, input_dimension=1),
+          ts.OutputIndexMap(stride=3, input_dimension=2),
+      ],
+  )
+
+
+def test_stride_scalar():
+  x = ts.IndexTransform(
+      input_inclusive_min=[0, 2, 1],
+      input_inclusive_max=[6, 5, 8],
+      input_labels=["x", "y", "z"])
+
+  expr = ts.d["x", "z"].stride[3]
+  assert repr(expr) == "d['x','z'].stride[3]"
+  assert x[expr] == ts.IndexTransform(
+      input_inclusive_min=[0, 2, 1],
+      input_inclusive_max=[2, 5, 2],
+      input_labels=["x", "y", "z"],
+      output=[
+          ts.OutputIndexMap(stride=3, input_dimension=0),
+          ts.OutputIndexMap(stride=1, input_dimension=1),
+          ts.OutputIndexMap(stride=3, input_dimension=2),
+      ],
+  )
+
+
 def test_label_single():
   x = ts.IndexTransform(input_shape=[2, 3], input_labels=["x", "y"])
 
@@ -112,6 +161,18 @@ def test_label_vector():
       input_shape=[2, 3], input_labels=["a", "b"])
 
 
+def test_label_wrong_number():
+  transform = ts.IndexTransform(3)
+  with pytest.raises(IndexError):
+    transform[ts.d[0].label["x", "y"]]
+
+
+def test_label_duplicates():
+  transform = ts.IndexTransform(3)[ts.d[0].label["x"]]
+  with pytest.raises(IndexError):
+    transform[ts.d[1].label["x"]]
+
+
 def test_add_new():
   x = ts.IndexTransform(input_shape=[2, 3], input_labels=["x", "y"])
   expr = ts.d[0, -2:][ts.newaxis]
@@ -129,6 +190,13 @@ def test_add_new():
   )
 
 
+def test_add_new_invalid_rank():
+  x = ts.IndexTransform(input_shape=[2, 3], input_labels=["x", "y"])
+  expr = ts.d[0:32][ts.newaxis]
+  with pytest.raises(IndexError):
+    x[expr]
+
+
 def test_diagonal():
   x = ts.IndexTransform(input_shape=[2, 3], input_labels=["x", "y"])
   expr = ts.d["x", "y"].diagonal
@@ -207,9 +275,9 @@ def test_transpose_empty():
 
 def test_transpose_label_target():
   x = ts.IndexTransform(input_labels=["x", "y", "z"])
-  with pytest.raises(
-      TypeError, match="Target dimensions cannot be specified by label"):
-    ts.d["x", "y"].transpose["x"]
+  with pytest.raises(IndexError,
+                     match="Target dimensions cannot be specified by label"):
+    x[ts.d["x", "y"].transpose["x"]]
 
 
 def test_transpose_move_to_back():
@@ -576,7 +644,7 @@ def test_dimension_selection_dimrange_index_invalid_start():
   expr = ts.d[5:][...]
   with pytest.raises(
       IndexError,
-      match=re.escape("Dimension index 5 is outside valid range [-3, 3)."),
+      match=re.escape("Dimension index 5 is outside valid range [-3, 3)"),
   ):
     x[expr]
 
@@ -587,7 +655,7 @@ def test_dimension_selection_dimrange_index_invalid_stop():
   with pytest.raises(
       IndexError,
       match=re.escape(
-          "Dimension exclusive stop index 5 is outside valid range [-4, 3]."),
+          "Dimension exclusive stop index 5 is outside valid range [-4, 3]"),
   ):
     x[expr]
 
