@@ -17,15 +17,17 @@
 
 #include <type_traits>
 
+#include "absl/status/status.h"
 #include "tensorstore/data_type.h"
 #include "tensorstore/internal/elementwise_function.h"
-#include "tensorstore/internal/preprocessor.h"
+#include "tensorstore/internal/preprocessor/defer.h"
+#include "tensorstore/internal/preprocessor/expand.h"
 
 namespace tensorstore {
 
 template <typename From, typename To>
 struct ConvertDataType {
-  void operator()(const From* from, To* to, Status* status) const {
+  void operator()(const From* from, To* to, absl::Status* status) const {
     *to = static_cast<To>(*from);
   }
 };
@@ -39,38 +41,38 @@ struct DataTypeConversionTraits {
 /// `bool`-valued metafunction that specifies whether a conversion is allowed
 /// from a compile-time data type of `From` to a compile-time data type of `To`.
 ///
-/// If either type is `void`, the conversion is permitted at compile time (but
-/// may fail at run time).  Otherwise, the conversion is allowed if, and only
-/// if, `DataTypeConversionTraits<From,To>::flags` includes `kSupported` and
-/// `AdditionalFlags`.
+/// .. warning::
+///
+///    If either `From` or `To` is `void`, the conversion is permitted at
+///    compile time but may fail at run time.
 ///
 /// \tparam From Unqualified element type, or `void` if unknown.
 /// \tparam To Unqualified element type, or `void` if unknown.
 /// \tparam AdditionalFlags Additional flags required,
 ///     e.g. `DataTypeConversionFlags::kSafeAndImplicit`.
+/// \relates DataType
 template <typename From, typename To,
           DataTypeConversionFlags AdditionalFlags = DataTypeConversionFlags{}>
-struct IsDataTypeConversionSupported
-    : public std::integral_constant<
-          bool, ((DataTypeConversionTraits<From, To>::flags &
-                  (DataTypeConversionFlags::kSupported | AdditionalFlags)) ==
-                 (DataTypeConversionFlags::kSupported | AdditionalFlags))> {};
+constexpr inline bool IsDataTypeConversionSupported =
+    ((DataTypeConversionTraits<From, To>::flags &
+      (DataTypeConversionFlags::kSupported | AdditionalFlags)) ==
+     (DataTypeConversionFlags::kSupported | AdditionalFlags));
 
 template <typename From, DataTypeConversionFlags AdditionalFlags>
-struct IsDataTypeConversionSupported<From, void, AdditionalFlags>
-    : public std::true_type {};
+constexpr inline bool
+    IsDataTypeConversionSupported<From, void, AdditionalFlags> = true;
 
 template <typename To, DataTypeConversionFlags AdditionalFlags>
-struct IsDataTypeConversionSupported<void, To, AdditionalFlags>
-    : public std::true_type {};
+constexpr inline bool IsDataTypeConversionSupported<void, To, AdditionalFlags> =
+    true;
 
 template <typename T, DataTypeConversionFlags AdditionalFlags>
-struct IsDataTypeConversionSupported<T, T, AdditionalFlags>
-    : public std::true_type {};
+constexpr inline bool IsDataTypeConversionSupported<T, T, AdditionalFlags> =
+    true;
 
 template <DataTypeConversionFlags AdditionalFlags>
-struct IsDataTypeConversionSupported<void, void, AdditionalFlags>
-    : public std::true_type {};
+constexpr inline bool
+    IsDataTypeConversionSupported<void, void, AdditionalFlags> = true;
 
 namespace internal {
 
@@ -80,12 +82,6 @@ extern const std::array<DataTypeOperations::CanonicalConversionOperations,
                         kNumDataTypeIds>
     canonical_data_type_conversions;
 
-struct DataTypeConversionLookupResult {
-  /// Valid only if the `flags` value includes `kSupported`.
-  ElementwiseClosure<2, Status*> closure;
-  DataTypeConversionFlags flags;
-};
-
 /// Returns the data type converter.
 ///
 /// If `from == to`, `flags` is set to
@@ -112,8 +108,8 @@ DataTypeConversionLookupResult GetDataTypeConverter(DataType from, DataType to);
 ///     specified, `kSupported` is always assumed.
 /// \dchecks `from.valid()`
 /// \dchecks `to.valid()`
-/// \returns `Status()` if the conversion is supported with the specified
-///     `required_flags`.
+/// \returns `absl::OkStatus()` if the conversion is supported with the
+///     specified `required_flags`.
 /// \error `absl::StatusCode::kInvalidArgument` if the conversion is not
 ///     supported.
 Result<DataTypeConversionLookupResult> GetDataTypeConverterOrError(
@@ -129,10 +125,10 @@ std::enable_if_t<((DataTypeConversionTraits<From, To>::flags &
                     DataTypeConversionFlags::kCanReinterpretCast)) ==
                       DataTypeConversionFlags::kSupported &&
                   !std::is_same_v<From, To>),
-                 internal::ElementwiseFunction<2, Status*>>
+                 internal::ElementwiseFunction<2, absl::Status*>>
 GetConvertFunction() {
   return internal::SimpleElementwiseFunction<
-      ConvertDataType<From, To>(From, const To), Status*>();
+      ConvertDataType<From, To>(From, const To), absl::Status*>();
 }
 
 template <typename From, typename To>
@@ -141,7 +137,7 @@ std::enable_if_t<((DataTypeConversionTraits<From, To>::flags &
                     DataTypeConversionFlags::kCanReinterpretCast)) !=
                       DataTypeConversionFlags::kSupported ||
                   std::is_same_v<From, To>),
-                 internal::ElementwiseFunction<2, Status*>>
+                 internal::ElementwiseFunction<2, absl::Status*>>
 GetConvertFunction() {
   return {};
 }
@@ -150,12 +146,12 @@ template <typename From>
 constexpr internal::DataTypeOperations::CanonicalConversionOperations
 GetConvertToCanonicalOperations() {
   return {
-      /*.convert=*/MapCanonicalDataTypes([](auto data_type) {
-        using X = typename decltype(data_type)::Element;
+      /*.convert=*/MapCanonicalDataTypes([](auto dtype) {
+        using X = typename decltype(dtype)::Element;
         return GetConvertFunction<From, X>();
       }),
-      /*.flags=*/MapCanonicalDataTypes([](auto data_type) {
-        using X = typename decltype(data_type)::Element;
+      /*.flags=*/MapCanonicalDataTypes([](auto dtype) {
+        using X = typename decltype(dtype)::Element;
         return DataTypeConversionTraits<From, X>::flags;
       }),
   };
@@ -163,7 +159,7 @@ GetConvertToCanonicalOperations() {
 
 }  // namespace internal_data_type
 
-/// Define conversion traits between canonical data types.
+// Define conversion traits between canonical data types.
 
 namespace internal_data_type {
 
@@ -294,7 +290,7 @@ TENSORSTORE_INTERNAL_DEFINE_CONVERT_TRAITS(
         DataTypeConversionFlags::kCanReinterpretCast |
         DataTypeConversionFlags::kSafeAndImplicit);
 
-/// Define conversion flags from bool to every other canonical type.
+// Define conversion flags from bool to every other canonical type.
 #define TENSORSTORE_INTERNAL_DEFINE_CONVERSIONS_FROM_BOOL(T, ...) \
   TENSORSTORE_INTERNAL_DEFINE_CONVERT_TRAITS(                     \
       T, bool_t,                                                  \
@@ -310,8 +306,8 @@ TENSORSTORE_FOR_EACH_COMPLEX_DATA_TYPE(
 TENSORSTORE_INTERNAL_DEFINE_CONVERSIONS_FROM_BOOL(json_t)
 #undef TENSORSTORE_INTERNAL_DEFINE_CONVERSIONS_FROM_BOOL
 
-/// Define conversion flags from canonical integer types to every other
-/// canonical type.
+// Define conversion flags from canonical integer types to every other
+// canonical type.
 #define TENSORSTORE_INTERNAL_DEFINE_CONVERSIONS_FROM_INT(X, ...)           \
   TENSORSTORE_PP_DEFER(TENSORSTORE_FOR_EACH_INTEGER_DATA_TYPE_ID)          \
   ()(TENSORSTORE_PP_DEFER(TENSORSTORE_INTERNAL_INHERITED_CONVERT), X,      \
@@ -335,8 +331,8 @@ TENSORSTORE_PP_EXPAND(TENSORSTORE_FOR_EACH_INTEGER_DATA_TYPE(
     TENSORSTORE_INTERNAL_DEFINE_CONVERSIONS_FROM_INT))
 #undef TENSORSTORE_INTERNAL_DEFINE_CONVERSIONS_FROM_INT
 
-/// Define conversion flags from canonical float types to every other canonical
-/// type.
+// Define conversion flags from canonical float types to every other canonical
+// type.
 #define TENSORSTORE_INTERNAL_DEFINE_CONVERSIONS_FROM_FLOAT(X, ...)       \
   TENSORSTORE_PP_DEFER(TENSORSTORE_FOR_EACH_INTEGER_DATA_TYPE_ID)        \
   ()(TENSORSTORE_INTERNAL_DEFINE_CONVERT_TRAITS, X,                      \
@@ -360,8 +356,8 @@ TENSORSTORE_PP_EXPAND(TENSORSTORE_FOR_EACH_FLOAT_DATA_TYPE(
     TENSORSTORE_INTERNAL_DEFINE_CONVERSIONS_FROM_FLOAT))
 #undef TENSORSTORE_INTERNAL_DEFINE_CONVERSIONS_FROM_FLOAT
 
-/// Define conversion flags from canonical complex types to every other
-/// canonical type.
+// Define conversion flags from canonical complex types to every other
+// canonical type.
 #define TENSORSTORE_INTERNAL_DEFINE_CONVERSIONS_FROM_COMPLEX(X, ...) \
   TENSORSTORE_PP_DEFER(TENSORSTORE_FOR_EACH_INTEGER_DATA_TYPE_ID)    \
   ()(TENSORSTORE_INTERNAL_DEFINE_CONVERT_TRAITS, X,                  \
@@ -385,7 +381,7 @@ template <typename T>
 struct DataTypeConversionTraits<std::complex<T>, json_t>
     : public DataTypeConversionTraits<T, json_t> {};
 
-/// Define conversion flags from json_t to every other canonical type.
+// Define conversion flags from json_t to every other canonical type.
 TENSORSTORE_INTERNAL_DEFINE_CONVERT_TRAITS(bool, json_t,
                                            DataTypeConversionFlags::kSupported)
 TENSORSTORE_INTERNAL_DEFINE_CONVERT_TRAITS(string_t, json_t,
@@ -402,25 +398,25 @@ TENSORSTORE_FOR_EACH_FLOAT_DATA_TYPE(  //
 
 // TODO(jbms): support JSON -> complex conversion
 
-/// ustring_t -> string_t conversion: converts to UTF8 encoding.
+// ustring_t -> string_t conversion: converts to UTF8 encoding.
 TENSORSTORE_INTERNAL_DEFINE_CONVERT_TRAITS(
     string_t, ustring_t,
     DataTypeConversionFlags::kSupported |
         DataTypeConversionFlags::kSafeAndImplicit |
         DataTypeConversionFlags::kCanReinterpretCast);
 
-/// string_t -> ustring_t conversion: validates UTF-8 encoding.
+// string_t -> ustring_t conversion: validates UTF-8 encoding.
 TENSORSTORE_INTERNAL_DEFINE_CONVERT_TRAITS(ustring_t, string_t,
                                            DataTypeConversionFlags::kSupported);
 
-/// string_t -> json_t conversion: validates UTF-8 encoding
+// string_t -> json_t conversion: validates UTF-8 encoding
 TENSORSTORE_INTERNAL_DEFINE_CONVERT_TRAITS(json_t, string_t,
                                            DataTypeConversionFlags::kSupported);
 
 // TODO(jbms): Define string_t and ustring_t -> number, complex, bool
 // conversions
 
-/// ustring_t -> json_t conversion always succeeds.
+// ustring_t -> json_t conversion always succeeds.
 TENSORSTORE_INTERNAL_DEFINE_CONVERT_TRAITS(
     json_t, ustring_t,
     DataTypeConversionFlags::kSupported |
