@@ -26,25 +26,33 @@
 #include "tensorstore/util/result.h"
 #include "tensorstore/util/status.h"
 #include "tensorstore/util/status_testutil.h"
-#include "tensorstore/util/to_string.h"
+#include "tensorstore/util/str_cat.h"
 
 namespace {
 
-using namespace tensorstore::data_types;  // NOLINT
-using tensorstore::DataTypeConversionFlags;
-using tensorstore::DataTypeConversionTraits;
-using tensorstore::DataTypeOf;
-using tensorstore::Index;
-using tensorstore::IsDataTypeConversionSupported;
-using tensorstore::MatchesStatus;
-using tensorstore::Result;
-using tensorstore::Status;
-using tensorstore::StrCat;
-using tensorstore::internal::GetDataTypeConverter;
-using tensorstore::internal::GetDataTypeConverterOrError;
-using tensorstore::internal::GetElementCopyErrorStatus;
-using tensorstore::internal::IterationBufferKind;
-using tensorstore::internal::IterationBufferPointer;
+using ::tensorstore::bfloat16_t;
+using ::tensorstore::bool_t;
+using ::tensorstore::complex128_t;
+using ::tensorstore::complex64_t;
+using ::tensorstore::DataTypeConversionFlags;
+using ::tensorstore::DataTypeConversionTraits;
+using ::tensorstore::dtype_v;
+using ::tensorstore::float16_t;
+using ::tensorstore::float32_t;
+using ::tensorstore::float64_t;
+using ::tensorstore::Index;
+using ::tensorstore::IsDataTypeConversionSupported;
+using ::tensorstore::json_t;
+using ::tensorstore::MatchesStatus;
+using ::tensorstore::Result;
+using ::tensorstore::StrCat;
+using ::tensorstore::string_t;
+using ::tensorstore::ustring_t;
+using ::tensorstore::internal::GetDataTypeConverter;
+using ::tensorstore::internal::GetDataTypeConverterOrError;
+using ::tensorstore::internal::GetElementCopyErrorStatus;
+using ::tensorstore::internal::IterationBufferKind;
+using ::tensorstore::internal::IterationBufferPointer;
 
 constexpr DataTypeConversionFlags kSupported =
     DataTypeConversionFlags::kSupported;
@@ -59,28 +67,28 @@ template <typename From, typename To>
 void TestUnsupported() {
   static_assert(DataTypeConversionTraits<From, To>::flags ==
                 DataTypeConversionFlags{});
-  static_assert(!IsDataTypeConversionSupported<From, To>::value);
-  auto r = GetDataTypeConverter(DataTypeOf<From>(), DataTypeOf<To>());
+  static_assert(!IsDataTypeConversionSupported<From, To>);
+  auto r = GetDataTypeConverter(dtype_v<From>, dtype_v<To>);
   EXPECT_EQ(DataTypeConversionFlags{}, r.flags);
 }
 
 template <typename To, typename From>
 Result<To> TestConversion(
     From from, DataTypeConversionFlags flags = DataTypeConversionFlags{}) {
-  SCOPED_TRACE(StrCat("TestConversion<To=", DataTypeOf<To>(),
-                      ", From=", DataTypeOf<From>(), ">")
-                   .c_str());
+  SCOPED_TRACE(
+      StrCat("TestConversion<To=", dtype_v<To>, ", From=", dtype_v<From>, ">")
+          .c_str());
   flags = flags | kSupported;
-  if (!std::is_same<To, From>::value) {
+  if constexpr (!std::is_same_v<To, From>) {
     EXPECT_EQ(flags, (DataTypeConversionTraits<From, To>::flags));
   }
   EXPECT_EQ(!!(flags & kSafeAndImplicit),
-            (IsDataTypeConversionSupported<From, To, kSafeAndImplicit>::value));
-  EXPECT_TRUE((IsDataTypeConversionSupported<From, To>::value));
-  auto r = GetDataTypeConverter(DataTypeOf<From>(), DataTypeOf<To>());
+            (IsDataTypeConversionSupported<From, To, kSafeAndImplicit>));
+  EXPECT_TRUE((IsDataTypeConversionSupported<From, To>));
+  auto r = GetDataTypeConverter(dtype_v<From>, dtype_v<To>);
   EXPECT_EQ(flags, r.flags);
   To value;
-  Status status;
+  absl::Status status;
   if ((*r.closure.function)[IterationBufferKind::kContiguous](
           r.closure.context, 1, IterationBufferPointer(&from, Index(0)),
           IterationBufferPointer(&value, Index(0)), &status) != 1) {
@@ -112,6 +120,8 @@ TEST(DataTypeConversionTest, Bool) {
   EXPECT_EQ(1u, TestConversion<uint64_t>(true, kSafeAndImplicit));
   EXPECT_EQ(0, TestConversion<float16_t>(false, kSafeAndImplicit));
   EXPECT_EQ(1, TestConversion<float16_t>(true, kSafeAndImplicit));
+  EXPECT_EQ(0, TestConversion<bfloat16_t>(false, kSafeAndImplicit));
+  EXPECT_EQ(1, TestConversion<bfloat16_t>(true, kSafeAndImplicit));
   EXPECT_EQ(0, TestConversion<float32_t>(false, kSafeAndImplicit));
   EXPECT_EQ(1, TestConversion<float32_t>(true, kSafeAndImplicit));
   EXPECT_EQ(0, TestConversion<float64_t>(false, kSafeAndImplicit));
@@ -155,6 +165,7 @@ TEST(DataTypeConversionTest, Int8) {
   EXPECT_EQ(uint64_t(neg), TestConversion<uint64_t>(neg));
   EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos));
   EXPECT_EQ(float16_t(neg), TestConversion<float16_t>(neg, kSafeAndImplicit));
+  EXPECT_EQ(bfloat16_t(neg), TestConversion<bfloat16_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(float32_t(neg), TestConversion<float32_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(float64_t(neg), TestConversion<float64_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(complex64_t(float32_t(neg)),
@@ -191,6 +202,7 @@ TEST(DataTypeConversionTest, Uint8) {
   EXPECT_EQ(uint64_t(neg), TestConversion<uint64_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos, kSafeAndImplicit));
   EXPECT_EQ(float16_t(neg), TestConversion<float16_t>(neg, kSafeAndImplicit));
+  EXPECT_EQ(bfloat16_t(neg), TestConversion<bfloat16_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(float32_t(neg), TestConversion<float32_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(float64_t(neg), TestConversion<float64_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(complex64_t(float32_t(neg)),
@@ -225,6 +237,7 @@ TEST(DataTypeConversionTest, Int16) {
   EXPECT_EQ(uint64_t(neg), TestConversion<uint64_t>(neg));
   EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos));
   EXPECT_EQ(float16_t(neg), TestConversion<float16_t>(neg));
+  EXPECT_EQ(bfloat16_t(neg), TestConversion<bfloat16_t>(neg));
   EXPECT_EQ(float32_t(neg), TestConversion<float32_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(float64_t(neg), TestConversion<float64_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(complex64_t(float32_t(neg)),
@@ -261,6 +274,7 @@ TEST(DataTypeConversionTest, Uint16) {
   EXPECT_EQ(uint64_t(neg), TestConversion<uint64_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos, kSafeAndImplicit));
   EXPECT_EQ(float16_t(neg), TestConversion<float16_t>(neg));
+  EXPECT_EQ(bfloat16_t(neg), TestConversion<bfloat16_t>(neg));
   EXPECT_EQ(float32_t(neg), TestConversion<float32_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(float64_t(neg), TestConversion<float64_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(complex64_t(float32_t(neg)),
@@ -294,7 +308,10 @@ TEST(DataTypeConversionTest, Int32) {
   EXPECT_EQ(uint32_t(pos), TestConversion<uint32_t>(pos, kCanReinterpretCast));
   EXPECT_EQ(uint64_t(neg), TestConversion<uint64_t>(neg));
   EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos));
-  EXPECT_EQ(float16_t(neg), TestConversion<float16_t>(neg));
+  EXPECT_EQ(float16_t(static_cast<float>(neg)),
+            TestConversion<float16_t>(neg));  // inexact
+  EXPECT_EQ(bfloat16_t(static_cast<float>(neg)),
+            TestConversion<bfloat16_t>(neg));  // inexact
   EXPECT_EQ(float32_t(neg), TestConversion<float32_t>(neg));
   EXPECT_EQ(float64_t(neg), TestConversion<float64_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(complex64_t(float32_t(neg)), TestConversion<complex64_t>(neg));
@@ -329,7 +346,10 @@ TEST(DataTypeConversionTest, Uint32) {
                 pos, kSafeAndImplicit | kIdentity | kCanReinterpretCast));
   EXPECT_EQ(uint64_t(neg), TestConversion<uint64_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos, kSafeAndImplicit));
-  EXPECT_EQ(float16_t(neg), TestConversion<float16_t>(neg));
+  EXPECT_EQ(float16_t(static_cast<float>(neg)),
+            TestConversion<float16_t>(neg));  // inexact
+  EXPECT_EQ(bfloat16_t(static_cast<float>(neg)),
+            TestConversion<bfloat16_t>(neg));  // inexact
   EXPECT_EQ(float32_t(neg), TestConversion<float32_t>(neg));
   EXPECT_EQ(float64_t(neg), TestConversion<float64_t>(neg, kSafeAndImplicit));
   EXPECT_EQ(complex64_t(float32_t(neg)), TestConversion<complex64_t>(neg));
@@ -362,7 +382,10 @@ TEST(DataTypeConversionTest, Int64) {
   EXPECT_EQ(uint32_t(pos), TestConversion<uint32_t>(pos));
   EXPECT_EQ(uint64_t(neg), TestConversion<uint64_t>(neg, kCanReinterpretCast));
   EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos, kCanReinterpretCast));
-  EXPECT_EQ(float16_t(neg), TestConversion<float16_t>(neg));
+  EXPECT_EQ(float16_t(static_cast<float>(neg)),
+            TestConversion<float16_t>(neg));  // inexact
+  EXPECT_EQ(bfloat16_t(static_cast<float>(neg)),
+            TestConversion<bfloat16_t>(neg));  // inexact
   EXPECT_EQ(float32_t(neg), TestConversion<float32_t>(neg));
   EXPECT_EQ(float64_t(neg), TestConversion<float64_t>(neg));
   EXPECT_EQ(complex64_t(float32_t(neg)), TestConversion<complex64_t>(neg));
@@ -396,7 +419,10 @@ TEST(DataTypeConversionTest, Uint64) {
   EXPECT_EQ(uint64_t(pos),
             TestConversion<uint64_t>(
                 pos, kCanReinterpretCast | kSafeAndImplicit | kIdentity));
-  EXPECT_EQ(float16_t(neg), TestConversion<float16_t>(neg));
+  EXPECT_EQ(float16_t(static_cast<float>(neg)),
+            TestConversion<float16_t>(neg));  // inexact
+  EXPECT_EQ(bfloat16_t(static_cast<float>(neg)),
+            TestConversion<bfloat16_t>(neg));  // inexact
   EXPECT_EQ(float32_t(neg), TestConversion<float32_t>(neg));
   EXPECT_EQ(float64_t(neg), TestConversion<float64_t>(neg));
   EXPECT_EQ(complex64_t(float32_t(neg)), TestConversion<complex64_t>(neg));
@@ -426,6 +452,39 @@ TEST(DataTypeConversionTest, Float16) {
   EXPECT_EQ(float16_t(pos),
             TestConversion<float16_t>(
                 pos, kSafeAndImplicit | kIdentity | kCanReinterpretCast));
+  EXPECT_EQ(bfloat16_t(pos), TestConversion<bfloat16_t>(pos));
+  EXPECT_EQ(float32_t(pos), TestConversion<float32_t>(pos, kSafeAndImplicit));
+  EXPECT_EQ(float64_t(pos), TestConversion<float64_t>(pos, kSafeAndImplicit));
+  EXPECT_EQ(complex64_t(float32_t(pos)),
+            TestConversion<complex64_t>(pos, kSafeAndImplicit));
+  EXPECT_EQ(complex128_t(float64_t(pos)),
+            TestConversion<complex128_t>(pos, kSafeAndImplicit));
+  EXPECT_EQ("42.5", TestConversion<string_t>(pos));
+  EXPECT_EQ(ustring_t{"42.5"}, TestConversion<ustring_t>(pos));
+  EXPECT_EQ(json_t(42.5), TestConversion<json_t>(pos, kSafeAndImplicit));
+}
+
+TEST(DataTypeConversionTest, Bfloat16) {
+  using T = bfloat16_t;
+  const T pos(42.5);
+  EXPECT_EQ(false, TestConversion<bool_t>(T(0)));
+  EXPECT_EQ(true, TestConversion<bool_t>(pos));
+  EXPECT_EQ(int8_t(pos), TestConversion<int8_t>(pos));
+  EXPECT_EQ(int16_t(pos), TestConversion<int16_t>(pos));
+  EXPECT_EQ(int32_t(pos), TestConversion<int32_t>(pos));
+  EXPECT_EQ(int64_t(pos), TestConversion<int64_t>(pos));
+  EXPECT_EQ(uint8_t(pos), TestConversion<uint8_t>(pos));
+  EXPECT_EQ(uint8_t(pos), TestConversion<uint8_t>(pos));
+  EXPECT_EQ(uint16_t(pos), TestConversion<uint16_t>(pos));
+  EXPECT_EQ(uint16_t(pos), TestConversion<uint16_t>(pos));
+  EXPECT_EQ(uint32_t(pos), TestConversion<uint32_t>(pos));
+  EXPECT_EQ(uint32_t(pos), TestConversion<uint32_t>(pos));
+  EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos));
+  EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos));
+  EXPECT_EQ(float16_t(pos), TestConversion<float16_t>(pos));
+  EXPECT_EQ(bfloat16_t(pos),
+            TestConversion<bfloat16_t>(
+                pos, kSafeAndImplicit | kIdentity | kCanReinterpretCast));
   EXPECT_EQ(float32_t(pos), TestConversion<float32_t>(pos, kSafeAndImplicit));
   EXPECT_EQ(float64_t(pos), TestConversion<float64_t>(pos, kSafeAndImplicit));
   EXPECT_EQ(complex64_t(float32_t(pos)),
@@ -460,6 +519,7 @@ TEST(DataTypeConversionTest, Float32) {
   EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos));
   EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos));
   EXPECT_EQ(float16_t(pos), TestConversion<float16_t>(pos));
+  EXPECT_EQ(bfloat16_t(pos), TestConversion<bfloat16_t>(pos));
   EXPECT_EQ(float32_t(pos),
             TestConversion<float32_t>(
                 pos, kSafeAndImplicit | kIdentity | kCanReinterpretCast));
@@ -491,6 +551,7 @@ TEST(DataTypeConversionTest, Float64) {
   EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos));
   EXPECT_EQ(uint64_t(pos), TestConversion<uint64_t>(pos));
   EXPECT_EQ(float16_t(pos), TestConversion<float16_t>(pos));
+  EXPECT_EQ(bfloat16_t(pos), TestConversion<bfloat16_t>(pos));
   EXPECT_EQ(float32_t(pos), TestConversion<float32_t>(pos));
   EXPECT_EQ(float64_t(pos),
             TestConversion<float64_t>(
@@ -519,6 +580,7 @@ TEST(DataTypeConversionTest, Complex64) {
   EXPECT_EQ(uint64_t(value.real()), TestConversion<uint64_t>(value));
   EXPECT_EQ(uint64_t(value.real()), TestConversion<uint64_t>(value));
   EXPECT_EQ(float16_t(value.real()), TestConversion<float16_t>(value));
+  EXPECT_EQ(bfloat16_t(value.real()), TestConversion<bfloat16_t>(value));
   EXPECT_EQ(float32_t(value.real()), TestConversion<float32_t>(value));
   EXPECT_EQ(float64_t(value.real()), TestConversion<float64_t>(value));
   EXPECT_EQ(complex64_t(value),
@@ -549,6 +611,7 @@ TEST(DataTypeConversionTest, Complex128) {
   EXPECT_EQ(uint64_t(value.real()), TestConversion<uint64_t>(value));
   EXPECT_EQ(uint64_t(value.real()), TestConversion<uint64_t>(value));
   EXPECT_EQ(float16_t(value.real()), TestConversion<float16_t>(value));
+  EXPECT_EQ(bfloat16_t(value.real()), TestConversion<bfloat16_t>(value));
   EXPECT_EQ(float32_t(value.real()), TestConversion<float32_t>(value));
   EXPECT_EQ(float64_t(value.real()), TestConversion<float64_t>(value));
   EXPECT_EQ(complex64_t(value), TestConversion<complex64_t>(value));
@@ -576,6 +639,7 @@ TEST(DataTypeConversionTest, String) {
   TestUnsupported<T, int64_t>();
   TestUnsupported<T, uint64_t>();
   TestUnsupported<T, float16_t>();
+  TestUnsupported<T, bfloat16_t>();
   TestUnsupported<T, float32_t>();
   TestUnsupported<T, float64_t>();
   TestUnsupported<T, complex64_t>();
@@ -606,6 +670,7 @@ TEST(DataTypeConversionTest, Ustring) {
   TestUnsupported<T, int64_t>();
   TestUnsupported<T, uint64_t>();
   TestUnsupported<T, float16_t>();
+  TestUnsupported<T, bfloat16_t>();
   TestUnsupported<T, float32_t>();
   TestUnsupported<T, float64_t>();
   TestUnsupported<T, complex64_t>();
@@ -641,6 +706,8 @@ TEST(DataTypeConversionTest, Json) {
               MatchesStatus(absl::StatusCode::kInvalidArgument));
   EXPECT_THAT(TestConversion<float16_t>(json_t(nullptr)),
               MatchesStatus(absl::StatusCode::kInvalidArgument));
+  EXPECT_THAT(TestConversion<bfloat16_t>(json_t(nullptr)),
+              MatchesStatus(absl::StatusCode::kInvalidArgument));
   EXPECT_THAT(TestConversion<float32_t>(json_t(nullptr)),
               MatchesStatus(absl::StatusCode::kInvalidArgument));
   EXPECT_THAT(TestConversion<float64_t>(json_t(nullptr)),
@@ -669,6 +736,8 @@ TEST(DataTypeConversionTest, Json) {
             TestConversion<uint64_t>(json_t(40123567891234)));
   EXPECT_EQ(float16_t(42.5), TestConversion<float16_t>(json_t(42.5)));
   EXPECT_EQ(float16_t(42.5), TestConversion<float16_t>(json_t("42.5")));
+  EXPECT_EQ(bfloat16_t(42.5), TestConversion<bfloat16_t>(json_t(42.5)));
+  EXPECT_EQ(bfloat16_t(42.5), TestConversion<bfloat16_t>(json_t("42.5")));
   EXPECT_EQ(float32_t(42.5), TestConversion<float32_t>(json_t(42.5)));
   EXPECT_EQ(float64_t(42.5), TestConversion<float64_t>(json_t(42.5)));
   EXPECT_EQ(float64_t(42.5), TestConversion<float64_t>(json_t("42.5")));
@@ -690,24 +759,24 @@ TEST(DataTypeConversionTest, Json) {
 }
 
 TEST(GetDataTypeConverterOrErrorTest, Basic) {
-  EXPECT_EQ(Status(),
-            GetStatus(GetDataTypeConverterOrError(DataTypeOf<std::int32_t>(),
-                                                  DataTypeOf<std::int32_t>())));
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
+            GetStatus(GetDataTypeConverterOrError(dtype_v<std::int32_t>,
+                                                  dtype_v<std::int32_t>)));
+  EXPECT_EQ(absl::OkStatus(),
             GetStatus(GetDataTypeConverterOrError(
-                DataTypeOf<int32_t>(), DataTypeOf<int32_t>(), kIdentity)));
-  EXPECT_EQ(Status(), GetStatus(GetDataTypeConverterOrError(
-                          DataTypeOf<int32_t>(), DataTypeOf<int64_t>(),
-                          kSafeAndImplicit)));
-  EXPECT_EQ(Status(), GetStatus(GetDataTypeConverterOrError(
-                          DataTypeOf<int32_t>(), DataTypeOf<uint32_t>(),
-                          kCanReinterpretCast)));
-  EXPECT_THAT(GetDataTypeConverterOrError(DataTypeOf<json_t>(),
-                                          DataTypeOf<complex64_t>()),
-              MatchesStatus(absl::StatusCode::kInvalidArgument,
-                            "Cannot convert json -> complex64"));
+                dtype_v<int32_t>, dtype_v<int32_t>, kIdentity)));
+  EXPECT_EQ(absl::OkStatus(),
+            GetStatus(GetDataTypeConverterOrError(
+                dtype_v<int32_t>, dtype_v<int64_t>, kSafeAndImplicit)));
+  EXPECT_EQ(absl::OkStatus(),
+            GetStatus(GetDataTypeConverterOrError(
+                dtype_v<int32_t>, dtype_v<uint32_t>, kCanReinterpretCast)));
   EXPECT_THAT(
-      GetDataTypeConverterOrError(DataTypeOf<uint32_t>(), DataTypeOf<int32_t>(),
+      GetDataTypeConverterOrError(dtype_v<json_t>, dtype_v<complex64_t>),
+      MatchesStatus(absl::StatusCode::kInvalidArgument,
+                    "Cannot convert json -> complex64"));
+  EXPECT_THAT(
+      GetDataTypeConverterOrError(dtype_v<uint32_t>, dtype_v<int32_t>,
                                   kSafeAndImplicit),
       MatchesStatus(
           absl::StatusCode::kInvalidArgument,
