@@ -18,12 +18,17 @@
 
 #include <gmock/gmock.h>
 #include <gtest/gtest.h>
-#include "absl/meta/type_traits.h"
 #include "tensorstore/context.h"
+#include "tensorstore/driver/driver_testutil.h"
 #include "tensorstore/index_space/dim_expression.h"
 #include "tensorstore/internal/elementwise_function.h"
+#include "tensorstore/internal/json_binding/gtest.h"
+#include "tensorstore/internal/json_gtest.h"
 #include "tensorstore/open.h"
 #include "tensorstore/open_mode.h"
+#include "tensorstore/serialization/serialization.h"
+#include "tensorstore/serialization/test_util.h"
+#include "tensorstore/spec.h"
 #include "tensorstore/tensorstore.h"
 #include "tensorstore/util/executor.h"
 #include "tensorstore/util/status.h"
@@ -31,20 +36,27 @@
 
 namespace {
 
-using tensorstore::Context;
-using tensorstore::CopyProgress;
-using tensorstore::CopyProgressFunction;
-using tensorstore::Index;
-using tensorstore::MatchesStatus;
-using tensorstore::offset_origin;
-using tensorstore::ReadProgress;
-using tensorstore::ReadProgressFunction;
-using tensorstore::ReadWriteMode;
-using tensorstore::Status;
-using tensorstore::TensorStore;
-using tensorstore::WriteProgress;
-using tensorstore::WriteProgressFunction;
-using tensorstore::zero_origin;
+using ::tensorstore::ChunkLayout;
+using ::tensorstore::Context;
+using ::tensorstore::CopyProgress;
+using ::tensorstore::CopyProgressFunction;
+using ::tensorstore::DimensionIndex;
+using ::tensorstore::Index;
+using ::tensorstore::MatchesJson;
+using ::tensorstore::MatchesStatus;
+using ::tensorstore::offset_origin;
+using ::tensorstore::ReadProgress;
+using ::tensorstore::ReadProgressFunction;
+using ::tensorstore::ReadWriteMode;
+using ::tensorstore::TensorStore;
+using ::tensorstore::WriteProgress;
+using ::tensorstore::WriteProgressFunction;
+using ::tensorstore::zero_origin;
+using ::tensorstore::internal::TestSpecSchema;
+using ::tensorstore::internal::TestTensorStoreCreateCheckSchema;
+
+constexpr const char kMismatchRE[] = ".* mismatch with target dimension .*";
+constexpr const char kOutsideValidRangeRE[] = ".* is outside valid range .*";
 
 namespace driver_tests {
 
@@ -57,14 +69,13 @@ TEST(ArrayDriverTest, Read) {
           .value();
   std::vector<ReadProgress> read_progress;
   auto dest_array = tensorstore::AllocateArray<int>(array.domain());
-  auto future = tensorstore::internal::DriverRead(
+  TENSORSTORE_ASSERT_OK(tensorstore::internal::DriverRead(
       /*executor=*/tensorstore::InlineExecutor{},
       /*source=*/transformed_driver,
       /*target=*/dest_array,
       {/*.progress_function=*/[&read_progress](ReadProgress progress) {
         read_progress.push_back(progress);
-      }});
-  EXPECT_EQ(Status(), GetStatus(future.result()));
+      }}));
   EXPECT_EQ(array, dest_array);
   EXPECT_THAT(read_progress, ::testing::ElementsAre(ReadProgress{6, 6}));
 }
@@ -77,16 +88,17 @@ TEST(ArrayDriverTest, ReadIntoNewArray) {
       tensorstore::internal::MakeArrayDriver<offset_origin>(context, array)
           .value();
   std::vector<ReadProgress> read_progress;
-  auto future = tensorstore::internal::DriverRead(
-      /*executor=*/tensorstore::InlineExecutor{},
-      /*source=*/transformed_driver,
-      /*target_type=*/array.data_type(),
-      /*target_layout_order=*/tensorstore::c_order,
-      {/*.progress_function=*/[&read_progress](ReadProgress progress) {
-        read_progress.push_back(progress);
-      }});
-  EXPECT_EQ(Status(), GetStatus(future.result()));
-  EXPECT_EQ(array, future.value());
+  EXPECT_THAT(
+      tensorstore::internal::DriverReadIntoNewArray(
+          /*executor=*/tensorstore::InlineExecutor{},
+          /*source=*/transformed_driver,
+          /*target_dtype=*/array.dtype(),
+          /*target_layout_order=*/tensorstore::c_order,
+          {/*.progress_function=*/[&read_progress](ReadProgress progress) {
+            read_progress.push_back(progress);
+          }})
+          .result(),
+      ::testing::Optional(array));
   EXPECT_THAT(read_progress, ::testing::ElementsAre(ReadProgress{6, 6}));
 }
 
@@ -110,7 +122,7 @@ TEST(ArrayDriverTest, ReadDomainMismatch) {
         read_progress.push_back(progress);
       }});
   EXPECT_THAT(GetStatus(future.result()),
-              MatchesStatus(absl::StatusCode::kInvalidArgument, "Mismatch .*"));
+              MatchesStatus(absl::StatusCode::kInvalidArgument, kMismatchRE));
   EXPECT_THAT(read_progress, ::testing::ElementsAre());
 }
 
@@ -139,7 +151,7 @@ TEST(ArrayDriverTest, ReadCopyTransformError) {
   // validated when copying from the ReadChunk to the target array.
   EXPECT_THAT(GetStatus(future.result()),
               MatchesStatus(absl::StatusCode::kOutOfRange,
-                            "Index 1 is outside valid range .*"));
+                            ".* is outside valid range .*"));
   EXPECT_THAT(read_progress, ::testing::ElementsAre());
 }
 
@@ -147,7 +159,7 @@ TEST(ArrayDriverTest, Write) {
   auto array =
       tensorstore::MakeOffsetArray<int>({1, 2}, {{1, 2, 3}, {4, 5, 6}});
   auto context = Context::Default();
-  auto [driver, transform] =
+  auto [driver, transform, transaction] =
       tensorstore::internal::MakeArrayDriver<offset_origin>(context, array)
           .value();
   std::vector<WriteProgress> write_progress;
@@ -161,53 +173,21 @@ TEST(ArrayDriverTest, Write) {
       {/*.progress_function=*/[&write_progress](WriteProgress progress) {
         write_progress.push_back(progress);
       }});
-  EXPECT_EQ(Status(), GetStatus(write_result.copy_future.result()));
-  EXPECT_EQ(Status(), GetStatus(write_result.commit_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.copy_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.commit_future.result()));
   EXPECT_EQ(tensorstore::MakeOffsetArray<int>({1, 2}, {{1, 2, 3}, {4, 7, 8}}),
             array);
   EXPECT_THAT(write_progress, ::testing::ElementsAre(WriteProgress{2, 2, 0},
                                                      WriteProgress{2, 2, 2}));
 }
 
-/// Tests calling Write with an invalid source array that yields an error when
-/// MakeNormalizedTransformedArray is called.
-TEST(ArrayDriverTest, WriteInvalidSourceTransform) {
-  auto array =
-      tensorstore::MakeOffsetArray<int>({1, 2}, {{1, 2, 3}, {4, 5, 6}});
-  auto context = Context::Default();
-  auto [driver, transform] =
-      tensorstore::internal::MakeArrayDriver<offset_origin>(context, array)
-          .value();
-  std::vector<WriteProgress> write_progress;
-  auto write_result = tensorstore::internal::DriverWrite(
-      /*executor=*/tensorstore::InlineExecutor{},
-      /*source=*/
-      tensorstore::TransformedArray(
-          tensorstore::MakeOffsetArray<int>({2, 3}, {{7, 8}}),
-          tensorstore::IdentityTransform(tensorstore::BoxView({2, 3}, {2, 2}))),
-      /*target=*/
-      {driver, ChainResult(transform, tensorstore::Dims(0, 1).SizedInterval(
-                                          {2, 3}, {1, 2}))
-                   .value()},
-      {/*.progress_function=*/[&write_progress](WriteProgress progress) {
-        write_progress.push_back(progress);
-      }});
-  EXPECT_THAT(GetStatus(write_result.copy_future.result()),
-              MatchesStatus(absl::StatusCode::kOutOfRange,
-                            "Propagated bounds .* for dimension 0 are "
-                            "incompatible with existing bounds .*"));
-  EXPECT_EQ(GetStatus(write_result.commit_future.result()),
-            GetStatus(write_result.copy_future.result()));
-  EXPECT_THAT(write_progress, ::testing::ElementsAre());
-}
-
 /// Tests calling Write with a source domain that does not match the destination
 /// domain.
 TEST(ArrayDriverTest, WriteDomainMismatch) {
   auto array =
       tensorstore::MakeOffsetArray<int>({1, 2}, {{1, 2, 3}, {4, 5, 6}});
   auto context = Context::Default();
-  auto [driver, transform] =
+  auto [driver, transform, transaction] =
       tensorstore::internal::MakeArrayDriver<offset_origin>(context, array)
           .value();
   std::vector<WriteProgress> write_progress;
@@ -222,9 +202,9 @@ TEST(ArrayDriverTest, WriteDomainMismatch) {
         write_progress.push_back(progress);
       }});
   EXPECT_THAT(GetStatus(write_result.copy_future.result()),
-              MatchesStatus(absl::StatusCode::kInvalidArgument, "Mismatch .*"));
+              MatchesStatus(absl::StatusCode::kInvalidArgument, kMismatchRE));
   EXPECT_THAT(GetStatus(write_result.commit_future.result()),
-              MatchesStatus(absl::StatusCode::kInvalidArgument, "Mismatch .*"));
+              MatchesStatus(absl::StatusCode::kInvalidArgument, kMismatchRE));
   EXPECT_THAT(write_progress, ::testing::ElementsAre());
 }
 
@@ -232,12 +212,12 @@ TEST(ArrayDriverTest, Copy) {
   auto context = Context::Default();
   auto array_a =
       tensorstore::MakeOffsetArray<int>({1, 2}, {{1, 2, 3}, {4, 5, 6}});
-  auto [driver_a, transform_a] =
+  auto [driver_a, transform_a, transaction_a] =
       tensorstore::internal::MakeArrayDriver<offset_origin>(context, array_a)
           .value();
   auto array_b =
       tensorstore::MakeOffsetArray<int>({1, 4}, {{7, 7, 7}, {7, 7, 7}});
-  auto [driver_b, transform_b] =
+  auto [driver_b, transform_b, transaction_b] =
       tensorstore::internal::MakeArrayDriver<offset_origin>(context, array_b)
           .value();
   std::vector<CopyProgress> progress;
@@ -255,8 +235,8 @@ TEST(ArrayDriverTest, Copy) {
       {/*.progress_function=*/[&progress](CopyProgress p) {
         progress.push_back(p);
       }});
-  EXPECT_EQ(Status(), GetStatus(write_result.copy_future.result()));
-  EXPECT_EQ(Status(), GetStatus(write_result.commit_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.copy_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.commit_future.result()));
   EXPECT_EQ(tensorstore::MakeOffsetArray<int>({1, 4}, {{7, 1, 3}, {7, 4, 6}}),
             array_b);
   EXPECT_THAT(progress, ::testing::ElementsAre(CopyProgress{4, 4, 0, 0},
@@ -285,7 +265,7 @@ TEST(ArrayDriverTest, CopyDomainMismatch) {
         progress.push_back(p);
       }});
   EXPECT_THAT(GetStatus(write_result.copy_future.result()),
-              MatchesStatus(absl::StatusCode::kInvalidArgument, "Mismatch .*"));
+              MatchesStatus(absl::StatusCode::kInvalidArgument, kMismatchRE));
   EXPECT_EQ(GetStatus(write_result.copy_future.result()),
             GetStatus(write_result.commit_future.result()));
   EXPECT_THAT(progress, ::testing::ElementsAre());
@@ -315,7 +295,7 @@ TEST(FromArrayTest, ResolveBounds) {
 
   auto store2 = ResolveBounds(store).value();
   EXPECT_EQ(array.domain(), store2.domain().box());
-  EXPECT_EQ(tensorstore::DataTypeOf<int>(), store2.data_type());
+  EXPECT_EQ(tensorstore::dtype_v<int>, store2.dtype());
   EXPECT_THAT(store2.domain().labels(), ::testing::ElementsAre("", ""));
 }
 
@@ -324,17 +304,15 @@ TEST(FromArrayTest, Read) {
       tensorstore::MakeOffsetArray<int>({1, 2}, {{1, 2, 3}, {4, 5, 6}});
   auto context = Context::Default();
   auto store = tensorstore::FromArray(context, array).value();
-  static_assert(std::is_same<TensorStore<int, 2, ReadWriteMode::read_write>,
-                             decltype(store)>::value,
-                "");
+  static_assert(std::is_same_v<TensorStore<int, 2, ReadWriteMode::read_write>,
+                               decltype(store)>);
   std::vector<ReadProgress> read_progress;
   auto dest_array = tensorstore::AllocateArray<int>(array.domain());
-  auto future =
+  TENSORSTORE_ASSERT_OK(
       Read(store, dest_array,
            ReadProgressFunction{[&read_progress](ReadProgress progress) {
              read_progress.push_back(progress);
-           }});
-  EXPECT_EQ(Status(), GetStatus(future.result()));
+           }}));
   EXPECT_EQ(array, dest_array);
   EXPECT_THAT(read_progress, ::testing::ElementsAre(ReadProgress{6, 6}));
 }
@@ -346,7 +324,7 @@ TEST(FromArrayTest, ReadBroadcast) {
   auto store = tensorstore::FromArray(context, array).value();
   auto dest_array = tensorstore::AllocateArray<int>({2, 2, 3});
   auto future = Read(store, dest_array);
-  EXPECT_EQ(Status(), GetStatus(future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(future.result()));
   EXPECT_EQ(tensorstore::MakeArray<int>(
                 {{{1, 2, 3}, {4, 5, 6}}, {{1, 2, 3}, {4, 5, 6}}}),
             dest_array);
@@ -361,10 +339,10 @@ TEST(FromArrayTest, ReadAlignByLabel) {
   auto context = Context::Default();
   auto store = tensorstore::FromArray(context, array).value();
   auto dest_array = tensorstore::AllocateArray<int>({3, 2});
-  auto future =
+  tensorstore::Future<void> future =
       Read(ChainResult(store, tensorstore::AllDims().Label("x", "y")),
            ChainResult(dest_array, tensorstore::AllDims().Label("y", "x")));
-  EXPECT_EQ(Status(), GetStatus(future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(future.result()));
   EXPECT_EQ(tensorstore::MakeArray<int>({{1, 4}, {2, 5}, {3, 6}}), dest_array);
 }
 
@@ -380,7 +358,7 @@ TEST(FromArrayTest, ReadIntoNewArray) {
           ReadProgressFunction{[&read_progress](ReadProgress progress) {
             read_progress.push_back(progress);
           }}});
-  EXPECT_EQ(Status(), GetStatus(future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(future.result()));
   auto read_array = future.value();
   EXPECT_EQ(array, read_array);
   EXPECT_THAT(read_progress, ::testing::ElementsAre(ReadProgress{6, 6}));
@@ -402,7 +380,7 @@ TEST(FromArrayTest, ReadDomainMismatch) {
              read_progress.push_back(progress);
            }});
   EXPECT_THAT(GetStatus(future.result()),
-              MatchesStatus(absl::StatusCode::kInvalidArgument, "Mismatch .*"));
+              MatchesStatus(absl::StatusCode::kInvalidArgument, kMismatchRE));
   EXPECT_THAT(read_progress, ::testing::ElementsAre());
 }
 
@@ -420,9 +398,9 @@ TEST(FromArrayTest, ReadCopyTransformError) {
       ReadProgressFunction{[&read_progress](ReadProgress progress) {
         read_progress.push_back(progress);
       }});
-  EXPECT_THAT(GetStatus(future.result()),
-              MatchesStatus(absl::StatusCode::kOutOfRange,
-                            "Index 1 is outside valid range .*"));
+  EXPECT_THAT(
+      GetStatus(future.result()),
+      MatchesStatus(absl::StatusCode::kOutOfRange, kOutsideValidRangeRE));
   EXPECT_THAT(read_progress, ::testing::ElementsAre());
 }
 
@@ -438,8 +416,8 @@ TEST(FromArrayTest, Write) {
       WriteProgressFunction{[&write_progress](WriteProgress progress) {
         write_progress.push_back(progress);
       }});
-  EXPECT_EQ(Status(), GetStatus(write_result.copy_future.result()));
-  EXPECT_EQ(Status(), GetStatus(write_result.commit_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.copy_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.commit_future.result()));
   EXPECT_EQ(tensorstore::MakeOffsetArray<int>({1, 2}, {{1, 2, 3}, {4, 7, 8}}),
             array);
   EXPECT_THAT(write_progress, ::testing::ElementsAre(WriteProgress{2, 2, 0},
@@ -455,37 +433,12 @@ TEST(FromArrayTest, WriteBroadcast) {
       tensorstore::MakeScalarArray<int>(42),
       ChainResult(store,
                   tensorstore::Dims(0, 1).SizedInterval({2, 3}, {1, 2})));
-  EXPECT_EQ(Status(), GetStatus(write_result.copy_future.result()));
-  EXPECT_EQ(Status(), GetStatus(write_result.commit_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.copy_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.commit_future.result()));
   EXPECT_EQ(tensorstore::MakeOffsetArray<int>({1, 2}, {{1, 2, 3}, {4, 42, 42}}),
             array);
 }
 
-/// Tests calling Write with an invalid source array that yields an error when
-/// MakeNormalizedTransformedArray is called.
-TEST(FromArrayTest, WriteInvalidSourceTransform) {
-  auto array =
-      tensorstore::MakeOffsetArray<int>({1, 2}, {{1, 2, 3}, {4, 5, 6}});
-  auto context = Context::Default();
-  auto store = tensorstore::FromArray(context, array);
-  std::vector<WriteProgress> write_progress;
-  auto write_result = Write(
-      tensorstore::TransformedArray(
-          tensorstore::MakeOffsetArray<int>({2, 3}, {{7, 8}}),
-          tensorstore::IdentityTransform(tensorstore::BoxView({2, 3}, {2, 2}))),
-      ChainResult(store, tensorstore::Dims(0, 1).SizedInterval({2, 3}, {1, 2})),
-      WriteProgressFunction{[&write_progress](WriteProgress progress) {
-        write_progress.push_back(progress);
-      }});
-  EXPECT_THAT(GetStatus(write_result.copy_future.result()),
-              MatchesStatus(absl::StatusCode::kOutOfRange,
-                            "Propagated bounds .* for dimension 0 are "
-                            "incompatible with existing bounds .*"));
-  EXPECT_EQ(GetStatus(write_result.commit_future.result()),
-            GetStatus(write_result.copy_future.result()));
-  EXPECT_THAT(write_progress, ::testing::ElementsAre());
-}
-
 /// Tests calling Write with a source domain that does not match the destination
 /// domain.
 TEST(FromArrayTest, WriteDomainMismatch) {
@@ -501,9 +454,9 @@ TEST(FromArrayTest, WriteDomainMismatch) {
         write_progress.push_back(progress);
       }});
   EXPECT_THAT(GetStatus(write_result.copy_future.result()),
-              MatchesStatus(absl::StatusCode::kInvalidArgument, "Mismatch .*"));
+              MatchesStatus(absl::StatusCode::kInvalidArgument, kMismatchRE));
   EXPECT_THAT(GetStatus(write_result.commit_future.result()),
-              MatchesStatus(absl::StatusCode::kInvalidArgument, "Mismatch .*"));
+              MatchesStatus(absl::StatusCode::kInvalidArgument, kMismatchRE));
   EXPECT_THAT(write_progress, ::testing::ElementsAre());
 }
 
@@ -523,8 +476,8 @@ TEST(FromArrayTest, Copy) {
                                {1, 5}, {2, 2})),
       CopyProgressFunction{
           [&progress](CopyProgress p) { progress.push_back(p); }});
-  EXPECT_EQ(Status(), GetStatus(write_result.copy_future.result()));
-  EXPECT_EQ(Status(), GetStatus(write_result.commit_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.copy_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.commit_future.result()));
   EXPECT_EQ(tensorstore::MakeOffsetArray<int>({1, 4}, {{7, 1, 3}, {7, 4, 6}}),
             array_b);
   EXPECT_THAT(progress, ::testing::ElementsAre(CopyProgress{4, 4, 0, 0},
@@ -545,8 +498,8 @@ TEST(FromArrayTest, CopyBroadcast) {
                   tensorstore::Dims(0).IndexSlice(1)),
       ChainResult(store_b,
                   tensorstore::Dims(0, 1).SizedInterval({1, 5}, {2, 2})));
-  EXPECT_EQ(Status(), GetStatus(write_result.copy_future.result()));
-  EXPECT_EQ(Status(), GetStatus(write_result.commit_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.copy_future.result()));
+  EXPECT_EQ(absl::OkStatus(), GetStatus(write_result.commit_future.result()));
   EXPECT_EQ(tensorstore::MakeOffsetArray<int>({1, 4}, {{7, 1, 3}, {7, 1, 3}}),
             array_b);
 }
@@ -565,7 +518,7 @@ TEST(FromArrayTest, CopyDomainMismatch) {
         progress.push_back(p);
       }});
   EXPECT_THAT(GetStatus(write_result.copy_future.result()),
-              MatchesStatus(absl::StatusCode::kInvalidArgument, "Mismatch .*"));
+              MatchesStatus(absl::StatusCode::kInvalidArgument, kMismatchRE));
   EXPECT_EQ(GetStatus(write_result.copy_future.result()),
             GetStatus(write_result.commit_future.result()));
   EXPECT_THAT(progress, ::testing::ElementsAre());
@@ -576,7 +529,7 @@ TEST(FromArrayTest, ReadDataTypeConversion) {
   auto source = tensorstore::MakeArray<std::int32_t>({1, 2, 3});
   auto dest = tensorstore::AllocateArray<std::int64_t>({3});
   EXPECT_EQ(
-      Status(),
+      absl::OkStatus(),
       GetStatus(tensorstore::Read(tensorstore::FromArray(context, source), dest)
                     .result()));
   EXPECT_EQ(dest, tensorstore::MakeArray<std::int64_t>({1, 2, 3}));
@@ -599,7 +552,7 @@ TEST(FromArrayTest, WriteDataTypeConversion) {
   auto context = Context::Default();
   auto source = tensorstore::MakeArray<std::int32_t>({1, 2, 3});
   auto dest = tensorstore::AllocateArray<std::int64_t>({3});
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
             GetStatus(tensorstore::Write(source,
                                          tensorstore::FromArray(context, dest))
                           .commit_future.result()));
@@ -624,7 +577,7 @@ TEST(FromArrayTest, CopyDataTypeConversion) {
   auto context = Context::Default();
   auto source = tensorstore::MakeArray<std::int32_t>({1, 2, 3});
   auto dest = tensorstore::AllocateArray<std::int64_t>({3});
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
             GetStatus(tensorstore::Copy(tensorstore::FromArray(context, source),
                                         tensorstore::FromArray(context, dest))
                           .commit_future.result()));
@@ -646,6 +599,43 @@ TEST(FromArrayTest, CopyInvalidDataTypeConversion) {
           "Explicit data type conversion required to convert int32 -> int16"));
 }
 
+TEST(FromArrayTest, ChunkLayoutCOrder) {
+  auto array =
+      tensorstore::AllocateArray<float>({2, 3, 4}, tensorstore::c_order);
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(
+      auto store, tensorstore::FromArray(Context::Default(), array));
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto expected_layout,
+                                   ChunkLayout::FromJson({
+                                       {"grid_origin", {0, 0, 0}},
+                                       {"inner_order", {0, 1, 2}},
+                                   }));
+  EXPECT_THAT(store.chunk_layout(), ::testing::Optional(expected_layout));
+}
+
+TEST(FromArrayTest, ChunkLayoutFortranOrder) {
+  auto array =
+      tensorstore::AllocateArray<float>({2, 3, 4}, tensorstore::fortran_order);
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(
+      auto store, tensorstore::FromArray(Context::Default(), array));
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto expected_layout,
+                                   ChunkLayout::FromJson({
+                                       {"grid_origin", {0, 0, 0}},
+                                       {"inner_order", {2, 1, 0}},
+                                   }));
+  EXPECT_THAT(store.chunk_layout(), ::testing::Optional(expected_layout));
+}
+
+TEST(FromArrayTest, DimensionUnits) {
+  auto array =
+      tensorstore::MakeOffsetArray<int>({1, 2}, {{1, 2, 3}, {4, 5, 6}});
+  auto context = Context::Default();
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(
+      auto store, tensorstore::FromArray(context, array, {"4nm", "5nm"}));
+  EXPECT_THAT(store.dimension_units(),
+              ::testing::Optional(::testing::ElementsAre(
+                  tensorstore::Unit("4nm"), tensorstore::Unit("5nm"))));
+}
+
 }  // namespace frontend_tests
 
 namespace open_tests {
@@ -670,7 +660,7 @@ TYPED_TEST(OpenNumericTest, Roundtrip) {
       {"array",
        {{static_cast<T>(1), static_cast<T>(2), static_cast<T>(3)},
         {static_cast<T>(4), static_cast<T>(5), static_cast<T>(6)}}},
-      {"dtype", std::string(tensorstore::DataTypeOf<T>().name())},
+      {"dtype", std::string(tensorstore::dtype_v<T>.name())},
       {"transform",
        {{"input_inclusive_min", {1, 2}},
         {"input_exclusive_max", {3, 5}},
@@ -682,10 +672,10 @@ TYPED_TEST(OpenNumericTest, Roundtrip) {
             },
         }}},
   };
-  auto context = Context::Default();
-  auto store = tensorstore::Open(context, json_spec).value();
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto store,
+                                   tensorstore::Open(json_spec).result());
   EXPECT_THAT(store.spec().value().ToJson(tensorstore::IncludeDefaults{false}),
-              ::testing::Optional(json_spec));
+              ::testing::Optional(MatchesJson(json_spec)));
   EXPECT_EQ(
       tensorstore::MakeOffsetArray<TypeParam>({1, 2}, {{1, 2, 3}, {4, 5, 6}}),
       tensorstore::Read(store).value());
@@ -699,23 +689,37 @@ TEST(OpenTest, RoundtripString) {
       {"transform",
        {{"input_exclusive_max", {2, 3}}, {"input_inclusive_min", {0, 0}}}},
   };
-  auto context = Context::Default();
-  auto store = tensorstore::Open(context, json_spec).value();
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto store,
+                                   tensorstore::Open(json_spec).result());
   EXPECT_THAT(store.spec().value().ToJson(tensorstore::IncludeDefaults{false}),
               ::testing::Optional(json_spec));
-  EXPECT_EQ(
-      tensorstore::MakeArray<std::string>({{"a", "b", "c"}, {"d", "e", "f"}}),
-      tensorstore::Read(store).value());
+  EXPECT_THAT(tensorstore::Read(store).result(),
+              ::testing::Optional(tensorstore::MakeArray<std::string>(
+                  {{"a", "b", "c"}, {"d", "e", "f"}})));
+}
+
+TEST(OpenTest, RoundtripDimensionUnits) {
+  ::nlohmann::json json_spec{
+      {"driver", "array"},
+      {"array", {{"a", "b", "c"}, {"d", "e", "f"}}},
+      {"dtype", "string"},
+      {"schema", {{"dimension_units", {{4, "nm"}, {5, "nm"}}}}},
+      {"transform",
+       {{"input_exclusive_max", {2, 3}}, {"input_inclusive_min", {0, 0}}}},
+  };
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto store,
+                                   tensorstore::Open(json_spec).result());
+  EXPECT_THAT(store.spec().value().ToJson(tensorstore::IncludeDefaults{false}),
+              ::testing::Optional(MatchesJson(json_spec)));
 }
 
 TEST(OpenTest, InvalidConversion) {
-  ::nlohmann::json json_spec{
-      {"driver", "array"},
-      {"array", {"a"}},
-      {"dtype", "int32"},
-  };
-  auto context = Context::Default();
-  EXPECT_THAT(tensorstore::Open(context, json_spec).result(),
+  EXPECT_THAT(tensorstore::Open({
+                                    {"driver", "array"},
+                                    {"array", {"a"}},
+                                    {"dtype", "int32"},
+                                })
+                  .result(),
               MatchesStatus(absl::StatusCode::kInvalidArgument,
                             ".*: Expected integer .*, but received: \"a\""));
 }
@@ -747,26 +751,22 @@ TEST(FromArrayTest, Spec) {
             },
         }}},
   };
-  EXPECT_THAT(store.spec().value().ToJson(tensorstore::IncludeContext{false}),
-              ::testing::Optional(json_spec));
+  EXPECT_THAT(store.spec().value().ToJson(),
+              ::testing::Optional(MatchesJson(json_spec)));
 }
 
 TEST(OpenTest, MissingDataType) {
-  auto context = Context::Default();
-  EXPECT_THAT(tensorstore::Open(context,
-                                ::nlohmann::json{
+  EXPECT_THAT(tensorstore::Open({
                                     {"driver", "array"},
                                     {"array", {{1, 2, 3}, {4, 5, 6}}},
                                 })
                   .result(),
               MatchesStatus(absl::StatusCode::kInvalidArgument,
-                            "Data type must be specified"));
+                            "dtype must be specified"));
 }
 
 TEST(OpenTest, InvalidTransformRank) {
-  auto context = Context::Default();
-  EXPECT_THAT(tensorstore::Open(context,
-                                ::nlohmann::json{
+  EXPECT_THAT(tensorstore::Open({
                                     {"driver", "array"},
                                     {"array", {{1, 2, 3}, {4, 5, 6}}},
                                     {"dtype", "int32"},
@@ -784,10 +784,8 @@ TEST(OpenTest, InvalidTransformRank) {
 }
 
 TEST(OpenTest, InvalidRank) {
-  auto context = Context::Default();
   EXPECT_THAT(
-      tensorstore::Open(context,
-                        ::nlohmann::json{
+      tensorstore::Open({
                             {"driver", "array"},
                             {"array", {{1, 2, 3}, {4, 5, 6}}},
                             {"dtype", "int32"},
@@ -799,5 +797,154 @@ TEST(OpenTest, InvalidRank) {
                     "Array rank \\(2\\) does not match expected rank \\(3\\)"));
 }
 
+// Tests that copying from an array driver to itself does not lead to deadlock.
+TEST(CopyTest, SelfCopy) {
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(
+      auto store,
+      tensorstore::FromArray(Context::Default(),
+                             tensorstore::MakeArray<int>({1, 2, 3, 4})));
+  TENSORSTORE_EXPECT_OK(
+      tensorstore::Copy(store | tensorstore::Dims(0).SizedInterval(0, 2),
+                        store | tensorstore::Dims(0).SizedInterval(2, 2)));
+  EXPECT_EQ(tensorstore::Read(store).result(),
+            tensorstore::MakeArray<int>({1, 2, 1, 2}));
+}
+
+TEST(ArrayTest, SpecRankPropagation) {
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto spec, tensorstore::Spec::FromJson({
+                                                  {"driver", "array"},
+                                                  {"array", {1, 2, 3}},
+                                                  {"dtype", "int32"},
+                                              }));
+  EXPECT_EQ(1, spec.rank());
+}
+
+TEST(ArrayDriverHandle, OpenResolveBounds) {
+  ::nlohmann::json json_spec{
+      {"driver", "array"},
+      {"array", {{1, 2, 3}, {4, 5, 6}}},
+      {"dtype", "uint32"},
+  };
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto spec,
+                                   tensorstore::Spec::FromJson(json_spec));
+
+  tensorstore::TransactionalOpenOptions options;
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(
+      auto handle,
+      tensorstore::internal::OpenDriver(
+          std::move(tensorstore::internal_spec::SpecAccess::impl(spec)),
+          std::move(options))
+          .result());
+
+  // Validates that an identity transform of rank returns the array bounds.
+  auto transform_result =
+      handle.driver
+          ->ResolveBounds(
+              {}, tensorstore::IdentityTransform(handle.driver->rank()), {})
+          .result();
+
+  EXPECT_THAT(transform_result->input_origin(), ::testing::ElementsAre(0, 0));
+  EXPECT_THAT(transform_result->input_shape(), ::testing::ElementsAre(2, 3));
+}
+
+TEST(ArrayTest, SpecFromArray) {
+  auto orig_array = tensorstore::MakeOffsetArray<float>({2}, {1, 2, 3});
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto spec,
+                                   tensorstore::SpecFromArray(orig_array));
+  EXPECT_EQ(1, spec.rank());
+  EXPECT_EQ(tensorstore::dtype_v<float>, spec.dtype());
+  EXPECT_THAT(spec.ToJson(),
+              ::testing::Optional(MatchesJson({
+                  {"driver", "array"},
+                  {"array", {1, 2, 3}},
+                  {"dtype", "float32"},
+                  {"transform",
+                   {
+                       {"input_inclusive_min", {2}},
+                       {"input_exclusive_max", {5}},
+                       {"output", {{{"input_dimension", 0}, {"offset", -2}}}},
+                   }},
+              })));
+
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto store,
+                                   tensorstore::Open(spec).result());
+  EXPECT_THAT(tensorstore::Read(store).result(),
+              ::testing::Optional(orig_array));
+}
+
+TEST(SpecSchemaTest, Basic) {
+  TestSpecSchema(
+      {
+          {"driver", "array"},
+          {"array", {{1, 2, 3}, {4, 5, 6}}},
+          {"dtype", "float32"},
+          {"schema", {{"dimension_units", {"4nm", "5nm"}}}},
+      },
+      {
+          {"rank", 2},
+          {"dtype", "float32"},
+          {"domain", {{"shape", {2, 3}}}},
+          {"chunk_layout", {{"grid_origin", {0, 0}}, {"inner_order", {0, 1}}}},
+          {"dimension_units", {"4nm", "5nm"}},
+      });
+}
+
+TEST(CreateCheckSchemaTest, Basic) {
+  TestTensorStoreCreateCheckSchema(
+      {
+          {"driver", "array"},
+          {"array", {{1, 2, 3}, {4, 5, 6}}},
+          {"dtype", "float32"},
+          {"schema", {{"dimension_units", {"4nm", "5nm"}}}},
+      },
+      {
+          {"rank", 2},
+          {"dtype", "float32"},
+          {"domain", {{"shape", {2, 3}}}},
+          {"chunk_layout", {{"grid_origin", {0, 0}}, {"inner_order", {0, 1}}}},
+          {"dimension_units", {"4nm", "5nm"}},
+      });
+}
+
+TEST(ArrayTest, SpecFromArrayWithDimensionUnits) {
+  auto orig_array = tensorstore::MakeOffsetArray<float>({2}, {1, 2, 3});
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(
+      auto spec, tensorstore::SpecFromArray(orig_array, {"5nm"}));
+  EXPECT_EQ(1, spec.rank());
+  EXPECT_EQ(tensorstore::dtype_v<float>, spec.dtype());
+  EXPECT_THAT(spec.ToJson(),
+              ::testing::Optional(MatchesJson(::nlohmann::json{
+                  {"driver", "array"},
+                  {"array", {1, 2, 3}},
+                  {"dtype", "float32"},
+                  {"transform",
+                   {
+                       {"input_inclusive_min", {2}},
+                       {"input_exclusive_max", {5}},
+                       {"output", {{{"input_dimension", 0}, {"offset", -2}}}},
+                   }},
+                  {"schema", {{"dimension_units", {{5, "nm"}}}}},
+              })));
+
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto store,
+                                   tensorstore::Open(spec).result());
+  EXPECT_THAT(
+      store.dimension_units(),
+      ::testing::Optional(::testing::ElementsAre(tensorstore::Unit("5nm"))));
+}
+
+TEST(ArraySerializationTest, SerializationRoundTest) {
+  ::nlohmann::json json_spec{
+      {"driver", "array"},
+      {"array", 42},
+      {"dtype", "uint32"},
+      {"transform", {{"input_rank", 0}}},
+      {"context", {{"data_copy_concurrency", {{"limit", 1}}}}}};
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto spec,
+                                   tensorstore::Spec::FromJson(json_spec));
+  tensorstore::serialization::TestSerializationRoundTrip(spec);
+  tensorstore::TestJsonBinderRoundTrip<tensorstore::Spec>({{spec, json_spec}});
+}
+
 }  // namespace open_tests
 }  // namespace
