@@ -16,7 +16,7 @@
 #define TENSORSTORE_DRIVER_ZARR_COMPRESSOR_H_
 
 #include "tensorstore/internal/compression/json_specified_compressor.h"
-#include "tensorstore/internal/json_bindable.h"
+#include "tensorstore/internal/json_binding/bindable.h"
 
 namespace tensorstore {
 namespace internal_zarr {
