@@ -15,34 +15,18 @@
 #ifndef TENSORSTORE_DRIVER_ZARR_DRIVER_IMPL_H_
 #define TENSORSTORE_DRIVER_ZARR_DRIVER_IMPL_H_
 
-#include <memory>
 #include <string>
 
-#include "tensorstore/driver/kvs_backed_chunk_driver.h"
 #include "tensorstore/driver/zarr/metadata.h"
-#include "tensorstore/driver/zarr/spec.h"
 #include "tensorstore/index.h"
-#include "tensorstore/index_space/index_transform.h"
-#include "tensorstore/internal/chunk_cache.h"
-#include "tensorstore/kvstore/key_value_store.h"
-#include "tensorstore/open_mode.h"
-#include "tensorstore/spec.h"
-#include "tensorstore/util/bit_span.h"
 #include "tensorstore/util/span.h"
-#include "tensorstore/util/status.h"
 
 namespace tensorstore {
 namespace internal_zarr {
 
 /// Encodes a chunk grid index vector as a storage key suffix.
 std::string EncodeChunkIndices(span<const Index> indices,
-                               ChunkKeyEncoding key_encoding);
-
-/// Returns a `DataCacheState`.
-///
-/// This is exposed for testing only.
-internal_kvs_backed_chunk_driver::DataCacheState::Ptr MakeDataCacheState(
-    std::string key_prefix, ChunkKeyEncoding key_encoding);
+                               DimensionSeparator dimension_separator);
 
 }  // namespace internal_zarr
 }  // namespace tensorstore
