@@ -23,37 +23,37 @@
 #include "tensorstore/driver/zarr/compressor.h"
 #include "tensorstore/internal/container_to_shared.h"
 #include "tensorstore/internal/data_type_endian_conversion.h"
-#include "tensorstore/internal/json.h"
+#include "tensorstore/internal/flat_cord_builder.h"
+#include "tensorstore/internal/json_binding/dimension_indexed.h"
+#include "tensorstore/internal/json_binding/enum.h"
+#include "tensorstore/internal/json_binding/json_binding.h"
+#include "tensorstore/internal/json_binding/std_optional.h"
+#include "tensorstore/serialization/fwd.h"
+#include "tensorstore/serialization/json_bindable.h"
 
 namespace tensorstore {
 namespace internal_zarr {
 
-Result<ContiguousLayoutOrder> ParseOrder(const nlohmann::json& j) {
-  if (j.is_string()) {
-    std::string value = j.get<std::string>();
-    if (value == "C") return ContiguousLayoutOrder::c;
-    if (value == "F") return ContiguousLayoutOrder::fortran;
-  }
-  return absl::InvalidArgumentError(
-      StrCat("Expected \"C\" or \"F\", but received: ", j.dump()));
+namespace jb = internal_json_binding;
+TENSORSTORE_DEFINE_JSON_BINDER(
+    OrderJsonBinder,
+    jb::Enum<ContiguousLayoutOrder, std::string_view>({{c_order, "C"},
+                                                       {fortran_order, "F"}}))
+
+TENSORSTORE_DEFINE_JSON_BINDER(DimensionSeparatorJsonBinder,
+                               jb::Enum<DimensionSeparator, std::string_view>(
+                                   {{DimensionSeparator::kDotSeparated, "."},
+                                    {DimensionSeparator::kSlashSeparated,
+                                     "/"}}))
+
+void to_json(::nlohmann::json& out, DimensionSeparator value) {
+  DimensionSeparatorJsonBinder(/*is_loading=*/std::false_type{},
+                               /*options=*/internal_json_binding::NoOptions{},
+                               &value, &out)
+      .IgnoreError();
 }
 
-::nlohmann::json EncodeOrder(ContiguousLayoutOrder order) {
-  return order == c_order ? "C" : "F";
-}
-
-template <typename T>
-inline T ConvertFromScalarArray(ElementPointer<const void> x) {
-  TENSORSTORE_UNREACHABLE;
-}
-
-template <typename T, typename U0, typename... U>
-inline T ConvertFromScalarArray(ElementPointer<const void> x) {
-  if (x.data_type() == DataTypeOf<U0>()) {
-    return static_cast<T>(*static_cast<const U0*>(x.data()));
-  }
-  return ConvertFromScalarArray<T, U...>(x);
-}
+namespace {
 
 Result<double> DecodeFloat(const nlohmann::json& j) {
   double value;
@@ -72,6 +72,15 @@ Result<double> DecodeFloat(const nlohmann::json& j) {
   return value;
 }
 
+::nlohmann::json EncodeFloat(double value) {
+  if (std::isnan(value)) return "NaN";
+  if (value == std::numeric_limits<double>::infinity()) return "Infinity";
+  if (value == -std::numeric_limits<double>::infinity()) return "-Infinity";
+  return value;
+}
+
+}  // namespace
+
 Result<std::vector<SharedArray<const void>>> ParseFillValue(
     const nlohmann::json& j, const ZarrDType& dtype) {
   std::vector<SharedArray<const void>> fill_values;
@@ -85,40 +94,37 @@ Result<std::vector<SharedArray<const void>>> ParseFillValue(
       case 'f': {
         TENSORSTORE_ASSIGN_OR_RETURN(double value, DecodeFloat(j));
         fill_values[0] =
-            MakeCopy(MakeScalarArrayView(value), c_order, field.data_type)
-                .value();
+            MakeCopy(MakeScalarArrayView(value), c_order, field.dtype).value();
         return fill_values;
       }
       case 'i': {
         std::int64_t value;
-        const std::size_t num_bits = 8 * field.data_type->size - 1;
+        const std::size_t num_bits = 8 * field.dtype->size - 1;
         const std::uint64_t max_value = static_cast<std::int64_t>(
             (static_cast<std::uint64_t>(1) << num_bits) - 1);
         const std::int64_t min_value = static_cast<std::int64_t>(-1)
                                        << num_bits;
-        TENSORSTORE_RETURN_IF_ERROR(internal::JsonRequireInteger(
+        TENSORSTORE_RETURN_IF_ERROR(internal_json::JsonRequireInteger(
             j, &value, /*strict=*/true, min_value, max_value));
         fill_values[0] =
-            MakeCopy(MakeScalarArrayView(value), c_order, field.data_type)
-                .value();
+            MakeCopy(MakeScalarArrayView(value), c_order, field.dtype).value();
         return fill_values;
       }
       case 'u': {
         std::uint64_t value;
-        const std::size_t num_bits = 8 * field.data_type->size;
+        const std::size_t num_bits = 8 * field.dtype->size;
         const std::uint64_t max_value =
             (static_cast<std::uint64_t>(2) << (num_bits - 1)) - 1;
-        TENSORSTORE_RETURN_IF_ERROR(internal::JsonRequireInteger(
+        TENSORSTORE_RETURN_IF_ERROR(internal_json::JsonRequireInteger(
             j, &value, /*strict=*/true, 0, max_value));
         fill_values[0] =
-            MakeCopy(MakeScalarArrayView(value), c_order, field.data_type)
-                .value();
+            MakeCopy(MakeScalarArrayView(value), c_order, field.dtype).value();
         return fill_values;
       }
       case 'b': {
         bool value;
         TENSORSTORE_RETURN_IF_ERROR(
-            internal::JsonRequireValueAs(j, &value, /*strict=*/true));
+            internal_json::JsonRequireValueAs(j, &value, /*strict=*/true));
         fill_values[0] = MakeScalarArray<bool>(value);
         return fill_values;
       }
@@ -128,10 +134,10 @@ Result<std::vector<SharedArray<const void>>> ParseFillValue(
           // Fallthrough to allow base64 encoding.
           break;
         }
-        TENSORSTORE_RETURN_IF_ERROR(internal::JsonParseArray(
+        TENSORSTORE_RETURN_IF_ERROR(internal_json::JsonParseArray(
             j,
             [](std::ptrdiff_t size) {
-              return internal::JsonValidateArrayLength(size, 2);
+              return internal_json::JsonValidateArrayLength(size, 2);
             },
             [&](const ::nlohmann::json& v, std::ptrdiff_t i) {
               TENSORSTORE_ASSIGN_OR_RETURN(values[i], DecodeFloat(v));
@@ -139,7 +145,7 @@ Result<std::vector<SharedArray<const void>>> ParseFillValue(
             }));
         fill_values[0] =
             MakeCopy(MakeScalarArrayView(complex128_t(values[0], values[1])),
-                     c_order, field.data_type)
+                     c_order, field.dtype)
                 .value();
         return fill_values;
       }
@@ -156,7 +162,7 @@ Result<std::vector<SharedArray<const void>>> ParseFillValue(
   }
   for (size_t field_i = 0; field_i < dtype.fields.size(); ++field_i) {
     auto& field = dtype.fields[field_i];
-    DataType r = field.data_type;
+    DataType r = field.dtype;
     auto fill_value = AllocateArray(field.field_shape, ContiguousLayoutOrder::c,
                                     default_init, r);
     internal::DecodeArray(
@@ -170,13 +176,6 @@ Result<std::vector<SharedArray<const void>>> ParseFillValue(
   return fill_values;
 }
 
-::nlohmann::json EncodeFloat(double value) {
-  if (std::isnan(value)) return "NaN";
-  if (value == std::numeric_limits<double>::infinity()) return "Infinity";
-  if (value == -std::numeric_limits<double>::infinity()) return "-Infinity";
-  return value;
-}
-
 ::nlohmann::json EncodeFillValue(
     const ZarrDType& dtype, span<const SharedArray<const void>> fill_values) {
   assert(dtype.fields.size() == static_cast<size_t>(fill_values.size()));
@@ -216,7 +215,7 @@ Result<std::vector<SharedArray<const void>>> ParseFillValue(
     const auto& field = dtype.fields[field_i];
     const auto& fill_value = fill_values[field_i];
     if (!fill_value.valid()) return nullptr;
-    DataType r = field.data_type;
+    DataType r = field.dtype;
     Array<void> encoded_fill_value(
         {static_cast<void*>(buffer.data() + field.byte_offset), r},
         field.field_shape);
@@ -224,8 +223,8 @@ Result<std::vector<SharedArray<const void>>> ParseFillValue(
   }
   std::string b64_encoded;
   absl::Base64Escape(
-      absl::string_view(reinterpret_cast<const char*>(buffer.data()),
-                        buffer.size()),
+      std::string_view(reinterpret_cast<const char*>(buffer.data()),
+                       buffer.size()),
       &b64_encoded);
   return b64_encoded;
 }
@@ -252,6 +251,7 @@ Result<ZarrChunkLayout> ComputeChunkLayout(const ZarrDType& dtype,
     auto& field_layout = layout.fields[field_i];
     const DimensionIndex inner_rank = field.field_shape.size();
     const DimensionIndex total_rank = chunk_shape.size() + inner_rank;
+    TENSORSTORE_RETURN_IF_ERROR(ValidateRank(total_rank));
     const auto initialize_layout = [&](StridedLayout<>* strided_layout,
                                        Index outer_element_stride) {
       strided_layout->set_rank(total_rank);
@@ -259,7 +259,7 @@ Result<ZarrChunkLayout> ComputeChunkLayout(const ZarrDType& dtype,
                 std::copy(chunk_shape.begin(), chunk_shape.end(),
                           strided_layout->shape().begin()));
       // Compute strides for inner field dimensions.
-      ComputeStrides(ContiguousLayoutOrder::c, field.data_type->size,
+      ComputeStrides(ContiguousLayoutOrder::c, field.dtype->size,
                      strided_layout->shape().last(inner_rank),
                      strided_layout->byte_strides().last(inner_rank));
       // Compute strides for outer dimensions.
@@ -273,144 +273,91 @@ Result<ZarrChunkLayout> ComputeChunkLayout(const ZarrDType& dtype,
   return layout;
 }
 
-namespace {
-Status ParseIndexVector(const ::nlohmann::json& value,
-                        absl::optional<DimensionIndex>* rank,
-                        std::vector<Index>* vec, Index min_value,
-                        Index max_value) {
-  return internal::JsonParseArray(
-      value,
-      [&](std::ptrdiff_t size) {
-        if (*rank) {
-          TENSORSTORE_RETURN_IF_ERROR(
-              internal::JsonValidateArrayLength(size, **rank));
-        } else {
-          *rank = size;
+constexpr auto MetadataJsonBinder = [](auto maybe_optional) {
+  return [=](auto is_loading, const auto& options, auto* obj, auto* j) {
+    using T = internal::remove_cvref_t<decltype(*obj)>;
+    DimensionIndex* rank = nullptr;
+    if constexpr (is_loading) {
+      rank = &obj->rank;
+    }
+    auto ensure_dtype = [&]() -> Result<const ZarrDType*> {
+      if constexpr (std::is_same_v<T, ZarrMetadata>) {
+        return &obj->dtype;
+      } else if constexpr (std::is_same_v<T, ZarrPartialMetadata>) {
+        /// dtype is wrapped std::optional<>
+        if (!obj->dtype) {
+          return absl::InvalidArgumentError(
+              "must be specified in conjunction with \"dtype\"");
         }
-        vec->resize(size);
-        return absl::OkStatus();
-      },
-      [&](const ::nlohmann::json& v, std::ptrdiff_t i) {
-        return internal::JsonRequireInteger<Index>(
-            v, &(*vec)[i], /*strict=*/true, min_value, max_value);
-      });
-}
-}  // namespace
+        return &*obj->dtype;
+      }
+      TENSORSTORE_UNREACHABLE;
+    };
 
-Status ParseChunkShape(const ::nlohmann::json& value,
-                       absl::optional<DimensionIndex>* rank,
-                       std::vector<Index>* shape) {
-  return ParseIndexVector(value, rank, shape, /*min_value=*/1,
-                          /*max_value=*/kInfIndex);
-}
-
-Status ParseShape(const ::nlohmann::json& value,
-                  absl::optional<DimensionIndex>* rank,
-                  std::vector<Index>* shape) {
-  return ParseIndexVector(value, rank, shape, /*min_value=*/0,
-                          /*max_value=*/kInfIndex);
-}
-
-Status ParseFilters(const nlohmann::json& value) {
-  if (!value.is_null()) {
-    return absl::InvalidArgumentError("Filters not supported");
-  }
-  return absl::OkStatus();
-}
-
-Result<std::uint64_t> ParseZarrFormat(const nlohmann::json& value) {
-  int result;
-  if (auto status = internal::JsonRequireInteger(value, &result,
-                                                 /*strict=*/true, 2, 2);
-      status.ok()) {
-    return result;
-  } else {
-    return status;
-  }
-}
-
-Status ParseMetadata(const nlohmann::json& j, ZarrMetadata* metadata) {
-  absl::optional<DimensionIndex> rank;
-  TENSORSTORE_RETURN_IF_ERROR(internal::JsonValidateObjectMembers(
-      j, {"zarr_format", "shape", "chunks", "dtype", "compressor", "fill_value",
-          "order", "filters"}));
-  TENSORSTORE_RETURN_IF_ERROR(internal::JsonRequireObjectMember(
-      j, "zarr_format", [&](const ::nlohmann::json& value) {
-        TENSORSTORE_ASSIGN_OR_RETURN(metadata->zarr_format,
-                                     internal_zarr::ParseZarrFormat(value));
-        return absl::OkStatus();
-      }));
-
-  TENSORSTORE_RETURN_IF_ERROR(internal::JsonRequireObjectMember(
-      j, "shape", [&](const ::nlohmann::json& value) {
-        return ParseShape(value, &rank, &metadata->shape);
-      }));
-
-  TENSORSTORE_RETURN_IF_ERROR(internal::JsonRequireObjectMember(
-      j, "chunks", [&](const ::nlohmann::json& value) {
-        return ParseChunkShape(value, &rank, &metadata->chunks);
-      }));
-
-  TENSORSTORE_RETURN_IF_ERROR(internal::JsonRequireObjectMember(
-      j, "dtype", [&](const ::nlohmann::json& value) {
-        TENSORSTORE_ASSIGN_OR_RETURN(metadata->dtype, ParseDType(value));
-        return absl::OkStatus();
-      }));
-
-  TENSORSTORE_RETURN_IF_ERROR(internal::JsonRequireObjectMember(
-      j, "compressor", [&](const ::nlohmann::json& value) {
-        TENSORSTORE_ASSIGN_OR_RETURN(metadata->compressor,
-                                     Compressor::FromJson(value));
-        return absl::OkStatus();
-      }));
-
-  TENSORSTORE_RETURN_IF_ERROR(internal::JsonRequireObjectMember(
-      j, "fill_value", [&](const ::nlohmann::json& value) {
-        TENSORSTORE_ASSIGN_OR_RETURN(metadata->fill_values,
-                                     ParseFillValue(value, metadata->dtype));
-        return absl::OkStatus();
-      }));
-
-  TENSORSTORE_RETURN_IF_ERROR(internal::JsonRequireObjectMember(
-      j, "order", [&](const ::nlohmann::json& value) {
-        TENSORSTORE_ASSIGN_OR_RETURN(metadata->order, ParseOrder(value));
-        return absl::OkStatus();
-      }));
-
-  TENSORSTORE_RETURN_IF_ERROR(internal::JsonRequireObjectMember(
-      j, "filters", [&](const ::nlohmann::json& value) {
-        return internal_zarr::ParseFilters(value);
-      }));
-
-  TENSORSTORE_ASSIGN_OR_RETURN(
-      metadata->chunk_layout,
-      ComputeChunkLayout(metadata->dtype, metadata->order, metadata->chunks));
-
-  return absl::OkStatus();
-}
-
-void to_json(::nlohmann::json& out,  // NOLINT,
-             const ZarrMetadata& metadata) {
-  out = ::nlohmann::json::object_t{
-      {"zarr_format", metadata.zarr_format},
-      {"shape", metadata.shape},
-      {"chunks", metadata.chunks},
-      {"dtype", metadata.dtype},
-      {"compressor", ::nlohmann::json(metadata.compressor)},
-      {"fill_value", EncodeFillValue(metadata.dtype, metadata.fill_values)},
-      {"order", EncodeOrder(metadata.order)},
-      {"filters", nullptr},
+    return jb::Object(
+        jb::Member("zarr_format",
+                   jb::Projection(&T::zarr_format,
+                                  maybe_optional(jb::Integer<int>(2, 2)))),
+        jb::Member(
+            "shape",
+            jb::Projection(&T::shape, maybe_optional(jb::ShapeVector(rank)))),
+        jb::Member("chunks",
+                   jb::Projection(&T::chunks,
+                                  maybe_optional(jb::ChunkShapeVector(rank)))),
+        jb::Member("dtype", jb::Projection(&T::dtype)),
+        jb::Member("compressor", jb::Projection(&T::compressor)),
+        jb::Member("fill_value",
+                   jb::Projection(
+                       &T::fill_value,
+                       maybe_optional([&](auto is_loading, const auto& options,
+                                          auto* obj, auto* j) {
+                         TENSORSTORE_ASSIGN_OR_RETURN(auto* dtype,
+                                                      ensure_dtype());
+                         return FillValueJsonBinder(*dtype)(is_loading, options,
+                                                            obj, j);
+                       }))),
+        jb::Member("order",
+                   jb::Projection(&T::order, maybe_optional(OrderJsonBinder))),
+        jb::Member("filters", jb::Projection(&T::filters)),
+        jb::Member("dimension_separator",
+                   jb::Projection(&T::dimension_separator,
+                                  jb::Optional(DimensionSeparatorJsonBinder))),
+        [](auto is_loading, const auto& options, auto* obj, auto* j) {
+          if constexpr (std::is_same_v<T, ZarrMetadata>) {
+            return jb::DefaultBinder<>(is_loading, options, &obj->extra_members,
+                                       j);
+          } else {
+            return absl::OkStatus();
+          }
+        })(is_loading, options, obj, j);
   };
+};
+
+absl::Status ValidateMetadata(ZarrMetadata& metadata) {
+  TENSORSTORE_ASSIGN_OR_RETURN(
+      metadata.chunk_layout,
+      ComputeChunkLayout(metadata.dtype, metadata.order, metadata.chunks));
+  return absl::OkStatus();
 }
 
+TENSORSTORE_DEFINE_JSON_DEFAULT_BINDER(
+    ZarrMetadata, jb::Validate([](const auto& options,
+                                  auto* obj) { return ValidateMetadata(*obj); },
+                               MetadataJsonBinder(internal::identity{})))
+
+TENSORSTORE_DEFINE_JSON_DEFAULT_BINDER(ZarrPartialMetadata,
+                                       MetadataJsonBinder([](auto binder) {
+                                         return jb::Optional(binder);
+                                       }))
+
 // Two decoding strategies:  raw decoder and custom decoder.  Initially we will
 // only support raw decoder.
 
 Result<absl::InlinedVector<SharedArrayView<const void>, 1>> DecodeChunk(
-    const ZarrMetadata& metadata, std::string buffer) {
+    const ZarrMetadata& metadata, absl::Cord buffer) {
   const size_t num_fields = metadata.dtype.fields.size();
   if (metadata.compressor) {
-    std::string decoded;
+    absl::Cord decoded;
     TENSORSTORE_RETURN_IF_ERROR(metadata.compressor->Decode(
         buffer, &decoded, metadata.dtype.bytes_per_outer_element));
     buffer = std::move(decoded);
@@ -421,52 +368,116 @@ Result<absl::InlinedVector<SharedArrayView<const void>, 1>> DecodeChunk(
         "Uncompressed chunk is ", buffer.size(), " bytes, but should be ",
         metadata.chunk_layout.bytes_per_chunk, " bytes"));
   }
-  std::shared_ptr<void> buffer_ptr =
-      internal::ContainerToSharedDataPointerWithOffset(std::move(buffer));
   absl::InlinedVector<SharedArrayView<const void>, 1> field_arrays(num_fields);
+
+  bool must_copy = false;
+  // First, attempt to create arrays that reference the cord without copying.
   for (size_t field_i = 0; field_i < num_fields; ++field_i) {
     const auto& field = metadata.dtype.fields[field_i];
     const auto& field_layout = metadata.chunk_layout.fields[field_i];
-    SharedArrayView<void> array{
-        {AddByteOffset(buffer_ptr, field.byte_offset), field.data_type},
-        field_layout.encoded_chunk_layout};
-    DecodeArray(&array, field.endian, field_layout.decoded_chunk_layout);
-    field_arrays[field_i] = array;
+    field_arrays[field_i] = internal::TryViewCordAsArray(
+        buffer, field.byte_offset, field.dtype, field.endian,
+        field_layout.encoded_chunk_layout);
+    if (!field_arrays[field_i].valid()) {
+      must_copy = true;
+      break;
+    }
+  }
+  if (must_copy) {
+    auto flat_buffer = buffer.Flatten();
+    for (size_t field_i = 0; field_i < num_fields; ++field_i) {
+      const auto& field = metadata.dtype.fields[field_i];
+      const auto& field_layout = metadata.chunk_layout.fields[field_i];
+      ArrayView<const void> source_array{
+          ElementPointer<const void>(
+              static_cast<const void*>(flat_buffer.data() + field.byte_offset),
+              field.dtype),
+          field_layout.encoded_chunk_layout};
+      field_arrays[field_i] = internal::CopyAndDecodeArray(
+          source_array, field.endian, field_layout.decoded_chunk_layout);
+    }
   }
   return field_arrays;
 }
 
-Status EncodeChunk(const ZarrMetadata& metadata,
-                   span<const ArrayView<const void>> components,
-                   std::string* out) {
-  const size_t num_fields = metadata.dtype.fields.size();
-  out->resize(metadata.chunk_layout.bytes_per_chunk);
-  ByteStridedPointer<void> data_ptr = out->data();
-  for (size_t field_i = 0; field_i < num_fields; ++field_i) {
+namespace {
+bool SingleArrayMatchesEncodedRepresentation(
+    const ZarrMetadata& metadata,
+    const SharedArrayView<const void>& component) {
+  auto& field = metadata.dtype.fields[0];
+  if (field.endian != endian::native) return false;
+  return internal::RangesEqual(
+      component.byte_strides(),
+      metadata.chunk_layout.fields[0].encoded_chunk_layout.byte_strides());
+}
+
+absl::Cord MakeCordFromSharedPtr(std::shared_ptr<const void> ptr, size_t size) {
+  std::string_view s(static_cast<const char*>(ptr.get()), size);
+  return absl::MakeCordFromExternal(
+      s, [ptr = std::move(ptr)](std::string_view s) mutable { ptr.reset(); });
+}
+
+absl::Cord MakeCordFromContiguousArray(
+    const SharedArrayView<const void>& array) {
+  return MakeCordFromSharedPtr(array.pointer(),
+                               array.num_elements() * array.dtype().size());
+}
+
+absl::Cord CopyComponentsToEncodedLayout(
+    const ZarrMetadata& metadata,
+    span<const SharedArrayView<const void>> components) {
+  internal::FlatCordBuilder output_builder(
+      metadata.chunk_layout.bytes_per_chunk);
+  ByteStridedPointer<void> data_ptr = output_builder.data();
+  for (size_t field_i = 0; field_i < components.size(); ++field_i) {
     const auto& field = metadata.dtype.fields[field_i];
     const auto& field_layout = metadata.chunk_layout.fields[field_i];
-    ArrayView<void> encoded_array{
-        {data_ptr + field.byte_offset, field.data_type},
-        field_layout.encoded_chunk_layout};
+    ArrayView<void> encoded_array{{data_ptr + field.byte_offset, field.dtype},
+                                  field_layout.encoded_chunk_layout};
     internal::EncodeArray(components[field_i], encoded_array, field.endian);
   }
-  if (metadata.compressor) {
-    std::string encoded;
-    TENSORSTORE_RETURN_IF_ERROR(metadata.compressor->Encode(
-        *out, &encoded, metadata.dtype.bytes_per_outer_element));
-    *out = std::move(encoded);
+  return std::move(output_builder).Build();
+}
+}  // namespace
+
+Result<absl::Cord> EncodeChunk(
+    const ZarrMetadata& metadata,
+    span<const SharedArrayView<const void>> components) {
+  absl::Cord output;
+  if (components.size() == 1 &&
+      SingleArrayMatchesEncodedRepresentation(metadata, components[0])) {
+    output = MakeCordFromContiguousArray(components[0]);
+  } else {
+    output = CopyComponentsToEncodedLayout(metadata, components);
   }
-  return absl::OkStatus();
+  if (metadata.compressor) {
+    absl::Cord encoded;
+    TENSORSTORE_RETURN_IF_ERROR(metadata.compressor->Encode(
+        output, &encoded, metadata.dtype.bytes_per_outer_element));
+    return encoded;
+  }
+  return output;
 }
 
 bool IsMetadataCompatible(const ZarrMetadata& a, const ZarrMetadata& b) {
   // Rank must be the same.
   if (a.shape.size() != b.shape.size()) return false;
+
   auto a_json = ::nlohmann::json(a);
   auto b_json = ::nlohmann::json(b);
+
   // Shape is allowed to differ.
   a_json.erase("shape");
   b_json.erase("shape");
+
+  // Extra members are allowed to differ.
+  for (const auto& [key, value] : a.extra_members) {
+    a_json.erase(key);
+  }
+  for (const auto& [key, value] : b.extra_members) {
+    b_json.erase(key);
+  }
+
   return a_json == b_json;
 }
 
@@ -478,3 +489,8 @@ void EncodeCacheKeyAdl(std::string* out, const ZarrMetadata& metadata) {
 
 }  // namespace internal_zarr
 }  // namespace tensorstore
+
+TENSORSTORE_DEFINE_SERIALIZER_SPECIALIZATION(
+    tensorstore::internal_zarr::ZarrPartialMetadata,
+    tensorstore::serialization::JsonBindableSerializer<
+        tensorstore::internal_zarr::ZarrPartialMetadata>())
