@@ -23,37 +23,44 @@
 #include "tensorstore/array_testutil.h"
 #include "tensorstore/driver/zarr/metadata_testutil.h"
 #include "tensorstore/index.h"
+#include "tensorstore/internal/json_binding/gtest.h"
 #include "tensorstore/internal/json_gtest.h"
 #include "tensorstore/util/status.h"
 #include "tensorstore/util/status_testutil.h"
 
 namespace {
-using tensorstore::ContiguousLayoutOrder;
-using tensorstore::DataTypeOf;
-using tensorstore::MakeArray;
-using tensorstore::MakeScalarArray;
-using tensorstore::MatchesStatus;
-using tensorstore::Status;
-using tensorstore::internal_zarr::EncodeFillValue;
-using tensorstore::internal_zarr::ParseDType;
-using tensorstore::internal_zarr::ParseFillValue;
-using tensorstore::internal_zarr::ParseMetadata;
-using tensorstore::internal_zarr::ParseOrder;
-using tensorstore::internal_zarr::ZarrMetadata;
-using testing::ElementsAre;
+using ::tensorstore::bfloat16_t;
+using ::tensorstore::ContiguousLayoutOrder;
+using ::tensorstore::dtype_v;
+using ::tensorstore::float16_t;
+using ::tensorstore::MakeArray;
+using ::tensorstore::MakeScalarArray;
+using ::tensorstore::MatchesStatus;
+using ::tensorstore::internal_zarr::DimensionSeparator;
+using ::tensorstore::internal_zarr::DimensionSeparatorJsonBinder;
+using ::tensorstore::internal_zarr::EncodeFillValue;
+using ::tensorstore::internal_zarr::OrderJsonBinder;
+using ::tensorstore::internal_zarr::ParseDType;
+using ::tensorstore::internal_zarr::ParseFillValue;
+using ::tensorstore::internal_zarr::ZarrMetadata;
+using ::testing::ElementsAre;
 
-TEST(ParseOrderTest, Success) {
-  EXPECT_EQ(ContiguousLayoutOrder::c, ParseOrder("C"));
-  EXPECT_EQ(ContiguousLayoutOrder::fortran, ParseOrder("F"));
+TEST(OrderJsonBinderTest, Success) {
+  tensorstore::TestJsonBinderRoundTrip<ContiguousLayoutOrder>(
+      {
+          {ContiguousLayoutOrder::c, "C"},
+          {ContiguousLayoutOrder::fortran, "F"},
+      },
+      OrderJsonBinder);
 }
 
 TEST(ParseOrderTest, Failure) {
-  EXPECT_THAT(ParseOrder("x"),
-              MatchesStatus(absl::StatusCode::kInvalidArgument,
-                            "Expected \"C\" or \"F\", but received: \"x\""));
-  EXPECT_THAT(ParseOrder(3),
-              MatchesStatus(absl::StatusCode::kInvalidArgument,
-                            "Expected \"C\" or \"F\", but received: 3"));
+  tensorstore::TestJsonBinderFromJson<ContiguousLayoutOrder>(
+      {
+          {"x", MatchesStatus(absl::StatusCode::kInvalidArgument)},
+          {3, MatchesStatus(absl::StatusCode::kInvalidArgument)},
+      },
+      OrderJsonBinder);
 }
 
 void TestFillValueRoundTrip(
@@ -61,7 +68,8 @@ void TestFillValueRoundTrip(
     std::vector<tensorstore::SharedArray<const void>> fill_values,
     std::vector<tensorstore::ArrayMatcher> fill_values_matcher) {
   auto parsed_dtype = ParseDType(dtype);
-  ASSERT_EQ(Status(), GetStatus(parsed_dtype)) << "dtype=" << dtype.dump();
+  ASSERT_EQ(absl::OkStatus(), GetStatus(parsed_dtype))
+      << "dtype=" << dtype.dump();
   EXPECT_THAT(
       ParseFillValue(encoded_fill_value, *parsed_dtype),
       ::testing::Optional(::testing::ElementsAreArray(fill_values_matcher)))
@@ -85,15 +93,24 @@ void TestFillValueRoundTrip(
 
 template <typename FloatType>
 void TestFillValueRoundTripFloat(const ::nlohmann::json& dtype) {
-  TestFillValueRoundTrip(dtype, 3.5, {MakeScalarArray<FloatType>(3.5)});
-  TestFillValueRoundTrip(dtype, "Infinity",
-                         {MakeScalarArray<FloatType>(INFINITY)});
-  TestFillValueRoundTrip(dtype, "-Infinity",
-                         {MakeScalarArray<FloatType>(-INFINITY)});
-  TestFillValueRoundTrip(dtype, "NaN", {MakeScalarArray<FloatType>(NAN)},
-                         {tensorstore::MatchesScalarArray<FloatType>(
-                             ::testing::internal::FloatingEqMatcher<FloatType>(
-                                 NAN, /*nan_eq_nan=*/true))});
+  TestFillValueRoundTrip(
+      dtype, 3.5, {MakeScalarArray<FloatType>(static_cast<FloatType>(3.5))});
+  TestFillValueRoundTrip(
+      dtype, "Infinity",
+      {MakeScalarArray<FloatType>(static_cast<FloatType>(INFINITY))});
+  TestFillValueRoundTrip(
+      dtype, "-Infinity",
+      {MakeScalarArray<FloatType>(static_cast<FloatType>(-INFINITY))});
+  if constexpr (std::is_same_v<FloatType, float> ||
+                std::is_same_v<FloatType, double>) {
+    // `testing::internal::FloatingEqMatcher` only supports the builtin floating
+    // point types.
+    TestFillValueRoundTrip(
+        dtype, "NaN", {MakeScalarArray<FloatType>(static_cast<FloatType>(NAN))},
+        {tensorstore::MatchesScalarArray<FloatType>(
+            ::testing::internal::FloatingEqMatcher<FloatType>(
+                static_cast<FloatType>(NAN), /*nan_eq_nan=*/true))});
+  }
 }
 
 template <typename FloatType>
@@ -106,6 +123,9 @@ void TestFillValueRoundTripComplex(const ::nlohmann::json& dtype) {
 }
 
 TEST(ParseFillValueTest, FloatingPointSuccess) {
+  TestFillValueRoundTripFloat<float16_t>("<f2");
+  TestFillValueRoundTripFloat<float16_t>(">f2");
+  TestFillValueRoundTripFloat<bfloat16_t>("bfloat16");
   TestFillValueRoundTripFloat<float>("<f4");
   TestFillValueRoundTripFloat<float>(">f4");
   TestFillValueRoundTripFloat<double>("<f8");
@@ -206,17 +226,14 @@ TEST(ParseFillValueTest, IntegerFailure) {
       MatchesStatus(absl::StatusCode::kInvalidArgument,
                     "Expected integer in the range \\[0, 4294967295\\], "
                     "but received: \"x\""));
-  EXPECT_THAT(ParseFillValue("x", ParseDType("<i8").value()),
-              MatchesStatus(absl::StatusCode::kInvalidArgument,
-                            "Expected integer in the range "
-                            "\\[-9223372036854775808, 9223372036854775807\\], "
-                            "but received: \"x\""));
+  EXPECT_THAT(
+      ParseFillValue("x", ParseDType("<i8").value()),
+      MatchesStatus(absl::StatusCode::kInvalidArgument,
+                    "Expected 64-bit signed integer, but received: \"x\""));
   EXPECT_THAT(
       ParseFillValue("x", ParseDType("<u8").value()),
-      MatchesStatus(
-          absl::StatusCode::kInvalidArgument,
-          "Expected integer in the range \\[0, 18446744073709551615\\], "
-          "but received: \"x\""));
+      MatchesStatus(absl::StatusCode::kInvalidArgument,
+                    "Expected 64-bit unsigned integer, but received: \"x\""));
 }
 
 TEST(ParseFillValueTest, Base64Success) {
@@ -260,7 +277,7 @@ TEST(ParseFillValueTest, Base64Failure) {
 
 // Corresponds to the zarr test_encode_decode_array_1 test case.
 TEST(EncodeDecodeMetadataTest, Array1) {
-  absl::string_view metadata_text = R"(
+  std::string_view metadata_text = R"(
 {
         "chunks": [10],
         "compressor": {"id": "zlib", "level": 1},
@@ -275,16 +292,14 @@ TEST(EncodeDecodeMetadataTest, Array1) {
   nlohmann::json j = nlohmann::json::parse(metadata_text, nullptr,
                                            /*allow_exceptions=*/false);
   ASSERT_FALSE(j.is_discarded());
-  ZarrMetadata metadata;
-  EXPECT_EQ(Status(), ParseMetadata(j, &metadata));
-  EXPECT_EQ(2, metadata.zarr_format);
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto metadata, ZarrMetadata::FromJson(j));
   EXPECT_THAT(metadata.shape, ElementsAre(100));
   EXPECT_THAT(metadata.chunks, ElementsAre(10));
 
   EXPECT_FALSE(metadata.dtype.has_fields);
   EXPECT_EQ(1, metadata.dtype.fields.size());
-  EXPECT_EQ(DataTypeOf<double>(), metadata.dtype.fields[0].data_type);
-  EXPECT_FALSE(metadata.fill_values[0].valid());
+  EXPECT_EQ(dtype_v<double>, metadata.dtype.fields[0].dtype);
+  EXPECT_FALSE(metadata.fill_value[0].valid());
   EXPECT_EQ(tensorstore::endian::little, metadata.dtype.fields[0].endian);
   EXPECT_THAT(metadata.dtype.fields[0].field_shape, ElementsAre());
   EXPECT_EQ(0, metadata.dtype.fields[0].byte_offset);
@@ -306,7 +321,7 @@ TEST(EncodeDecodeMetadataTest, Array1) {
 // Corresponds to the zarr test_encode_decode_array_2 test case, except that
 // "filters" are `null`.
 TEST(EncodeDecodeMetadataTest, Array2) {
-  absl::string_view metadata_text = R"(
+  std::string_view metadata_text = R"(
 {
         "chunks": [10, 10],
         "compressor": {
@@ -327,16 +342,14 @@ TEST(EncodeDecodeMetadataTest, Array2) {
   nlohmann::json j = nlohmann::json::parse(metadata_text, nullptr,
                                            /*allow_exceptions=*/false);
   ASSERT_FALSE(j.is_discarded());
-  ZarrMetadata metadata;
-  EXPECT_EQ(Status(), ParseMetadata(j, &metadata));
-  EXPECT_EQ(2, metadata.zarr_format);
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto metadata, ZarrMetadata::FromJson(j));
   EXPECT_THAT(metadata.shape, ElementsAre(100, 100));
   EXPECT_THAT(metadata.chunks, ElementsAre(10, 10));
   EXPECT_TRUE(metadata.dtype.has_fields);
   EXPECT_EQ(2, metadata.dtype.fields.size());
-  EXPECT_EQ(DataTypeOf<std::int32_t>(), metadata.dtype.fields[0].data_type);
-  EXPECT_TRUE(metadata.fill_values[0].valid());
-  EXPECT_EQ(metadata.fill_values[0], MakeScalarArray<std::int32_t>(0));
+  EXPECT_EQ(dtype_v<std::int32_t>, metadata.dtype.fields[0].dtype);
+  EXPECT_TRUE(metadata.fill_value[0].valid());
+  EXPECT_EQ(metadata.fill_value[0], MakeScalarArray<std::int32_t>(0));
   EXPECT_EQ(tensorstore::endian::little, metadata.dtype.fields[0].endian);
   EXPECT_THAT(metadata.dtype.fields[0].field_shape, ElementsAre());
   EXPECT_EQ(0, metadata.dtype.fields[0].byte_offset);
@@ -347,9 +360,9 @@ TEST(EncodeDecodeMetadataTest, Array2) {
   EXPECT_EQ(tensorstore::StridedLayoutView<>({10, 10}, {4, 40}),
             metadata.chunk_layout.fields[0].decoded_chunk_layout);
 
-  EXPECT_EQ(DataTypeOf<char>(), metadata.dtype.fields[1].data_type);
-  EXPECT_TRUE(metadata.fill_values[1].valid());
-  EXPECT_EQ(metadata.fill_values[1],
+  EXPECT_EQ(dtype_v<char>, metadata.dtype.fields[1].dtype);
+  EXPECT_TRUE(metadata.fill_value[1].valid());
+  EXPECT_EQ(metadata.fill_value[1],
             tensorstore::MakeArray<char>({0, 0, 0, 0, 0, 0, 0, 0, 0, 0}));
   EXPECT_EQ(tensorstore::endian::native, metadata.dtype.fields[1].endian);
   EXPECT_THAT(metadata.dtype.fields[1].field_shape, ElementsAre(10));
@@ -372,7 +385,7 @@ TEST(EncodeDecodeMetadataTest, Array2) {
 // Corresponds to the zarr test_encode_decode_array_2 test case, except that
 // "filters" are `null` and the "fill_value" is changed to not be all zero.
 TEST(EncodeDecodeMetadataTest, Array2Modified) {
-  absl::string_view metadata_text = R"(
+  std::string_view metadata_text = R"(
 {
         "chunks": [10, 10],
         "compressor": {
@@ -393,16 +406,14 @@ TEST(EncodeDecodeMetadataTest, Array2Modified) {
   nlohmann::json j = nlohmann::json::parse(metadata_text, nullptr,
                                            /*allow_exceptions=*/false);
   ASSERT_FALSE(j.is_discarded());
-  ZarrMetadata metadata;
-  EXPECT_EQ(Status(), ParseMetadata(j, &metadata));
-  EXPECT_EQ(2, metadata.zarr_format);
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto metadata, ZarrMetadata::FromJson(j));
   EXPECT_THAT(metadata.shape, ElementsAre(100, 100));
   EXPECT_THAT(metadata.chunks, ElementsAre(10, 10));
   EXPECT_TRUE(metadata.dtype.has_fields);
   EXPECT_EQ(2, metadata.dtype.fields.size());
-  EXPECT_EQ(DataTypeOf<std::int32_t>(), metadata.dtype.fields[0].data_type);
-  EXPECT_TRUE(metadata.fill_values[0].valid());
-  EXPECT_EQ(metadata.fill_values[0], MakeScalarArray<std::int32_t>(123456789));
+  EXPECT_EQ(dtype_v<std::int32_t>, metadata.dtype.fields[0].dtype);
+  EXPECT_TRUE(metadata.fill_value[0].valid());
+  EXPECT_EQ(metadata.fill_value[0], MakeScalarArray<std::int32_t>(123456789));
   EXPECT_EQ(tensorstore::endian::little, metadata.dtype.fields[0].endian);
   EXPECT_THAT(metadata.dtype.fields[0].field_shape, ElementsAre());
   EXPECT_EQ(0, metadata.dtype.fields[0].byte_offset);
@@ -413,9 +424,9 @@ TEST(EncodeDecodeMetadataTest, Array2Modified) {
   EXPECT_EQ(tensorstore::StridedLayoutView<>({10, 10}, {4, 40}),
             metadata.chunk_layout.fields[0].decoded_chunk_layout);
 
-  EXPECT_EQ(DataTypeOf<char>(), metadata.dtype.fields[1].data_type);
-  EXPECT_TRUE(metadata.fill_values[1].valid());
-  EXPECT_EQ(metadata.fill_values[1],
+  EXPECT_EQ(dtype_v<char>, metadata.dtype.fields[1].dtype);
+  EXPECT_TRUE(metadata.fill_value[1].valid());
+  EXPECT_EQ(metadata.fill_value[1],
             tensorstore::MakeArray<char>(
                 {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j'}));
   EXPECT_EQ(tensorstore::endian::native, metadata.dtype.fields[1].endian);
@@ -438,7 +449,7 @@ TEST(EncodeDecodeMetadataTest, Array2Modified) {
 
 // Corresponds to the zarr test_encode_decode_array_structured test case.
 TEST(EncodeDecodeMetadataTest, ArrayStructured) {
-  absl::string_view metadata_text = R"(
+  std::string_view metadata_text = R"(
 {
         "chunks": [10],
         "compressor": {"id": "zlib", "level": 1},
@@ -453,15 +464,13 @@ TEST(EncodeDecodeMetadataTest, ArrayStructured) {
   nlohmann::json j = nlohmann::json::parse(metadata_text, nullptr,
                                            /*allow_exceptions=*/false);
   ASSERT_FALSE(j.is_discarded());
-  ZarrMetadata metadata;
-  EXPECT_EQ(Status(), ParseMetadata(j, &metadata));
-  EXPECT_EQ(2, metadata.zarr_format);
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto metadata, ZarrMetadata::FromJson(j));
   EXPECT_THAT(metadata.shape, ElementsAre(100));
   EXPECT_THAT(metadata.chunks, ElementsAre(10));
   EXPECT_TRUE(metadata.dtype.has_fields);
   EXPECT_EQ(3, metadata.dtype.fields.size());
-  EXPECT_EQ(DataTypeOf<std::int64_t>(), metadata.dtype.fields[0].data_type);
-  EXPECT_FALSE(metadata.fill_values[0].valid());
+  EXPECT_EQ(dtype_v<std::int64_t>, metadata.dtype.fields[0].dtype);
+  EXPECT_FALSE(metadata.fill_value[0].valid());
   EXPECT_EQ(tensorstore::endian::little, metadata.dtype.fields[0].endian);
   EXPECT_THAT(metadata.dtype.fields[0].field_shape, ElementsAre());
   EXPECT_EQ(0, metadata.dtype.fields[0].byte_offset);
@@ -472,8 +481,8 @@ TEST(EncodeDecodeMetadataTest, ArrayStructured) {
   EXPECT_EQ(tensorstore::StridedLayoutView<>({10}, {8}),
             metadata.chunk_layout.fields[0].decoded_chunk_layout);
 
-  EXPECT_EQ(DataTypeOf<double>(), metadata.dtype.fields[1].data_type);
-  EXPECT_FALSE(metadata.fill_values[1].valid());
+  EXPECT_EQ(dtype_v<double>, metadata.dtype.fields[1].dtype);
+  EXPECT_FALSE(metadata.fill_value[1].valid());
   EXPECT_EQ(tensorstore::endian::little, metadata.dtype.fields[1].endian);
   EXPECT_THAT(metadata.dtype.fields[1].field_shape, ElementsAre(10, 10));
   EXPECT_EQ(8, metadata.dtype.fields[1].byte_offset);
@@ -484,8 +493,8 @@ TEST(EncodeDecodeMetadataTest, ArrayStructured) {
   EXPECT_EQ(tensorstore::StridedLayoutView<>({10, 10, 10}, {800, 80, 8}),
             metadata.chunk_layout.fields[1].decoded_chunk_layout);
 
-  EXPECT_EQ(DataTypeOf<std::uint8_t>(), metadata.dtype.fields[2].data_type);
-  EXPECT_FALSE(metadata.fill_values[2].valid());
+  EXPECT_EQ(dtype_v<std::uint8_t>, metadata.dtype.fields[2].dtype);
+  EXPECT_FALSE(metadata.fill_value[2].valid());
   EXPECT_EQ(tensorstore::endian::native, metadata.dtype.fields[2].endian);
   EXPECT_THAT(metadata.dtype.fields[2].field_shape, ElementsAre(5, 10, 15));
   EXPECT_EQ(808, metadata.dtype.fields[2].byte_offset);
@@ -529,22 +538,20 @@ TEST(EncodeDecodeMetadataTest, FillValuesNan) {
     nlohmann::json j = nlohmann::json::parse(metadata_text, nullptr,
                                              /*allow_exceptions=*/false);
     ASSERT_FALSE(j.is_discarded());
-    ZarrMetadata metadata;
-    EXPECT_EQ(Status(), ParseMetadata(j, &metadata));
-    EXPECT_EQ(2, metadata.zarr_format);
+    TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto metadata, ZarrMetadata::FromJson(j));
     EXPECT_THAT(metadata.shape, ElementsAre(100));
     EXPECT_THAT(metadata.chunks, ElementsAre(10));
 
     EXPECT_FALSE(metadata.dtype.has_fields);
     EXPECT_EQ(1, metadata.dtype.fields.size());
-    EXPECT_EQ(DataTypeOf<double>(), metadata.dtype.fields[0].data_type);
-    EXPECT_TRUE(metadata.fill_values[0].valid());
-    EXPECT_THAT(metadata.fill_values[0].shape(), ElementsAre());
+    EXPECT_EQ(dtype_v<double>, metadata.dtype.fields[0].dtype);
+    EXPECT_TRUE(metadata.fill_value[0].valid());
+    EXPECT_THAT(metadata.fill_value[0].shape(), ElementsAre());
     if (std::isnan(pair.first)) {
       EXPECT_TRUE(std::isnan(
-          *static_cast<const double*>(metadata.fill_values[0].data())));
+          *static_cast<const double*>(metadata.fill_value[0].data())));
     } else {
-      EXPECT_EQ(MakeScalarArray<double>(pair.first), metadata.fill_values[0]);
+      EXPECT_EQ(MakeScalarArray<double>(pair.first), metadata.fill_value[0]);
     }
     EXPECT_EQ(tensorstore::endian::little, metadata.dtype.fields[0].endian);
     EXPECT_THAT(metadata.dtype.fields[0].field_shape, ElementsAre());
@@ -580,21 +587,19 @@ void EncodeDecodeMetadataTestArrayComplex(std::string zarr_dtype) {
                    {"order", "F"},
                    {"shape", {100, 100}},
                    {"zarr_format", 2}};
-  ZarrMetadata metadata;
-  EXPECT_EQ(Status(), ParseMetadata(j, &metadata));
-  EXPECT_EQ(2, metadata.zarr_format);
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto metadata, ZarrMetadata::FromJson(j));
   EXPECT_THAT(metadata.shape, ElementsAre(100, 100));
   EXPECT_THAT(metadata.chunks, ElementsAre(10, 10));
 
   EXPECT_FALSE(metadata.dtype.has_fields);
   EXPECT_EQ(1, metadata.dtype.fields.size());
-  ASSERT_EQ(DataTypeOf<T>(), metadata.dtype.fields[0].data_type);
-  EXPECT_TRUE(metadata.fill_values[0].valid());
-  EXPECT_THAT(metadata.fill_values[0].shape(), ElementsAre());
-  EXPECT_TRUE(std::isnan(
-      static_cast<const T*>(metadata.fill_values[0].data())->real()));
+  ASSERT_EQ(dtype_v<T>, metadata.dtype.fields[0].dtype);
+  EXPECT_TRUE(metadata.fill_value[0].valid());
+  EXPECT_THAT(metadata.fill_value[0].shape(), ElementsAre());
+  EXPECT_TRUE(
+      std::isnan(static_cast<const T*>(metadata.fill_value[0].data())->real()));
   EXPECT_EQ(-1.0f,
-            static_cast<const T*>(metadata.fill_values[0].data())->imag());
+            static_cast<const T*>(metadata.fill_value[0].data())->imag());
   EXPECT_EQ(tensorstore::endian::little, metadata.dtype.fields[0].endian);
   EXPECT_THAT(metadata.dtype.fields[0].field_shape, ElementsAre());
   EXPECT_EQ(0, metadata.dtype.fields[0].byte_offset);
@@ -624,7 +629,7 @@ TEST(EncodeDecodeMetadataTest, ArrayComplex16) {
 }
 
 TEST(ParseMetadataTest, Simple) {
-  absl::string_view metadata_text = R"(
+  std::string_view metadata_text = R"(
 {
     "chunks": [
         100
@@ -643,13 +648,11 @@ TEST(ParseMetadataTest, Simple) {
   nlohmann::json j = nlohmann::json::parse(metadata_text, nullptr,
                                            /*allow_exceptions=*/false);
   ASSERT_FALSE(j.is_discarded());
-  ZarrMetadata metadata;
-  EXPECT_EQ(Status(), ParseMetadata(j, &metadata));
-  EXPECT_EQ(2, metadata.zarr_format);
+  TENSORSTORE_ASSERT_OK_AND_ASSIGN(auto metadata, ZarrMetadata::FromJson(j));
   EXPECT_THAT(metadata.shape, ElementsAre(1111));
   EXPECT_THAT(metadata.chunks, ElementsAre(100));
   EXPECT_EQ(1, metadata.dtype.fields.size());
-  EXPECT_EQ(DataTypeOf<std::int8_t>(), metadata.dtype.fields[0].data_type);
+  EXPECT_EQ(dtype_v<std::int8_t>, metadata.dtype.fields[0].dtype);
   EXPECT_EQ(tensorstore::endian::native, metadata.dtype.fields[0].endian);
   EXPECT_THAT(metadata.dtype.fields[0].field_shape, ElementsAre());
   EXPECT_EQ(0, metadata.dtype.fields[0].byte_offset);
@@ -662,19 +665,52 @@ TEST(ParseMetadataTest, Simple) {
 }
 
 TEST(ParseMetadataTest, InvalidChunks) {
-  ZarrMetadata metadata;
-  // Chunk dimensions must be > 0.
-  EXPECT_THAT(
-      ParseMetadata(::nlohmann::json{{"chunks", {0}},
-                                     {"compressor", nullptr},
-                                     {"dtype", "|i1"},
-                                     {"fill_value", 0},
-                                     {"filters", nullptr},
-                                     {"order", "F"},
-                                     {"shape", {10}},
-                                     {"zarr_format", 2}},
-                    &metadata),
-      MatchesStatus(absl::StatusCode::kInvalidArgument, ".*\"chunks\".*"));
+  tensorstore::TestJsonBinderFromJson<ZarrMetadata>({
+      // Chunk dimensions must be > 0.
+      {{{"chunks", {0}},
+        {"compressor", nullptr},
+        {"dtype", "|i1"},
+        {"fill_value", 0},
+        {"filters", nullptr},
+        {"order", "F"},
+        {"shape", {10}},
+        {"zarr_format", 2}},
+
+       MatchesStatus(absl::StatusCode::kInvalidArgument, ".*\"chunks\".*")},
+  });
+}
+
+TEST(ParseMetadataTest, InvalidRank) {
+  tensorstore::TestJsonBinderFromJson<ZarrMetadata>({
+      {{{"chunks", ::nlohmann::json::array_t(33, 1)},
+        {"compressor", nullptr},
+        {"dtype", "|i1"},
+        {"fill_value", 0},
+        {"filters", nullptr},
+        {"order", "F"},
+        {"shape", ::nlohmann::json::array_t(33, 10)},
+        {"zarr_format", 2}},
+       MatchesStatus(absl::StatusCode::kInvalidArgument,
+                     ".*: Rank 33 is outside valid range \\[0, 32\\]")},
+  });
+}
+
+TEST(DimensionSeparatorTest, JsonBinderTest) {
+  tensorstore::TestJsonBinderRoundTrip<DimensionSeparator>(
+      {
+          {DimensionSeparator::kDotSeparated, "."},
+          {DimensionSeparator::kSlashSeparated, "/"},
+      },
+      DimensionSeparatorJsonBinder);
+}
+
+TEST(DimensionSeparatorTest, JsonBinderTestInvalid) {
+  tensorstore::TestJsonBinderFromJson<DimensionSeparator>(
+      {
+          {"x", MatchesStatus(absl::StatusCode::kInvalidArgument)},
+          {3, MatchesStatus(absl::StatusCode::kInvalidArgument)},
+      },
+      DimensionSeparatorJsonBinder);
 }
 
 }  // namespace
