@@ -12,16 +12,18 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-#include <ostream>
-#include <system_error>  // NOLINT
-
 #include "tensorstore/index_interval.h"
+
+#include <ostream>
+
+#include "absl/status/status.h"
 #include "tensorstore/internal/integer_overflow.h"
+#include "tensorstore/serialization/serialization.h"
 #include "tensorstore/util/division.h"
 #include "tensorstore/util/quote_string.h"
 #include "tensorstore/util/result.h"
 #include "tensorstore/util/status.h"
-#include "tensorstore/util/to_string.h"
+#include "tensorstore/util/str_cat.h"
 
 namespace tensorstore {
 
@@ -30,7 +32,7 @@ Result<IndexInterval> IndexInterval::Closed(Index inclusive_min,
   if (!ValidClosed(inclusive_min, inclusive_max)) {
     return absl::InvalidArgumentError(
         StrCat("(", inclusive_min, ", ", inclusive_max,
-               ") do not specify a valid closed index interval."));
+               ") do not specify a valid closed index interval"));
   }
   return UncheckedClosed(inclusive_min, inclusive_max);
 }
@@ -40,7 +42,7 @@ Result<IndexInterval> IndexInterval::HalfOpen(Index inclusive_min,
   if (!ValidHalfOpen(inclusive_min, exclusive_max)) {
     return absl::InvalidArgumentError(
         StrCat("(", inclusive_min, ", ", exclusive_max,
-               ") do not specify a valid half-open index interval."));
+               ") do not specify a valid half-open index interval"));
   }
   return UncheckedHalfOpen(inclusive_min, exclusive_max);
 }
@@ -49,7 +51,7 @@ Result<IndexInterval> IndexInterval::Sized(Index inclusive_min, Index size) {
   if (!ValidSized(inclusive_min, size)) {
     return absl::InvalidArgumentError(
         StrCat("(", inclusive_min, ", ", size,
-               ") do not specify a valid sized index interval."));
+               ") do not specify a valid sized index interval"));
   }
   return UncheckedSized(inclusive_min, size);
 }
@@ -129,6 +131,14 @@ bool AreCompatibleOrUnbounded(IndexInterval a, IndexInterval b) {
          (a_upper == b_upper || a_upper == kInfIndex || b_upper == kInfIndex);
 }
 
+IndexInterval Hull(IndexInterval a, IndexInterval b) {
+  if (a.empty()) return b;
+  if (b.empty()) return a;
+  const Index lower = std::min(a.inclusive_min(), b.inclusive_min());
+  const Index upper = std::max(a.inclusive_max(), b.inclusive_max());
+  return IndexInterval::UncheckedClosed(lower, upper);
+}
+
 IndexInterval Intersect(IndexInterval a, IndexInterval b) {
   const Index lower = std::max(a.inclusive_min(), b.inclusive_min());
   const Index upper = std::min(a.inclusive_max(), b.inclusive_max());
@@ -136,8 +146,44 @@ IndexInterval Intersect(IndexInterval a, IndexInterval b) {
   return IndexInterval::UncheckedSized(lower, size);
 }
 
+OptionallyImplicitIndexInterval Hull(OptionallyImplicitIndexInterval a,
+                                     OptionallyImplicitIndexInterval b) {
+  IndexInterval interval = Hull(a.interval(), b.interval());
+  bool implicit_lower = (a.inclusive_min() == b.inclusive_min())
+                            ? (a.implicit_lower() && b.implicit_lower())
+                            : (interval.inclusive_min() == a.inclusive_min()
+                                   ? a.implicit_lower()
+                                   : b.implicit_lower());
+  bool implicit_upper = (a.inclusive_max() == b.inclusive_max())
+                            ? (a.implicit_upper() && b.implicit_upper())
+                            : (a.inclusive_max() == interval.inclusive_max()
+                                   ? a.implicit_upper()
+                                   : b.implicit_upper());
+
+  return OptionallyImplicitIndexInterval{interval, implicit_lower,
+                                         implicit_upper};
+}
+
 OptionallyImplicitIndexInterval Intersect(OptionallyImplicitIndexInterval a,
                                           OptionallyImplicitIndexInterval b) {
+  IndexInterval interval = Intersect(a.interval(), b.interval());
+  bool implicit_lower = (a.inclusive_min() == b.inclusive_min())
+                            ? (a.implicit_lower() && b.implicit_lower())
+                            : (interval.inclusive_min() == a.inclusive_min()
+                                   ? a.implicit_lower()
+                                   : b.implicit_lower());
+  bool implicit_upper = (a.inclusive_max() == b.inclusive_max())
+                            ? (a.implicit_upper() && b.implicit_upper())
+                            : (a.inclusive_max() == interval.inclusive_max()
+                                   ? a.implicit_upper()
+                                   : b.implicit_upper());
+
+  return OptionallyImplicitIndexInterval{interval, implicit_lower,
+                                         implicit_upper};
+}
+
+OptionallyImplicitIndexInterval IntersectPreferringExplicit(
+    OptionallyImplicitIndexInterval a, OptionallyImplicitIndexInterval b) {
   const Index inclusive_min =
       a.implicit_lower() == b.implicit_lower()
           ? std::max(a.inclusive_min(), b.inclusive_min())
@@ -148,18 +194,11 @@ OptionallyImplicitIndexInterval Intersect(OptionallyImplicitIndexInterval a,
           ? std::min(a.inclusive_max(), b.inclusive_max())
           : std::min(a.effective_interval().inclusive_max(),
                      b.effective_interval().inclusive_max());
-  return {IndexInterval::UncheckedClosed(
-              inclusive_min, std::max(inclusive_min - 1, inclusive_max)),
-          a.implicit_lower() && b.implicit_lower(),
-          a.implicit_upper() && b.implicit_upper()};
-}
-
-IndexInterval Hull(IndexInterval a, IndexInterval b) {
-  if (a.empty()) return b;
-  if (b.empty()) return a;
-  const Index lower = std::min(a.inclusive_min(), b.inclusive_min());
-  const Index upper = std::max(a.inclusive_max(), b.inclusive_max());
-  return IndexInterval::UncheckedClosed(lower, upper);
+  return OptionallyImplicitIndexInterval{
+      IndexInterval::UncheckedClosed(
+          inclusive_min, std::max(inclusive_min - 1, inclusive_max)),
+      a.implicit_lower() && b.implicit_lower(),
+      a.implicit_upper() && b.implicit_upper()};
 }
 
 bool ContainsOrUnbounded(IndexInterval outer, IndexInterval inner) {
@@ -169,60 +208,78 @@ bool ContainsOrUnbounded(IndexInterval outer, IndexInterval inner) {
           inner.inclusive_max() <= outer.inclusive_max());
 }
 
-Result<IndexInterval> ShiftInterval(IndexInterval interval, Index offset) {
-  if (!IsFiniteIndex(offset)) {
-    return absl::OutOfRangeError(StrCat("Index offset ", offset,
-                                        " is outside valid range ",
-                                        IndexInterval::FiniteRange(), "."));
-  }
+Result<IndexInterval> ShiftInterval(IndexInterval interval, Index min_offset,
+                                    Index max_offset) {
   Index inclusive_min;
   if (interval.inclusive_min() == -kInfIndex) {
     inclusive_min = -kInfIndex;
-  } else {
-    inclusive_min = interval.inclusive_min() + offset;
-    if (!IsFiniteIndex(inclusive_min)) {
-      return absl::InvalidArgumentError(StrCat(
-          "Shifted inclusive_min value ", inclusive_min,
-          " is outside valid range ", IndexInterval::FiniteRange(), "."));
-    }
+  } else if (internal::AddOverflow(interval.inclusive_min(), min_offset,
+                                   &inclusive_min) ||
+             !IsFiniteIndex(inclusive_min)) {
+    return absl::InvalidArgumentError(
+        StrCat(interval.inclusive_min(), " + ", min_offset,
+               " is outside valid range ", IndexInterval::FiniteRange()));
   }
   Index inclusive_max;
   if (interval.inclusive_max() == kInfIndex) {
     inclusive_max = kInfIndex;
-  } else {
-    inclusive_max = interval.inclusive_max() + offset;
-    if (!IsFiniteIndex(inclusive_max)) {
-      return absl::InvalidArgumentError(StrCat(
-          "Shifted inclusive_max value ", inclusive_max,
-          " is outside valid range ", IndexInterval::FiniteRange(), "."));
-    }
+  } else if (internal::AddOverflow(interval.inclusive_max(), max_offset,
+                                   &inclusive_max) ||
+             !IsFiniteIndex(inclusive_max)) {
+    return absl::InvalidArgumentError(
+        StrCat(interval.inclusive_max(), " + ", max_offset,
+               " is outside valid range ", IndexInterval::FiniteRange()));
   }
   return IndexInterval::UncheckedClosed(inclusive_min, inclusive_max);
 }
 
+Result<IndexInterval> ShiftIntervalBackward(IndexInterval interval,
+                                            Index min_offset,
+                                            Index max_offset) {
+  // If `min_offset` or `max_offset` equals `std::numeric_limits<Index>::min()`,
+  // then wrapping negation is a no-op (normal `operator-` in that case is
+  // undefined behavior due to signed integer overflow).  However, it turns out
+  // `ShiftInterval` still gives the correct result in the case that the offset
+  // is incorrectly wrapped, because the offset is either ignored due to the
+  // corresponding bound being infinite, or it leads to overflow.  In the case
+  // of overflow, the error message will include the incorrectly-wrapped value,
+  // but that is not terribly important.
+  return ShiftInterval(
+      interval, internal::wrap_on_overflow::Multiply(min_offset, Index(-1)),
+      internal::wrap_on_overflow::Multiply(max_offset, Index(-1)));
+}
+
+Result<IndexInterval> ShiftInterval(IndexInterval interval, Index offset) {
+  return ShiftInterval(interval, offset, offset);
+}
+
+Result<IndexInterval> ShiftIntervalBackward(IndexInterval interval,
+                                            Index offset) {
+  return ShiftIntervalBackward(interval, offset, offset);
+}
+
 Result<IndexInterval> ShiftIntervalTo(IndexInterval interval, Index origin) {
   if (!IsFiniteIndex(origin)) {
     return absl::OutOfRangeError(StrCat("Origin ", origin,
                                         " is outside valid range ",
-                                        IndexInterval::FiniteRange(), "."));
+                                        IndexInterval::FiniteRange()));
   }
   if (interval.inclusive_min() == -kInfIndex) {
     return absl::InvalidArgumentError(
-        StrCat("Interval ", interval, " is not bounded below."));
+        StrCat("Interval ", interval, " is not bounded below"));
   }
   // Guaranteed not to overflow because `IsFiniteIndex(origin) == true`.
   Index offset;
-  const bool overflow =
+  [[maybe_unused]] const bool overflow =
       internal::SubOverflow(origin, interval.inclusive_min(), &offset);
-  ABSL_ASSERT(!overflow);
+  assert(!overflow);
   return ShiftInterval(interval, offset);
 }
 
-Status CheckContains(IndexInterval interval, Index index) {
+absl::Status CheckContains(IndexInterval interval, Index index) {
   if (Contains(interval, index)) return absl::OkStatus();
-  return Status(
-      absl::StatusCode::kOutOfRange,
-      StrCat("Index ", index, " is outside valid range ", interval, "."));
+  return absl::OutOfRangeError(
+      StrCat("Index ", index, " is outside valid range ", interval));
 }
 
 Result<std::pair<OptionallyImplicitIndexInterval, Index>> ExtractStridedSlice(
@@ -235,14 +292,13 @@ Result<std::pair<OptionallyImplicitIndexInterval, Index>> ExtractStridedSlice(
 
   // Check for 0 and std::numeric_limits<Index>::min(), which are both invalid.
   if (stride == 0 || stride == std::numeric_limits<Index>::min()) {
-    return absl::InvalidArgumentError(StrCat("Invalid stride ", stride, "."));
+    return absl::InvalidArgumentError(StrCat("Invalid stride ", stride));
   }
   if (start == kImplicit) {
     start = stride > 0 ? orig.inclusive_min() : orig.inclusive_max();
   } else {
     if (!IsValidIndex(start)) {
-      return absl::InvalidArgumentError(
-          StrCat("Invalid start index ", start, "."));
+      return absl::InvalidArgumentError(StrCat("Invalid start index ", start));
     }
     orig.implicit_lower() = false;
   }
@@ -253,9 +309,8 @@ Result<std::pair<OptionallyImplicitIndexInterval, Index>> ExtractStridedSlice(
       inclusive_stop = stride > 0 ? orig.inclusive_max() : orig.inclusive_min();
     } else {
       if (size < 0) {
-        return Status(
-            absl::StatusCode::kInvalidArgument,
-            StrCat("Negative size ", size, " specified for sized interval."));
+        return absl::InvalidArgumentError(
+            StrCat("Negative size ", size, " specified for sized interval"));
       }
       orig.implicit_upper() = false;
       if (size == 0) {
@@ -266,7 +321,7 @@ Result<std::pair<OptionallyImplicitIndexInterval, Index>> ExtractStridedSlice(
         if (internal::MulOverflow(stride, size - 1, &inclusive_stop) ||
             internal::AddOverflow(start, inclusive_stop, &inclusive_stop)) {
           return absl::OutOfRangeError(
-              StrCat("Integer overflow computing slice result."));
+              StrCat("Integer overflow computing slice result"));
         }
       }
     }
@@ -279,16 +334,16 @@ Result<std::pair<OptionallyImplicitIndexInterval, Index>> ExtractStridedSlice(
           !IsFiniteIndex(stop_or_size)) {
         inclusive_stop = stop_or_size;
       } else {
-        ABSL_ASSERT(interval_form == IntervalForm::half_open);
+        assert(interval_form == IntervalForm::half_open);
         // Can't overflow since `IsFiniteIndex(stop_or_size) == true`.
         inclusive_stop = stop_or_size + (stride > 0 ? -1 : 1);
       }
     }
   }
   if (std::abs(stride) != 1 && !IsFiniteIndex(start)) {
-    return absl::InvalidArgumentError(
-        StrCat("Slicing with non-unit stride of ", stride,
-               " requires a finite start index."));
+    return absl::InvalidArgumentError(StrCat("Slicing with non-unit stride of ",
+                                             stride,
+                                             " requires a finite start index"));
   }
   Index adjusted_inclusive_min, adjusted_inclusive_max;
   if (stride > 0) {
@@ -305,7 +360,7 @@ Result<std::pair<OptionallyImplicitIndexInterval, Index>> ExtractStridedSlice(
   if (!Contains(constraint, adjusted_interval)) {
     return absl::OutOfRangeError(StrCat("Slice interval ", adjusted_interval,
                                         " is not contained within domain ",
-                                        constraint, "."));
+                                        constraint));
   }
 
   Index new_start = start / stride;
@@ -314,7 +369,7 @@ Result<std::pair<OptionallyImplicitIndexInterval, Index>> ExtractStridedSlice(
           ? kInfIndex + 1 - new_start
           : CeilOfRatio(adjusted_interval.size(), std::abs(stride));
   orig.interval() = IndexInterval::UncheckedSized(new_start, new_size);
-  return {absl::in_place, orig, start};
+  return {std::in_place, orig, start};
 }
 
 Result<std::pair<OptionallyImplicitIndexInterval, Index>>
@@ -336,12 +391,12 @@ ExtractSizedStridedSlice(OptionallyImplicitIndexInterval orig, Index start,
   return ExtractStridedSlice(orig, IntervalForm::sized, start, size, stride);
 }
 
-Status ComputeStridedSliceMap(OptionallyImplicitIndexInterval orig,
-                              IntervalForm interval_form,
-                              Index translate_origin_to, Index start,
-                              Index stop_or_size, Index stride,
-                              OptionallyImplicitIndexInterval* new_domain,
-                              Index* output_offset) {
+absl::Status ComputeStridedSliceMap(OptionallyImplicitIndexInterval orig,
+                                    IntervalForm interval_form,
+                                    Index translate_origin_to, Index start,
+                                    Index stop_or_size, Index stride,
+                                    OptionallyImplicitIndexInterval* new_domain,
+                                    Index* output_offset) {
   TENSORSTORE_ASSIGN_OR_RETURN(
       auto new_interval_and_adjusted_start,
       ExtractStridedSlice(orig, interval_form, start, stop_or_size, stride));
@@ -427,6 +482,15 @@ Result<OptionallyImplicitIndexInterval> GetAffineTransformDomain(
   return interval;
 }
 
+namespace {
+absl::Status GetAffineTransformError(IndexInterval interval, Index offset,
+                                     Index multiplier) {
+  return absl::InvalidArgumentError(tensorstore::StrCat(
+      "Integer overflow computing affine transform of domain ", interval,
+      " with offset ", offset, " and multiplier ", multiplier));
+}
+}  // namespace
+
 Result<IndexInterval> GetAffineTransformRange(IndexInterval interval,
                                               Index offset, Index multiplier) {
   const auto transform_bound_overflow = [&](Index* bound) {
@@ -441,11 +505,7 @@ Result<IndexInterval> GetAffineTransformRange(IndexInterval interval,
 
   Index lower = interval.inclusive_min(), upper = interval.inclusive_max();
   if (transform_bound_overflow(&lower) || transform_bound_overflow(&upper)) {
-    return Status(
-        absl::StatusCode::kInvalidArgument,
-        StrCat("Integer overflow computing affine transform of domain ",
-               interval, " with offset ", offset, " and multiplier ",
-               multiplier));
+    return GetAffineTransformError(interval, offset, multiplier);
   }
   if (interval.empty()) {
     return IndexInterval::UncheckedSized(lower, 0);
@@ -457,6 +517,35 @@ Result<IndexInterval> GetAffineTransformRange(IndexInterval interval,
   return IndexInterval::UncheckedClosed(lower, upper);
 }
 
+Result<IndexInterval> GetAffineTransformInverseDomain(IndexInterval interval,
+                                                      Index offset,
+                                                      Index divisor) {
+  TENSORSTORE_ASSIGN_OR_RETURN(
+      auto new_interval, GetAffineTransformRange(interval, offset, divisor));
+  if (new_interval.empty()) return new_interval;
+  if (divisor > 0 && new_interval.inclusive_max() != kInfIndex) {
+    Index new_inclusive_max;
+    if (internal::AddOverflow(new_interval.inclusive_max(), divisor - 1,
+                              &new_inclusive_max) ||
+        !IsFiniteIndex(new_inclusive_max)) {
+      return GetAffineTransformError(interval, offset, divisor);
+    }
+    return IndexInterval::UncheckedClosed(new_interval.inclusive_min(),
+                                          new_inclusive_max);
+  }
+  if (divisor < 0 && new_interval.inclusive_min() != -kInfIndex) {
+    Index new_inclusive_min;
+    if (internal::AddOverflow(new_interval.inclusive_min(), divisor + 1,
+                              &new_inclusive_min) ||
+        !IsFiniteIndex(new_inclusive_min)) {
+      return GetAffineTransformError(interval, offset, divisor);
+    }
+    return IndexInterval::UncheckedClosed(new_inclusive_min,
+                                          new_interval.inclusive_max());
+  }
+  return new_interval;
+}
+
 Result<OptionallyImplicitIndexInterval> GetAffineTransformRange(
     OptionallyImplicitIndexInterval interval, Index offset, Index multiplier) {
   TENSORSTORE_ASSIGN_OR_RETURN(
@@ -468,4 +557,67 @@ Result<OptionallyImplicitIndexInterval> GetAffineTransformRange(
   return interval;
 }
 
+Result<std::string_view> MergeDimensionLabels(std::string_view a,
+                                              std::string_view b) {
+  if (a.empty()) return b;
+  if (b.empty()) return a;
+  if (a == b) return a;
+  return absl::InvalidArgumentError("Dimension labels do not match");
+}
+
+Result<OptionallyImplicitIndexInterval> MergeOptionallyImplicitIndexIntervals(
+    OptionallyImplicitIndexInterval a, OptionallyImplicitIndexInterval b) {
+  if (a == b) return a;
+  Index inclusive_min, inclusive_max;
+  bool implicit_lower, implicit_upper;
+  if (a.inclusive_min() == -kInfIndex && a.implicit_lower() == true) {
+    inclusive_min = b.inclusive_min();
+    implicit_lower = b.implicit_lower();
+  } else if (b.inclusive_min() == -kInfIndex && b.implicit_lower() == true) {
+    inclusive_min = a.inclusive_min();
+    implicit_lower = a.implicit_lower();
+  } else if (a.inclusive_min() != b.inclusive_min()) {
+    return absl::InvalidArgumentError("Lower bounds do not match");
+  } else {
+    inclusive_min = a.inclusive_min();
+    implicit_lower = a.implicit_lower() && b.implicit_lower();
+  }
+  if (a.inclusive_max() == kInfIndex && a.implicit_upper() == true) {
+    inclusive_max = b.inclusive_max();
+    implicit_upper = b.implicit_upper();
+  } else if (b.inclusive_max() == kInfIndex && b.implicit_upper() == true) {
+    inclusive_max = a.inclusive_max();
+    implicit_upper = a.implicit_upper();
+  } else if (a.inclusive_max() != b.inclusive_max()) {
+    return absl::InvalidArgumentError("Upper bounds do not match");
+  } else {
+    inclusive_max = a.inclusive_max();
+    implicit_upper = a.implicit_upper() && b.implicit_upper();
+  }
+  TENSORSTORE_ASSIGN_OR_RETURN(
+      auto interval, IndexInterval::Closed(inclusive_min, inclusive_max));
+  return OptionallyImplicitIndexInterval{interval, implicit_lower,
+                                         implicit_upper};
+}
+
+namespace serialization {
+
+bool Serializer<IndexInterval>::Encode(EncodeSink& sink,
+                                       const IndexInterval& value) {
+  return serialization::EncodeTuple(sink, value.inclusive_min(), value.size());
+}
+
+bool Serializer<IndexInterval>::Decode(DecodeSource& source,
+                                       IndexInterval& value) {
+  Index inclusive_min, size;
+  if (!serialization::DecodeTuple(source, inclusive_min, size)) {
+    return false;
+  }
+  TENSORSTORE_ASSIGN_OR_RETURN(value, IndexInterval::Sized(inclusive_min, size),
+                               (source.Fail(_), false));
+  return true;
+}
+
+}  // namespace serialization
+
 }  // namespace tensorstore
