@@ -25,26 +25,24 @@
 
 namespace {
 
-using tensorstore::AllDims;
-using tensorstore::DimensionIndex;
-using tensorstore::Dims;
-using tensorstore::Index;
-using tensorstore::IndexInterval;
-using tensorstore::IndexTransformBuilder;
-using tensorstore::MakeArray;
-using tensorstore::SharedArrayView;
-using tensorstore::span;
-using tensorstore::internal_index_space::EquivalentIndices;
-using tensorstore::internal_index_space::TestDimExpression;
+using ::tensorstore::AllDims;
+using ::tensorstore::DimensionIndex;
+using ::tensorstore::Dims;
+using ::tensorstore::Index;
+using ::tensorstore::IndexInterval;
+using ::tensorstore::IndexTransformBuilder;
+using ::tensorstore::MakeArray;
+using ::tensorstore::SharedArrayView;
+using ::tensorstore::span;
+using ::tensorstore::internal_index_space::EquivalentIndices;
+using ::tensorstore::internal_index_space::TestDimExpression;
 
 TEST(IndexArraySliceTest, Example) {
   const auto original_transform = IndexTransformBuilder<3, 3>()
                                       .input_origin({0, 2, 0})
                                       .input_shape({7, 4, 10})
                                       .input_labels({"x", "y", "z"})
-                                      .output_single_input_dimension(0, 0, 1, 0)
-                                      .output_single_input_dimension(1, 0, 1, 1)
-                                      .output_single_input_dimension(2, 0, 1, 2)
+                                      .output_identity_transform()
                                       .Finalize()
                                       .value();
   const auto expected_new_transform =
@@ -55,7 +53,7 @@ TEST(IndexArraySliceTest, Example) {
           .output_index_array(
               0, 0, 1, MakeArray<Index>({{{1}, {2}, {3}}, {{4}, {5}, {6}}}),
               IndexInterval::Sized(0, 7))
-          .output_single_input_dimension(1, 0, 1, 2)
+          .output_single_input_dimension(1, 2)
           .output_index_array(
               2, 0, 1, MakeArray<Index>({{{7}, {8}, {9}}, {{0}, {1}, {2}}}),
               IndexInterval::Sized(0, 10))
@@ -96,9 +94,7 @@ TEST(IndexVectorArraySliceTest, Example) {
                                       .input_origin({0, 2, 0})
                                       .input_shape({7, 4, 10})
                                       .input_labels({"x", "y", "z"})
-                                      .output_single_input_dimension(0, 0, 1, 0)
-                                      .output_single_input_dimension(1, 0, 1, 1)
-                                      .output_single_input_dimension(2, 0, 1, 2)
+                                      .output_identity_transform()
                                       .Finalize()
                                       .value();
   const auto expected_new_transform =
@@ -109,7 +105,7 @@ TEST(IndexVectorArraySliceTest, Example) {
           .output_index_array(
               0, 0, 1, MakeArray<Index>({{{1}, {2}, {3}}, {{4}, {5}, {6}}}),
               IndexInterval::Sized(0, 7))
-          .output_single_input_dimension(1, 0, 1, 2)
+          .output_single_input_dimension(1, 2)
           .output_index_array(
               2, 0, 1, MakeArray<Index>({{{7}, {8}, {9}}, {{0}, {1}, {2}}}),
               IndexInterval::Sized(0, 10))
@@ -152,9 +148,7 @@ TEST(IndexArrayOuterIndexArraySliceTest, Example) {
                                       .input_origin({4, 2, 0})
                                       .input_shape({5, 4, 10})
                                       .input_labels({"x", "y", "z"})
-                                      .output_single_input_dimension(0, 0, 1, 0)
-                                      .output_single_input_dimension(1, 0, 1, 1)
-                                      .output_single_input_dimension(2, 0, 1, 2)
+                                      .output_identity_transform()
                                       .Finalize()
                                       .value();
   const auto expected_new_transform =
@@ -164,7 +158,7 @@ TEST(IndexArrayOuterIndexArraySliceTest, Example) {
           .input_labels({"", "y", "", ""})
           .output_index_array(0, 0, 1, MakeArray<Index>({{{{6}}}, {{{7}}}}),
                               IndexInterval::Sized(4, 5))
-          .output_single_input_dimension(1, 0, 1, 1)
+          .output_single_input_dimension(1, 1)
           .output_index_array(2, 0, 1, MakeArray<Index>({{{{2, 3}, {4, 5}}}}),
                               IndexInterval::Sized(0, 10))
           .Finalize()
@@ -214,7 +208,7 @@ TEST(IndexArraySliceTest, OneDOutputOneDArray) {
           .input_shape({2, 200})
           .output_index_array(0, 0, 1, MakeArray<Index>({{1}, {2}}),
                               IndexInterval::Sized(-10, 20))
-          .output_single_input_dimension(1, 0, 1, 1)
+          .output_single_input_dimension(1, 1)
           .Finalize()
           .value(),
       /*expected_new_transform=*/
@@ -279,7 +273,7 @@ TEST(IndexArraySliceTest, OneElementIndexArray) {
           .input_origin({0, 0, -100})
           .input_shape({1, 1, 200})
           .output_constant(0, 5)
-          .output_single_input_dimension(1, 0, 1, 2)
+          .output_single_input_dimension(1, 2)
           .Finalize()
           .value(),
       /*expected_new_transform=*/
@@ -317,7 +311,7 @@ TEST(IndexArraySliceTest, OneDOutputOneDArrayLabeled) {
                         .output_index_array(0, 0, 1,
                                             MakeArray<Index>({{1}, {2}}),
                                             IndexInterval::Sized(-10, 20))
-                        .output_single_input_dimension(1, 0, 1, 1)
+                        .output_single_input_dimension(1, 1)
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -422,21 +416,30 @@ TEST(IndexArraySliceTest, ErrorHandling) {
           .IndexArraySlice(MakeArray<Index>({1, 2}), MakeArray<Index>({3, 4})),
       absl::StatusCode::kInvalidArgument,
       "Number of selected dimensions \\(1\\) does not equal number of index "
-      "arrays \\(2\\)\\.");
+      "arrays \\(2\\)");
 
   TestDimExpressionError(
       IndexTransformBuilder<1, 0>().Finalize().value(),
       Dims(span<const DimensionIndex>())
           .IndexArraySlice(span<const SharedArrayView<const Index>>()),
       absl::StatusCode::kInvalidArgument,
-      "At least one index array must be specified\\.");
+      "At least one index array must be specified");
   TestDimExpressionError(
       IndexTransformBuilder<2, 0>().Finalize().value(),
       Dims(0, 1).IndexArraySlice(MakeArray<Index>({1, 2}),
                                  MakeArray<Index>({3, 4, 5})),
       absl::StatusCode::kInvalidArgument,
-      "Index arrays cannot be broadcast to a common shape: \\{2\\} vs "
-      "\\{3\\}\\.");
+      "Index arrays with shapes \\{2\\}, \\{3\\} cannot be broadcast "
+      "to a common shape");
+}
+
+TEST(IndexArraySliceTest, InvalidRank) {
+  auto index_array = tensorstore::AllocateArray<Index>(
+      std::vector<Index>(32, 1), tensorstore::c_order, tensorstore::value_init);
+  TestDimExpressionError(tensorstore::IdentityTransform(2),
+                         Dims(0).IndexArraySlice(index_array),
+                         absl::StatusCode::kInvalidArgument,
+                         "Rank 33 is outside valid range \\[0, 32\\]");
 }
 
 TEST(IndexVectorArraySliceTest, OneDOutputOneDArray) {
@@ -459,7 +462,7 @@ TEST(IndexVectorArraySliceTest, OneDOutputOneDArray) {
                         .output_index_array(0, 0, 1,
                                             MakeArray<Index>({{1}, {2}}),
                                             IndexInterval::Sized(-10, 20))
-                        .output_single_input_dimension(1, 0, 1, 1)
+                        .output_single_input_dimension(1, 1)
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -493,7 +496,7 @@ TEST(IndexVectorArraySliceTest, OneElementIndexArray) {
           .input_origin({0, -100})
           .input_shape({1, 200})
           .output_constant(0, 1)
-          .output_single_input_dimension(1, 0, 1, 1)
+          .output_single_input_dimension(1, 1)
           .Finalize()
           .value(),
       /*expected_new_transform=*/
@@ -530,7 +533,7 @@ TEST(IndexVectorArraySliceTest, OneDOutputOneDArrayLabeled) {
           .input_labels({"index", "y"})
           .output_index_array(0, 0, 1, MakeArray<Index>({{1}, {2}}),
                               IndexInterval::Sized(-10, 20))
-          .output_single_input_dimension(1, 0, 1, 1)
+          .output_single_input_dimension(1, 1)
           .Finalize()
           .value(),
       /*expected_new_transform=*/
@@ -594,13 +597,28 @@ TEST(IndexVectorArraySliceTest, ErrorHandling) {
       Dims(0).IndexVectorArraySlice(MakeArray<Index>({1, 2}), 0),
       absl::StatusCode::kInvalidArgument,
       "Number of selected dimensions \\(1\\) does not equal index vector "
-      "length \\(2\\)\\.");
+      "length \\(2\\)");
 
   TestDimExpressionError(
       IndexTransformBuilder<2, 0>().Finalize().value(),
       Dims(0).IndexVectorArraySlice(MakeArray<Index>({1, 2}), 1),
       absl::StatusCode::kInvalidArgument,
-      "Dimension index 1 is outside valid range \\[-1, 1\\)\\.");
+      "Dimension index 1 is outside valid range \\[-1, 1\\)");
+}
+
+TEST(IndexVectorArraySliceTest, InvalidRank) {
+  TestDimExpressionError(
+      tensorstore::IdentityTransform(4),
+      Dims(0, 1).IndexVectorArraySlice(
+          tensorstore::AllocateArray<Index>({1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  //
+                                             1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  //
+                                             1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  //
+                                             1, 2},
+                                            tensorstore::c_order,
+                                            tensorstore::default_init),
+          -1),
+      absl::StatusCode::kInvalidArgument,
+      "Rank 33 is outside valid range \\[0, 32\\]");
 }
 
 TEST(OuterIndexArraySliceTest, Integration) {
@@ -627,7 +645,7 @@ TEST(OuterIndexArraySliceTest, Integration) {
                             MakeArray<Index>({{{{3}}, {{4}}, {{5}}},
                                               {{{8}}, {{9}}, {{10}}}}),
                             IndexInterval::Sized(-10, 21))
-                        .output_single_input_dimension(1, 0, 1, 2)
+                        .output_single_input_dimension(1, 2)
                         .output_index_array(2, 0, 1,
                                             MakeArray<Index>({{{{1, 2}}}}),
                                             IndexInterval::Sized(-2, 15))
@@ -680,7 +698,7 @@ TEST(OuterIndexArraySliceTest, OneElementIndexArray) {
                             MakeArray<Index>({{{{3}}, {{4}}, {{5}}},
                                               {{{8}}, {{9}}, {{10}}}}),
                             IndexInterval::Sized(-10, 21))
-                        .output_single_input_dimension(1, 0, 1, 2)
+                        .output_single_input_dimension(1, 2)
                         .output_constant(2, 1)
                         .Finalize()
                         .value(),
@@ -709,7 +727,17 @@ TEST(OuterIndexArraySliceTest, ErrorHandling) {
                                 MakeArray<Index>({3, 4})),
       absl::StatusCode::kInvalidArgument,
       "Number of selected dimensions \\(1\\) does not equal number of index "
-      "arrays \\(2\\)\\.");
+      "arrays \\(2\\)");
+}
+
+TEST(OuterIndexArraySliceTest, InvalidRank) {
+  auto index_array = tensorstore::AllocateArray<Index>(
+      std::vector<Index>(17, 1), tensorstore::c_order, tensorstore::value_init);
+  TestDimExpressionError(
+      tensorstore::IdentityTransform(2),
+      Dims(0, 1).OuterIndexArraySlice(index_array, index_array),
+      absl::StatusCode::kInvalidArgument,
+      "Rank 34 is outside valid range \\[0, 32\\]");
 }
 
 }  // namespace
