@@ -18,7 +18,7 @@
 #include <type_traits>
 
 #include "absl/container/inlined_vector.h"
-#include "absl/meta/type_traits.h"
+#include "absl/status/status.h"
 #include "tensorstore/index_space/index_transform.h"
 #include "tensorstore/index_space/internal/deep_copy_transform_rep_ptr.h"
 #include "tensorstore/index_space/internal/transform_rep.h"
@@ -29,20 +29,35 @@ namespace tensorstore {
 
 namespace internal_index_space {
 
-using BuilderFlags = unsigned int;
+enum class BuilderFlags : unsigned int {
+  kDefault = 0,
 
-/// Indicates that `input_origin` was called.
-constexpr BuilderFlags kSetLower = 1;
+  /// Indicates that `input_origin` was called.
+  kSetLower = 1,
 
-/// Indicates that `implicit_lower_bounds` was called.
-constexpr BuilderFlags kSetImplicitLower = 2;
+  /// Indicates that `implicit_lower_bounds` was called.
+  kSetImplicitLower = 2,
 
-/// Indicates that `input_shape`, `input_inclusive_max` or `input_exclusive_max`
-/// was called.
-constexpr BuilderFlags kSetUpper = 4;
+  /// Indicates that `input_shape`, `input_inclusive_max` or
+  /// `input_exclusive_max` was called.
+  kSetUpper = 4,
 
-/// Indicates that `implicit_upper_bounds` was called.
-constexpr BuilderFlags kSetImplicitUpper = 8;
+  /// Indicates that `implicit_upper_bounds` was called.
+  kSetImplicitUpper = 8
+};
+
+inline BuilderFlags operator|(BuilderFlags a, BuilderFlags b) {
+  return static_cast<BuilderFlags>(static_cast<unsigned int>(a) |
+                                   static_cast<unsigned int>(b));
+}
+inline BuilderFlags& operator|=(BuilderFlags& a, BuilderFlags b) {
+  return a = a | b;
+}
+
+inline BuilderFlags operator&(BuilderFlags a, BuilderFlags b) {
+  return static_cast<BuilderFlags>(static_cast<unsigned int>(a) &
+                                   static_cast<unsigned int>(b));
+}
 
 /// Specifies how to initialize an OutputIndexMap (without the offset or stride
 /// values).
@@ -67,7 +82,7 @@ class OutputIndexMapInitializer {
                                                   offset_origin>& index_array,
                             Result<IndexInterval> bounds)
       : index_array(index_array), index_array_bounds(bounds) {}
-  absl::optional<DimensionIndex> input_dimension;
+  std::optional<DimensionIndex> input_dimension;
   SharedArray<const Index, dynamic_rank, offset_origin> index_array;
   /// We store a `Result<IndexInterval>` rather than just a plain
   /// `IndexInterval` in order to allow an expression like
@@ -80,21 +95,17 @@ class OutputIndexMapInitializer {
 template <typename Range, typename Element>
 void AssignRange(const Range& range, span<Element> dest);
 
-template <typename Range>
-void AssignRange(const Range& range, BitSpan<std::uint64_t> dest);
-
 /// Bool-valued metafunction that evaluates to `true` if `Range` is not
 /// `span`-compatible, or `Range` is `span`-compatible with a static
 /// extent compatible with `StaticExtent`.
 template <std::ptrdiff_t StaticExtent, typename Range, typename = void>
-struct IsStaticExtentCompatibleWithRange : public std::true_type {};
+constexpr inline bool IsStaticExtentCompatibleWithRange = true;
 
 template <std::ptrdiff_t StaticExtent, typename Range>
-struct IsStaticExtentCompatibleWithRange<
-    StaticExtent, Range, absl::void_t<internal::ConstSpanType<Range>>>
-    : public std::integral_constant<
-          bool, IsRankExplicitlyConvertible(
-                    StaticExtent, internal::ConstSpanType<Range>::extent)> {};
+constexpr inline bool IsStaticExtentCompatibleWithRange<
+    StaticExtent, Range, std::void_t<internal::ConstSpanType<Range>>> =
+    RankConstraint::EqualOrUnspecified(StaticExtent,
+                                       internal::ConstSpanType<Range>::extent);
 
 }  // namespace internal_index_space
 
@@ -111,9 +122,10 @@ struct IsStaticExtentCompatibleWithRange<
 /// To use this class, call (a subset of) the setter methods `input_origin`,
 /// `input_shape`, `input_exclusive_max`, `input_inclusive_max`
 /// `implicit_lower_bounds`, `implicit_upper_bounds`, `input_labels`,
-/// `output_constant`, `output_single_input_dimension`, `output_index_array` to
-/// specify the input space and the output index maps.  For convenience, all of
-/// these setter methods return a reference to `*this` and can be chained.
+/// `output_map`, `output_constant`, `output_single_input_dimension`,
+/// `output_index_array` to specify the input space and the output index maps.
+/// For convenience, all of these setter methods return a reference to `*this`
+/// and can be chained.
 ///
 /// The `input_origin`, `input_shape`, `input_inclusive_max`,
 /// `input_exclusive_max`, and `input_labels` methods can be called with vectors
@@ -124,7 +136,7 @@ struct IsStaticExtentCompatibleWithRange<
 /// be called at most once, and only one of `input_shape`,
 /// `input_inclusive_max`, and `input_exclusive_max` should be called.
 ///
-/// The `output_constant`, `output_single_input_dimension`, and
+/// The `output_map`, `output_constant`, `output_single_input_dimension`, and
 /// `output_index_array` methods specify the output index map for just the
 /// single output dimension specified by the first argument, `output_dim` (and
 /// override the output index map specified by any prior call to a `output_*`
@@ -171,59 +183,73 @@ struct IsStaticExtentCompatibleWithRange<
 /// at most once, and either returns a valid index transform, or an error if an
 /// invalid input domain or output index map was specified.
 ///
-/// Examples:
+/// .. example:: Examples
 ///
-///     IndexTransform<3, 2> t = IndexTransformBuilder<3, 2>()
-///         .input_origin({1, 2, 3})
-///         .input_inclusive_max({5, 7, 9})
-///         .output_single_input_dimension(0, 1, 2, 0)
-///         .output_constant(1, 5)
-///         .Finalize()
-///         .value();
+///     .. code-block:: cpp
 ///
-/// Creates an index transform from an rank-3 input domain
-/// `[1, 5], [2, 7], [3, 9]` to a rank 2 output space, where
-/// `output[0] = 1 + 2 * input[0]` and `output[1] = 5`.  Both the input and
-/// output ranks are specified at compile time.
+///        IndexTransform<3, 2> t = IndexTransformBuilder<3, 2>()
+///            .input_origin({1, 2, 3})
+///            .input_inclusive_max({5, 7, 9})
+///            .output_single_input_dimension(0, 1, 2, 0)
+///            .output_constant(1, 5)
+///            .Finalize()
+///            .value();
 ///
-///     IndexTransform<> t = IndexTransformBuilder<>(3, 2)
-///         .input_origin({1, 2, 3})
-///         .input_exclusive_max({6, 8, 10})
-///         .output_single_input_dimension(0, 1, 2, 0)
-///         .output_constant(1, 5)
-///         .Finalize()
-///         .value();
+///    Creates an index transform from an rank-3 input domain
+///    `[1, 5], [2, 7], [3, 9]` to a rank 2 output space, where
+///    `output[0] = 1 + 2 * input[0]` and `output[1] = 5`.  Both the input and
+///    output ranks are specified at compile time.
 ///
-/// Same as previous example, except that both the input and output ranks are
-/// specified at run time, and the upper bound of the input domain is specified
-/// using the `input_exclusive_max` method.
+///    .. code-block:: cpp
 ///
-///     IndexTransform<3, 2> t = IndexTransformBuilder<3, 2>()
-///         .input_origin({1, 2, 3})
-///         .input_shape({5, 6, 3})
-///         .input_labels({"a", "b", "x"})
-///         .output_single_input_dimension(0, 1, 2, 0)
-///         .output_index_array(1, 0, 1, MakeArray<Index>({{{5, 10, 13}}}))
-///         .Finalize()
-///         .value();
+///       IndexTransform<> t = IndexTransformBuilder<>(3, 2)
+///           .input_origin({1, 2, 3})
+///           .input_exclusive_max({6, 8, 10})
+///           .output_single_input_dimension(0, 1, 2, 0)
+///           .output_constant(1, 5)
+///           .Finalize()
+///           .value();
 ///
-/// Creates an index transform from a rank-3 input domain
-/// `"a": [1, 5], "b": [2, 7], "x": [3, 5]` to a rank 2 output space, where
-/// `output[0] = 1 + 2 * input[0]` and `output[1] = {5, 10, 13}[input[2] - 3]`.
-/// Both the input and output ranks are specified at compile time.
+///    Same as previous example, except that both the input and output ranks are
+///    specified at run time, and the upper bound of the input domain is
+///    specified using the `input_exclusive_max` method.
 ///
-/// \remark Invalid arguments specified to IndexTransformBuilder are handled in
-///     two different ways.  Calling `input_origin`, `input_shape`,
-///     `input_inclusive_max`, `input_exclusive_max`, or `input_labels` with a
-///     sequence of length not equal to `input_rank()`, and calling `output_*`
-///     with an invalid `output_dim` are fatal errors.  Specifying an invalid
-///     input domain or invalid output index map is not a fatal error; the
-///     Finalize method will simply return an error `Result` if the invalid
-///     input domain or output index map has not been overridden with a valid
-///     one prior to the call to Finalize.
+///    .. code-block:: cpp
+///
+///       IndexTransform<3, 2> t = IndexTransformBuilder<3, 2>()
+///           .input_origin({1, 2, 3})
+///           .input_shape({5, 6, 3})
+///           .input_labels({"a", "b", "x"})
+///           .output_single_input_dimension(0, 1, 2, 0)
+///           .output_index_array(1, 0, 1, MakeArray<Index>({{{5, 10, 13}}}))
+///           .Finalize()
+///           .value();
+///
+///    Creates an index transform from a rank-3 input domain
+///    ``"a": [1, 5], "b": [2, 7], "x": [3, 5]`` to a rank 2 output space, where
+///    ``output[0] = 1 + 2 * input[0]`` and
+///    ``output[1] = {5, 10, 13}[input[2] - 3]``.  Both the input and output
+///    ranks are specified at compile time.
+///
+/// .. note::
+///
+///    Invalid arguments specified to `IndexTransformBuilder` are handled in two
+///    different ways.  Calling `input_origin`, `input_shape`,
+///    `input_inclusive_max`, `input_exclusive_max`, or `input_labels` with a
+///    sequence of length not equal to `input_rank()`, and calling `output_*`
+///    with an invalid `output_dim` are fatal errors.  Specifying an invalid
+///    input domain or invalid output index map is not a fatal error; the
+///    `Finalize` method will simply return an error `Result` if the invalid
+///    input domain or output index map has not been overridden with a valid one
+///    prior to the call to `Finalize`.
+///
+/// \relates IndexTransform
 template <DimensionIndex InputRank = dynamic_rank,
           DimensionIndex OutputRank = dynamic_rank>
 class IndexTransformBuilder {
+  static_assert(RankConstraint(InputRank).valid());
+  static_assert(RankConstraint(OutputRank).valid());
+
  public:
   /// Constructs an invalid `IndexTransformBuilder`.
   ///
@@ -287,7 +313,7 @@ class IndexTransformBuilder {
   ///
   /// \pre `valid() == true`
   span<Index, InputRank> input_origin() {
-    flags_ |= internal_index_space::kSetLower;
+    flags_ |= internal_index_space::BuilderFlags::kSetLower;
     return {rep_->input_origin().data(), input_rank()};
   }
 
@@ -300,9 +326,9 @@ class IndexTransformBuilder {
   /// \remarks Calling this method after it has already been called simply
   ///     overrides the previous value.
   template <typename Indices>
-  absl::enable_if_t<internal_index_space::IsStaticExtentCompatibleWithRange<
-                        InputRank, Indices>::value,
-                    IndexTransformBuilder&>
+  std::enable_if_t<internal_index_space::IsStaticExtentCompatibleWithRange<
+                       InputRank, Indices>,
+                   IndexTransformBuilder&>
   input_origin(const Indices& indices) {
     internal_index_space::AssignRange(indices, span<Index>(input_origin()));
     return *this;
@@ -324,7 +350,7 @@ class IndexTransformBuilder {
   ///
   /// \pre `valid() == true`
   span<Index, InputRank> input_shape() {
-    flags_ |= internal_index_space::kSetUpper;
+    flags_ |= internal_index_space::BuilderFlags::kSetUpper;
     interval_form_ = IntervalForm::sized;
     return {rep_->input_shape().data(), input_rank()};
   }
@@ -339,9 +365,9 @@ class IndexTransformBuilder {
   /// \remarks Calling this method after the upper bound of the input domain has
   ///     already been specified simply overrides the previous value.
   template <typename Indices>
-  absl::enable_if_t<internal_index_space::IsStaticExtentCompatibleWithRange<
-                        InputRank, Indices>::value,
-                    IndexTransformBuilder&>
+  std::enable_if_t<internal_index_space::IsStaticExtentCompatibleWithRange<
+                       InputRank, Indices>,
+                   IndexTransformBuilder&>
   input_shape(const Indices& indices) {
     internal_index_space::AssignRange(indices, span<Index>(input_shape()));
     return *this;
@@ -362,7 +388,7 @@ class IndexTransformBuilder {
   ///
   /// \pre `valid() == true`
   span<Index, InputRank> input_exclusive_max() {
-    flags_ |= internal_index_space::kSetUpper;
+    flags_ |= internal_index_space::BuilderFlags::kSetUpper;
     interval_form_ = IntervalForm::half_open;
     return {rep_->input_shape().data(), input_rank()};
   }
@@ -377,9 +403,9 @@ class IndexTransformBuilder {
   /// \remarks Calling this method after the upper bound of the input domain has
   ///     already been specified simply overrides the previous value.
   template <typename Indices>
-  absl::enable_if_t<internal_index_space::IsStaticExtentCompatibleWithRange<
-                        InputRank, Indices>::value,
-                    IndexTransformBuilder&>
+  std::enable_if_t<internal_index_space::IsStaticExtentCompatibleWithRange<
+                       InputRank, Indices>,
+                   IndexTransformBuilder&>
   input_exclusive_max(const Indices& indices) {
     internal_index_space::AssignRange(indices,
                                       span<Index>(input_exclusive_max()));
@@ -402,7 +428,7 @@ class IndexTransformBuilder {
   ///
   /// \pre `valid() == true`
   span<Index, InputRank> input_inclusive_max() {
-    flags_ |= internal_index_space::kSetUpper;
+    flags_ |= internal_index_space::BuilderFlags::kSetUpper;
     interval_form_ = IntervalForm::closed;
     return {rep_->input_shape().data(), input_rank()};
   }
@@ -417,9 +443,9 @@ class IndexTransformBuilder {
   /// \remarks Calling this method after the upper bound of the input domain has
   ///     already been specified simply overrides the previous value.
   template <typename Indices>
-  absl::enable_if_t<internal_index_space::IsStaticExtentCompatibleWithRange<
-                        InputRank, Indices>::value,
-                    IndexTransformBuilder&>
+  std::enable_if_t<internal_index_space::IsStaticExtentCompatibleWithRange<
+                       InputRank, Indices>,
+                   IndexTransformBuilder&>
   input_inclusive_max(const Indices& indices) {
     internal_index_space::AssignRange(indices,
                                       span<Index>(input_inclusive_max()));
@@ -450,15 +476,25 @@ class IndexTransformBuilder {
   ///     input domain have already been specified simply overrides the previous
   ///     values.
   template <typename BoxLike>
-  absl::enable_if_t<
-      IsBoxLikeImplicitlyConvertibleToRank<BoxLike, InputRank>::value,
-      IndexTransformBuilder&>
+  std::enable_if_t<IsBoxLikeImplicitlyConvertibleToRank<BoxLike, InputRank>,
+                   IndexTransformBuilder&>
   input_bounds(const BoxLike& box) {
-    input_origin(box.origin());
-    input_shape(box.shape());
+    this->input_bounds().DeepAssign(box);
     return *this;
   }
 
+  /// Returns the mutable `input_bounds` box, and marks the lower and upper
+  /// bounds as having been set.
+  ///
+  /// \pre `valid() == true`
+  MutableBoxView<InputRank> input_bounds() {
+    flags_ |= (internal_index_space::BuilderFlags::kSetLower |
+               internal_index_space::BuilderFlags::kSetUpper);
+    interval_form_ = IntervalForm::sized;
+    return MutableBoxView<InputRank>(input_rank(), rep_->input_origin().data(),
+                                     rep_->input_shape().data());
+  }
+
   /// Specifies the input domain.
   ///
   /// \pre `valid() == true`
@@ -484,15 +520,15 @@ class IndexTransformBuilder {
   /// Sets the `input_labels` vector the specified value.
   ///
   /// \param labels A sequence with `value_type` convertible to
-  ///     `absl::string_view` specifying the label for each input dimension.
+  ///     `std::string_view` specifying the label for each input dimension.
   /// \pre `valid() == true`
   /// \checks The size of the `labels` vector must equal `input_rank()`.
   /// \remarks Calling this method after it has already been called simply
   ///     overrides the previous value.
   template <typename Labels>
-  absl::enable_if_t<internal_index_space::IsStaticExtentCompatibleWithRange<
-                        InputRank, Labels>::value,
-                    IndexTransformBuilder&>
+  std::enable_if_t<internal_index_space::IsStaticExtentCompatibleWithRange<
+                       InputRank, Labels>,
+                   IndexTransformBuilder&>
   input_labels(const Labels& labels) {
     internal_index_space::AssignRange(labels,
                                       span<std::string>(input_labels()));
@@ -504,7 +540,7 @@ class IndexTransformBuilder {
   ///
   /// \pre `valid() == true`
   template <std::size_t N>
-  IndexTransformBuilder& input_labels(const absl::string_view (&labels)[N]) {
+  IndexTransformBuilder& input_labels(const std::string_view (&labels)[N]) {
     static_assert(InputRank == dynamic_rank || InputRank == N, "");
     return input_labels(span(labels));
   }
@@ -513,23 +549,18 @@ class IndexTransformBuilder {
   /// having been set.
   ///
   /// \pre `valid() == true`
-  BitSpan<std::uint64_t, InputRank> implicit_lower_bounds() {
-    flags_ |= internal_index_space::kSetImplicitLower;
-    return {rep_->implicit_lower_bounds(input_rank()).begin(), input_rank()};
+  DimensionSet& implicit_lower_bounds() {
+    flags_ |= internal_index_space::BuilderFlags::kSetImplicitLower;
+    return rep_->implicit_lower_bounds;
   }
 
   /// Sets the `implicit_lower_bounds` bit-vector to the specified value.
   ///
-  /// \param indices A sequence with `value_type` convertible to `bool`
-  ///     specifying whether the lower bound is implicit for each input
-  ///     dimension.
   /// \pre `valid() == true`
-  /// \checks The size of the `indices` sequence must equal `input_rank()`.
   /// \remarks Calling this method after it has already been called simply
   ///     overrides the previous value.
-  template <typename X>
-  IndexTransformBuilder& implicit_lower_bounds(const X& x) {
-    internal_index_space::AssignRange(x, implicit_lower_bounds());
+  IndexTransformBuilder& implicit_lower_bounds(DimensionSet x) {
+    implicit_lower_bounds() = x;
     return *this;
   }
 
@@ -539,31 +570,27 @@ class IndexTransformBuilder {
   /// \pre `valid() == true`
   template <std::size_t N>
   IndexTransformBuilder& implicit_lower_bounds(const bool (&x)[N]) {
-    static_assert(InputRank == dynamic_rank || InputRank == N, "");
-    return implicit_lower_bounds(span(x));
+    static_assert(InputRank == dynamic_rank || InputRank == N);
+    TENSORSTORE_CHECK(N == input_rank() && "range size mismatch");
+    return implicit_lower_bounds(DimensionSet(x));
   }
 
   /// Returns the mutable `implicit_upper_bounds` bit vector, and marks it as
   /// having been set.
   ///
   /// \pre `valid() == true`
-  BitSpan<std::uint64_t, InputRank> implicit_upper_bounds() {
-    flags_ |= internal_index_space::kSetImplicitUpper;
-    return {rep_->implicit_upper_bounds(input_rank()).begin(), input_rank()};
+  DimensionSet& implicit_upper_bounds() {
+    flags_ |= internal_index_space::BuilderFlags::kSetImplicitUpper;
+    return rep_->implicit_upper_bounds;
   }
 
   /// Sets the `implicit_upper_bounds` bit-vector to the specified value.
   ///
-  /// \param indices A sequence with `value_type` convertible to `bool`
-  ///     specifying whether the upper bound is implicit for each input
-  ///     dimension.
   /// \pre `valid() == true`
-  /// \checks The size of the `indices` sequence must equal `input_rank()`.
   /// \remarks Calling this method after it has already been called simply
   ///     overrides the previous value.
-  template <typename X>
-  IndexTransformBuilder& implicit_upper_bounds(const X& x) {
-    internal_index_space::AssignRange(x, implicit_upper_bounds());
+  IndexTransformBuilder& implicit_upper_bounds(DimensionSet x) {
+    implicit_upper_bounds() = x;
     return *this;
   }
 
@@ -573,8 +600,45 @@ class IndexTransformBuilder {
   /// \pre `valid() == true`
   template <std::size_t N>
   IndexTransformBuilder& implicit_upper_bounds(const bool (&x)[N]) {
-    static_assert(InputRank == dynamic_rank || InputRank == N, "");
-    return implicit_upper_bounds(span(x));
+    static_assert(InputRank == dynamic_rank || InputRank == N);
+    TENSORSTORE_CHECK(N == input_rank() && "range size mismatch");
+    return implicit_upper_bounds(DimensionSet(x));
+  }
+
+  /// Sets the output index map for output dimension `output_dim` to a copy of
+  /// `map` from an existing index transform.
+  ///
+  /// \param output_dim The output dimension for which to set the output index
+  ///     map.
+  /// \param map The output index map to copy.
+  /// \pre `valid() == true`
+  /// \checks `0 <= output_dim && output_dim < output_rank()`.
+  /// \remarks This method overrides any output index map for `output_dim`
+  ///     specified by a previous `output_*(output_dim, ...)` call.
+  template <DimensionIndex OtherInputRank>
+  IndexTransformBuilder& output_map(DimensionIndex output_dim,
+                                    OutputIndexMapRef<OtherInputRank> map) {
+    switch (map.method()) {
+      case OutputIndexMethod::constant:
+        AssignOutput(output_dim, map.offset(), 0,
+                     internal_index_space::OutputIndexMapInitializer());
+        break;
+      case OutputIndexMethod::single_input_dimension:
+        AssignOutput(output_dim, map.offset(), map.stride(),
+                     internal_index_space::OutputIndexMapInitializer(
+                         map.input_dimension()));
+        break;
+      case OutputIndexMethod::array: {
+        auto index_array = map.index_array();
+        AssignOutput(
+            output_dim, map.offset(), map.stride(),
+            internal_index_space::OutputIndexMapInitializer(
+                index_array.shared_array_ref(), index_array.index_range()));
+
+        break;
+      }
+    }
+    return *this;
   }
 
   /// Sets the output index map for output dimension `output_dim` to a
@@ -593,7 +657,7 @@ class IndexTransformBuilder {
   /// \pre `valid() == true`
   /// \checks `0 <= output_dim && output_dim < output_rank()`.
   /// \remarks This method overrides any output index map for `output_dim`
-  ///     specified by a previous `output(output_dim, ...)` call.
+  ///     specified by a previous `output_*(output_dim, ...)` call.
   IndexTransformBuilder& output_constant(DimensionIndex output_dim,
                                          Index offset) {
     AssignOutput(output_dim, offset, 0,
@@ -620,7 +684,7 @@ class IndexTransformBuilder {
   /// \pre `valid() == true`
   /// \checks `0 <= output_dim && output_dim < output_rank()`.
   /// \remarks This method overrides any output index map for `output_dim`
-  ///     specified by a previous `output(output_dim, ...)` call.
+  ///     specified by a previous `output_*(output_dim, ...)` call.
   IndexTransformBuilder& output_single_input_dimension(
       DimensionIndex output_dim, Index offset, Index stride,
       DimensionIndex input_dim) {
@@ -672,6 +736,8 @@ class IndexTransformBuilder {
   ///     specified, this output index map is considered invalid (resulting in
   ///     `Finalize` returning an error).
   /// \pre `valid() == true`
+  /// \remarks This method overrides any output index map for `output_dim`
+  ///     specified by a previous `output_*(output_dim, ...)` call.
   IndexTransformBuilder& output_index_array(
       DimensionIndex output_dim, Index offset, Index stride,
       const SharedArrayView<const Index, dynamic_rank, offset_origin>&
@@ -683,6 +749,35 @@ class IndexTransformBuilder {
     return *this;
   }
 
+  /// Sets the first `min(input_rank(), output_rank())` output dimensions to
+  /// identity map to input dimensions.
+  ///
+  /// \pre `valid() == true`
+  IndexTransformBuilder& output_identity_transform() {
+    for (DimensionIndex i = 0, rank = std::min(DimensionIndex(input_rank()),
+                                               DimensionIndex(output_rank()));
+         i < rank; ++i) {
+      this->output_single_input_dimension(i, i);
+    }
+    return *this;
+  }
+
+  /// Copies the first `min(output_rank(), output_maps.size())` output maps from
+  /// `output_maps`.
+  ///
+  /// \pre `valid() == true`
+  template <DimensionIndex OtherInputRank, DimensionIndex OtherOutputRank>
+  IndexTransformBuilder& output_maps(
+      OutputIndexMapRange<OtherInputRank, OtherOutputRank> output_maps) {
+    for (DimensionIndex i = 0,
+                        rank = std::min(DimensionIndex(output_rank()),
+                                        DimensionIndex(output_maps.size()));
+         i < rank; ++i) {
+      this->output_map(i, output_maps[i]);
+    }
+    return *this;
+  }
+
   /// Validates and returns the specified index transform.
   ///
   /// \pre `valid() == true`
@@ -702,8 +797,9 @@ class IndexTransformBuilder {
                           ? internal::kNumInlinedDims
                           : (OutputRank == 0 ? 1 : OutputRank)>
       output_index_maps_;
-  IntervalForm interval_form_;
-  internal_index_space::BuilderFlags flags_;
+  IntervalForm interval_form_ = IntervalForm::sized;
+  internal_index_space::BuilderFlags flags_ =
+      internal_index_space::BuilderFlags::kDefault;
 
   void AssignOutput(
       DimensionIndex output_dim, Index offset, Index stride,
@@ -717,6 +813,8 @@ class IndexTransformBuilder {
   }
 };
 
+IndexTransformBuilder()->IndexTransformBuilder<>;
+
 namespace internal_index_space {
 
 template <typename Range, typename Element>
@@ -733,20 +831,6 @@ void AssignRange(const Range& range, span<Element> dest) {
   TENSORSTORE_CHECK(it == last && "range size mismatch");
 }
 
-template <typename Range>
-void AssignRange(const Range& range, BitSpan<std::uint64_t> dest) {
-  using std::begin;
-  using std::end;
-  auto it = begin(range);
-  auto last = end(range);
-  for (std::ptrdiff_t i = 0; i < dest.size(); ++i) {
-    TENSORSTORE_CHECK(it != last && "range size mismatch");
-    dest[i] = *it;
-    ++it;
-  }
-  TENSORSTORE_CHECK(it == last && "range size mismatch");
-}
-
 /// Initializes all output `offset` and `stride` values with `0`.
 ///
 /// \dchecks data != nullptr
@@ -761,10 +845,11 @@ void InitializeTransformRepForBuilder(TransformRep* data);
 /// \param output_index_maps Array of length `data->output_rank()`.
 /// \param interval_form Specifies the interpretation of `data->input_shape()`.
 /// \param flags Specifies which fields have been set.
-/// \returns `Status()` if the resulting transform is valid, or an error Status
+/// \returns `absl::Status()` if the resulting transform is valid, or an error
+/// absl::Status
 ///     otherwise.  In the case of an error return, the output index maps of
 ///     `data` are left in an unspecified state.
-Status SetOutputIndexMapsAndValidateTransformRep(
+absl::Status SetOutputIndexMapsAndValidateTransformRep(
     TransformRep* data, span<const OutputIndexMapInitializer> output_index_maps,
     IntervalForm interval_form, BuilderFlags flags);
 
@@ -777,7 +862,7 @@ IndexTransformBuilder<InputRank, OutputRank>::IndexTransformBuilder(
                .release(),
            internal::adopt_object_ref),
       output_index_maps_(output_rank),
-      flags_(0) {
+      flags_(internal_index_space::BuilderFlags::kDefault) {
   rep_->input_rank = input_rank;
   rep_->output_rank = output_rank;
   internal_index_space::InitializeTransformRepForBuilder(rep_.get());
