@@ -16,17 +16,19 @@
 
 #include <system_error>  // NOLINT
 
+#include "absl/status/status.h"
 #include "tensorstore/util/status.h"
-#include "tensorstore/util/to_string.h"
+#include "tensorstore/util/str_cat.h"
 
 namespace tensorstore {
 namespace internal_index_space {
 
-Status CheckIndexVectorSize(IndexVectorOrScalar indices, DimensionIndex size) {
+absl::Status CheckIndexVectorSize(IndexVectorOrScalarView indices,
+                                  DimensionIndex size) {
   if (indices.pointer && indices.size_or_scalar != size)
     return absl::InvalidArgumentError(StrCat(
         "Number of dimensions (", size, ") does not match number of indices (",
-        indices.size_or_scalar, ")."));
+        indices.size_or_scalar, ")"));
   return absl::OkStatus();
 }
 
