@@ -75,9 +75,11 @@ bool BroadcastShapes(span<const Index> source_shape, span<Index> result_shape) {
 /// \returns The transform where the output dimensions specified in
 ///     `*dimensions` are mapped using the specified index arrays, and the
 ///     remaining output dimensions are identity mapped.
+/// \error `absl::StatusCode::kInvalidArgument` if the resultant input rank is
+///     invalid.
 template <typename GetNewDimensionShapeFn, typename GetIndexArrayBasePointerFn,
           typename GetIndexArrayByteStrideFn>
-TransformRep::Ptr<> MakeTransformFromJointIndexArrays(
+Result<TransformRep::Ptr<>> MakeTransformFromJointIndexArrays(
     DimensionIndex num_new_dims, TransformRep* orig_transform,
     DimensionIndexBuffer* dimensions,
     GetNewDimensionShapeFn get_new_dimension_bounds,
@@ -87,14 +89,15 @@ TransformRep::Ptr<> MakeTransformFromJointIndexArrays(
   const DimensionIndex output_rank = orig_transform->input_rank;
   const DimensionIndex input_rank =
       output_rank - dimensions->size() + num_new_dims;
+  TENSORSTORE_RETURN_IF_ERROR(ValidateRank(input_rank));
   auto result = TransformRep::Allocate(input_rank, output_rank);
   result->input_rank = input_rank;
   result->output_rank = output_rank;
-  // Set all bounds to explicit, and all labels to explicit.  These defaults are
-  // only used for dimensions corresponding to the domain of the index arrays.
-  // For identity-mapped dimensions, the defaults are overridden.
-  std::fill_n(result->implicit_bitvector_base(),
-              result->implicit_bitvector_storage().size(), 0);
+  // Set all bounds to explicit.  These defaults are only used for dimensions
+  // corresponding to the domain of the index arrays.  For identity-mapped
+  // dimensions, the defaults are overridden.
+  result->implicit_lower_bounds = false;
+  result->implicit_upper_bounds = false;
   span<OutputIndexMap> maps = result->output_index_maps().first(output_rank);
   const DimensionIndex num_preserved_dims = output_rank - num_indexed_dims;
   // Set all output dimensions to single_input_dimension index method.  The
@@ -105,45 +108,28 @@ TransformRep::Ptr<> MakeTransformFromJointIndexArrays(
   for (DimensionIndex output_dim = 0; output_dim < output_rank; ++output_dim) {
     maps[output_dim].SetSingleInputDimension(0);
   }
-
   const auto input_domain = result->input_domain(input_rank);
-  bool is_singleton_index_array = true;
   // Sets the input domain for dimensions corresponding to the domain of the
   // index arrays.
   for (DimensionIndex new_dim = 0; new_dim < num_new_dims; ++new_dim) {
-    const IndexInterval bounds = input_domain[new_dim] =
-        get_new_dimension_bounds(new_dim);
-    if (bounds.size() != 1) is_singleton_index_array = false;
+    input_domain[new_dim] = get_new_dimension_bounds(new_dim);
   }
-  if (is_singleton_index_array) {
-    // Index array actually has only a single value; therefore, we can just set
-    // all specified array-indexed dimensions to have constant output index
-    // maps.
-    for (DimensionIndex indexed_dim = 0; indexed_dim < num_indexed_dims;
-         ++indexed_dim) {
-      const DimensionIndex output_dim = (*dimensions)[indexed_dim];
-      auto& map = maps[output_dim];
-      map.SetConstant();
-      map.offset() = *get_index_array_base_pointer(indexed_dim);
-    }
-  } else {
-    // Sets all array-indexed dimensions to have array output index maps.
-    for (DimensionIndex indexed_dim = 0; indexed_dim < num_indexed_dims;
-         ++indexed_dim) {
-      const DimensionIndex output_dim = (*dimensions)[indexed_dim];
-      auto& map = maps[output_dim];
-      map.offset() = 0;
-      map.stride() = 1;
-      auto& index_array_data = map.SetArrayIndexing(input_rank);
-      std::fill_n(index_array_data.byte_strides + num_new_dims,
-                  num_preserved_dims, 0);
-      for (DimensionIndex new_dim = 0; new_dim < num_new_dims; ++new_dim) {
-        index_array_data.byte_strides[new_dim] =
-            get_index_array_byte_stride(indexed_dim, new_dim);
-      }
-      index_array_data.element_pointer =
-          get_index_array_base_pointer(indexed_dim);
+  // Sets all array-indexed dimensions to have array output index maps.
+  for (DimensionIndex indexed_dim = 0; indexed_dim < num_indexed_dims;
+       ++indexed_dim) {
+    const DimensionIndex output_dim = (*dimensions)[indexed_dim];
+    auto& map = maps[output_dim];
+    map.offset() = 0;
+    map.stride() = 1;
+    auto& index_array_data = map.SetArrayIndexing(input_rank);
+    std::fill_n(index_array_data.byte_strides + num_new_dims,
+                num_preserved_dims, 0);
+    for (DimensionIndex new_dim = 0; new_dim < num_new_dims; ++new_dim) {
+      index_array_data.byte_strides[new_dim] =
+          get_index_array_byte_stride(indexed_dim, new_dim);
     }
+    index_array_data.element_pointer =
+        get_index_array_base_pointer(indexed_dim);
   }
   // Sets the output index maps for output dimensions not indexed by the index
   // array to be identity maps, and copies the input dimension fields from the
@@ -159,9 +145,13 @@ TransformRep::Ptr<> MakeTransformFromJointIndexArrays(
         orig_transform->input_dimension(output_dim);
     ++input_dim;
   }
+  if (IsDomainExplicitlyEmpty(result.get())) {
+    ReplaceAllIndexArrayMapsWithConstantMaps(result.get());
+  }
   dimensions->resize(num_new_dims);
   std::iota(dimensions->begin(), dimensions->end(),
             static_cast<DimensionIndex>(0));
+  internal_index_space::DebugCheckInvariants(result.get());
   return result;
 }
 
@@ -170,25 +160,34 @@ Result<TransformRep::Ptr<>> MakeTransformFromIndexArrays(
     span<const SharedArrayView<const Index>> index_arrays) {
   const DimensionIndex num_indexed_dims = dimensions->size();
   if (index_arrays.size() != num_indexed_dims) {
-    return absl::InvalidArgumentError(
-        StrCat("Number of selected dimensions (", num_indexed_dims,
-               ") does not equal number of index arrays (", index_arrays.size(),
-               ")."));
+    return absl::InvalidArgumentError(StrCat(
+        "Number of selected dimensions (", num_indexed_dims,
+        ") does not equal number of index arrays (", index_arrays.size(), ")"));
   }
   if (index_arrays.empty()) {
     return absl::InvalidArgumentError(
-        StrCat("At least one index array must be specified."));
+        StrCat("At least one index array must be specified"));
   }
   absl::FixedArray<Index, internal::kNumInlinedDims> shape(
       index_arrays[0].rank(), 1);
+
+  bool error = false;
   for (DimensionIndex i = 0; i < index_arrays.size(); ++i) {
     if (!BroadcastShapes(index_arrays[i].shape(), shape)) {
-      return Status(
-          absl::StatusCode::kInvalidArgument,
-          StrCat("Index arrays cannot be broadcast to a common shape: ",
-                 span(shape), " vs ", index_arrays[i].shape(), "."));
+      error = true;
     }
   }
+  if (error) {
+    std::string shape_msg;
+    for (DimensionIndex i = 0; i < index_arrays.size(); ++i) {
+      StrAppend(&shape_msg, (shape_msg.empty() ? "" : ", "),
+                index_arrays[i].shape());
+    }
+    return absl::InvalidArgumentError(
+        StrCat("Index arrays with shapes ", shape_msg,
+               " cannot be broadcast to a common shape"));
+  }
+
   const DimensionIndex num_new_dims = shape.size();
   const auto get_new_dimension_bounds = [&](DimensionIndex new_dim) {
     return IndexInterval::UncheckedSized(0, shape[new_dim]);
@@ -212,24 +211,24 @@ Result<TransformRep::Ptr<>> MakeTransformFromOuterIndexArrays(
     span<const SharedArrayView<const Index>> index_arrays) {
   const DimensionIndex num_indexed_dims = dimensions->size();
   if (index_arrays.size() != num_indexed_dims) {
-    return absl::InvalidArgumentError(
-        StrCat("Number of selected dimensions (", num_indexed_dims,
-               ") does not equal number of index arrays (", index_arrays.size(),
-               ")."));
+    return absl::InvalidArgumentError(StrCat(
+        "Number of selected dimensions (", num_indexed_dims,
+        ") does not equal number of index arrays (", index_arrays.size(), ")"));
   }
   const DimensionIndex output_rank = orig_transform->input_rank;
   DimensionIndex input_rank = output_rank - num_indexed_dims;
   for (const auto& index_array : index_arrays) {
     input_rank += index_array.rank();
   }
+  TENSORSTORE_RETURN_IF_ERROR(ValidateRank(input_rank));
   auto result = TransformRep::Allocate(input_rank, output_rank);
   result->input_rank = input_rank;
   result->output_rank = output_rank;
-  // Set all bounds to explicit, and all labels to explicit.  These defaults are
-  // only used for dimensions corresponding to the domain of the index arrays.
-  // For identity-mapped dimensions, the defaults are overridden.
-  std::fill_n(result->implicit_bitvector_base(),
-              result->implicit_bitvector_storage().size(), 0);
+  // Set all bounds to explicit.  These defaults are only used for dimensions
+  // corresponding to the domain of the index arrays.  For identity-mapped
+  // dimensions, the defaults are overridden.
+  result->implicit_lower_bounds = false;
+  result->implicit_upper_bounds = false;
   absl::FixedArray<DimensionIndex, internal::kNumInlinedDims>
       index_array_start_dim(num_indexed_dims);
   absl::FixedArray<DimensionIndex, internal::kNumInlinedDims> index_array_order(
@@ -264,6 +263,7 @@ Result<TransformRep::Ptr<>> MakeTransformFromOuterIndexArrays(
         if (array.num_elements() == 1) {
           map.SetConstant();
           map.offset() = *array.data();
+          map.stride() = 0;
         } else {
           auto& index_array_data = map.SetArrayIndexing(input_rank);
           index_array_data.element_pointer = array.element_pointer();
@@ -285,6 +285,9 @@ Result<TransformRep::Ptr<>> MakeTransformFromOuterIndexArrays(
     map.SetSingleInputDimension(input_dim);
     ++input_dim;
   }
+  if (IsDomainExplicitlyEmpty(result.get())) {
+    ReplaceAllIndexArrayMapsWithConstantMaps(result.get());
+  }
   // Sets `dimensions` to the new input dimensions corresponding to the index
   // array domains.
   dimensions->clear();
@@ -299,6 +302,7 @@ Result<TransformRep::Ptr<>> MakeTransformFromOuterIndexArrays(
       dimensions->push_back(input_dim);
     }
   }
+  internal_index_space::DebugCheckInvariants(result.get());
   return result;
 }
 
@@ -314,7 +318,7 @@ Result<TransformRep::Ptr<>> MakeTransformFromIndexVectorArray(
     return absl::InvalidArgumentError(
         StrCat("Number of selected dimensions (", num_indexed_dims,
                ") does not equal index vector length (",
-               index_vector_array.shape()[vector_dimension], ")."));
+               index_vector_array.shape()[vector_dimension], ")"));
   }
   const DimensionIndex num_new_dims = index_vector_array.rank() - 1;
   const auto get_index_vector_array_dim = [&](DimensionIndex new_dim) {
@@ -346,8 +350,8 @@ Result<TransformRep::Ptr<>> MakeTransformFromIndexVectorArray(
 
 Result<IndexTransform<>> ApplyIndexArraySlice(
     IndexTransform<> transform, DimensionIndexBuffer* dimensions,
-    span<const SharedArrayView<const Index>> index_arrays,
-    bool outer_indexing) {
+    span<const SharedArrayView<const Index>> index_arrays, bool outer_indexing,
+    bool domain_only) {
   TENSORSTORE_ASSIGN_OR_RETURN(
       auto other_transform,
       outer_indexing
@@ -359,14 +363,14 @@ Result<IndexTransform<>> ApplyIndexArraySlice(
       auto new_rep,
       ComposeTransforms(TransformAccess::rep(transform),
                         /*can_move_from_b_to_c=*/false, other_transform.get(),
-                        /*can_move_from_a_to_b=*/true));
+                        /*can_move_from_a_to_b=*/true, domain_only));
   return TransformAccess::Make<IndexTransform<>>(std::move(new_rep));
 }
 
 Result<IndexTransform<>> ApplyIndexVectorArraySlice(
     IndexTransform<> transform, DimensionIndexBuffer* dimensions,
     DimensionIndex vector_dimension,
-    const SharedArrayView<const Index>& index_vector_array) {
+    const SharedArrayView<const Index>& index_vector_array, bool domain_only) {
   TENSORSTORE_ASSIGN_OR_RETURN(auto other_transform,
                                MakeTransformFromIndexVectorArray(
                                    TransformAccess::rep(transform), dimensions,
@@ -375,7 +379,7 @@ Result<IndexTransform<>> ApplyIndexVectorArraySlice(
       auto new_rep,
       ComposeTransforms(TransformAccess::rep(transform),
                         /*can_move_from_b_to_c=*/false, other_transform.get(),
-                        /*can_move_from_a_to_b=*/true));
+                        /*can_move_from_a_to_b=*/true, domain_only));
   return TransformAccess::Make<IndexTransform<>>(std::move(new_rep));
 }
 
