@@ -14,6 +14,7 @@
 
 #include "tensorstore/util/internal/iterate.h"
 
+#include "absl/status/status.h"
 #include "tensorstore/array.h"
 #include "tensorstore/index_space/index_transform.h"
 #include "tensorstore/index_space/internal/iterate_impl.h"
@@ -34,8 +35,9 @@ void MarkSingletonDimsAsSkippable(
     span<const Index> input_shape,
     input_dimension_iteration_flags::Bitmask* input_dimension_flags) {
   for (DimensionIndex i = 0; i < input_shape.size(); ++i) {
-    if (input_shape[i] <= 1)
+    if (input_shape[i] == 1) {
       input_dimension_flags[i] = input_dimension_iteration_flags::can_skip;
+    }
   }
 }
 
@@ -49,9 +51,9 @@ namespace {
 ///
 /// If `UseStridedLayout == false`, `transform` must not be `nullptr`, and
 /// `(array.element_pointer(), transform)` is assumed to be a valid
-/// `NormalizedTransformedArray`.
+/// `TransformedArray`.
 template <bool UseStridedLayout>
-Status InitializeSingleArrayIterationStateImpl(
+absl::Status InitializeSingleArrayIterationStateImpl(
     OffsetArrayView<const void, (UseStridedLayout ? dynamic_rank : 0)> array,
     TransformRep* transform, const Index* iteration_origin,
     const Index* iteration_shape, SingleArrayIterationState* single_array_state,
@@ -72,7 +74,7 @@ Status InitializeSingleArrayIterationStateImpl(
         const DimensionIndex input_dim = output_dim;
         const Index byte_stride = array.byte_strides()[output_dim];
         single_array_state->input_byte_strides[input_dim] = byte_stride;
-        if (iteration_shape[input_dim] > 1) {
+        if (iteration_shape[input_dim] != 1) {
           input_dimension_flags[input_dim] |=
               input_dimension_iteration_flags::strided;
         }
@@ -188,7 +190,11 @@ Status InitializeSingleArrayIterationStateImpl(
         // The index array has only a single distinct value; therefore, we treat
         // it as a constant output index map.
         const Index index = *index_array_pointer;
-        TENSORSTORE_RETURN_IF_ERROR(CheckContains(index_bounds, index));
+        TENSORSTORE_RETURN_IF_ERROR(
+            CheckContains(index_bounds, index),
+            MaybeAnnotateStatus(
+                _, StrCat("In index array map for output dimension ",
+                          output_dim)));
         single_array_state->base_pointer +=
             internal::wrap_on_overflow::Multiply(
                 byte_stride,
@@ -205,20 +211,25 @@ Status InitializeSingleArrayIterationStateImpl(
         single_array_state->index_array_output_byte_strides[index_array_num] =
             internal::wrap_on_overflow::Multiply(byte_stride, output_stride);
 
-        TENSORSTORE_RETURN_IF_ERROR(ValidateIndexArrayBounds(
-            index_bounds,
-            ArrayView<const Index>(index_array_pointer.get(),
-                                   StridedLayoutView<dynamic_rank>(
-                                       input_rank, iteration_shape,
-                                       index_array_data.byte_strides))));
+        TENSORSTORE_RETURN_IF_ERROR(
+            ValidateIndexArrayBounds(
+                index_bounds,
+                ArrayView<const Index>(index_array_pointer.get(),
+                                       StridedLayoutView<dynamic_rank>(
+                                           input_rank, iteration_shape,
+                                           index_array_data.byte_strides))),
+            MaybeAnnotateStatus(
+                _, StrCat("In index array map for output dimension ",
+                          output_dim)));
       }
     }
   }
   return {};
 }
+
 }  // namespace
 
-Status InitializeSingleArrayIterationState(
+absl::Status InitializeSingleArrayIterationState(
     OffsetArrayView<const void> array, TransformRep* transform,
     const Index* iteration_origin, const Index* iteration_shape,
     SingleArrayIterationState* single_array_state,
@@ -228,7 +239,7 @@ Status InitializeSingleArrayIterationState(
       input_dimension_flags);
 }
 
-Status InitializeSingleArrayIterationState(
+absl::Status InitializeSingleArrayIterationState(
     ElementPointer<const void> element_pointer, TransformRep* transform,
     const Index* iteration_origin, const Index* iteration_shape,
     SingleArrayIterationState* single_array_state,
@@ -294,7 +305,8 @@ void FillOffsetsArray(span<Index> offsets, span<const Index> position,
     for (Index j = 0; j < offsets.size(); ++j) {
       offsets[j] = internal::wrap_on_overflow::Add(
           offsets[j],
-          internal::wrap_on_overflow::Multiply(final_input_dim_byte_stride, j));
+          internal::wrap_on_overflow::Multiply(
+              final_input_dim_byte_stride, j + final_input_dim_start_position));
     }
   }
 }
@@ -303,8 +315,9 @@ template <std::size_t Arity>
 ArrayIterateResult IterateUsingSimplifiedLayout(
     const SimplifiedDimensionIterationOrder& layout,
     span<const Index> input_shape,
-    internal::ElementwiseClosure<Arity, Status*> closure, Status* status,
-    span<absl::optional<SingleArrayIterationState>, Arity> single_array_states,
+    internal::ElementwiseClosure<Arity, absl::Status*> closure,
+    absl::Status* status,
+    span<std::optional<SingleArrayIterationState>, Arity> single_array_states,
     std::array<std::ptrdiff_t, Arity> element_sizes) {
   const Index final_indexed_dim_size =
       layout.simplified_shape[layout.pure_strided_start_dim - 1];
@@ -399,8 +412,9 @@ ArrayIterateResult IterateUsingSimplifiedLayout(
   template ArrayIterateResult IterateUsingSimplifiedLayout<Arity>(           \
       const SimplifiedDimensionIterationOrder& layout,                       \
       span<const Index> input_shape,                                         \
-      internal::ElementwiseClosure<Arity, Status*> closure, Status* status,  \
-      span<absl::optional<SingleArrayIterationState>, Arity>                 \
+      internal::ElementwiseClosure<Arity, absl::Status*> closure,            \
+      absl::Status* status,                                                  \
+      span<std::optional<SingleArrayIterationState>, Arity>                  \
           single_array_states,                                               \
       std::array<std::ptrdiff_t, Arity> element_sizes);
 TENSORSTORE_INTERNAL_FOR_EACH_ARITY(
@@ -409,7 +423,7 @@ TENSORSTORE_INTERNAL_FOR_EACH_ARITY(
 
 }  // namespace internal_index_space
 
-Status ValidateIndexArrayBounds(
+absl::Status ValidateIndexArrayBounds(
     IndexInterval bounds,
     ArrayView<const Index, dynamic_rank, offset_origin> index_array) {
   const auto finite_bounds = FiniteSubset(bounds);
