@@ -16,6 +16,7 @@
 #define TENSORSTORE_INDEX_SPACE_INTERNAL_ITERATE_IMPL_H_
 
 #include "absl/container/fixed_array.h"
+#include "absl/status/status.h"
 #include "tensorstore/array.h"
 #include "tensorstore/index.h"
 #include "tensorstore/index_space/internal/transform_rep.h"
@@ -137,7 +138,7 @@ struct SingleArrayIterationState {
 ///     contained within the domain of the array.
 /// \error `absl::StatusCode::kInvalidArgument` if integer overflow occurs
 ///     computing the result.
-Status InitializeSingleArrayIterationState(
+absl::Status InitializeSingleArrayIterationState(
     OffsetArrayView<const void> array, TransformRep* transform,
     const Index* iteration_origin, const Index* iteration_shape,
     SingleArrayIterationState* single_array_state,
@@ -145,7 +146,7 @@ Status InitializeSingleArrayIterationState(
 
 /// Equivalent to calling the overload above with an `array` with an element
 /// pointer of `element_pointer`, an unbounded domain, and all 1 byte strides.
-Status InitializeSingleArrayIterationState(
+absl::Status InitializeSingleArrayIterationState(
     ElementPointer<const void> element_pointer, TransformRep* transform,
     const Index* iteration_origin, const Index* iteration_shape,
     SingleArrayIterationState* single_array_state,
@@ -182,7 +183,7 @@ struct SimplifiedDimensionIterationOrder : public DimensionIterationOrder {
 
 template <std::size_t Arity>
 struct OrderTransformedArrayDimensionsByStrides {
-  span<const absl::optional<SingleArrayIterationState>, Arity>
+  span<const std::optional<SingleArrayIterationState>, Arity>
       single_array_states;
 
   bool operator()(DimensionIndex input_dim_a,
@@ -295,7 +296,7 @@ DimensionIterationOrder ComputeDimensionIterationOrder(
 
 template <std::size_t Arity>
 DimensionIterationOrder ComputeDimensionIterationOrder(
-    span<const absl::optional<SingleArrayIterationState>, Arity>
+    span<const std::optional<SingleArrayIterationState>, Arity>
         single_array_states,
     span<const input_dimension_iteration_flags::Bitmask> input_dimension_flags,
     LayoutOrderConstraint order_constraint) {
@@ -309,7 +310,7 @@ SimplifiedDimensionIterationOrder SimplifyDimensionIterationOrder(
     const DimensionIterationOrder& original_layout,
     span<const Index> input_shape,
     CanCombineDimensions can_combine_dimensions) {
-  ABSL_ASSERT(original_layout.pure_strided_start_dim > 0);
+  assert(original_layout.pure_strided_start_dim > 0);
 
   SimplifiedDimensionIterationOrder result(
       original_layout.pure_strided_end_dim);
@@ -348,7 +349,7 @@ SimplifiedDimensionIterationOrder SimplifyDimensionIterationOrder(
 /// Predicate for use with SimplifyDimensionIterationOrder.
 template <std::size_t Arity>
 struct CanCombineTransformedArrayDimensions {
-  span<const absl::optional<SingleArrayIterationState>, Arity>
+  span<const std::optional<SingleArrayIterationState>, Arity>
       single_array_states;
 
   bool operator()(DimensionIndex prev_input_dim, DimensionIndex input_dim,
@@ -379,7 +380,7 @@ template <std::size_t Arity>
 SimplifiedDimensionIterationOrder SimplifyDimensionIterationOrder(
     const DimensionIterationOrder& original_layout,
     span<const Index> input_shape,
-    span<const absl::optional<SingleArrayIterationState>, Arity>
+    span<const std::optional<SingleArrayIterationState>, Arity>
         single_array_states) {
   return SimplifyDimensionIterationOrder(
       original_layout, input_shape,
@@ -390,8 +391,9 @@ template <std::size_t Arity>
 ArrayIterateResult IterateUsingSimplifiedLayout(
     const SimplifiedDimensionIterationOrder& layout,
     span<const Index> input_shape,
-    internal::ElementwiseClosure<Arity, Status*> closure, Status* status,
-    span<absl::optional<SingleArrayIterationState>, Arity> single_array_states,
+    internal::ElementwiseClosure<Arity, absl::Status*> closure,
+    absl::Status* status,
+    span<std::optional<SingleArrayIterationState>, Arity> single_array_states,
     std::array<std::ptrdiff_t, Arity> element_sizes);
 
 }  // namespace internal_index_space
