@@ -19,12 +19,15 @@
 #include <utility>
 
 #include "absl/container/fixed_array.h"
+#include "absl/status/status.h"
 #include "tensorstore/index_space/index_transform.h"
 #include "tensorstore/index_space/internal/transform_rep_impl.h"
+#include "tensorstore/internal/dimension_labels.h"
 #include "tensorstore/internal/integer_overflow.h"
 #include "tensorstore/internal/memory.h"
 #include "tensorstore/util/division.h"
 #include "tensorstore/util/quote_string.h"
+#include "tensorstore/util/str_cat.h"
 
 namespace tensorstore {
 namespace internal_index_space {
@@ -36,24 +39,22 @@ void FreeIndexArrayData(IndexArrayData* data) {
 }
 
 void CopyTrivialFields(TransformRep* source, TransformRep* dest) {
-  ABSL_ASSERT(dest->input_rank_capacity >= source->input_rank &&
-              dest->output_rank_capacity >= source->output_rank);
+  assert(dest->input_rank_capacity >= source->input_rank &&
+         dest->output_rank_capacity >= source->output_rank);
   const DimensionIndex input_rank = dest->input_rank = source->input_rank;
   dest->output_rank = source->output_rank;
   std::copy_n(source->input_origin().begin(), input_rank,
               dest->input_origin().begin());
   std::copy_n(source->input_shape().begin(), input_rank,
               dest->input_shape().begin());
-  dest->implicit_lower_bounds(input_rank)
-      .DeepAssign(source->implicit_lower_bounds(input_rank));
-  dest->implicit_upper_bounds(input_rank)
-      .DeepAssign(source->implicit_upper_bounds(input_rank));
+  dest->implicit_lower_bounds = source->implicit_lower_bounds;
+  dest->implicit_upper_bounds = source->implicit_upper_bounds;
 }
 
 }  // namespace
 
 void CopyInputLabels(TransformRep* source, TransformRep* dest, bool can_move) {
-  ABSL_ASSERT(dest->input_rank_capacity >= source->input_rank);
+  assert(dest->input_rank_capacity >= source->input_rank);
   const DimensionIndex input_rank = source->input_rank;
   if (can_move) {
     std::copy_n(std::make_move_iterator(source->input_labels().begin()),
@@ -116,7 +117,7 @@ IndexArrayData& OutputIndexMap::SetArrayIndexing(DimensionIndex rank) {
 
 IndexArrayData& OutputIndexMap::SetArrayIndexing(DimensionIndex rank,
                                                  const IndexArrayData& other) {
-  ABSL_ASSERT(other.rank_capacity >= rank);
+  assert(other.rank_capacity >= rank);
   auto& data = SetArrayIndexing(rank);
   data.element_pointer = other.element_pointer;
   data.index_range = other.index_range;
@@ -134,27 +135,16 @@ void OutputIndexMap::Assign(DimensionIndex rank, const OutputIndexMap& other) {
   stride_ = other.stride_;
 }
 
-// Singleton transform instance used when the input and output rank are both
-// zero.
-static TransformRep rank_zero_transform_data{0, 0, 0, 0, 1};
-
 TransformRep::Ptr<> TransformRep::Allocate(
     DimensionIndex input_rank_capacity, DimensionIndex output_rank_capacity) {
-  TENSORSTORE_CHECK(
-      input_rank_capacity >= 0 && output_rank_capacity >= 0 &&
-      input_rank_capacity <= std::numeric_limits<std::int32_t>::max() &&
-      output_rank_capacity <= std::numeric_limits<std::int32_t>::max());
-  if (input_rank_capacity == 0 && output_rank_capacity == 0) {
-    return TransformRep::Ptr<>(&rank_zero_transform_data);
-  }
+  TENSORSTORE_CHECK(input_rank_capacity >= 0 && output_rank_capacity >= 0 &&
+                    input_rank_capacity <= kMaxRank &&
+                    output_rank_capacity <= kMaxRank);
   const size_t total_size =
       // header size
       sizeof(TransformRep) +
       // size of OutputIndexMap array
       sizeof(OutputIndexMap) * output_rank_capacity +
-      // size of implicit_bitvector_storage
-      sizeof(std::uint64_t) * CeilOfRatio(input_rank_capacity * 2,
-                                          static_cast<DimensionIndex>(64)) +
       // size of input_origin, input_shape, and input_labels arrays
       input_rank_capacity * (sizeof(Index) * 2 + sizeof(std::string));
   char* base_ptr = static_cast<char*>(::operator new(total_size));
@@ -176,7 +166,7 @@ void DestroyLabelFields(TransformRep* ptr) {
 }
 
 void TransformRep::Free(TransformRep* ptr) {
-  ABSL_ASSERT(ptr->reference_count == 0);
+  assert(ptr->reference_count == 0);
   DestroyLabelFields(ptr);
   std::destroy_n(ptr->output_index_maps().begin(), ptr->output_rank_capacity);
   ::operator delete(static_cast<void*>(ptr->output_index_maps().data()));
@@ -206,10 +196,8 @@ void CopyTransformRepDomain(TransformRep* source, TransformRep* dest) {
               dest->input_origin().begin());
   std::copy_n(source->input_shape().begin(), input_rank,
               dest->input_shape().begin());
-  dest->implicit_lower_bounds(input_rank)
-      .DeepAssign(source->implicit_lower_bounds(input_rank));
-  dest->implicit_upper_bounds(input_rank)
-      .DeepAssign(source->implicit_upper_bounds(input_rank));
+  dest->implicit_lower_bounds = source->implicit_lower_bounds;
+  dest->implicit_upper_bounds = source->implicit_upper_bounds;
   std::copy_n(source->input_labels().begin(), input_rank,
               dest->input_labels().begin());
 }
@@ -221,24 +209,76 @@ void MoveTransformRep(TransformRep* source, TransformRep* dest) {
   CopyInputLabels(source, dest, /*can_move=*/true);
 }
 
-TransformRep::Ptr<> MutableRep(TransformRep::Ptr<> ptr) {
-  if (!ptr || ptr.get() == &rank_zero_transform_data || ptr->is_unique()) {
+void ResetOutputIndexMaps(TransformRep* ptr) {
+  auto output_index_maps = ptr->output_index_maps();
+  for (DimensionIndex output_dim = 0, output_rank = ptr->output_rank;
+       output_dim < output_rank; ++output_dim) {
+    output_index_maps[output_dim].SetConstant();
+  }
+  ptr->output_rank = 0;
+}
+
+TransformRep::Ptr<> MutableRep(TransformRep::Ptr<> ptr, bool domain_only) {
+  if (!ptr) return ptr;
+  if (ptr->is_unique()) {
+    if (domain_only) {
+      ResetOutputIndexMaps(ptr.get());
+      ptr->output_rank = 0;
+    }
     return ptr;
   }
-  auto new_rep = TransformRep::Allocate(ptr->input_rank, ptr->output_rank);
-  CopyTransformRep(ptr.get(), new_rep.get());
-  return new_rep;
+  if (domain_only) {
+    auto new_rep = TransformRep::Allocate(ptr->input_rank, 0);
+    CopyTransformRepDomain(ptr.get(), new_rep.get());
+    new_rep->output_rank = 0;
+    internal_index_space::DebugCheckInvariants(new_rep.get());
+    return new_rep;
+  } else {
+    auto new_rep = TransformRep::Allocate(ptr->input_rank, ptr->output_rank);
+    CopyTransformRep(ptr.get(), new_rep.get());
+    internal_index_space::DebugCheckInvariants(new_rep.get());
+    return new_rep;
+  }
 }
 
 TransformRep::Ptr<> NewOrMutableRep(TransformRep* ptr,
                                     DimensionIndex input_rank_capacity,
-                                    DimensionIndex output_rank_capacity) {
-  ABSL_ASSERT(ptr);
+                                    DimensionIndex output_rank_capacity,
+                                    bool domain_only) {
+  assert(ptr);
   if (ptr->input_rank_capacity >= input_rank_capacity &&
       ptr->output_rank_capacity >= output_rank_capacity && ptr->is_unique()) {
+    if (domain_only) {
+      ResetOutputIndexMaps(ptr);
+    }
     return TransformRep::Ptr<>(ptr);
   } else {
-    return TransformRep::Allocate(input_rank_capacity, output_rank_capacity);
+    return TransformRep::Allocate(input_rank_capacity,
+                                  domain_only ? 0 : output_rank_capacity);
+  }
+}
+
+bool IsDomainExplicitlyEmpty(TransformRep* ptr) {
+  DimensionSet implicit_dims = ptr->implicit_dimensions();
+
+  const Index* input_shape = ptr->input_shape().data();
+  for (DimensionIndex input_dim = 0, input_rank = ptr->input_rank;
+       input_dim < input_rank; ++input_dim) {
+    if (!implicit_dims[input_dim] && input_shape[input_dim] == 0) {
+      return true;
+    }
+  }
+  return false;
+}
+
+void ReplaceAllIndexArrayMapsWithConstantMaps(TransformRep* ptr) {
+  for (DimensionIndex output_dim = 0, output_rank = ptr->output_rank;
+       output_dim < output_rank; ++output_dim) {
+    auto& map = ptr->output_index_maps()[output_dim];
+    if (map.method() != OutputIndexMethod::array) continue;
+    map.SetConstant();
+    map.offset() = 0;
+    map.stride() = 0;
   }
 }
 
@@ -283,14 +323,8 @@ bool AreDomainsEqual(TransformRep* a, TransformRep* b) {
   const DimensionIndex input_rank = a->input_rank;
   const BoxView<> input_domain_a = a->input_domain(input_rank);
   if (input_domain_a != b->input_domain(input_rank)) return false;
-  auto implicit_lower_bounds_a = a->implicit_lower_bounds(input_rank);
-  auto implicit_upper_bounds_a = a->implicit_upper_bounds(input_rank);
-  if (!std::equal(implicit_lower_bounds_a.begin(),
-                  implicit_lower_bounds_a.end(),
-                  b->implicit_lower_bounds(input_rank).begin()) ||
-      !std::equal(implicit_upper_bounds_a.begin(),
-                  implicit_upper_bounds_a.end(),
-                  b->implicit_upper_bounds(input_rank).begin())) {
+  if (a->implicit_lower_bounds != b->implicit_lower_bounds ||
+      a->implicit_upper_bounds != b->implicit_upper_bounds) {
     return false;
   }
   span<const std::string> input_labels_a = a->input_labels().first(input_rank);
@@ -410,14 +444,14 @@ Result<Index> OutputIndexMap::operator()(
       break;
     case OutputIndexMethod::single_input_dimension: {
       const DimensionIndex input_dim = input_dimension();
-      ABSL_ASSERT(input_dim >= 0 && input_dim < input_indices.size());
+      assert(input_dim >= 0 && input_dim < input_indices.size());
       base_output_index = input_indices[input_dim];
       break;
     }
     case OutputIndexMethod::array: {
       const IndexArrayData& data = index_array_data();
-      ABSL_ASSERT(data.element_pointer &&
-                  input_indices.size() <= data.rank_capacity);
+      assert(data.element_pointer &&
+             input_indices.size() <= data.rank_capacity);
       base_output_index =
           data.element_pointer.byte_strided_pointer()[IndexInnerProduct(
               input_indices.size(), input_indices.data(), data.byte_strides)];
@@ -431,10 +465,11 @@ Result<Index> OutputIndexMap::operator()(
   return base_output_index * stride() + offset();
 }
 
-Status TransformIndices(TransformRep* data, span<const Index> input_indices,
-                        span<Index> output_indices) {
-  ABSL_ASSERT(data && data->input_rank == input_indices.size() &&
-              data->output_rank == output_indices.size());
+absl::Status TransformIndices(TransformRep* data,
+                              span<const Index> input_indices,
+                              span<Index> output_indices) {
+  assert(data && data->input_rank == input_indices.size() &&
+         data->output_rank == output_indices.size());
   const DimensionIndex output_rank = data->output_rank;
   const DimensionIndex input_rank = data->input_rank;
   span<const OutputIndexMap> output_index_maps =
@@ -457,9 +492,10 @@ Status TransformIndices(TransformRep* data, span<const Index> input_indices,
   return absl::OkStatus();
 }
 
-Status ReplaceZeroRankIndexArrayIndexMap(Index index, IndexInterval bounds,
-                                         Index* output_offset,
-                                         Index* output_stride) {
+absl::Status ReplaceZeroRankIndexArrayIndexMap(Index index,
+                                               IndexInterval bounds,
+                                               Index* output_offset,
+                                               Index* output_stride) {
   TENSORSTORE_RETURN_IF_ERROR(CheckContains(bounds, index));
   Index new_offset;
   if (internal::MulOverflow(index, *output_stride, &new_offset) ||
@@ -473,8 +509,8 @@ Status ReplaceZeroRankIndexArrayIndexMap(Index index, IndexInterval bounds,
 
 TransformRep::Ptr<> GetSubDomain(TransformRep* rep,
                                  span<const DimensionIndex> dims) {
-  ABSL_ASSERT(rep);
-  const DimensionIndex old_rank = rep->input_rank;
+  assert(rep);
+  [[maybe_unused]] const DimensionIndex old_rank = rep->input_rank;
   const DimensionIndex new_rank = dims.size();
   auto new_rep = TransformRep::Allocate(new_rank, 0);
   new_rep->output_rank = 0;
@@ -484,7 +520,7 @@ TransformRep::Ptr<> GetSubDomain(TransformRep* rep,
 #endif
   for (DimensionIndex new_dim = 0; new_dim < dims.size(); ++new_dim) {
     const DimensionIndex old_dim = dims[new_dim];
-    ABSL_ASSERT(old_dim >= 0 && old_dim < old_rank);
+    assert(old_dim >= 0 && old_dim < old_rank);
 #ifndef NDEBUG
     assert(!seen_dims[old_dim]);
     seen_dims[old_dim] = true;
@@ -494,28 +530,114 @@ TransformRep::Ptr<> GetSubDomain(TransformRep* rep,
   return new_rep;
 }
 
-Status ValidateLabelsAreUnique(span<const std::string> labels) {
-  // TODO(jbms): Consider using a hash set instead.
-  absl::FixedArray<absl::string_view, internal::kNumInlinedDims> sorted_labels(
-      labels.begin(), labels.end());
-  std::sort(sorted_labels.begin(), sorted_labels.end());
-  size_t i;
-  for (i = 1; i < sorted_labels.size() && sorted_labels[i].empty(); ++i)
-    continue;
-  for (; i < sorted_labels.size(); ++i) {
-    absl::string_view label = sorted_labels[i];
-    if (label == sorted_labels[i - 1]) {
-      return absl::InvalidArgumentError(
-          StrCat("Dimension label ", QuoteString(label), " is not unique"));
-    }
-  }
-  return absl::OkStatus();
-}
-
 bool IsUnlabeled(span<const std::string> labels) {
   return std::all_of(labels.begin(), labels.end(),
-                     [](absl::string_view s) { return s.empty(); });
+                     [](std::string_view s) { return s.empty(); });
 }
 
+DimensionSet GetIndexArrayInputDimensions(TransformRep* transform) {
+  DimensionSet set;
+  const DimensionIndex output_rank = transform->output_rank;
+  const DimensionIndex input_rank = transform->input_rank;
+  auto output_maps = transform->output_index_maps();
+  for (DimensionIndex output_dim = 0; output_dim < output_rank; ++output_dim) {
+    auto& map = output_maps[output_dim];
+    if (map.method() != OutputIndexMethod::array) continue;
+    const auto& index_array_data = map.index_array_data();
+    for (DimensionIndex input_dim = 0; input_dim < input_rank; ++input_dim) {
+      if (index_array_data.byte_strides[input_dim] != 0) {
+        set[input_dim] = true;
+      }
+    }
+  }
+  return set;
+}
+
+TransformRep::Ptr<> WithImplicitDimensions(TransformRep::Ptr<> transform,
+                                           DimensionSet implicit_lower_bounds,
+                                           DimensionSet implicit_upper_bounds,
+                                           bool domain_only) {
+  transform = MutableRep(std::move(transform), domain_only);
+  if (!domain_only && (implicit_lower_bounds || implicit_upper_bounds)) {
+    // Ensure all dimensions on which an index array depends remain explicit.
+    auto index_array_dims =
+        internal_index_space::GetIndexArrayInputDimensions(transform.get());
+    implicit_lower_bounds &= ~index_array_dims;
+    implicit_upper_bounds &= ~index_array_dims;
+  }
+  const auto mask = DimensionSet::UpTo(transform->input_rank);
+  transform->implicit_lower_bounds = implicit_lower_bounds & mask;
+  transform->implicit_upper_bounds = implicit_upper_bounds & mask;
+  return transform;
+}
+
+#ifndef NDEBUG
+void DebugCheckInvariants(TransformRep* rep) {
+  assert(rep);
+  assert(rep->reference_count > 0);
+  const DimensionIndex input_rank = rep->input_rank,
+                       output_rank = rep->output_rank;
+  assert(rep->input_rank_capacity <= kMaxRank);
+  assert(rep->output_rank_capacity <= kMaxRank);
+  assert(input_rank <= rep->input_rank_capacity);
+  assert(output_rank <= rep->output_rank_capacity);
+  assert(input_rank >= 0);
+  assert(output_rank >= 0);
+  const auto mask = DimensionSet::UpTo(rep->input_rank);
+  assert((rep->implicit_lower_bounds & mask) == rep->implicit_lower_bounds);
+  assert((rep->implicit_upper_bounds & mask) == rep->implicit_upper_bounds);
+  TENSORSTORE_CHECK_OK(internal::ValidateDimensionLabelsAreUnique(
+      rep->input_labels().first(input_rank)));
+  auto input_origin = rep->input_origin().data();
+  auto input_shape = rep->input_shape().data();
+  for (DimensionIndex input_dim = 0; input_dim < input_rank; ++input_dim) {
+    TENSORSTORE_CHECK_OK(
+        IndexInterval::Sized(input_origin[input_dim], input_shape[input_dim]));
+  }
+  const bool is_domain_explicitly_empty = IsDomainExplicitlyEmpty(rep);
+  const auto implicit_dims = rep->implicit_dimensions();
+  for (DimensionIndex output_dim = 0; output_dim < output_rank; ++output_dim) {
+    auto& map = rep->output_index_maps()[output_dim];
+    switch (map.method()) {
+      case OutputIndexMethod::constant: {
+        assert(map.stride() == 0);
+        break;
+      }
+      case OutputIndexMethod::single_input_dimension: {
+        const DimensionIndex input_dim = map.input_dimension();
+        assert(input_dim >= 0 && input_dim < input_rank);
+        assert(map.stride() != 0);
+        break;
+      }
+      case OutputIndexMethod::array: {
+        assert(map.stride() != 0);
+        const auto& index_array_data = map.index_array_data();
+        assert(index_array_data.rank_capacity >= input_rank);
+        assert(index_array_data.rank_capacity <= kMaxRank);
+        assert(!is_domain_explicitly_empty);
+        for (DimensionIndex input_dim = 0; input_dim < input_rank;
+             ++input_dim) {
+          const Index byte_stride = index_array_data.byte_strides[input_dim];
+          if (byte_stride == 0) continue;
+          const auto bounds = IndexInterval::UncheckedSized(
+              input_origin[input_dim], input_shape[input_dim]);
+          assert(IsFinite(bounds));
+          assert(!implicit_dims[input_dim]);
+        }
+        break;
+      }
+    }
+  }
+
+  // Validate that no index arrays remain allocated
+  for (DimensionIndex output_dim = output_rank,
+                      output_rank_capacity = rep->output_rank_capacity;
+       output_dim < output_rank_capacity; ++output_dim) {
+    assert(rep->output_index_maps()[output_dim].method() !=
+           OutputIndexMethod::array);
+  }
+}
+#endif
+
 }  // namespace internal_index_space
 }  // namespace tensorstore
