@@ -21,6 +21,7 @@
 #include <memory>
 #include <string>
 
+#include "absl/status/status.h"
 #include "tensorstore/array.h"
 #include "tensorstore/box.h"
 #include "tensorstore/index.h"
@@ -30,7 +31,7 @@
 #include "tensorstore/internal/intrusive_ptr.h"
 #include "tensorstore/internal/string_like.h"
 #include "tensorstore/rank.h"
-#include "tensorstore/util/bit_span.h"
+#include "tensorstore/util/dimension_set.h"
 #include "tensorstore/util/division.h"
 #include "tensorstore/util/element_pointer.h"
 #include "tensorstore/util/span.h"
@@ -41,6 +42,9 @@ template <DimensionIndex InputRank, DimensionIndex OutputRank,
           ContainerKind CKind>
 class IndexTransform;
 
+template <DimensionIndex Rank, ContainerKind CKind>
+class IndexDomain;
+
 namespace internal_index_space {
 
 struct IndexArrayData {
@@ -53,6 +57,24 @@ struct IndexArrayData {
   IndexInterval index_range;
   DimensionIndex rank_capacity;
   Index byte_strides[];
+
+  StridedLayoutView<dynamic_rank, offset_origin> layout(
+      BoxView<> input_domain) const {
+    assert(rank_capacity >= input_domain.rank());
+    return StridedLayoutView<dynamic_rank, offset_origin>(
+        input_domain.rank(), input_domain.origin().data(),
+        input_domain.shape().data(), byte_strides);
+  }
+
+  ArrayView<const Index, dynamic_rank, offset_origin> array_view(
+      BoxView<> input_domain) const {
+    return {element_pointer, layout(input_domain)};
+  }
+
+  SharedArrayView<const Index, dynamic_rank, offset_origin> shared_array_view(
+      BoxView<> input_domain) const {
+    return {element_pointer, layout(input_domain)};
+  }
 };
 // Alignment of IndexArrayData must be >= 2 to ensure that the low bit of
 // pointers to IndexArrayData is always 0.
@@ -96,23 +118,23 @@ class OutputIndexMap {
   }
 
   OutputIndexMethod method() const {
-    return value_ == 0 ? OutputIndexMethod::constant
-                       : value_ & 1 ? OutputIndexMethod::single_input_dimension
-                                    : OutputIndexMethod::array;
+    return value_ == 0  ? OutputIndexMethod::constant
+           : value_ & 1 ? OutputIndexMethod::single_input_dimension
+                        : OutputIndexMethod::array;
   }
 
   DimensionIndex input_dimension() const {
-    ABSL_ASSERT(method() == OutputIndexMethod::single_input_dimension);
+    assert(method() == OutputIndexMethod::single_input_dimension);
     return static_cast<DimensionIndex>(value_ >> 1);
   }
 
   const IndexArrayData& index_array_data() const {
-    ABSL_ASSERT(method() == OutputIndexMethod::array);
+    assert(method() == OutputIndexMethod::array);
     return *reinterpret_cast<IndexArrayData*>(value_);
   }
 
   IndexArrayData& index_array_data() {
-    ABSL_ASSERT(method() == OutputIndexMethod::array);
+    assert(method() == OutputIndexMethod::array);
     return *reinterpret_cast<IndexArrayData*>(value_);
   }
 
@@ -165,8 +187,6 @@ class InputDimensionRef;
 ///     TransformRep header;
 ///     Index input_origin[input_rank_capacity];
 ///     Index input_shape[input_rank_capacity];
-///     std::uint64_t implicit_bitvector_storage[
-///         CeilOfRatio(2 * input_rank_capacity, 64)];
 ///     std::string input_labels[input_rank_capacity];
 ///
 /// The representation is reference counted, and IndexTransform provide a safe,
@@ -184,23 +204,34 @@ class InputDimensionRef;
 struct TransformRep {
   /// The input rank.
   /// \invariant `0 <= input_rank && input_rank <= input_rank_capacity`.
-  std::int32_t input_rank;
+  std::int16_t input_rank;
 
   /// The output rank.
   /// \invariant `0 <= output_rank && output_rank <= output_rank_capacity`.
-  std::int32_t output_rank;
+  std::int16_t output_rank;
 
   /// The length of the `input_origin`, `input_shape`, and `input_labels` arrays
   /// that immediately follow this TransformRep header in memory.
   ///
-  /// \invariant `0 <= input_rank_capacity`.
-  std::int32_t input_rank_capacity;
+  /// \invariant `0 <= input_rank_capacity <= kMaxRank`.
+  std::int16_t input_rank_capacity;
 
   /// The length of the `output_index_maps` array that precedes this
   /// TransformRep header in memory.
   ///
-  /// \invariant `0 <= output_rank_capacity`.
-  std::int32_t output_rank_capacity;
+  /// \invariant `0 <= output_rank_capacity <= kMaxRank`.
+  std::int16_t output_rank_capacity;
+
+  /// Set of dimensions with implicit lower/upper bounds.
+  DimensionSet implicit_lower_bounds;
+  DimensionSet implicit_upper_bounds;
+
+  /// Dimensions for which the lower or upper bound is implicit.
+  DimensionSet implicit_dimensions() const {
+    return implicit_lower_bounds | implicit_upper_bounds;
+  }
+
+  static_assert(kMaxRank * 2 <= 64);
 
   /// Reference count.
   ///
@@ -245,40 +276,10 @@ struct TransformRep {
   ///
   /// \dchecks `0 <= rank && rank <= input_rank_capacity`.
   MutableBoxView<> input_domain(DimensionIndex rank) {
-    ABSL_ASSERT(0 <= rank && rank <= input_rank_capacity);
+    assert(0 <= rank && rank <= input_rank_capacity);
     return MutableBoxView<>(rank, input_origin().data(), input_shape().data());
   }
 
-  /// Returns the base pointer for the bit vector containing
-  /// `implicit_lower_bounds` and `implicit_upper_bounds`.
-  std::uint64_t* implicit_bitvector_base() {
-    return reinterpret_cast<std::uint64_t*>(input_shape().end());
-  }
-
-  /// Returns the span that stores the bit vector containing
-  /// `implicit_lower_bounds` and `implicit_upper_bounds`.
-  span<std::uint64_t> implicit_bitvector_storage() {
-    return {implicit_bitvector_base(),
-            CeilOfRatio(static_cast<DimensionIndex>(input_rank_capacity) * 2,
-                        static_cast<DimensionIndex>(64))};
-  }
-
-  /// Returns the `implicit_lower_bounds` of length `rank`.
-  ///
-  /// \dchecks `0 <= rank && rank <= input_rank_capacity`.
-  BitSpan<std::uint64_t> implicit_lower_bounds(DimensionIndex rank) {
-    ABSL_ASSERT(0 <= rank && rank <= input_rank_capacity);
-    return {implicit_bitvector_base(), 0, rank};
-  }
-
-  /// Returns the `implicit_upper_bounds` of length `rank`.
-  ///
-  /// \dchecks `0 <= rank && rank <= input_rank_capacity`.
-  BitSpan<std::uint64_t> implicit_upper_bounds(DimensionIndex rank) {
-    ABSL_ASSERT(0 <= rank && rank <= input_rank_capacity);
-    return {implicit_bitvector_base(), input_rank_capacity, rank};
-  }
-
   /// Returns the `output_index_maps` array of length `output_rank_capacity`.
   ///
   /// \invariant For `0 <= i && i < output_rank`:
@@ -298,9 +299,8 @@ struct TransformRep {
 
   /// Returns the `input_labels` array of length `input_rank_capacity`.
   span<std::string> input_labels() {
-    return span(
-        reinterpret_cast<std::string*>(implicit_bitvector_storage().end()),
-        input_rank_capacity);
+    return span(reinterpret_cast<std::string*>(input_shape().end()),
+                input_rank_capacity);
   }
 
   /// Free a TransformRep allocated by Allocate.
@@ -324,14 +324,15 @@ struct TransformRep {
   };
 
   template <ContainerKind CKind = container>
-  using Ptr = absl::conditional_t<
+  using Ptr = std::conditional_t<
       CKind == view, TransformRep*,
       internal::IntrusivePtr<TransformRep, IntrusivePtrTraits<>>>;
 
   /// Allocate a new TransformRep with the specified capacities and type.
   ///
-  /// Sets the `input_rank_capacity` and `output_rank_capacity` fields, but does
-  /// not initialize any other fields, including `input_rank` and `output_rank`.
+  /// Sets the `input_rank_capacity` and `output_rank_capacity` fields, and
+  /// zero-initializes `implicit_lower_bounds` and `implicit_lower_bounds`, but
+  /// does not initialize `input_rank` and `output_rank`.
   /// The `output_index_maps` array and `input_labels` array are default
   /// initialized.
   ///
@@ -346,6 +347,23 @@ struct TransformRep {
                         DimensionIndex output_rank_capacity);
 };
 
+#ifdef NDEBUG
+// Don't check invariants when not in debug mode.
+inline void DebugCheckInvariants(TransformRep* rep) {}
+#else
+/// Checks invariants of the representation in debug mode.
+///
+/// This must be called any time the internal representation of an index
+/// transform is manipulated directly.
+void DebugCheckInvariants(TransformRep* rep);
+#endif
+
+inline void NormalizeImplicitBounds(TransformRep& rep) {
+  const auto mask = DimensionSet::UpTo(rep.input_rank);
+  rep.implicit_lower_bounds &= mask;
+  rep.implicit_upper_bounds &= mask;
+}
+
 // Check that OutputIndexMap and std::string don't have a greater alignment
 // value than Index, as that would require more complicated logic for accessing
 // the variable length fields than is currently implemented.  In practice these
@@ -379,12 +397,12 @@ class InputDimensionRef {
     return {optionally_implicit_domain(), label()};
   }
 
-  BitRef<std::uint64_t> implicit_lower_bound() const {
-    return rep_->implicit_lower_bounds(rep_->input_rank_capacity)[input_dim_];
+  DimensionSet::reference implicit_lower_bound() const {
+    return rep_->implicit_lower_bounds[input_dim_];
   }
 
-  BitRef<std::uint64_t> implicit_upper_bound() const {
-    return rep_->implicit_upper_bounds(rep_->input_rank_capacity)[input_dim_];
+  DimensionSet::reference implicit_upper_bound() const {
+    return rep_->implicit_upper_bounds[input_dim_];
   }
 
   std::string& label() const { return rep_->input_labels()[input_dim_]; }
@@ -407,7 +425,7 @@ class InputDimensionRef {
     domain() = other.interval();
     implicit_lower_bound() = other.implicit_lower();
     implicit_upper_bound() = other.implicit_upper();
-    // TODO(jbms): simplify this once absl::string_view has been replaced by
+    // TODO(jbms): simplify this once std::string_view has been replaced by
     // std::string_view.
     label().assign(other.label().begin(), other.label().end());
     return *this;
@@ -435,7 +453,7 @@ class InputDimensionsView {
   DimensionIndex size() const { return size_; }
 
   InputDimensionRef operator[](DimensionIndex i) const {
-    ABSL_ASSERT(i >= 0 && i <= size_);
+    assert(i >= 0 && i <= size_);
     return InputDimensionRef(rep_, i);
   }
 
@@ -446,13 +464,13 @@ class InputDimensionsView {
 
 inline ABSL_ATTRIBUTE_ALWAYS_INLINE InputDimensionsView
 TransformRep::all_input_dimensions(DimensionIndex rank) {
-  ABSL_ASSERT(rank >= 0 && rank <= input_rank_capacity);
+  assert(rank >= 0 && rank <= input_rank_capacity);
   return InputDimensionsView(this, rank);
 }
 
 inline ABSL_ATTRIBUTE_ALWAYS_INLINE InputDimensionRef
 TransformRep::input_dimension(DimensionIndex i) {
-  ABSL_ASSERT(i >= 0 && i <= input_rank_capacity);
+  assert(i >= 0 && i <= input_rank_capacity);
   return InputDimensionRef(this, i);
 }
 
@@ -486,8 +504,16 @@ void MoveTransformRep(TransformRep* source, TransformRep* dest);
 /// If `ptr->reference_count == 1`, returns `ptr`.  Otherwise, returns a
 /// newly-allocated copy.
 ///
+/// \param ptr Existing transform.
+/// \param domain_only If true, the returned transform will have an output rank
+///     of 0.
 /// \dchecks `ptr != nullptr`
-TransformRep::Ptr<> MutableRep(TransformRep::Ptr<> ptr);
+TransformRep::Ptr<> MutableRep(TransformRep::Ptr<> ptr,
+                               bool domain_only = false);
+
+/// Resets all output index maps to constant maps, freeing any associated index
+/// arrays.
+void ResetOutputIndexMaps(TransformRep* ptr);
 
 /// Returns a transform with the specified capacities.
 ///
@@ -501,18 +527,30 @@ TransformRep::Ptr<> MutableRep(TransformRep::Ptr<> ptr);
 /// \param ptr Non-null pointer to an existing transform.
 /// \param input_rank_capacity Required input rank capacity.
 /// \param output_rank_capacity Required output rank capacity.
+/// \param domain_only If true, `output_rank_capacity` is ignored and taken to
+///     be 0 instead.  Additionally, if the existing `ptr` is returned, all of
+///     its output index maps are reset in order to free any associated index
+///     arrays.
 /// \dchecks `ptr != nullptr`
 /// \returns A non-null transform pointer.
 TransformRep::Ptr<> NewOrMutableRep(TransformRep* ptr,
                                     DimensionIndex input_rank_capacity,
-                                    DimensionIndex output_rank_capacity);
+                                    DimensionIndex output_rank_capacity,
+                                    bool domain_only = false);
 
-/// Validates that non-empty labels are unique.
+/// Returns `true` if at least one dimension has an explicit zero-size domain,
+/// i.e. the domain is empty and it cannot be resized to become non-empty.
 ///
-/// \param labels The sequence of labels to validate.
-/// \returns `Status()` if valid.
-/// \error `absl::StatusCode::kInvalidArgument` if there is a non-unique label.
-Status ValidateLabelsAreUnique(span<const std::string> labels);
+/// Transforms for which this is true must not contain any index arrays, as the
+/// index array output map representation does not support index arrays that
+/// contain zero elements.
+bool IsDomainExplicitlyEmpty(TransformRep* ptr);
+
+/// Replaces all index array maps with constant maps to 0.
+///
+/// This is used in the case that `IsDomainExplicitlyEmpty(ptr) == true` to
+/// ensure the invariant is satisfied.
+void ReplaceAllIndexArrayMapsWithConstantMaps(TransformRep* ptr);
 
 /// Checks if the domains of two transforms are equal.
 ///
@@ -542,6 +580,17 @@ void PrintToOstream(std::ostream& os, TransformRep* transform);
 /// \param transform Pointer to the representation, may be `nullptr`.
 void PrintDomainToOstream(std::ostream& os, TransformRep* transform);
 
+/// Returns the set of input dimensions on which any index array output index
+/// map depends.
+DimensionSet GetIndexArrayInputDimensions(TransformRep* transform);
+
+/// Returns a copy of `transform` with `implicit_lower_bounds` and
+/// `implicit_upper_bounds` set to the specified values.
+TransformRep::Ptr<> WithImplicitDimensions(TransformRep::Ptr<> transform,
+                                           DimensionSet implicit_lower_bounds,
+                                           DimensionSet implicit_upper_bounds,
+                                           bool domain_only);
+
 /// Computes the `output_indices` corresponding to the given `input_indices`.
 ///
 /// \dchecks `data != nullptr`.
@@ -552,8 +601,9 @@ void PrintDomainToOstream(std::ostream& os, TransformRep* transform);
 ///     within the domain (implicit bounds are ignored).
 /// \error `absl::StatusCode::kOutOfRange` if an array output index map results
 ///     in an index outside its `index_range` constraint.
-Status TransformIndices(TransformRep* data, span<const Index> input_indices,
-                        span<Index> output_indices);
+absl::Status TransformIndices(TransformRep* data,
+                              span<const Index> input_indices,
+                              span<Index> output_indices);
 
 /// Returns a transform with `input_rank==dims.size()` and `output_rank==0` in
 /// which dimension `i` has the domain of dimension `dims[i]` of `*rep`.
@@ -583,21 +633,30 @@ class TransformAccess {
   }
 
   template <ContainerKind TargetCKind, typename T>
-  static absl::enable_if_t<TargetCKind == view, TransformRep::Ptr<TargetCKind>>
-  rep_ptr(const T& x) {
-    return rep(x);
-  }
-
-  template <ContainerKind TargetCKind, typename T>
-  static absl::enable_if_t<TargetCKind == container,
-                           TransformRep::Ptr<TargetCKind>>
-  rep_ptr(T&& x) {
-    return TransformRep::Ptr<>(std::forward<T>(x).rep_);
+  static auto rep_ptr(T&& x) {
+    if constexpr (TargetCKind == view) {
+      return rep(x);
+    } else {
+      return TransformRep::Ptr<>(std::forward<T>(x).rep_);
+    }
   }
 
   template <typename T>
-  static auto transform(T&& x) -> decltype((std::declval<T>().transform_)) {
-    return std::forward<T>(x).transform_;
+  static decltype(auto) rep_ptr(T&& x) {
+    return (std::forward<T>(x).rep_);
+  }
+
+  template <DimensionIndex Rank, ContainerKind CKind>
+  static IndexTransform<Rank, dynamic_rank, view> transform(
+      const IndexDomain<Rank, CKind>& x) {
+    return Make<IndexTransform<Rank, dynamic_rank, view>>(rep(x));
+  }
+
+  template <DimensionIndex Rank>
+  static IndexTransform<Rank, dynamic_rank, container> transform(
+      IndexDomain<Rank, container>&& x) {
+    return Make<IndexTransform<Rank, dynamic_rank, container>>(
+        std::move(x.rep_));
   }
 
   template <typename T>
