@@ -12,11 +12,13 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-/// Tests for the DimExpression::{Translate,}{Closed,HalfOpen,Sized}Interval
+/// Tests for the
+/// DimExpression::{{Translate,}{Closed,HalfOpen,Sized}Interval,Stride}
 /// operations.
 
 #include <gmock/gmock.h>
 #include <gtest/gtest.h>
+#include "tensorstore/box.h"
 #include "tensorstore/index_space/dim_expression.h"
 #include "tensorstore/index_space/index_transform_builder.h"
 #include "tensorstore/index_space/internal/dim_expression_testutil.h"
@@ -24,27 +26,27 @@
 
 namespace {
 
-using tensorstore::AllDims;
-using tensorstore::Dims;
-using tensorstore::Index;
-using tensorstore::IndexTransformBuilder;
-using tensorstore::kImplicit;
-using tensorstore::kInfIndex;
-using tensorstore::kInfSize;
-using tensorstore::MakeArray;
-using tensorstore::StrCat;
-using tensorstore::internal_index_space::EquivalentIndices;
-using tensorstore::internal_index_space::TestDimExpression;
-using tensorstore::internal_index_space::TestDimExpressionError;
+using ::tensorstore::AllDims;
+using ::tensorstore::BoxView;
+using ::tensorstore::Dims;
+using ::tensorstore::Index;
+using ::tensorstore::IndexInterval;
+using ::tensorstore::IndexTransformBuilder;
+using ::tensorstore::kImplicit;
+using ::tensorstore::kInfIndex;
+using ::tensorstore::kInfSize;
+using ::tensorstore::MakeArray;
+using ::tensorstore::StrCat;
+using ::tensorstore::internal_index_space::EquivalentIndices;
+using ::tensorstore::internal_index_space::TestDimExpression;
+using ::tensorstore::internal_index_space::TestDimExpressionError;
 
 TEST(ClosedIntervalTest, Example) {
   const auto original_transform = IndexTransformBuilder<3, 3>()
                                       .input_origin({0, 2, 0})
                                       .input_shape({7, 4, 10})
                                       .input_labels({"x", "y", "z"})
-                                      .output_single_input_dimension(0, 0, 1, 0)
-                                      .output_single_input_dimension(1, 0, 1, 1)
-                                      .output_single_input_dimension(2, 0, 1, 2)
+                                      .output_identity_transform()
                                       .Finalize()
                                       .value();
   const auto expected_new_transform =
@@ -52,8 +54,8 @@ TEST(ClosedIntervalTest, Example) {
           .input_origin({1, 2, -4})
           .input_shape({4, 4, 3})
           .input_labels({"x", "y", "z"})
-          .output_single_input_dimension(0, 0, 1, 0)
-          .output_single_input_dimension(1, 0, 1, 1)
+          .output_single_input_dimension(0, 0)
+          .output_single_input_dimension(1, 1)
           .output_single_input_dimension(2, 0, -2, 2)
           .Finalize()
           .value();
@@ -83,9 +85,7 @@ TEST(ClosedIntervalTest, ExampleWithOffset) {
                                       .input_origin({0, 2, 0})
                                       .input_shape({7, 4, 10})
                                       .input_labels({"x", "y", "z"})
-                                      .output_single_input_dimension(0, 0, 1, 0)
-                                      .output_single_input_dimension(1, 0, 1, 1)
-                                      .output_single_input_dimension(2, 0, 1, 2)
+                                      .output_identity_transform()
                                       .Finalize()
                                       .value();
   const auto expected_new_transform =
@@ -93,8 +93,8 @@ TEST(ClosedIntervalTest, ExampleWithOffset) {
           .input_origin({1, 2, -4})
           .input_shape({4, 4, 4})
           .input_labels({"x", "y", "z"})
-          .output_single_input_dimension(0, 0, 1, 0)
-          .output_single_input_dimension(1, 0, 1, 1)
+          .output_single_input_dimension(0, 0)
+          .output_single_input_dimension(1, 1)
           .output_single_input_dimension(2, 1, -2, 2)
           .Finalize()
           .value();
@@ -116,8 +116,8 @@ TEST(HalfOpenIntervalTest, Example) {
           .input_origin({1, 2, -4})
           .input_shape({3, 4, 3})
           .input_labels({"x", "y", "z"})
-          .output_single_input_dimension(0, 0, 1, 0)
-          .output_single_input_dimension(1, 0, 1, 1)
+          .output_single_input_dimension(0, 0)
+          .output_single_input_dimension(1, 1)
           .output_single_input_dimension(2, 0, -2, 2)
           .Finalize()
           .value();
@@ -126,9 +126,7 @@ TEST(HalfOpenIntervalTest, Example) {
           .input_origin({0, 2, 0})
           .input_shape({7, 4, 10})
           .input_labels({"x", "y", "z"})
-          .output_single_input_dimension(0, 0, 1, 0)
-          .output_single_input_dimension(1, 0, 1, 1)
-          .output_single_input_dimension(2, 0, 1, 2)
+          .output_identity_transform()
           .Finalize()
           .value(),
       /*expression=*/Dims(0, 2).HalfOpenInterval({1, 8}, {4, 3}, {1, -2}),
@@ -144,8 +142,8 @@ TEST(SizedIntervalTest, Example) {
           .input_origin({1, 2, -4})
           .input_shape({3, 4, 2})
           .input_labels({"x", "y", "z"})
-          .output_single_input_dimension(0, 0, 1, 0)
-          .output_single_input_dimension(1, 0, 1, 1)
+          .output_single_input_dimension(0, 0)
+          .output_single_input_dimension(1, 1)
           .output_single_input_dimension(2, 0, -2, 2)
           .Finalize()
           .value();
@@ -154,9 +152,7 @@ TEST(SizedIntervalTest, Example) {
           .input_origin({0, 2, 0})
           .input_shape({7, 4, 10})
           .input_labels({"x", "y", "z"})
-          .output_single_input_dimension(0, 0, 1, 0)
-          .output_single_input_dimension(1, 0, 1, 1)
-          .output_single_input_dimension(2, 0, 1, 2)
+          .output_identity_transform()
           .Finalize()
           .value(),
       /*expression=*/Dims(0, 2).SizedInterval({1, 8}, {3, 2}, {1, -2}),
@@ -179,7 +175,7 @@ TEST(ClosedIntervalTest, OneDimensionalConstantNonStrided) {
                     IndexTransformBuilder<1, 1>()
                         .input_origin({-3})
                         .input_shape({8})
-                        .output_single_input_dimension(0, 0, 1, 0)
+                        .output_identity_transform()
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -257,7 +253,7 @@ TEST(ClosedIntervalTest, OneDimensionalSingleInputDimensionNonStrided) {
                     IndexTransformBuilder<1, 1>()
                         .input_origin({-3})
                         .input_shape({8})
-                        .output_single_input_dimension(0, 0, 1, 0)
+                        .output_identity_transform()
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -310,7 +306,7 @@ TEST(ClosedIntervalTest, OneDimensionalArrayNonStrided) {
       IndexTransformBuilder<1, 1>()
           .input_origin({-1})
           .input_shape({3})
-          .output_single_input_dimension(0, 0, 1, 0)
+          .output_identity_transform()
           .Finalize()
           .value(),
       /*expected_new_transform=*/
@@ -323,6 +319,73 @@ TEST(ClosedIntervalTest, OneDimensionalArrayNonStrided) {
       /*equivalent_indices=*/{{{1}, {1}}});
 }
 
+TEST(ClosedIntervalTest, OneDimensionalArrayNonStridedZeroElements) {
+  TestDimExpression(
+      /*original_transform=*/IndexTransformBuilder<1, 1>()
+          .input_origin({-2})
+          .input_shape({4})
+          .output_index_array(0, 3, 2, MakeArray<Index>({6, 5, 4, 3}))
+          .Finalize()
+          .value(),
+      /*expression=*/AllDims().ClosedInterval(-1, -2),
+      /*expected_new_dimension_selection=*/{0},
+      /*expected_identity_new_transform=*/
+      IndexTransformBuilder<1, 1>()
+          .input_origin({-1})
+          .input_shape({0})
+          .output_identity_transform()
+          .Finalize()
+          .value(),
+      /*expected_new_transform=*/
+      IndexTransformBuilder<1, 1>()
+          .input_origin({-1})
+          .input_shape({0})
+          .output_constant(0, 0)
+          .Finalize()
+          .value(),
+      /*equivalent_indices=*/{});
+}
+
+TEST(ClosedIntervalTest, OneDimensionalArrayNonStridedOneElement) {
+  TestDimExpression(
+      /*original_transform=*/IndexTransformBuilder<1, 1>()
+          .input_origin({-2})
+          .input_shape({4})
+          .output_index_array(0, 3, 2, MakeArray<Index>({6, 5, 4, 3}))
+          .Finalize()
+          .value(),
+      /*expression=*/AllDims().ClosedInterval(-1, -1),
+      /*expected_new_dimension_selection=*/{0},
+      /*expected_identity_new_transform=*/
+      IndexTransformBuilder<1, 1>()
+          .input_origin({-1})
+          .input_shape({1})
+          .output_identity_transform()
+          .Finalize()
+          .value(),
+      /*expected_new_transform=*/
+      IndexTransformBuilder<1, 1>()
+          .input_origin({-1})
+          .input_shape({1})
+          .output_constant(0, 13)
+          .Finalize()
+          .value(),
+      /*equivalent_indices=*/{{{-1}, {-1}}});
+}
+
+TEST(ClosedIntervalTest, OneDimensionalArrayNonStridedInvalidOneElement) {
+  TestDimExpressionError(
+      IndexTransformBuilder<1, 1>()
+          .input_origin({-2})
+          .input_shape({4})
+          .output_index_array(0, 3, 2, MakeArray<Index>({6, 5, 4, 3}),
+                              IndexInterval::Closed(3, 4))
+          .Finalize()
+          .value(),
+      AllDims().ClosedInterval(-1, -1), absl::StatusCode::kOutOfRange,
+      "Index 5 is outside valid range \\[3, 5\\)");
+}
+
 TEST(SliceTranslateClosedIntervalTest, OneDimensionalArrayNonStrided) {
   TestDimExpression(
       /*original_transform=*/IndexTransformBuilder<1, 1>()
@@ -401,10 +464,7 @@ TEST(ClosedIntervalTest, DimSubset) {
                     IndexTransformBuilder<4, 4>()
                         .input_origin({-5, 2, 2, -kInfIndex})
                         .input_shape({16, 2, 3, kInfIndex + 7})
-                        .output_single_input_dimension(0, 0, 1, 0)
-                        .output_single_input_dimension(1, 0, 1, 1)
-                        .output_single_input_dimension(2, 0, 1, 2)
-                        .output_single_input_dimension(3, 0, 1, 3)
+                        .output_identity_transform()
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -482,7 +542,7 @@ TEST(SliceClosedIntervalTest, UnboundedStart) {
                     IndexTransformBuilder<1, 1>()
                         .input_origin({5})
                         .input_shape({5})
-                        .output_single_input_dimension(0, 0, 1, 0)
+                        .output_identity_transform()
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -530,7 +590,7 @@ TEST(SliceHalfOpenIntervalTest, OneDimensionalUnstrided) {
                     IndexTransformBuilder<1, 1>()
                         .input_origin({6})
                         .input_shape({4})
-                        .output_single_input_dimension(0, 0, 1, 0)
+                        .output_identity_transform()
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -554,7 +614,7 @@ TEST(SliceHalfOpenIntervalTest, OneDimensionalUnstridedUnboundedStart) {
                     IndexTransformBuilder<1, 1>()
                         .input_origin({5})
                         .input_shape({5})
-                        .output_single_input_dimension(0, 0, 1, 0)
+                        .output_identity_transform()
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -578,7 +638,7 @@ TEST(SliceHalfOpenIntervalTest, OneDimensionalUnstridedUnboundedStop) {
                     IndexTransformBuilder<1, 1>()
                         .input_origin({6})
                         .input_shape({9})
-                        .output_single_input_dimension(0, 0, 1, 0)
+                        .output_identity_transform()
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -619,7 +679,7 @@ TEST(SliceHalfOpenIntervalTest, ErrorHandling) {
       IndexTransformBuilder<1, 0>().Finalize().value(),
       Dims(0).HalfOpenInterval(6, std::numeric_limits<Index>::min() + 1),
       absl::StatusCode::kInvalidArgument,
-      StrCat(".* do not specify a valid closed index interval\\."));
+      StrCat(".* do not specify a valid closed index interval"));
 }
 
 TEST(SliceClosedIntervalTest, ErrorHandling) {
@@ -630,7 +690,7 @@ TEST(SliceClosedIntervalTest, ErrorHandling) {
                              .value(),
                          Dims(0).ClosedInterval(6, 10, 0),
                          absl::StatusCode::kInvalidArgument,
-                         ".*Invalid stride 0\\.");
+                         ".*Invalid stride 0");
 
   TestDimExpressionError(
       IndexTransformBuilder<1, 0>()
@@ -639,13 +699,13 @@ TEST(SliceClosedIntervalTest, ErrorHandling) {
           .Finalize()
           .value(),
       Dims(0).ClosedInterval(6, 4), absl::StatusCode::kInvalidArgument,
-      ".*\\(6, 4\\) do not specify a valid closed index interval\\.");
+      ".*\\(6, 4\\) do not specify a valid closed index interval");
 
   TestDimExpressionError(
       IndexTransformBuilder<1, 0>().input_shape({10}).Finalize().value(),
       Dims(0).ClosedInterval(-kInfIndex, 4, 2),
       absl::StatusCode::kInvalidArgument,
-      ".*Slicing with non-unit stride of 2 requires a finite start index\\.");
+      ".*Slicing with non-unit stride of 2 requires a finite start index");
 
   TestDimExpressionError(
       IndexTransformBuilder<1, 0>()
@@ -655,7 +715,7 @@ TEST(SliceClosedIntervalTest, ErrorHandling) {
           .value(),
       Dims(0).ClosedInterval(kInfIndex, 4, -2),
       absl::StatusCode::kInvalidArgument,
-      ".*Slicing with non-unit stride of -2 requires a finite start index\\.");
+      ".*Slicing with non-unit stride of -2 requires a finite start index");
 
   TestDimExpressionError(IndexTransformBuilder<1, 0>()
                              .input_origin({5})
@@ -665,7 +725,7 @@ TEST(SliceClosedIntervalTest, ErrorHandling) {
                          Dims(0).ClosedInterval(6, 15),
                          absl::StatusCode::kOutOfRange,
                          ".*Slice interval \\[6, 16\\) is not "
-                         "contained within domain \\[5, 15\\)\\.");
+                         "contained within domain \\[5, 15\\)");
 
   TestDimExpressionError(
       IndexTransformBuilder<1, 1>()
@@ -676,7 +736,7 @@ TEST(SliceClosedIntervalTest, ErrorHandling) {
           .Finalize()
           .value(),
       Dims(0).ClosedInterval(5, 10, 3), absl::StatusCode::kInvalidArgument,
-      "Integer overflow computing offset for output dimension 0\\.");
+      "Integer overflow computing offset for output dimension 0");
 
   TestDimExpressionError(
       IndexTransformBuilder<1, 1>()
@@ -687,7 +747,7 @@ TEST(SliceClosedIntervalTest, ErrorHandling) {
           .Finalize()
           .value(),
       Dims(0).ClosedInterval(5, 10, 3), absl::StatusCode::kInvalidArgument,
-      "Integer overflow computing offset for output dimension 0\\.");
+      "Integer overflow computing offset for output dimension 0");
 
   TestDimExpressionError(
       IndexTransformBuilder<1, 1>()
@@ -698,14 +758,14 @@ TEST(SliceClosedIntervalTest, ErrorHandling) {
           .Finalize()
           .value(),
       Dims(0).ClosedInterval(5, 10, 2), absl::StatusCode::kInvalidArgument,
-      "Integer overflow computing stride for output dimension 0\\.");
+      "Integer overflow computing stride for output dimension 0");
 }
 
 TEST(SliceTranslateClosedIntervalTest, ErrorHandling) {
   TestDimExpressionError(IndexTransformBuilder<1, 0>().Finalize().value(),
                          Dims(0).TranslateClosedInterval(-kInfIndex, 100),
                          absl::StatusCode::kInvalidArgument,
-                         ".*Interval \\(-inf, 101\\) is not bounded below\\.");
+                         ".*Interval \\(-inf, 101\\) is not bounded below");
 }
 
 TEST(SliceSizedIntervalTest, ErrorHandling) {
@@ -716,15 +776,15 @@ TEST(SliceSizedIntervalTest, ErrorHandling) {
                              .value(),
                          Dims(0).SizedInterval(6, -2),
                          absl::StatusCode::kInvalidArgument,
-                         ".*Negative size -2 specified for sized interval\\.");
+                         ".*Negative size -2 specified for sized interval");
   TestDimExpressionError(IndexTransformBuilder<1, 0>().Finalize().value(),
                          Dims(0).SizedInterval(6, kInfIndex - 1, 100),
                          absl::StatusCode::kOutOfRange,
-                         ".*Integer overflow computing slice result\\.");
+                         ".*Integer overflow computing slice result");
   TestDimExpressionError(IndexTransformBuilder<1, 0>().Finalize().value(),
                          Dims(0).SizedInterval(6, kInfSize - 1),
                          absl::StatusCode::kOutOfRange,
-                         ".*Integer overflow computing slice result\\.");
+                         ".*Integer overflow computing slice result");
 }
 
 TEST(SliceSizedIntervalTest, OneDimensionalUnstrided) {
@@ -739,7 +799,7 @@ TEST(SliceSizedIntervalTest, OneDimensionalUnstrided) {
                     IndexTransformBuilder<1, 1>()
                         .input_origin({6})
                         .input_shape({3})
-                        .output_single_input_dimension(0, 0, 1, 0)
+                        .output_identity_transform()
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -763,7 +823,7 @@ TEST(SliceSizedIntervalTest, OneDimensionalUnstridedUnboundedMin) {
                     IndexTransformBuilder<1, 1>()
                         .input_origin({5})
                         .input_shape({3})
-                        .output_single_input_dimension(0, 0, 1, 0)
+                        .output_identity_transform()
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -787,7 +847,7 @@ TEST(SliceSizedIntervalTest, OneDimensionalUnstridedUnboundedSize) {
                     IndexTransformBuilder<1, 1>()
                         .input_origin({6})
                         .input_shape({9})
-                        .output_single_input_dimension(0, 0, 1, 0)
+                        .output_identity_transform()
                         .Finalize()
                         .value(),
                     /*expected_new_transform=*/
@@ -917,4 +977,137 @@ TEST(SliceSizedIntervalTest, OneDimensionalNegativeStridedUnboundedSize) {
                     /*equivalent_indices=*/{});
 }
 
+TEST(StrideTest, Example) {
+  const auto original_transform = IndexTransformBuilder<3, 3>()
+                                      .input_origin({0, 2, 1})
+                                      .input_inclusive_max({6, 5, 8})
+                                      .input_labels({"x", "y", "z"})
+                                      .output_identity_transform()
+                                      .Finalize()
+                                      .value();
+  const auto expected_new_transform =
+      IndexTransformBuilder<3, 3>()
+          .input_origin({-3, 2, 1})
+          .input_inclusive_max({0, 5, 2})
+          .input_labels({"x", "y", "z"})
+          .output_single_input_dimension(0, 0, -2, 0)
+          .output_single_input_dimension(1, 1)
+          .output_single_input_dimension(2, 0, 3, 2)
+          .Finalize()
+          .value();
+  const EquivalentIndices equivalent_indices = {
+      {{4, 3, 3}, {-2, 3, 1}},
+  };
+  TestDimExpression(
+      /*original_transform=*/original_transform,
+      /*expression=*/Dims(0, 2).Stride({-2, 3}),
+      /*expected_new_dimension_selection=*/{0, 2},
+      /*expected_identity_new_transform=*/expected_new_transform,
+      /*expected_new_transform=*/expected_new_transform,
+      /*equivalent_indices=*/equivalent_indices);
+
+  // Test using labels to select dimensions.
+  TestDimExpression(
+      /*original_transform=*/original_transform,
+      /*expression=*/Dims("x", "z").Stride({-2, 3}),
+      /*expected_new_dimension_selection=*/{0, 2},
+      /*expected_identity_new_transform=*/expected_new_transform,
+      /*expected_new_transform=*/expected_new_transform,
+      /*equivalent_indices=*/equivalent_indices);
+}
+
+TEST(StrideTest, ErrorZeroStride) {
+  TestDimExpressionError(
+      IndexTransformBuilder<1, 0>().Finalize().value(), Dims(0).Stride(0),
+      absl::StatusCode::kInvalidArgument,
+      StrCat("Applying stride to input dimension 0: Stride must be non-zero"));
+}
+
+TEST(StrideTest, ErrorStrideOverflow) {
+  TestDimExpressionError(
+      IndexTransformBuilder<1, 1>()
+          .output_single_input_dimension(0, 0,
+                                         std::numeric_limits<Index>::min(), 0)
+          .Finalize()
+          .value(),
+      Dims(0).Stride(std::numeric_limits<Index>::min()),
+      absl::StatusCode::kInvalidArgument,
+      StrCat("Integer overflow computing stride for output dimension 0"));
+}
+
+TEST(BoxSliceTest, Example) {
+  const auto original_transform = IndexTransformBuilder<3, 3>()
+                                      .input_origin({0, 2, 0})
+                                      .input_inclusive_max({6, 5, 9})
+                                      .input_labels({"x", "y", "z"})
+                                      .output_identity_transform()
+                                      .Finalize()
+                                      .value();
+  const auto expected_new_transform = IndexTransformBuilder<3, 3>()
+                                          .input_origin({1, 2, 4})
+                                          .input_inclusive_max({3, 5, 7})
+                                          .input_labels({"x", "y", "z"})
+                                          .output_identity_transform()
+                                          .Finalize()
+                                          .value();
+  const EquivalentIndices equivalent_indices = {
+      {{1, 3, 4}, {1, 3, 4}},
+  };
+  TestDimExpression(
+      /*original_transform=*/original_transform,
+      /*expression=*/Dims(0, 2).BoxSlice(BoxView({1, 4}, {3, 4})),
+      /*expected_new_dimension_selection=*/{0, 2},
+      /*expected_identity_new_transform=*/expected_new_transform,
+      /*expected_new_transform=*/expected_new_transform,
+      /*equivalent_indices=*/equivalent_indices);
+
+  // Test using labels to select dimensions.
+  TestDimExpression(
+      /*original_transform=*/original_transform,
+      /*expression=*/Dims("x", "z").BoxSlice(BoxView({1, 4}, {3, 4})),
+      /*expected_new_dimension_selection=*/{0, 2},
+      /*expected_identity_new_transform=*/expected_new_transform,
+      /*expected_new_transform=*/expected_new_transform,
+      /*equivalent_indices=*/equivalent_indices);
+}
+
+TEST(TranslateBoxSliceTest, Example) {
+  const auto original_transform = IndexTransformBuilder<3, 3>()
+                                      .input_origin({0, 2, 0})
+                                      .input_inclusive_max({6, 5, 9})
+                                      .input_labels({"x", "y", "z"})
+                                      .output_identity_transform()
+                                      .Finalize()
+                                      .value();
+  const auto expected_new_transform =
+      IndexTransformBuilder<3, 3>()
+          .input_origin({0, 2, 0})
+          .input_inclusive_max({2, 5, 3})
+          .input_labels({"x", "y", "z"})
+          .output_single_input_dimension(0, 1, 1, 0)
+          .output_single_input_dimension(1, 1)
+          .output_single_input_dimension(2, 4, 1, 2)
+          .Finalize()
+          .value();
+  const EquivalentIndices equivalent_indices = {
+      {{1, 3, 4}, {0, 3, 0}},
+  };
+  TestDimExpression(
+      /*original_transform=*/original_transform,
+      /*expression=*/Dims(0, 2).TranslateBoxSlice(BoxView({1, 4}, {3, 4})),
+      /*expected_new_dimension_selection=*/{0, 2},
+      /*expected_identity_new_transform=*/expected_new_transform,
+      /*expected_new_transform=*/expected_new_transform,
+      /*equivalent_indices=*/equivalent_indices);
+
+  // Test using labels to select dimensions.
+  TestDimExpression(
+      /*original_transform=*/original_transform,
+      /*expression=*/Dims("x", "z").TranslateBoxSlice(BoxView({1, 4}, {3, 4})),
+      /*expected_new_dimension_selection=*/{0, 2},
+      /*expected_identity_new_transform=*/expected_new_transform,
+      /*expected_new_transform=*/expected_new_transform,
+      /*equivalent_indices=*/equivalent_indices);
+}
+
 }  // namespace
