@@ -17,26 +17,29 @@
 #include <gmock/gmock.h>
 #include <gtest/gtest.h>
 #include "tensorstore/index_space/dim_expression.h"
+#include "tensorstore/index_space/index_domain_builder.h"
 #include "tensorstore/index_space/index_transform_builder.h"
-#include "tensorstore/internal/json.h"
+#include "tensorstore/internal/json_binding/gtest.h"
+#include "tensorstore/internal/json_binding/json_binding.h"
 #include "tensorstore/internal/json_gtest.h"
-#include "tensorstore/util/result_util.h"
+#include "tensorstore/util/result.h"
 #include "tensorstore/util/status.h"
 #include "tensorstore/util/status_testutil.h"
 
 namespace {
 
-using tensorstore::DimensionIndex;
-using tensorstore::dynamic_rank;
-using tensorstore::Index;
-using tensorstore::IndexTransform;
-using tensorstore::IndexTransformBuilder;
-using tensorstore::IndexTransformSpec;
-using tensorstore::kInfIndex;
-using tensorstore::kInfSize;
-using tensorstore::MatchesStatus;
-using tensorstore::Result;
-using tensorstore::internal::ParseJson;
+using ::tensorstore::DimensionIndex;
+using ::tensorstore::dynamic_rank;
+using ::tensorstore::Index;
+using ::tensorstore::IndexInterval;
+using ::tensorstore::IndexTransform;
+using ::tensorstore::IndexTransformBuilder;
+using ::tensorstore::kInfIndex;
+using ::tensorstore::kInfSize;
+using ::tensorstore::MatchesJson;
+using ::tensorstore::MatchesStatus;
+using ::tensorstore::Result;
+using ::tensorstore::internal::ParseJson;
 
 IndexTransform<> MakeExampleTransform() {
   return tensorstore::IndexTransformBuilder<4, 3>()
@@ -70,7 +73,8 @@ IndexTransform<> MakeUnlabeledExampleTransform() {
                               {{1}},
                               {{2}},
                               {{3}},
-                          }}))
+                          }}),
+                          IndexInterval::Closed(1, 2))
       .Finalize()
       .value();
 }
@@ -83,7 +87,11 @@ IndexTransform<> MakeUnlabeledExampleTransform() {
     "output": [
         {"offset": 3},
         {"stride": 2, "input_dimension": 2},
-        {"offset": 7, "index_array": [[ [[1]], [[2]], [[3]] ]]}
+        {
+            "offset": 7,
+            "index_array": [[ [[1]], [[2]], [[3]] ]],
+            "index_array_bounds": [1, 2]
+        }
     ]
 }
 )");
@@ -113,6 +121,78 @@ TEST(ToJsonTest, Labeled) {
   EXPECT_EQ(MakeLabeledExampleJson(), ::nlohmann::json(MakeExampleTransform()));
 }
 
+TEST(IndexTransformJsonBinderTest, IndexArrayOutOfBounds) {
+  tensorstore::TestJsonBinderRoundTrip<IndexTransform<>>({
+      // Index array does not contain out-of-bounds index.
+      {IndexTransformBuilder(1, 1)
+           .input_shape({3})
+           .output_index_array(0, 0, 1,
+                               tensorstore::MakeArray<Index>({1, 2, 3}))
+           .Finalize()
+           .value(),
+       {
+           {"input_inclusive_min", {0}},
+           {"input_exclusive_max", {3}},
+           {"output",
+            {
+                {{"index_array", {1, 2, 3}}},
+            }},
+       }},
+      // Index array contains out-of-bounds index and `index_range` is
+      // bounded.
+      {IndexTransformBuilder(1, 1)
+           .input_shape({3})
+           .output_index_array(0, 0, 1,
+                               tensorstore::MakeArray<Index>({1, 2, 3}),
+                               IndexInterval::UncheckedClosed(1, 2))
+           .Finalize()
+           .value(),
+       {
+           {"input_inclusive_min", {0}},
+           {"input_exclusive_max", {3}},
+           {"output",
+            {
+                {{"index_array", {1, 2, 3}}, {"index_array_bounds", {1, 2}}},
+            }},
+       }},
+      // Index array contains out-of-bounds index, but `index_range` is
+      // unbounded anyway.
+      {IndexTransformBuilder(1, 1)
+           .input_shape({3})
+           .output_index_array(
+               0, 0, 1, tensorstore::MakeArray<Index>({1, kInfIndex + 1, 3}))
+           .Finalize()
+           .value(),
+       {
+           {"input_inclusive_min", {0}},
+           {"input_exclusive_max", {3}},
+           {"output",
+            {
+                {{"index_array", {1, kInfIndex + 1, 3}}},
+            }},
+       }},
+  });
+  tensorstore::TestJsonBinderToJson<IndexTransform<>>({
+      // Because index array does not contain an out-of-bounds index, the
+      // `index_range` of `[1, 3]` is not encoded.
+      {IndexTransformBuilder(1, 1)
+           .input_shape({3})
+           .output_index_array(0, 0, 1,
+                               tensorstore::MakeArray<Index>({1, 2, 3}),
+                               IndexInterval::Closed(1, 3))
+           .Finalize()
+           .value(),
+       ::testing::Optional(MatchesJson(::nlohmann::json{
+           {"input_inclusive_min", {0}},
+           {"input_exclusive_max", {3}},
+           {"output",
+            {
+                {{"index_array", {1, 2, 3}}},
+            }},
+       }))},
+  });
+}
+
 TEST(ToJsonTest, NullTransform) {
   EXPECT_TRUE(::nlohmann::json(tensorstore::IndexTransform<>()).is_discarded());
 }
@@ -150,7 +230,7 @@ TEST(ToJsonTest, Translation) {
 void TestRoundTripJson(const ::nlohmann::json& json) {
   SCOPED_TRACE(json.dump());
   auto parse_result = tensorstore::ParseIndexTransform(json);
-  ASSERT_EQ(tensorstore::Status(), GetStatus(parse_result));
+  ASSERT_EQ(absl::OkStatus(), GetStatus(parse_result));
   EXPECT_EQ(json, ::nlohmann::json(*parse_result));
 }
 
@@ -158,11 +238,42 @@ TEST(RoundTripJsonTest, Labels) {
   TestRoundTripJson({
       {"input_inclusive_min", {1}},
       {"input_exclusive_max", {3}},
-      {"input_labels", {"x"}},
   });
   TestRoundTripJson({
       {"input_inclusive_min", {1}},
       {"input_exclusive_max", {3}},
+      {"input_labels", {"x"}},
+  });
+}
+
+TEST(RoundTripJsonTest, Rank0) {
+  TestRoundTripJson({
+      {"input_rank", 0},
+  });
+}
+
+TEST(RoundTripJsonTest, Input1Output0) {
+  TestRoundTripJson({
+      {"input_rank", 1},
+      {"output", ::nlohmann::json::array_t()},
+  });
+}
+
+TEST(RoundTripJsonTest, LabelsOnly) {
+  TestRoundTripJson({
+      {"input_labels", {"x", "y", "z"}},
+  });
+}
+
+TEST(RoundTripJsonTest, MinOnlyNotImplicit) {
+  TestRoundTripJson({
+      {"input_inclusive_min", {"-inf"}},
+  });
+}
+
+TEST(RoundTripJsonTest, MaxOnlyNotImplicit) {
+  TestRoundTripJson({
+      {"input_exclusive_max", {"+inf"}},
   });
 }
 
@@ -184,9 +295,8 @@ TEST(ParseIndexTransformTest, DynamicFromUnlabeled) {
 
 TEST(ParseIndexTransformTest, Static) {
   auto t = tensorstore::ParseIndexTransform<4, 3>(MakeLabeledExampleJson());
-  static_assert(std::is_same<decltype(t),
-                             Result<tensorstore::IndexTransform<4, 3>>>::value,
-                "");
+  static_assert(
+      std::is_same_v<decltype(t), Result<tensorstore::IndexTransform<4, 3>>>);
   EXPECT_EQ(MakeExampleTransform(), t);
 }
 
@@ -195,8 +305,7 @@ TEST(ParseIndexTransformTest, IdentityTransformExclusiveMax) {
   EXPECT_EQ(tensorstore::IndexTransformBuilder<>(2, 2)
                 .input_origin({1, 2})
                 .input_exclusive_max({5, kInfIndex + 1})
-                .output_single_input_dimension(0, 0)
-                .output_single_input_dimension(1, 1)
+                .output_identity_transform()
                 .Finalize()
                 .value(),
             tensorstore::ParseIndexTransform(ParseJson(R"(
@@ -211,8 +320,7 @@ TEST(ParseIndexTransformTest, IdentityTransformInclusiveMax) {
   EXPECT_EQ(tensorstore::IndexTransformBuilder<>(2, 2)
                 .input_origin({1, 2})
                 .input_inclusive_max({5, kInfIndex})
-                .output_single_input_dimension(0, 0)
-                .output_single_input_dimension(1, 1)
+                .output_identity_transform()
                 .Finalize()
                 .value(),
             tensorstore::ParseIndexTransform(ParseJson(R"(
@@ -227,8 +335,7 @@ TEST(ParseIndexTransformTest, IdentityTransformShape) {
   EXPECT_EQ(tensorstore::IndexTransformBuilder<>(2, 2)
                 .input_origin({1, 2})
                 .input_shape({5, kInfSize})
-                .output_single_input_dimension(0, 0)
-                .output_single_input_dimension(1, 1)
+                .output_identity_transform()
                 .Finalize()
                 .value(),
             tensorstore::ParseIndexTransform(ParseJson(R"(
@@ -241,8 +348,7 @@ TEST(ParseIndexTransformTest, IdentityTransformShape) {
 
 TEST(ParseIndexTransformTest, IdentityTransformInputRank) {
   EXPECT_EQ(tensorstore::IndexTransformBuilder<>(2, 2)
-                .output_single_input_dimension(0, 0)
-                .output_single_input_dimension(1, 1)
+                .output_identity_transform()
                 .Finalize()
                 .value(),
             tensorstore::ParseIndexTransform(ParseJson(R"(
@@ -257,7 +363,7 @@ TEST(ParseIndexTransformTest, StaticInputRankMismatch) {
       (tensorstore::ParseIndexTransform<3, 3>(MakeLabeledExampleJson())),
       MatchesStatus(absl::StatusCode::kInvalidArgument,
                     "Error parsing index transform from JSON: "  //
-                    "Expected input rank to be 3, but is: 4"));
+                    "Expected input_rank to be 3, but is: 4"));
 }
 
 TEST(ParseIndexTransformTest, StaticOutputRankMismatch) {
@@ -284,7 +390,7 @@ TEST(ParseIndexTransformTest, MissingInputRank) {
           "Error parsing index transform from JSON: "  //
           "At least one of \"input_rank\", \"input_inclusive_min\", "
           "\"input_shape\", \"input_inclusive_max\", \"input_exclusive_max\", "
-          "or \"input_labels\" must be specified"));
+          "\"input_labels\" members must be specified"));
 }
 
 TEST(ParseIndexTransformTest, InvalidInputRank) {
@@ -323,10 +429,8 @@ TEST(ParseIndexTransformTest, ExclusiveMaxAndInclusiveMax) {
 )")),
       MatchesStatus(absl::StatusCode::kInvalidArgument,
                     "Error parsing index transform from JSON: "  //
-                    "Error parsing object member "
-                    "\"input_(exclusive_max|inclusive_max)\": "  //
                     "At most one of \"input_shape\", \"input_inclusive_max\", "
-                    "and \"input_exclusive_max\" members must be specified"));
+                    "\"input_exclusive_max\" members is allowed"));
 }
 
 TEST(ParseIndexTransformTest, ExclusiveMaxAndShape) {
@@ -338,12 +442,10 @@ TEST(ParseIndexTransformTest, ExclusiveMaxAndShape) {
     "input_shape": [5, 10]
 }
 )")),
-      MatchesStatus(
-          absl::StatusCode::kInvalidArgument,
-          "Error parsing index transform from JSON: "                      //
-          "Error parsing object member \"input_(exclusive_max|shape)\": "  //
-          "At most one of \"input_shape\", \"input_inclusive_max\", and "
-          "\"input_exclusive_max\" members must be specified"));
+      MatchesStatus(absl::StatusCode::kInvalidArgument,
+                    "Error parsing index transform from JSON: "  //
+                    "At most one of \"input_shape\", \"input_inclusive_max\", "
+                    "\"input_exclusive_max\" members is allowed"));
 }
 
 TEST(ParseIndexTransformTest, InclusiveMaxAndShape) {
@@ -355,12 +457,10 @@ TEST(ParseIndexTransformTest, InclusiveMaxAndShape) {
     "input_shape": [5, 10]
 }
 )")),
-      MatchesStatus(
-          absl::StatusCode::kInvalidArgument,
-          "Error parsing index transform from JSON: "                      //
-          "Error parsing object member \"input_(inclusive_max|shape)\": "  //
-          "At most one of \"input_shape\", \"input_inclusive_max\", and "
-          "\"input_exclusive_max\" members must be specified"));
+      MatchesStatus(absl::StatusCode::kInvalidArgument,
+                    "Error parsing index transform from JSON: "  //
+                    "At most one of \"input_shape\", \"input_inclusive_max\", "
+                    "\"input_exclusive_max\" members is allowed"));
 }
 
 // Tests that omitting the `"output"` member when `output_rank` is specified
@@ -377,8 +477,7 @@ TEST(ParseIndexTransformTest, MissingOutputs) {
   EXPECT_EQ((tensorstore::IndexTransformBuilder<2, 2>()
                  .input_origin({1, 2})
                  .input_exclusive_max({5, 10})
-                 .output_single_input_dimension(0, 0)
-                 .output_single_input_dimension(1, 1)
+                 .output_identity_transform()
                  .Finalize()
                  .value()),
             (tensorstore::ParseIndexTransform<dynamic_rank, 2>(json)));
@@ -587,7 +686,7 @@ TEST(ParseIndexTransformTest, InvalidOutputArray) {
                     "Error parsing value at position 0: "               //
                     "Error parsing object member \"index_array\": "     //
                     "Error parsing array element at position \\{\\}: "  //
-                    "Expected 64-bit signed integer, but received: \"a\""));
+                    ".* received: \"a\""));
 }
 
 TEST(ParseIndexTransformTest, InvalidOutputInputAndArray) {
@@ -601,14 +700,12 @@ TEST(ParseIndexTransformTest, InvalidOutputInputAndArray) {
     ]
 }
 )")),
-      MatchesStatus(
-          absl::StatusCode::kInvalidArgument,
-          "Error parsing index transform from JSON: "                        //
-          "Error parsing object member \"output\": "                         //
-          "Error parsing value at position 0: "                              //
-          "Error parsing object member \"(index_array|input_dimension)\": "  //
-          "At most one of \"input_dimension\" and \"index_array\" "
-          "must be specified"));
+      MatchesStatus(absl::StatusCode::kInvalidArgument,
+                    "Error parsing index transform from JSON: "             //
+                    "Error parsing object member \"output\": "              //
+                    "Error parsing value at position 0: "                   //
+                    "At most one of \"input_dimension\", \"index_array\" "  //
+                    "members is allowed"));
 }
 
 TEST(ParseIndexTransformTest, DuplicateLabels) {
@@ -618,39 +715,176 @@ TEST(ParseIndexTransformTest, DuplicateLabels) {
 }
 )")),
               MatchesStatus(absl::StatusCode::kInvalidArgument,
-                            "Error parsing index transform from JSON: "  //
-                            "Dimension label \"x\" is not unique"));
+                            "Error parsing index transform from JSON: "       //
+                            "Error parsing object member \"input_labels\": "  //
+                            "Dimension label.*"));
 }
 
-void TestIndexTransformSpecRoundTrip(IndexTransformSpec spec,
-                                     ::nlohmann::json json) {
-  SCOPED_TRACE(json.dump());
-  namespace jb = tensorstore::internal::json_binding;
-  const auto binder = jb::Object(tensorstore::IndexTransformSpecBinder);
-  EXPECT_THAT(jb::ToJson(spec, binder), ::testing::Optional(json));
-  EXPECT_THAT(jb::FromJson<IndexTransformSpec>(json, binder),
-              ::testing::Optional(spec));
+TEST(IndexDomainJsonBinderTest, Simple) {
+  tensorstore::TestJsonBinderRoundTrip<tensorstore::IndexDomain<>>({
+      {tensorstore::IndexDomainBuilder<4>()
+           .origin({-kInfIndex, 7, -kInfIndex, 8})
+           .exclusive_max({kInfIndex + 1, 10, kInfIndex + 1, 17})
+           .implicit_lower_bounds({0, 0, 1, 1})
+           .implicit_upper_bounds({0, 0, 1, 1})
+           .labels({"x", "y", "z", "t"})
+           .Finalize()
+           .value(),
+       {
+           {"inclusive_min", {"-inf", 7, {"-inf"}, {8}}},
+           {"exclusive_max", {"+inf", 10, {"+inf"}, {17}}},
+           {"labels", {"x", "y", "z", "t"}},
+       }},
+  });
+  tensorstore::TestJsonBinderFromJson<tensorstore::IndexDomain<>>({
+      {{
+           {"rank", 33},
+       },
+       MatchesStatus(
+           absl::StatusCode::kInvalidArgument,
+           "Error parsing index domain from JSON: "
+           "Error parsing object member \"rank\": "
+           "Expected integer in the range \\[0, 32\\], but received: 33")},
+      {{
+           {"shape", {1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  //
+                      1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  //
+                      1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  //
+                      1, 1, 1}},
+       },
+       MatchesStatus(absl::StatusCode::kInvalidArgument,
+                     "Error parsing index domain from JSON: "
+                     "Error parsing object member \"shape\": "
+                     "Rank 33 is outside valid range \\[0, 32\\]")},
+      {{
+           {"labels", {"", "", "", "", "", "", "", "", "", "",  //
+                       "", "", "", "", "", "", "", "", "", "",  //
+                       "", "", "", "", "", "", "", "", "", "",  //
+                       "", "", ""}},
+       },
+       MatchesStatus(absl::StatusCode::kInvalidArgument,
+                     "Error parsing index domain from JSON: "
+                     "Error parsing object member \"labels\": "
+                     "Rank 33 is outside valid range \\[0, 32\\]")},
+      {{
+           {"inclusive_min", {"-inf", 7, {"-inf"}, {8}}},
+           {"exclusive_max", {"+inf", 10, {"+inf"}, {17}}},
+           {"labels", {"x", "y", "z", "t"}},
+           {"output", "abc"},
+       },
+       MatchesStatus(absl::StatusCode::kInvalidArgument,
+                     "Error parsing index domain from JSON: "
+                     "Object includes extra members: \"output\"")},
+  });
 }
 
-TEST(IndexTransformSpecTest, JsonBinding) {
-  TestIndexTransformSpecRoundTrip(IndexTransformSpec(),
-                                  ::nlohmann::json::object());
+TEST(IndexBinderTest, Basic) {
+  using ::tensorstore::kMaxFiniteIndex;
+  tensorstore::TestJsonBinderRoundTrip<Index>(
+      {
+          {-kInfIndex, "-inf"},
+          {+kInfIndex, "+inf"},
+          {-kMaxFiniteIndex, -kMaxFiniteIndex},
+          {0, 0},
+          {5, 5},
+          {+kMaxFiniteIndex, +kMaxFiniteIndex},
+      },
+      tensorstore::internal_json_binding::IndexBinder);
+  tensorstore::TestJsonBinderFromJson<Index>(
+      {
+          {"abc", MatchesStatus(absl::StatusCode::kInvalidArgument)},
+          {-kInfIndex, ::testing::Optional(MatchesJson(-kInfIndex))},
+          {+kInfIndex, ::testing::Optional(MatchesJson(+kInfIndex))},
+          {-kInfIndex - 1, MatchesStatus(absl::StatusCode::kInvalidArgument)},
+          {kInfIndex + 1, MatchesStatus(absl::StatusCode::kInvalidArgument)},
+      },
+      tensorstore::internal_json_binding::IndexBinder);
+}
 
-  TestIndexTransformSpecRoundTrip(IndexTransformSpec(3),
-                                  ::nlohmann::json{{"rank", 3}});
+TEST(IndexIntervalBinderTest, Basic) {
+  using ::tensorstore::IndexInterval;
+  tensorstore::TestJsonBinderRoundTrip<IndexInterval>({
+      {IndexInterval::UncheckedClosed(5, 10), {5, 10}},
+      {IndexInterval(), {"-inf", "+inf"}},
+      {IndexInterval::UncheckedClosed(5, 4), {5, 4}},
+      {IndexInterval::UncheckedClosed(-kInfIndex, 20), {"-inf", 20}},
+      {IndexInterval::UncheckedClosed(20, +kInfIndex), {20, "+inf"}},
+  });
+  tensorstore::TestJsonBinderFromJson<IndexInterval>({
+      {"abc", MatchesStatus(absl::StatusCode::kInvalidArgument)},
+      {{-kInfIndex - 1, 10}, MatchesStatus(absl::StatusCode::kInvalidArgument)},
+      {{10, 5}, MatchesStatus(absl::StatusCode::kInvalidArgument)},
+  });
+}
 
-  TestIndexTransformSpecRoundTrip(
-      IndexTransformSpec(IndexTransformBuilder<>(2, 1)
-                             .input_shape({2, 3})
-                             .output_single_input_dimension(0, 0)
-                             .Finalize()
-                             .value()),
-      ::nlohmann::json{{"transform",
-                        {
-                            {"input_exclusive_max", {2, 3}},
-                            {"input_inclusive_min", {0, 0}},
-                            {"output", {{{"input_dimension", 0}}}},
-                        }}});
+TEST(ConstrainedRankJsonBinderTest, RoundTripNoConstraintIncludeDefaults) {
+  tensorstore::TestJsonBinderRoundTrip<DimensionIndex>(
+      {
+          {5, 5},
+          {dynamic_rank, ::nlohmann::json::value_t::discarded},
+      },
+      tensorstore::internal_json_binding::ConstrainedRankJsonBinder);
+}
+
+TEST(ConstrainedRankJsonBinderTest, RoundTripNoConstraintExcludeDefaults) {
+  tensorstore::TestJsonBinderRoundTrip<DimensionIndex>(
+      {
+          {5, 5},
+          {dynamic_rank, ::nlohmann::json::value_t::discarded},
+      },
+      tensorstore::internal_json_binding::ConstrainedRankJsonBinder,
+      tensorstore::IncludeDefaults{false});
+}
+
+TEST(ConstrainedRankJsonBinderTest, RoundTripRankConstraintIncludeDefaults) {
+  tensorstore::TestJsonBinderRoundTrip<DimensionIndex>(
+      {
+          {30, 30},
+      },
+      tensorstore::internal_json_binding::ConstrainedRankJsonBinder,
+      tensorstore::JsonSerializationOptions{tensorstore::RankConstraint{30},
+                                            tensorstore::IncludeDefaults{true}},
+      tensorstore::RankConstraint{30});
+}
+
+TEST(ConstrainedRankJsonBinderTest, FromJsonRankConstraint) {
+  tensorstore::TestJsonBinderFromJson<DimensionIndex>(
+      {
+          {30, ::testing::Optional(30)},
+          {::nlohmann::json::value_t::discarded, ::testing::Optional(30)},
+          {5, MatchesStatus(absl::StatusCode::kInvalidArgument,
+                            "Expected 30, but received: 5")},
+      },
+      tensorstore::internal_json_binding::ConstrainedRankJsonBinder,
+      tensorstore::RankConstraint{30});
+}
+
+TEST(ConstrainedRankJsonBinderTest, ToJsonRankConstraintIncludeDefaults) {
+  tensorstore::TestJsonBinderToJson<DimensionIndex>(
+      {
+          {30, ::testing::Optional(MatchesJson(30))},
+          {dynamic_rank, ::testing::Optional(MatchesJson(
+                             ::nlohmann::json::value_t::discarded))},
+          {5, MatchesStatus(absl::StatusCode::kInvalidArgument,
+                            "Expected 30, but received: 5")},
+      },
+      tensorstore::internal_json_binding::ConstrainedRankJsonBinder,
+      tensorstore::JsonSerializationOptions{
+          tensorstore::RankConstraint{30}, tensorstore::IncludeDefaults{true}});
+}
+
+TEST(ConstrainedRankJsonBinderTest, ToJsonRankConstraintExcludeDefaults) {
+  tensorstore::TestJsonBinderToJson<DimensionIndex>(
+      {
+          {30, ::testing::Optional(
+                   MatchesJson(::nlohmann::json::value_t::discarded))},
+          {dynamic_rank, ::testing::Optional(MatchesJson(
+                             ::nlohmann::json::value_t::discarded))},
+          {5, MatchesStatus(absl::StatusCode::kInvalidArgument,
+                            "Expected 30, but received: 5")},
+      },
+      tensorstore::internal_json_binding::ConstrainedRankJsonBinder,
+      tensorstore::JsonSerializationOptions{tensorstore::IncludeDefaults{false},
+                                            tensorstore::RankConstraint{30}});
 }
 
 }  // namespace
