@@ -26,65 +26,65 @@
 
 namespace {
 
-using tensorstore::DimensionIndex;
-using tensorstore::IdentityTransform;
-using tensorstore::Index;
-using tensorstore::IndexInterval;
-using tensorstore::IndexTransformBuilder;
-using tensorstore::IndexTransformView;
-using tensorstore::kImplicit;
-using tensorstore::kInfIndex;
-using tensorstore::MakeArray;
-using tensorstore::MatchesStatus;
-using tensorstore::span;
-using tensorstore::Status;
-using tensorstore::StrCat;
+using ::tensorstore::DimensionIndex;
+using ::tensorstore::IdentityTransform;
+using ::tensorstore::Index;
+using ::tensorstore::IndexInterval;
+using ::tensorstore::IndexTransformBuilder;
+using ::tensorstore::IndexTransformView;
+using ::tensorstore::kImplicit;
+using ::tensorstore::kInfIndex;
+using ::tensorstore::MakeArray;
+using ::tensorstore::MatchesStatus;
+using ::tensorstore::span;
+using ::tensorstore::StrCat;
+using ::tensorstore::internal::GetLValue;
+using ::tensorstore::internal_index_space::ValidateInputDimensionResize;
 
 TEST(ValidateInputDimensionResizeTest, ValidArguments) {
-  using tensorstore::internal_index_space::ValidateInputDimensionResize;
-  using OIII = tensorstore::OptionallyImplicitIndexInterval;
+  using OIII = ::tensorstore::OptionallyImplicitIndexInterval;
 
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
             ValidateInputDimensionResize(  //
                 OIII{},
                 /*requested_inclusive_min=*/kImplicit,
                 /*requested_exclusive_max=*/kImplicit));
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
             ValidateInputDimensionResize(  //
                 OIII{},
                 /*requested_inclusive_min=*/1,
                 /*requested_exclusive_max=*/1));
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
             ValidateInputDimensionResize(  //
                 OIII{},
                 /*requested_inclusive_min=*/kImplicit,
                 /*requested_exclusive_max=*/1));
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
             ValidateInputDimensionResize(  //
                 OIII{},
                 /*requested_inclusive_min=*/1,
                 /*requested_exclusive_max=*/kImplicit));
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
             ValidateInputDimensionResize(  //
                 OIII{},
                 /*requested_inclusive_min=*/-kInfIndex + 1,
                 /*requested_exclusive_max=*/kImplicit));
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
             ValidateInputDimensionResize(  //
                 OIII{},
                 /*requested_inclusive_min=*/kImplicit,
                 /*requested_exclusive_max=*/+kInfIndex));
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
             ValidateInputDimensionResize(  //
                 OIII{},
                 /*requested_inclusive_min=*/kImplicit,
                 /*requested_exclusive_max=*/-kInfIndex + 2));
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
             ValidateInputDimensionResize(  //
                 OIII{IndexInterval::UncheckedClosed(1, 10), false, false},
                 /*requested_inclusive_min=*/kImplicit,
                 /*requested_exclusive_max=*/kImplicit));
-  EXPECT_EQ(Status(),
+  EXPECT_EQ(absl::OkStatus(),
             ValidateInputDimensionResize(  //
                 OIII{IndexInterval::UncheckedClosed(1, 10), true, true},
                 /*requested_inclusive_min=*/-kInfIndex,
@@ -92,7 +92,6 @@ TEST(ValidateInputDimensionResizeTest, ValidArguments) {
 }
 
 TEST(ValidateInputDimensionResizeTest, InvalidArguments) {
-  using tensorstore::internal_index_space::ValidateInputDimensionResize;
   using OIII = tensorstore::OptionallyImplicitIndexInterval;
   EXPECT_THAT(ValidateInputDimensionResize(  //
                   OIII{},
@@ -168,7 +167,7 @@ TEST(PropagateInputDomainResizeToOutputTest, ResizedSingleInputDimension) {
     Index new_output_exclusive_max[1];
     bool is_noop;
     ASSERT_EQ(
-        Status(),
+        absl::OkStatus(),
         tensorstore::PropagateInputDomainResizeToOutput(
             transform, requested_input_inclusive_min,
             requested_input_exclusive_max, can_resize_tied_bounds,
@@ -210,7 +209,7 @@ TEST(PropagateInputDomainResizeToOutputTest,
     Index new_output_exclusive_max[2];
     bool is_noop;
     ASSERT_EQ(
-        Status(),
+        absl::OkStatus(),
         tensorstore::PropagateInputDomainResizeToOutput(
             transform, requested_input_inclusive_min,
             requested_input_exclusive_max, can_resize_tied_bounds,
@@ -265,7 +264,7 @@ TEST(PropagateInputDomainResizeToOutputTest,
   Index new_output_exclusive_max[2];
   bool is_noop;
   ASSERT_EQ(
-      Status(),
+      absl::OkStatus(),
       tensorstore::PropagateInputDomainResizeToOutput(
           transform, requested_input_inclusive_min,
           requested_input_exclusive_max, /*can_resize_tied_bounds=*/true,
@@ -308,7 +307,7 @@ TEST(PropagateInputDomainResizeToOutputTest,
     Index new_output_exclusive_max[2];
     bool is_noop;
     ASSERT_EQ(
-        Status(),
+        absl::OkStatus(),
         tensorstore::PropagateInputDomainResizeToOutput(
             transform, requested_input_inclusive_min,
             requested_input_exclusive_max, can_resize_tied_bounds,
@@ -346,7 +345,6 @@ TEST(PropagateInputDomainResizeToOutputTest,
 }
 
 TEST(PropagateInputDomainResizeToOutputTest, InvalidArguments) {
-  using tensorstore::internal::GetLValue;
   bool is_noop;
   const auto test_resize = [&](bool can_resize_tied_bounds) {
     // `can_resize_tied_bounds` should have no effect on these cases.
@@ -370,7 +368,7 @@ TEST(PropagateInputDomainResizeToOutputTest, InvalidArguments) {
     EXPECT_THAT(
         tensorstore::PropagateInputDomainResizeToOutput(
             IndexTransformBuilder<>(1, 2)
-                .output_single_input_dimension(0, 0)
+                .output_identity_transform()
                 .output_constant(1, kInfIndex)
                 .Finalize()
                 .value(),
@@ -527,21 +525,19 @@ TEST(PropagateInputDomainResizeToOutputTest, InvalidArguments) {
                     StrCat("Error propagating bounds for output dimension 0 "
                            "from existing bounds for input dimension 1: .*")));
 
-  // TODO(jbms): If support for treating rank-0 index arrays like constant maps
-  // is added, this test will also need to change.
   EXPECT_THAT(
       tensorstore::PropagateInputDomainResizeToOutput(
-          IndexTransformBuilder<>(1, 2)
-              .input_origin({1})
-              .input_shape({10})
-              .implicit_lower_bounds({1})
-              .implicit_upper_bounds({1})
-              .output_index_array(0, 0, 1, MakeArray<Index>({5}))
-              .output_single_input_dimension(1, 0)
+          IndexTransformBuilder(2, 2)
+              .input_origin({0, 1})
+              .input_shape({2, 10})
+              .implicit_lower_bounds({0, 1})
+              .implicit_upper_bounds({0, 1})
+              .output_index_array(0, 0, 1, MakeArray<Index>({{5}, {6}}))
+              .output_single_input_dimension(1, 1)
               .Finalize()
               .value(),
-          /*requested_input_inclusive_min=*/span<const Index>({2}),
-          /*requested_input_exclusive_max=*/span<const Index>({7}),
+          /*requested_input_inclusive_min=*/span<const Index>({kImplicit, 2}),
+          /*requested_input_exclusive_max=*/span<const Index>({kImplicit, 7}),
           /*can_resize_tied_bounds=*/false,
           /*output_inclusive_min_constraint=*/
           GetLValue(std::array<Index, 2>()),
