@@ -15,17 +15,18 @@
 #include "tensorstore/internal/grid_partition_impl.h"
 
 #include <algorithm>
+#include <cassert>
 #include <memory>
 #include <numeric>
 #include <optional>
 #include <utility>
 #include <vector>
 
-#include "absl/base/macros.h"
 #include "absl/container/fixed_array.h"
 #include "absl/container/flat_hash_map.h"
 #include "absl/container/inlined_vector.h"
 #include "absl/hash/hash.h"
+#include "absl/status/status.h"
 #include "tensorstore/array.h"
 #include "tensorstore/box.h"
 #include "tensorstore/index.h"
@@ -40,7 +41,7 @@
 #include "tensorstore/util/result.h"
 #include "tensorstore/util/span.h"
 #include "tensorstore/util/status.h"
-#include "tensorstore/util/to_string.h"
+#include "tensorstore/util/str_cat.h"
 
 namespace tensorstore {
 namespace internal_grid_partition {
@@ -52,15 +53,15 @@ IndexTransformGridPartition::IndexTransformGridPartition(
 SharedArray<const Index, 2>
 IndexTransformGridPartition::IndexArraySet::partition_input_indices(
     Index partition_i) const {
-  ABSL_ASSERT(partition_i >= 0 && partition_i < num_partitions());
+  assert(partition_i >= 0 && partition_i < num_partitions());
   SharedArray<const Index, 2> result;
   const Index start = grid_cell_partition_offsets[partition_i];
   const Index end =
       static_cast<size_t>(partition_i + 1) == grid_cell_partition_offsets.size()
           ? partitioned_input_indices.shape()[0]
           : grid_cell_partition_offsets[partition_i + 1];
-  ABSL_ASSERT(start >= 0 && start < partitioned_input_indices.shape()[0]);
-  ABSL_ASSERT(end > start && end <= partitioned_input_indices.shape()[0]);
+  assert(start >= 0 && start < partitioned_input_indices.shape()[0]);
+  assert(end > start && end <= partitioned_input_indices.shape()[0]);
   result.pointer() =
       std::shared_ptr<const Index>(partitioned_input_indices.pointer(),
                                    &partitioned_input_indices(start, 0));
@@ -92,19 +93,19 @@ namespace {
 ///     as a temporary buffer.  The spans passed to `set_callback` will refer to
 ///     memory in this buffer.
 /// \param set_callback Function with a signature compatible with:
-///     `Status (span<const DimensionIndex> input_dims,
+///     `absl::Status (span<const DimensionIndex> input_dims,
 ///              span<const DimensionIndex> grid_dims,
 ///              bool has_array)`.  This function is called for each connected.
 ///     Any error returned causes iteration to stop.
 /// \return The error value returned by the last call to `set_callback`, or
-///     `Status()` on success.
+///     `absl::Status()` on success.
 template <typename SetCallbackFn>
-Status ForEachConnectedSet(span<const DimensionIndex> grid_output_dimensions,
-                           OutputIndexMapRange<> output_index_maps,
-                           span<DimensionIndex> temp_buffer,
-                           SetCallbackFn set_callback) {
+absl::Status ForEachConnectedSet(
+    span<const DimensionIndex> grid_output_dimensions,
+    OutputIndexMapRange<> output_index_maps, span<DimensionIndex> temp_buffer,
+    SetCallbackFn set_callback) {
   const DimensionIndex input_rank = output_index_maps.input_rank();
-  ABSL_ASSERT(temp_buffer.size() >= input_rank + grid_output_dimensions.size());
+  assert(temp_buffer.size() >= input_rank + grid_output_dimensions.size());
   const span<DimensionIndex> input_dims = temp_buffer.first(input_rank);
   const span<DimensionIndex> grid_dims =
       temp_buffer.subspan(input_rank, grid_output_dimensions.size());
@@ -144,9 +145,9 @@ Status ForEachConnectedSet(span<const DimensionIndex> grid_output_dimensions,
   /// \returns `true` if, and only if, any additional input dimensions were
   ///     added to the current set.
   const auto add_grid_dim_to_current_set = [&](DimensionIndex grid_i) -> bool {
-    ABSL_ASSERT(grid_i >= 0 && grid_i < num_dependent_grid_dims);
+    assert(grid_i >= 0 && grid_i < num_dependent_grid_dims);
     const DimensionIndex grid_dim = grid_dims[grid_i];
-    ABSL_ASSERT(grid_dim >= 0 && grid_dim < grid_output_dimensions.size());
+    assert(grid_dim >= 0 && grid_dim < grid_output_dimensions.size());
     const DimensionIndex output_dim = grid_output_dimensions[grid_dim];
     const OutputIndexMapRef<> map = output_index_maps[output_dim];
     switch (map.method()) {
@@ -188,9 +189,9 @@ Status ForEachConnectedSet(span<const DimensionIndex> grid_output_dimensions,
   /// If the dependency is due to an `array` output index map, also sets
   /// `current_set_has_array` to `true`.
   const auto is_grid_dim_in_set = [&](DimensionIndex grid_i) -> DimensionIndex {
-    ABSL_ASSERT(grid_i >= 0 && grid_i < num_dependent_grid_dims);
+    assert(grid_i >= 0 && grid_i < num_dependent_grid_dims);
     const DimensionIndex grid_dim = grid_dims[grid_i];
-    ABSL_ASSERT(grid_dim >= 0 && grid_dim < grid_output_dimensions.size());
+    assert(grid_dim >= 0 && grid_dim < grid_output_dimensions.size());
     const DimensionIndex output_dim = grid_output_dimensions[grid_dim];
     const OutputIndexMapRef<> map = output_index_maps[output_dim];
     switch (map.method()) {
@@ -204,12 +205,11 @@ Status ForEachConnectedSet(span<const DimensionIndex> grid_output_dimensions,
       case OutputIndexMethod::array: {
         const OutputIndexMapRef<>::IndexArrayView index_array =
             map.index_array();
-        if (std::any_of(
-                input_dims.begin() + input_dim_set_begin,
-                input_dims.begin() + input_dim_set_end,
-                [&](DimensionIndex input_i) {
-                  return index_array.byte_strides()[input_dims[input_i]] != 0;
-                })) {
+        if (std::any_of(input_dims.begin() + input_dim_set_begin,
+                        input_dims.begin() + input_dim_set_end,
+                        [&](DimensionIndex input_dim) {
+                          return index_array.byte_strides()[input_dim] != 0;
+                        })) {
           current_set_has_array = true;
           return true;
         }
@@ -310,14 +310,14 @@ OutputIt FillWithTiledStridedRange(T start, T size, Stride stride,
 ///     to be filled with the output indices.
 /// \param output_stride Stride (in elements, not bytes) of the innermost
 ///     dimension of `output_indices`.
-/// \returns `Status()` on success.
+/// \returns `absl::Status()` on success.
 /// \error `absl::StatusCode::kOutOfRange` if `map` has an invalid `offset`.
 /// \error `absl::StatusCode::kInvalidArgument` if integer overflow occurs.
-Status GenerateSingleInputDimensionOutputIndices(
+absl::Status GenerateSingleInputDimensionOutputIndices(
     OutputIndexMapRef<> map, span<const DimensionIndex> input_dims,
     IndexTransformView<> index_transform, Index* output_indices,
     Index output_stride) {
-  ABSL_ASSERT(map.method() == OutputIndexMethod::single_input_dimension);
+  assert(map.method() == OutputIndexMethod::single_input_dimension);
   const DimensionIndex single_input_dim = map.input_dimension();
   const IndexInterval domain = index_transform.input_domain()[single_input_dim];
   const Index stride = map.stride();
@@ -367,15 +367,14 @@ Status GenerateSingleInputDimensionOutputIndices(
 ///     to be filled with the output indices.
 /// \param output_stride Stride (in elements, not bytes) of the innermost
 ///     dimension of `output_indices`.
-/// \returns `Status()` on success.
+/// \returns `absl::Status()` on success.
 /// \error `absl::StatusCode::kOutOfRange` if `map` contains an invalid index.
 /// \error `absl::StatusCode::kInvalidArgument` if integer overflow occurs.
-Status GenerateIndexArrayOutputIndices(OutputIndexMapRef<> map,
-                                       span<const DimensionIndex> input_dims,
-                                       IndexTransformView<> index_transform,
-                                       Index* output_indices,
-                                       Index output_stride) {
-  ABSL_ASSERT(map.method() == OutputIndexMethod::array);
+absl::Status GenerateIndexArrayOutputIndices(
+    OutputIndexMapRef<> map, span<const DimensionIndex> input_dims,
+    IndexTransformView<> index_transform, Index* output_indices,
+    Index output_stride) {
+  assert(map.method() == OutputIndexMethod::array);
   absl::FixedArray<Index, internal::kNumInlinedDims> output_byte_strides(
       index_transform.input_rank(), 0);
   DimensionIndex byte_stride = sizeof(Index) * output_stride;
@@ -407,24 +406,6 @@ Status GenerateIndexArrayOutputIndices(OutputIndexMapRef<> map,
   return absl::OkStatus();
 }
 
-/// Converts output indices to grid indices of a regular grid.
-///
-/// \param output_indices[in,out] Non-null pointer to array of shape
-///     `{num_positions}` with stride (in elements, not bytes) `stride` between
-///     consecutive elements.  On invocation, contains the output indices.  On
-///     return, contains the corresponding grid indices.
-/// \param stride Stride of `output_indices` array.
-/// \param num_positions Extent of `output_indices` array.
-/// \param cell_size The size of the grid cell.
-void ConvertOutputIndicesToCellIndices(Index* output_indices, Index stride,
-                                       Index num_positions, Index cell_size) {
-  for (Index* end = output_indices + num_positions * stride;
-       output_indices != end; output_indices += stride) {
-    auto& x = *output_indices;
-    x = FloorOfRatio(x, cell_size);
-  }
-}
-
 /// Computes the product of `input_shape[d]` for `d` in `dims`.
 ///
 /// \error `absl::StatusCode::kInvalidArgument` if integer overflow occurs.
@@ -464,8 +445,8 @@ Result<Index> ProductOfIndirectExtents(span<const Index> input_shape,
 Result<std::vector<Index>> GenerateIndexArraySetGridCellIndices(
     span<const DimensionIndex> grid_dims, span<const DimensionIndex> input_dims,
     span<const DimensionIndex> grid_output_dimensions,
-    span<const Index> grid_cell_shape, IndexTransformView<> index_transform,
-    Index num_positions) {
+    OutputToGridCellFn output_to_grid_cell,
+    IndexTransformView<> index_transform, Index num_positions) {
   // Logically represents a row-major array of shape `{num_positions,
   // grid_dims.size()}` containing the partial grid cell index vectors for each
   // position.
@@ -485,13 +466,18 @@ Result<std::vector<Index>> GenerateIndexArraySetGridCellIndices(
           map, input_dims, index_transform, cur_cell_indices,
           grid_dims.size()));
     } else {
-      ABSL_ASSERT(map.method() == OutputIndexMethod::array);
+      assert(map.method() == OutputIndexMethod::array);
       TENSORSTORE_RETURN_IF_ERROR(
           GenerateIndexArrayOutputIndices(map, input_dims, index_transform,
                                           cur_cell_indices, grid_dims.size()));
     }
-    ConvertOutputIndicesToCellIndices(cur_cell_indices, grid_dims.size(),
-                                      num_positions, grid_cell_shape[grid_dim]);
+
+    // Convert the output indices to grid cell indices
+    for (Index* end = cur_cell_indices + num_positions * grid_dims.size();
+         cur_cell_indices != end; cur_cell_indices += grid_dims.size()) {
+      *cur_cell_indices =
+          output_to_grid_cell(grid_dim, *cur_cell_indices, nullptr);
+    }
   }
   return temp_cell_indices;
 }
@@ -650,7 +636,7 @@ IndirectVectorMap PartitionIndexArraySetGridCellIndexVectors(
     for (Index& position_i_or_offset : *grid_cell_partition_offsets) {
       const Index position_i = position_i_or_offset;
       auto it = cells.find(position_i);
-      ABSL_ASSERT(it != cells.end());
+      assert(it != cells.end());
       auto& count_or_offset = it->second;
       const Index count = count_or_offset;
       position_i_or_offset = count_or_offset = offset;
@@ -693,7 +679,7 @@ SharedArray<Index, 2> GenerateIndexArraySetPartitionedInputIndices(
   Index position_i = 0;
   IterateOverIndexRange(partial_input_domain, [&](span<const Index> indices) {
     auto it = cells.find(position_i);
-    ABSL_ASSERT(it != cells.end());
+    assert(it != cells.end());
     auto& offset = it->second;
     std::copy(indices.begin(), indices.end(),
               partitioned_input_indices.data() + offset * input_dims.size());
@@ -732,10 +718,11 @@ SharedArray<Index, 2> GenerateIndexArraySetPartitionedInputIndices(
 /// \error `absl::StatusCode::kInvalidArgument` if integer overflow occurs.
 /// \error `absl::StatusCode::kOutOfRange` if an index array contains an
 ///     out-of-bounds index.
-Status FillIndexArraySetData(
+absl::Status FillIndexArraySetData(
     IndexTransformGridPartition::IndexArraySet* index_array_set,
     span<const DimensionIndex> grid_output_dimensions,
-    span<const Index> grid_cell_shape, IndexTransformView<> index_transform) {
+    OutputToGridCellFn output_to_grid_cell,
+    IndexTransformView<> index_transform) {
   // Compute the total number of distinct partial input index vectors in the
   // input domain subset.  This allows us to terminate early if it equals 0, and
   // avoids the need for computing it and checking for overflow in each of the
@@ -755,7 +742,7 @@ Status FillIndexArraySetData(
       std::vector<Index> temp_cell_indices,
       GenerateIndexArraySetGridCellIndices(
           index_array_set->grid_dimensions, index_array_set->input_dimensions,
-          grid_output_dimensions, grid_cell_shape, index_transform,
+          grid_output_dimensions, output_to_grid_cell, index_transform,
           num_positions));
 
   // Compute `index_array_set->grid_cell_indices`, the sorted array of the
@@ -796,19 +783,20 @@ Status FillIndexArraySetData(
 /// \error `absl::StatusCode::kInvalidArgument` if integer overflow occurs.
 /// \error `absl::StatusCode::kOutOfRange` if an index array contains an
 ///     out-of-bounds index.
-Status GenerateIndexTransformGridPartitionData(
+absl::Status GenerateIndexTransformGridPartitionData(
     span<const DimensionIndex> grid_output_dimensions,
-    span<const Index> grid_cell_shape, IndexTransformView<> index_transform,
-    IndexTransformGridPartition* output) {
+    OutputToGridCellFn output_to_grid_cell,
+    IndexTransformView<> index_transform, IndexTransformGridPartition* output) {
   return ForEachConnectedSet(
       grid_output_dimensions, index_transform.output_index_maps(),
       output->temp_buffer_,
       [&](span<const DimensionIndex> input_dims,
-          span<const DimensionIndex> grid_dims, bool has_array) -> Status {
+          span<const DimensionIndex> grid_dims,
+          bool has_array) -> absl::Status {
         if (!has_array) {
           // The connected set contains only `single_input_dimension`
           // dependencies.
-          ABSL_ASSERT(input_dims.size() == 1);
+          assert(input_dims.size() == 1);
           output->strided_sets_.push_back({grid_dims, input_dims[0]});
           return absl::OkStatus();
         }
@@ -818,18 +806,17 @@ Status GenerateIndexTransformGridPartitionData(
         set->input_dimensions = input_dims;
         set->grid_dimensions = grid_dims;
         return FillIndexArraySetData(set, grid_output_dimensions,
-                                     grid_cell_shape, index_transform);
+                                     output_to_grid_cell, index_transform);
       });
 }
 }  // namespace
 
-Status PrePartitionIndexTransformOverRegularGrid(
+absl::Status PrePartitionIndexTransformOverGrid(
     IndexTransformView<> index_transform,
     span<const DimensionIndex> grid_output_dimensions,
-    span<const Index> grid_cell_shape,
-    absl::optional<IndexTransformGridPartition>* result) {
-  ABSL_ASSERT(result != nullptr);
-  ABSL_ASSERT(grid_output_dimensions.size() == grid_cell_shape.size());
+    OutputToGridCellFn output_to_grid_cell,
+    std::optional<IndexTransformGridPartition>* result) {
+  assert(result != nullptr);
   const DimensionIndex input_rank = index_transform.input_rank();
 
   // Check that the input domains are all bounded.
@@ -859,9 +846,9 @@ Status PrePartitionIndexTransformOverRegularGrid(
   }
 
   // Compute the IndexTransformGridPartition structure.
-  result->emplace(index_transform.input_rank(), grid_cell_shape.size());
+  result->emplace(index_transform.input_rank(), grid_output_dimensions.size());
   return internal_grid_partition::GenerateIndexTransformGridPartitionData(
-      grid_output_dimensions, grid_cell_shape, index_transform,
+      grid_output_dimensions, output_to_grid_cell, index_transform,
       &result->value());
 }
 
