@@ -29,6 +29,7 @@
 
 #include <memory>
 
+#include "absl/status/status.h"
 #include "tensorstore/data_type.h"
 #include "tensorstore/index.h"
 #include "tensorstore/internal/arena.h"
@@ -52,7 +53,7 @@ namespace internal {
 /// While not explicitly exposed as part of the interface, the following
 /// additional attributes are implicitly associated with an `NDIterator`:
 ///
-///   - `DataType data_type`: the type of the data yielded by this iterator.
+///   - `DataType dtype`: the type of the data yielded by this iterator.
 ///
 ///   - `IterationBufferKind buffer_kind`: the type of buffer supported by this
 ///     iterator.
@@ -94,11 +95,11 @@ class NDIterator {
   ///     `<= max_block_size`.
   /// \param pointer[in,out] If the implicit attribute `external == true`,
   ///     `*pointer` must be an already-constructed buffer of kind
-  ///     `buffer_kind`, data type `data_type`, and length `block_size` to fill
+  ///     `buffer_kind`, data type `dtype`, and length `block_size` to fill
   ///     with the data starting at the specified position.  If
   ///     `external == false`, the existing value of `*pointer` is ignored and
   ///     `*pointer` must be assigned to a buffer of kind `buffer_kind`, length
-  ///     `block_size`, and data type `data_type` containing the data starting
+  ///     `block_size`, and data type `dtype` containing the data starting
   ///     at `indices`.  The returned buffer is invalidated upon the next call
   ///     to `GetBlock`, `UpdateBlock`, or if the `NDIterator` is destroyed.
   /// \param status[out] Non-null pointer to location in which an error may be
@@ -113,7 +114,7 @@ class NDIterator {
   /// \remark The default implementation just returns `block_size`.  Except for
   ///     write-only iterators with `external == true`, it must be overridden.
   virtual Index GetBlock(span<const Index> indices, Index block_size,
-                         IterationBufferPointer* pointer, Status* status);
+                         IterationBufferPointer* pointer, absl::Status* status);
 
   /// Updates the block at the specified location.
   ///
@@ -135,7 +136,8 @@ class NDIterator {
   /// \remark The default implementation just returns `block_size`, and need not
   ///     be overridden by read-only iterators.
   virtual Index UpdateBlock(span<const Index> indices, Index block_size,
-                            IterationBufferPointer pointer, Status* status);
+                            IterationBufferPointer pointer,
+                            absl::Status* status);
 
   /// Needed by `VirtualDestroyDeleter`.
   virtual void Destroy() = 0;
@@ -189,6 +191,10 @@ class NDIterator {
 /// where combining with other dimensions may be possible regardless of the
 /// iteration direction, e.g. data generated on the fly during iteration.
 ///
+/// The `DirectionPref::kForwardRequired` value is intended to be used for
+/// special cases where the NDIterable implementation may be simplified by not
+/// having to consider skipped / backward iteration for a dimension.
+///
 /// For the example above, the result is:
 ///
 ///     | Dimension index | Size | Byte stride | DirectionPref |
@@ -203,33 +209,30 @@ class NDIterator {
 /// iterable.
 class NDIterableLayoutConstraint {
  public:
-  /// Specifies preferences for iterating over a given dimension.
+  /// Specifies preferences for iterating over a given dimension.  Precedence is
+  /// according to magnitude.
   enum class DirectionPref {
     /// Forward is more likely to permit combining.  This dimension cannot be
     /// skipped.
-    kForward = 1,
+    kForward = 2,
+    /// Forward direction is required for iteration.  This dimension cannot be
+    /// skipped, even if the extent is 1.
+    kForwardRequired = 3,
     /// Backward is more likely to permit combining.  This dimension cannot be
     /// skipped.
-    kBackward = -1,
+    kBackward = -2,
     /// Either direction is equally good.  This dimension cannot be skipped.
-    kEither = 0,
+    kEither = 1,
     /// Either direction is equally good.  This dimension can be skipped.
-    kCanSkip = 2,
+    kCanSkip = 0,
   };
 
-  /// Returns the combined `DirectionPref`.
+  /// Returns the most restrictive `DirectionPref`, or `a` if both are equally
+  /// restrictive.
   static inline DirectionPref CombineDirectionPrefs(DirectionPref a,
                                                     DirectionPref b) {
-    switch (a) {
-      case DirectionPref::kForward:
-      case DirectionPref::kBackward:
-        return a;
-      case DirectionPref::kCanSkip:
-        return b;
-      case DirectionPref::kEither:
-        return (b == DirectionPref::kCanSkip) ? DirectionPref::kEither : b;
-    }
-    TENSORSTORE_UNREACHABLE;  // COV_NF_LINE
+    return (std::abs(static_cast<int>(a)) >= std::abs(static_cast<int>(b))) ? a
+                                                                            : b;
   }
 
   /// Returns the iteration direction (`+1` for forward, `-1` for backward) for
@@ -278,7 +281,7 @@ class NDIterableBufferConstraint : public NDIterableLayoutConstraint {
 
     /// The iteration direction for each original dimension.  A value of `+1`
     /// means forward, a value of `-1` means reverse.  A value of `0` is only
-    /// valid when it corresponds to a dummy iteration dimension of `-1`.
+    /// valid when it corresponds to a skipped (but not a combined) dimension.
     span<const int> directions;
 
     /// The sequence of simplified dimensions.  The first dimension is the
@@ -349,7 +352,7 @@ class NDIterable : public NDIterableBufferConstraint {
   virtual ArenaAllocator<> get_allocator() const = 0;
 
   /// Returns the data type.
-  virtual DataType data_type() const = 0;
+  virtual DataType dtype() const = 0;
 
   /// Combines an `IterationLayoutView` with a block size value.
   struct IterationBufferLayoutView : public IterationLayoutView {
@@ -372,8 +375,8 @@ class NDIterable : public NDIterableBufferConstraint {
   /// \pre `iteration_dimensions.size() == iteration_directions.size()`.
   /// \returns A non-null `NDIterator` with an associated `iteration_shape`
   ///     equal to the implicitly associated `shape`, indexed by
-  ///     `layout.iteration_dimensions`, and associated `data_type` equal to
-  ///     `this->data_type()`, an associated `buffer_kind` of
+  ///     `layout.iteration_dimensions`, and associated `dtype` equal to
+  ///     `this->dtype()`, an associated `buffer_kind` of
   ///     `layout.buffer_kind`, and an associated `max_block_size` of
   ///     `layout.block_size`.
   virtual NDIterator::Ptr GetIterator(
