@@ -35,18 +35,17 @@
 
 namespace {
 
-using tensorstore::Array;
-using tensorstore::DimensionIndex;
-using tensorstore::Index;
-using tensorstore::span;
-using tensorstore::Status;
-using tensorstore::StridedLayout;
-using tensorstore::internal::Arena;
-using tensorstore::internal::GetArrayNDIterable;
-using tensorstore::internal::IterationBufferKind;
-using tensorstore::internal::IterationBufferPointer;
-using tensorstore::internal::MultiNDIterator;
-using tensorstore::internal::NDIterable;
+using ::tensorstore::Array;
+using ::tensorstore::DimensionIndex;
+using ::tensorstore::Index;
+using ::tensorstore::span;
+using ::tensorstore::StridedLayout;
+using ::tensorstore::internal::Arena;
+using ::tensorstore::internal::GetArrayNDIterable;
+using ::tensorstore::internal::IterationBufferKind;
+using ::tensorstore::internal::IterationBufferPointer;
+using ::tensorstore::internal::MultiNDIterator;
+using ::tensorstore::internal::NDIterable;
 
 using DirectionPref = NDIterable::DirectionPref;
 
@@ -59,7 +58,7 @@ TEST(NDIterableArrayTest, Direct) {
   Array<std::uint8_t> array(data + 500,
                             StridedLayout<>({6, 3, 4, 5}, {-1, -6, 0, 3}));
   Arena arena;
-  auto iterable = GetArrayNDIterable(array, &arena);
+  auto iterable = GetArrayNDIterable(UnownedToShared(array), &arena);
   {
     std::vector<DirectionPref> direction_prefs(4, DirectionPref::kCanSkip);
     iterable->UpdateDirectionPrefs(direction_prefs.data());
@@ -177,12 +176,12 @@ TEST(NDIterableArrayTest, Direct) {
           /*.block_size=*/3},
          /*.buffer_kind=*/IterationBufferKind::kContiguous});
     IterationBufferPointer pointer;
-    Status status;
+    absl::Status status;
     EXPECT_EQ(3, iterator->GetBlock(span<const Index>({2, 3, 1}), 3, &pointer,
                                     &status));
     EXPECT_EQ(&array((6 - 1) - 1, (3 - 1) - 2, 0, 3), pointer.pointer.get());
     EXPECT_EQ(1, pointer.byte_stride);
-    EXPECT_EQ(Status(), status);
+    EXPECT_EQ(absl::OkStatus(), status);
   }
 
   {
@@ -194,13 +193,13 @@ TEST(NDIterableArrayTest, Direct) {
           /*.block_size=*/3},
          /*.buffer_kind=*/IterationBufferKind::kIndexed});
     IterationBufferPointer pointer;
-    Status status;
+    absl::Status status;
     EXPECT_EQ(3, iterator->GetBlock(span<const Index>({2, 3, 1}), 3, &pointer,
                                     &status));
     EXPECT_EQ(&array((6 - 1) - 1, (3 - 1) - 2, 0, 3), pointer.pointer.get());
     EXPECT_THAT(span<const Index>(pointer.byte_offsets, 3),
                 ::testing::ElementsAre(0, 1, 2));
-    EXPECT_EQ(Status(), status);
+    EXPECT_EQ(absl::OkStatus(), status);
   }
 }
 
@@ -223,16 +222,22 @@ TEST(NDIterableArrayTest, RankZero) {
   EXPECT_EQ(1, multi_iterator.block_size);
 
   EXPECT_EQ(1, multi_iterator.ResetAtBeginning());
-  Status status;
+  absl::Status status;
   EXPECT_TRUE(multi_iterator.GetBlock(1, &status));
   EXPECT_THAT(multi_iterator.position(), ::testing::ElementsAre(0));
-  EXPECT_EQ(Status(), status);
+  EXPECT_EQ(absl::OkStatus(), status);
   EXPECT_EQ(array.data(), multi_iterator.block_pointers()[0].pointer);
   EXPECT_EQ(0, multi_iterator.block_pointers()[0].byte_stride);
   EXPECT_EQ(0, multi_iterator.StepForward(1));
   EXPECT_THAT(multi_iterator.position(), ::testing::ElementsAre(1));
 }
 
+#ifndef TENSORSTORE_INTERNAL_NDITERABLE_TEST_UNIT_BLOCK_SIZE
+constexpr Index ExpectedBlockSize(Index block_size) { return block_size; }
+#else
+constexpr Index ExpectedBlockSize(Index block_size) { return 1; }
+#endif
+
 TEST(NDIterableArrayTest, RankOne) {
   auto array = tensorstore::MakeArray<int>({1, 2, 3, 4, 5});
 
@@ -249,13 +254,13 @@ TEST(NDIterableArrayTest, RankOne) {
               ::testing::ElementsAre(0));
   EXPECT_EQ(IterationBufferKind::kContiguous, multi_iterator.buffer_kind);
   EXPECT_EQ(false, multi_iterator.empty);
-  EXPECT_EQ(5, multi_iterator.block_size);
+  EXPECT_EQ(ExpectedBlockSize(5), multi_iterator.block_size);
 
-  EXPECT_EQ(5, multi_iterator.ResetAtBeginning());
-  Status status;
-  EXPECT_TRUE(multi_iterator.GetBlock(5, &status));
+  EXPECT_EQ(ExpectedBlockSize(5), multi_iterator.ResetAtBeginning());
+  absl::Status status;
+  EXPECT_TRUE(multi_iterator.GetBlock(ExpectedBlockSize(5), &status));
   EXPECT_THAT(multi_iterator.position(), ::testing::ElementsAre(0));
-  EXPECT_EQ(Status(), status);
+  EXPECT_EQ(absl::OkStatus(), status);
   EXPECT_EQ(array.data(), multi_iterator.block_pointers()[0].pointer);
   EXPECT_EQ(sizeof(int), multi_iterator.block_pointers()[0].byte_stride);
   EXPECT_EQ(0, multi_iterator.StepForward(5));
@@ -278,13 +283,13 @@ TEST(NDIterableArrayTest, RankTwoContiguous) {
               ::testing::ElementsAre(0, 1));
   EXPECT_EQ(IterationBufferKind::kContiguous, multi_iterator.buffer_kind);
   EXPECT_EQ(false, multi_iterator.empty);
-  EXPECT_EQ(6, multi_iterator.block_size);
+  EXPECT_EQ(ExpectedBlockSize(6), multi_iterator.block_size);
 
-  EXPECT_EQ(6, multi_iterator.ResetAtBeginning());
-  Status status;
-  EXPECT_TRUE(multi_iterator.GetBlock(6, &status));
+  EXPECT_EQ(ExpectedBlockSize(6), multi_iterator.ResetAtBeginning());
+  absl::Status status;
+  EXPECT_TRUE(multi_iterator.GetBlock(ExpectedBlockSize(6), &status));
   EXPECT_THAT(multi_iterator.position(), ::testing::ElementsAre(0));
-  EXPECT_EQ(Status(), status);
+  EXPECT_EQ(absl::OkStatus(), status);
   EXPECT_EQ(array.data(), multi_iterator.block_pointers()[0].pointer);
   EXPECT_EQ(sizeof(int), multi_iterator.block_pointers()[0].byte_stride);
   EXPECT_EQ(0, multi_iterator.StepForward(6));
@@ -308,25 +313,29 @@ TEST(NDIterableArrayTest, RankTwoTranspose) {
               ::testing::ElementsAre(1, 0));
   EXPECT_EQ(IterationBufferKind::kStrided, multi_iterator.buffer_kind);
   EXPECT_EQ(false, multi_iterator.empty);
-  EXPECT_EQ(2, multi_iterator.block_size);
+  EXPECT_EQ(ExpectedBlockSize(2), multi_iterator.block_size);
 
-  EXPECT_EQ(2, multi_iterator.ResetAtBeginning());
-  Status status;
+  EXPECT_EQ(ExpectedBlockSize(2), multi_iterator.ResetAtBeginning());
+  absl::Status status;
   EXPECT_THAT(multi_iterator.position(), ::testing::ElementsAre(0, 0));
+#ifdef TENSORSTORE_INTERNAL_NDITERABLE_TEST_UNIT_BLOCK_SIZE
+  GTEST_SKIP();
+#endif
+
   EXPECT_TRUE(multi_iterator.GetBlock(2, &status));
-  EXPECT_EQ(Status(), status);
+  EXPECT_EQ(absl::OkStatus(), status);
   EXPECT_EQ(&array(0, 0), multi_iterator.block_pointers()[0].pointer);
   EXPECT_EQ(sizeof(int) * 3, multi_iterator.block_pointers()[0].byte_stride);
   EXPECT_EQ(2, multi_iterator.StepForward(2));
   EXPECT_THAT(multi_iterator.position(), ::testing::ElementsAre(1, 0));
   EXPECT_TRUE(multi_iterator.GetBlock(2, &status));
-  EXPECT_EQ(Status(), status);
+  EXPECT_EQ(absl::OkStatus(), status);
   EXPECT_EQ(&array(0, 1), multi_iterator.block_pointers()[0].pointer);
   EXPECT_EQ(sizeof(int) * 3, multi_iterator.block_pointers()[0].byte_stride);
   EXPECT_EQ(2, multi_iterator.StepForward(2));
   EXPECT_THAT(multi_iterator.position(), ::testing::ElementsAre(2, 0));
   EXPECT_TRUE(multi_iterator.GetBlock(2, &status));
-  EXPECT_EQ(Status(), status);
+  EXPECT_EQ(absl::OkStatus(), status);
   EXPECT_EQ(&array(0, 2), multi_iterator.block_pointers()[0].pointer);
   EXPECT_EQ(sizeof(int) * 3, multi_iterator.block_pointers()[0].byte_stride);
   EXPECT_EQ(0, multi_iterator.StepForward(2));
@@ -342,7 +351,7 @@ TEST(NDIterableArrayTest, SkipSize1Dimension) {
 
   Array<unsigned char> array = {&data[150],
                                 StridedLayout<>({2, 1, 3}, {5, 10, -20})};
-  auto iterable = GetArrayNDIterable(array, &arena);
+  auto iterable = GetArrayNDIterable(UnownedToShared(array), &arena);
 
   MultiNDIterator<1, /*Full=*/true> multi_iterator(array.shape(), {},
                                                    {{iterable.get()}}, &arena);
@@ -361,7 +370,7 @@ TEST(NDIterableArrayTest, SkipZeroByteStride) {
   Arena arena;
 
   Array<unsigned char> array = {&data[150], StridedLayout<>({2, 3}, {5, 0})};
-  auto iterable = GetArrayNDIterable(array, &arena);
+  auto iterable = GetArrayNDIterable(UnownedToShared(array), &arena);
 
   MultiNDIterator<1, /*Full=*/true> multi_iterator(
       array.shape(), tensorstore::skip_repeated_elements, {{iterable.get()}},
@@ -400,7 +409,7 @@ TEST(NDIterableArrayTest, ReversedDimensions) {
       StridedLayout<>({orig_shape[2], orig_shape[0], orig_shape[1]},
                       {-orig_strides[2], orig_strides[0], -orig_strides[1]}));
   Arena arena;
-  auto iterable = GetArrayNDIterable(array, &arena);
+  auto iterable = GetArrayNDIterable(UnownedToShared(array), &arena);
   MultiNDIterator<1, /*Full=*/true> multi_iterator(
       array.shape(), tensorstore::skip_repeated_elements, {{iterable.get()}},
       &arena);
@@ -414,13 +423,13 @@ TEST(NDIterableArrayTest, ReversedDimensions) {
               ::testing::ElementsAre(1, 2, 0));
   EXPECT_EQ(IterationBufferKind::kContiguous, multi_iterator.buffer_kind);
   EXPECT_EQ(false, multi_iterator.empty);
-  EXPECT_EQ(3 * 4 * 5, multi_iterator.block_size);
+  EXPECT_EQ(ExpectedBlockSize(3 * 4 * 5), multi_iterator.block_size);
 
-  EXPECT_EQ(3 * 4 * 5, multi_iterator.ResetAtBeginning());
-  Status status;
+  EXPECT_EQ(ExpectedBlockSize(3 * 4 * 5), multi_iterator.ResetAtBeginning());
+  absl::Status status;
   EXPECT_THAT(multi_iterator.position(), ::testing::ElementsAre(0));
-  EXPECT_TRUE(multi_iterator.GetBlock(3 * 4 * 5, &status));
-  EXPECT_EQ(Status(), status);
+  EXPECT_TRUE(multi_iterator.GetBlock(ExpectedBlockSize(3 * 4 * 5), &status));
+  EXPECT_EQ(absl::OkStatus(), status);
   EXPECT_EQ(orig_array.byte_strided_pointer(),
             multi_iterator.block_pointers()[0].pointer);
   EXPECT_EQ(sizeof(int), multi_iterator.block_pointers()[0].byte_stride);
@@ -446,12 +455,16 @@ TEST(NDIterableArrayTest, MultipleArrays) {
               ::testing::ElementsAre(0, 1));
   EXPECT_EQ(IterationBufferKind::kStrided, multi_iterator.buffer_kind);
   EXPECT_EQ(false, multi_iterator.empty);
-  EXPECT_EQ(3, multi_iterator.block_size);
-  EXPECT_EQ(3, multi_iterator.ResetAtBeginning());
-  Status status;
+  EXPECT_EQ(ExpectedBlockSize(3), multi_iterator.block_size);
+
+  EXPECT_EQ(ExpectedBlockSize(3), multi_iterator.ResetAtBeginning());
+  absl::Status status;
   EXPECT_THAT(multi_iterator.position(), ::testing::ElementsAre(0, 0));
+#ifdef TENSORSTORE_INTERNAL_NDITERABLE_TEST_UNIT_BLOCK_SIZE
+  GTEST_SKIP();
+#endif
   EXPECT_TRUE(multi_iterator.GetBlock(3, &status));
-  EXPECT_EQ(Status(), status);
+  EXPECT_EQ(absl::OkStatus(), status);
   EXPECT_EQ(&array_a(0, 0), multi_iterator.block_pointers()[0].pointer);
   EXPECT_EQ(&array_b(0, 0), multi_iterator.block_pointers()[1].pointer);
   EXPECT_EQ(sizeof(int), multi_iterator.block_pointers()[0].byte_stride);
@@ -460,7 +473,7 @@ TEST(NDIterableArrayTest, MultipleArrays) {
   EXPECT_EQ(3, multi_iterator.StepForward(3));
   EXPECT_THAT(multi_iterator.position(), ::testing::ElementsAre(1, 0));
   EXPECT_TRUE(multi_iterator.GetBlock(3, &status));
-  EXPECT_EQ(Status(), status);
+  EXPECT_EQ(absl::OkStatus(), status);
   EXPECT_EQ(&array_a(1, 0), multi_iterator.block_pointers()[0].pointer);
   EXPECT_EQ(&array_b(1, 0), multi_iterator.block_pointers()[1].pointer);
   EXPECT_EQ(sizeof(int), multi_iterator.block_pointers()[0].byte_stride);
