@@ -20,11 +20,11 @@
 
 #include <algorithm>
 #include <cassert>
+#include <type_traits>
 #include <utility>
 
 #include "absl/container/fixed_array.h"
 #include "absl/container/inlined_vector.h"
-#include "absl/meta/type_traits.h"
 #include "tensorstore/index.h"
 #include "tensorstore/internal/arena.h"
 #include "tensorstore/internal/elementwise_function.h"
@@ -44,8 +44,8 @@ struct NDIterationSimplifiedLayoutInfo;
 /// or `NDIterationFullLayoutInfo` depending on the value of `Full`.
 template <bool Full = false>
 using NDIterationLayoutInfo =
-    absl::conditional_t<Full, NDIterationFullLayoutInfo,
-                        NDIterationSimplifiedLayoutInfo>;
+    std::conditional_t<Full, NDIterationFullLayoutInfo,
+                       NDIterationSimplifiedLayoutInfo>;
 
 /// Computes the simplified iteration layout for a given `iterable` and `shape`.
 ///
@@ -140,6 +140,10 @@ struct NDIterationBufferInfo {
 
 /// Computes the block size to use for iteration that is L1-cache efficient.
 ///
+/// For testing purposes, the behavior may be overridden to always return 1 by
+/// calling `SetNDIterableTestUnitBlockSize(true)` or defining the
+/// `TENSORSTORE_INTERNAL_NDITERABLE_TEST_UNIT_BLOCK_SIZE` preprocessor macro.
+///
 /// \param working_memory_bytes_per_element The number of bytes of temporary
 ///     buffer space required for each block element.
 /// \param iteration_shape The simplified iteration shape, must have
@@ -478,10 +482,11 @@ class NDIterationPositionStepper {
   Index block_size_;
 };
 
-/// Arena with a stack-allocated buffer of 48 KiB.
+/// Arena with a stack-allocated buffer of 32 KiB.
 class DefaultNDIterableArena {
  public:
-  DefaultNDIterableArena() : arena_(buffer_) {}
+  DefaultNDIterableArena()
+      : arena_(/*Workaround gcc -Wuninitialized*/ (buffer_[0] = 0, buffer_)) {}
 
   operator Arena*() { return &arena_; }
 
@@ -491,10 +496,23 @@ class DefaultNDIterableArena {
   }
 
  private:
-  unsigned char buffer_[48 * 1024];
+  unsigned char buffer_[32 * 1024];
   tensorstore::internal::Arena arena_;
 };
 
+#ifndef NDEBUG
+/// If `value` is `true`, forces `GetNDIterationBlockSize` to always return 1.
+///
+/// Note that if `TENSORSTORE_INTERNAL_NDITERABLE_TEST_UNIT_BLOCK_SIZE` is
+/// defined, `GetNDIterationBlockSize` always returns 1 and calling this
+/// function has no effect.
+///
+/// Should only be used for testing purposes, and should not be called
+/// concurrently from multiple threads, or while any other thread is executing
+/// NDIterable code.
+void SetNDIterableTestUnitBlockSize(bool value);
+#endif
+
 }  // namespace internal
 }  // namespace tensorstore
 
