@@ -15,9 +15,12 @@
 #ifndef TENSORSTORE_INTERNAL_TEST_UTIL_H_
 #define TENSORSTORE_INTERNAL_TEST_UTIL_H_
 
+#include <functional>
 #include <string>
+#include <string_view>
 
 #include "absl/status/status.h"
+#include "tensorstore/internal/source_location.h"
 
 namespace tensorstore {
 namespace internal {
@@ -41,6 +44,16 @@ class ScopedTemporaryDirectory {
   std::string path_;
 };
 
+/// Registers a GoogleTest case dynamically.
+void RegisterGoogleTestCaseDynamically(std::string test_suite_name,
+                                       std::string test_name,
+                                       std::function<void()> test_func,
+                                       SourceLocation loc);
+
+/// Returns a seed for a random number generator, which will either be obtained
+/// from `env_var` if specified, or randomly generated and logged.
+unsigned int GetRandomSeedForTest(const char* env_var);
+
 }  // namespace internal
 }  // namespace tensorstore
 #endif  // TENSORSTORE_INTERNAL_TEST_UTIL_H_
