@@ -13,34 +13,39 @@
 // limitations under the License.
 
 #include <errno.h>
-#include <string.h>
 
 #include <cstddef>
+#include <cstring>
 #include <fstream>
-#include <iostream>
-#include <optional>
 #include <string>
-#include <thread>  // NOLINT
+#include <tuple>
+#include <type_traits>
 #include <utility>
 #include <vector>
 
 #include <gmock/gmock.h>
 #include <gtest/gtest.h>
 #include "absl/status/status.h"
+#include "absl/strings/cord.h"
+#include "absl/synchronization/notification.h"
 #include <nlohmann/json.hpp>
 #include "tensorstore/context.h"
 #include "tensorstore/internal/file_io_concurrency_resource.h"
 #include "tensorstore/internal/test_util.h"
+#include "tensorstore/internal/thread.h"
 #include "tensorstore/kvstore/generation.h"
 #include "tensorstore/kvstore/generation_testutil.h"
-#include "tensorstore/kvstore/key_value_store.h"
-#include "tensorstore/kvstore/key_value_store_testutil.h"
-#include "tensorstore/util/execution.h"
+#include "tensorstore/kvstore/key_range.h"
+#include "tensorstore/kvstore/kvstore.h"
+#include "tensorstore/kvstore/test_util.h"
+#include "tensorstore/util/execution/execution.h"
+#include "tensorstore/util/execution/sender_testutil.h"
 #include "tensorstore/util/executor.h"
-#include "tensorstore/util/sender_testutil.h"
+#include "tensorstore/util/future.h"
+#include "tensorstore/util/result.h"
 #include "tensorstore/util/status.h"
 #include "tensorstore/util/status_testutil.h"
-#include "tensorstore/util/to_string.h"
+#include "tensorstore/util/str_cat.h"
 
 // Include system headers last to reduce impact of macros.
 #ifndef _WIN32
@@ -50,19 +55,19 @@
 
 namespace {
 
-using tensorstore::CompletionNotifyingReceiver;
-using tensorstore::Context;
-using tensorstore::KeyValueStore;
-using tensorstore::MatchesStatus;
-using tensorstore::Status;
-using tensorstore::StorageGeneration;
-using tensorstore::internal::MatchesKvsReadResultNotFound;
-using tensorstore::internal::MatchesTimestampedStorageGeneration;
+namespace kvstore = tensorstore::kvstore;
+using ::tensorstore::CompletionNotifyingReceiver;
+using ::tensorstore::Context;
+using ::tensorstore::KeyRange;
+using ::tensorstore::KvStore;
+using ::tensorstore::MatchesStatus;
+using ::tensorstore::StorageGeneration;
+using ::tensorstore::internal::MatchesKvsReadResultNotFound;
+using ::tensorstore::internal::MatchesTimestampedStorageGeneration;
+using ::testing::HasSubstr;
 
-KeyValueStore::Ptr GetStore(std::string root) {
-  return KeyValueStore::Open(Context::Default(),
-                             {{"driver", "file"}, {"path", root}})
-      .value();
+KvStore GetStore(std::string root) {
+  return kvstore::Open({{"driver", "file"}, {"path", root + "/"}}).value();
 }
 
 TEST(FileKeyValueStoreTest, Basic) {
@@ -77,49 +82,43 @@ TEST(FileKeyValueStoreTest, InvalidKey) {
   std::string root = tempdir.path() + "/root";
   auto store = GetStore(root);
 
-  EXPECT_THAT(store->Read("this_is_a_long_key").result(),
+  EXPECT_THAT(kvstore::Read(store, "this_is_a_long_key").result(),
               MatchesKvsReadResultNotFound());
   EXPECT_THAT(
-      GetStatus(store->Read("").result()),
+      kvstore::Read(store, "").result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
   EXPECT_THAT(
-      GetStatus(store->Read(std::string("\0", 1)).result()),
+      kvstore::Read(store, std::string("\0", 1)).result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
   EXPECT_THAT(
-      GetStatus(store->Write("", {}).result()),
+      kvstore::Write(store, "", {}).result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
   EXPECT_THAT(
-      GetStatus(store->Read("/").result()),
+      kvstore::Read(store, "/").result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
   EXPECT_THAT(
-      GetStatus(store->Read(".").result()),
+      kvstore::Read(store, ".").result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
   EXPECT_THAT(
-      GetStatus(store->Read("..").result()),
+      kvstore::Read(store, "..").result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
   EXPECT_THAT(
-      GetStatus(store->Read("a//b").result()),
+      kvstore::Read(store, "a/./b").result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
   EXPECT_THAT(
-      GetStatus(store->Read("a/./b").result()),
+      kvstore::Read(store, "a/../b").result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
   EXPECT_THAT(
-      GetStatus(store->Read("a/../b").result()),
+      kvstore::Read(store, "a/").result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
   EXPECT_THAT(
-      GetStatus(store->Read("a/").result()),
+      kvstore::Read(store, "a.__lock").result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
   EXPECT_THAT(
-      GetStatus(store->Read("/a").result()),
+      kvstore::Read(store, "a/b.__lock/c").result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
   EXPECT_THAT(
-      GetStatus(store->Read("a.__lock").result()),
-      MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
-  EXPECT_THAT(
-      GetStatus(store->Read("a/b.__lock/c").result()),
-      MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
-  EXPECT_THAT(
-      GetStatus(store->Read("///").result()),
+      kvstore::Read(store, "///").result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument, "Invalid key: .*"));
 }
 
@@ -144,12 +143,14 @@ TEST(FileKeyValueStoreTest, LockFiles) {
   std::string root = tempdir.path() + "/root";
   auto store = GetStore(root);
   TENSORSTORE_ASSERT_OK(
-      store->Write("a/foo", "xyz", {/*.if_equal=*/StorageGeneration::NoValue()})
+      kvstore::Write(store, "a/foo", absl::Cord("xyz"),
+                     {/*.if_equal=*/StorageGeneration::NoValue()})
           .result());
   EXPECT_THAT(GetDirectoryContents(root),
               ::testing::UnorderedElementsAre("a", "a/foo"));
   EXPECT_THAT(
-      store->Write("a/foo", "qqq", {/*.if_equal=*/StorageGeneration::NoValue()})
+      kvstore::Write(store, "a/foo", absl::Cord("qqq"),
+                     {/*.if_equal=*/StorageGeneration::NoValue()})
           .result(),
       MatchesTimestampedStorageGeneration(StorageGeneration::Unknown()));
 
@@ -164,11 +165,12 @@ TEST(FileKeyValueStoreTest, LockFiles) {
               ::testing::UnorderedElementsAre("a", "a/foo", "a/foo.__lock"));
 
   // Test that the lock file is not included in the `List` result.
-  EXPECT_THAT(ListFuture(store.get()).result(),
+  EXPECT_THAT(ListFuture(store).result(),
               ::testing::Optional(::testing::UnorderedElementsAre("a/foo")));
 
   // Test that a stale lock file does not interfere with writing.
-  TENSORSTORE_ASSERT_OK(store->Write("a/foo", "xyz").result());
+  TENSORSTORE_ASSERT_OK(
+      kvstore::Write(store, "a/foo", absl::Cord("xyz")).result());
 
   // Recreate the lock file.
   std::ofstream(root + "/a/foo.__lock");
@@ -176,7 +178,7 @@ TEST(FileKeyValueStoreTest, LockFiles) {
   // Test that the "a" prefix can be deleted despite the presence of the lock
   // file.  Only a single key, "a/foo" is removed.  The lock file should not be
   // included in the count.
-  EXPECT_THAT(store->DeletePrefix("a/").result(), ::testing::Optional(1));
+  TENSORSTORE_EXPECT_OK(kvstore::DeleteRange(store, KeyRange::Prefix("a/")));
   EXPECT_THAT(GetDirectoryContents(root), ::testing::UnorderedElementsAre());
 }
 
@@ -184,19 +186,23 @@ TEST(FileKeyValueStoreTest, NestedDirectories) {
   tensorstore::internal::ScopedTemporaryDirectory tempdir;
   std::string root = tempdir.path() + "/root";
   auto store = GetStore(root);
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/foo", "xyz").result()));
+  TENSORSTORE_EXPECT_OK(kvstore::Write(store, "a/foo", absl::Cord("xyz")));
 
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/ba/ccc/dddd", "xyz").result()));
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/ba/ccc/foo", "xyz").result()));
+  TENSORSTORE_EXPECT_OK(
+      kvstore::Write(store, "a/ba/ccc/dddd", absl::Cord("xyz")));
+  TENSORSTORE_EXPECT_OK(
+      kvstore::Write(store, "a/ba/ccc/foo", absl::Cord("xyz")));
   EXPECT_THAT(
-      GetStatus(store->Write("a/ba/ccc", "xyz").result()),
+      kvstore::Write(store, "a/ba/ccc", absl::Cord("xyz")).result(),
       ::testing::AnyOf(MatchesStatus(absl::StatusCode::kPermissionDenied),
                        MatchesStatus(absl::StatusCode::kFailedPrecondition)));
 }
 
 TEST(FileKeyValueStoreTest, ConcurrentWrites) {
   constexpr std::size_t num_threads = 4;
-  std::vector<std::thread> threads;
+  std::vector<tensorstore::internal::Thread> threads;
+  threads.reserve(num_threads);
+
   tensorstore::internal::ScopedTemporaryDirectory tempdir;
   std::string root = tempdir.path() + "/root";
   auto store = GetStore(root);
@@ -204,42 +210,45 @@ TEST(FileKeyValueStoreTest, ConcurrentWrites) {
   std::string initial_value;
   initial_value.resize(sizeof(std::size_t) * num_threads);
   auto initial_generation =
-      store->Write(key, {initial_value}).value().generation;
+      kvstore::Write(store, key, absl::Cord(initial_value)).value().generation;
   constexpr std::size_t num_iterations = 100;
   for (std::size_t thread_i = 0; thread_i < num_threads; ++thread_i) {
-    threads.push_back(std::thread([&, thread_i] {
-      StorageGeneration generation = initial_generation;
-      std::string value = initial_value;
-      for (std::size_t i = 0; i < num_iterations; ++i) {
-        const std::size_t value_offset = sizeof(std::size_t) * thread_i;
-        while (true) {
-          std::size_t x;
-          std::memcpy(&x, &value[value_offset], sizeof(std::size_t));
-          ASSERT_EQ(i, x);
-          std::string new_value = value;
-          x = i + 1;
-          std::memcpy(&new_value[value_offset], &x, sizeof(std::size_t));
-          auto write_result =
-              store->Write(key, new_value, {generation}).result();
-          ASSERT_EQ(Status(), GetStatus(write_result));
-          if (!StorageGeneration::IsUnknown(write_result->generation)) {
-            generation = write_result->generation;
-            value = new_value;
-            break;
+    threads.emplace_back(
+        tensorstore::internal::Thread({"concurrent_write"}, [&, thread_i] {
+          StorageGeneration generation = initial_generation;
+          std::string value = initial_value;
+          for (std::size_t i = 0; i < num_iterations; ++i) {
+            const std::size_t value_offset = sizeof(std::size_t) * thread_i;
+            while (true) {
+              std::size_t x;
+              std::memcpy(&x, &value[value_offset], sizeof(std::size_t));
+              ASSERT_EQ(i, x);
+              std::string new_value = value;
+              x = i + 1;
+              std::memcpy(&new_value[value_offset], &x, sizeof(std::size_t));
+              TENSORSTORE_ASSERT_OK_AND_ASSIGN(
+                  auto write_result,
+                  kvstore::Write(store, key, absl::Cord(new_value),
+                                 {generation})
+                      .result());
+              if (!StorageGeneration::IsUnknown(write_result.generation)) {
+                generation = write_result.generation;
+                value = new_value;
+                break;
+              }
+              TENSORSTORE_ASSERT_OK_AND_ASSIGN(
+                  auto read_result, kvstore::Read(store, key).result());
+              ASSERT_FALSE(read_result.aborted() || read_result.not_found());
+              value = std::string(read_result.value);
+              ASSERT_EQ(sizeof(std::size_t) * num_threads, value.size());
+              generation = read_result.stamp.generation;
+            }
           }
-          auto read_result = store->Read(key).result();
-          ASSERT_EQ(Status(), GetStatus(read_result));
-          ASSERT_FALSE(read_result->aborted() || read_result->not_found());
-          value = *read_result->value;
-          ASSERT_EQ(sizeof(std::size_t) * num_threads, value.size());
-          generation = read_result->generation.generation;
-        }
-      }
-    }));
+        }));
   }
-  for (auto& t : threads) t.join();
+  for (auto& t : threads) t.Join();
   {
-    auto read_result = store->Read(key).result();
+    auto read_result = kvstore::Read(store, key).result();
     ASSERT_TRUE(read_result);
     std::string expected_value;
     expected_value.resize(sizeof(std::size_t) * num_threads);
@@ -248,7 +257,7 @@ TEST(FileKeyValueStoreTest, ConcurrentWrites) {
       std::memcpy(const_cast<char*>(expected_value.data()),
                   expected_nums.data(), expected_value.size());
     }
-    EXPECT_THAT(read_result->value, ::testing::Optional(expected_value));
+    EXPECT_EQ(expected_value, read_result->value);
   }
 }
 
@@ -264,7 +273,8 @@ TEST(FileKeyValueStoreTest, Permissions) {
   tensorstore::internal::ScopedTemporaryDirectory tempdir;
   std::string root = tempdir.path() + "/root";
   auto store = GetStore(root);
-  EXPECT_EQ(Status(), GetStatus(store->Write("foo", "xyz").result()));
+  TENSORSTORE_ASSERT_OK(
+      kvstore::Write(store, "foo", absl::Cord("xyz")).result());
 
   // Remove write permission on directory.
   ASSERT_EQ(0, ::chmod(root.c_str(), 0500))
@@ -282,24 +292,25 @@ TEST(FileKeyValueStoreTest, Permissions) {
   RestoreWritePermission restore{root};
 
   // Read should still succeed.
-  EXPECT_EQ("xyz", store->Read("foo").value().value);
+  EXPECT_EQ("xyz", kvstore::Read(store, "foo").value().value);
 
   // Writing an existing key should fail.
-  EXPECT_THAT(store->Write("foo", "abc").result(),
+  EXPECT_THAT(kvstore::Write(store, "foo", absl::Cord("abc")).result(),
               MatchesStatus(absl::StatusCode::kPermissionDenied));
 
   // Value should not have changed.
-  EXPECT_EQ("xyz", store->Read("foo").value().value);
+  EXPECT_EQ("xyz", kvstore::Read(store, "foo").value().value);
 
   // Writing a new key should fail.
-  EXPECT_THAT(store->Write("bar", "abc").result(),
+  EXPECT_THAT(kvstore::Write(store, "bar", absl::Cord("abc")).result(),
               MatchesStatus(absl::StatusCode::kPermissionDenied));
 
   // Value should not exist.
-  EXPECT_THAT(store->Read("bar").result(), MatchesKvsReadResultNotFound());
+  EXPECT_THAT(kvstore::Read(store, "bar").result(),
+              MatchesKvsReadResultNotFound());
 
   // Delete should fail.
-  EXPECT_THAT(store->Delete("foo").result(),
+  EXPECT_THAT(kvstore::Delete(store, "foo").result(),
               MatchesStatus(absl::StatusCode::kPermissionDenied));
 
   // Remove read permission on file.
@@ -307,7 +318,7 @@ TEST(FileKeyValueStoreTest, Permissions) {
       << "Error " << errno << ": " << ::strerror(errno);
 
   // Read should fail.
-  EXPECT_THAT(store->Read("foo").result(),
+  EXPECT_THAT(kvstore::Read(store, "foo").result(),
               MatchesStatus(absl::StatusCode::kPermissionDenied));
 }
 #endif
@@ -316,25 +327,28 @@ TEST(FileKeyValueStoreTest, DeletePrefix) {
   tensorstore::internal::ScopedTemporaryDirectory tempdir;
   std::string root = tempdir.path() + "/root";
   auto store = GetStore(root);
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/b", "xyz").result()));
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/d", "xyz").result()));
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/c/x", "xyz").result()));
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/c/y", "xyz").result()));
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/c/z/e", "xyz").result()));
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/c/z/f", "xyz").result()));
+  tensorstore::internal::TestKeyValueStoreDeletePrefix(store);
+}
 
-  EXPECT_THAT(store->DeletePrefix("a/c").result(),
-              MatchesStatus(absl::StatusCode::kInvalidArgument));
+TEST(FileKeyValueStoreTest, DeleteRange) {
+  tensorstore::internal::ScopedTemporaryDirectory tempdir;
+  std::string root = tempdir.path() + "/root";
+  auto store = GetStore(root);
+  tensorstore::internal::TestKeyValueStoreDeleteRange(store);
+}
 
-  EXPECT_EQ(4, store->DeletePrefix("a/c/").result());
+TEST(FileKeyValueStoreTest, DeleteRangeToEnd) {
+  tensorstore::internal::ScopedTemporaryDirectory tempdir;
+  std::string root = tempdir.path() + "/root";
+  auto store = GetStore(root);
+  tensorstore::internal::TestKeyValueStoreDeleteRangeToEnd(store);
+}
 
-  EXPECT_EQ("xyz", store->Read("a/b").value().value);
-  EXPECT_EQ("xyz", store->Read("a/d").value().value);
-
-  EXPECT_THAT(store->Read("a/c/x").result(), MatchesKvsReadResultNotFound());
-  EXPECT_THAT(store->Read("a/c/y").result(), MatchesKvsReadResultNotFound());
-  EXPECT_THAT(store->Read("a/c/z/e").result(), MatchesKvsReadResultNotFound());
-  EXPECT_THAT(store->Read("a/c/z/f").result(), MatchesKvsReadResultNotFound());
+TEST(FileKeyValueStoreTest, DeleteRangeFromBeginning) {
+  tensorstore::internal::ScopedTemporaryDirectory tempdir;
+  std::string root = tempdir.path() + "/root";
+  auto store = GetStore(root);
+  tensorstore::internal::TestKeyValueStoreDeleteRangeFromBeginning(store);
 }
 
 TEST(FileKeyValueStoreTest, ListErrors) {
@@ -346,14 +360,15 @@ TEST(FileKeyValueStoreTest, ListErrors) {
     absl::Notification notification;
     std::vector<std::string> log;
     tensorstore::execution::submit(
-        store->List({"a"}),
+        kvstore::List(store, {KeyRange::Prefix("a//")}),
         CompletionNotifyingReceiver{&notification,
                                     tensorstore::LoggingReceiver{&log}});
     notification.WaitForNotification();
     EXPECT_THAT(log,
-                ::testing::ElementsAre("set_error: INVALID_ARGUMENT: "
-                                       "Prefix must be empty or end with '/'",
-                                       "set_stopping"));
+                ::testing::ElementsAre(
+                    "set_starting",
+                    HasSubstr("set_error: INVALID_ARGUMENT: Invalid key: "),
+                    "set_stopping"));
   }
 }
 
@@ -366,7 +381,7 @@ TEST(FileKeyValueStoreTest, List) {
     absl::Notification notification;
     std::vector<std::string> log;
     tensorstore::execution::submit(
-        store->List({""}),
+        kvstore::List(store, {}),
         CompletionNotifyingReceiver{&notification,
                                     tensorstore::LoggingReceiver{&log}});
     notification.WaitForNotification();
@@ -374,19 +389,19 @@ TEST(FileKeyValueStoreTest, List) {
                                             "set_stopping"));
   }
 
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/b", "xyz").result()));
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/d", "xyz").result()));
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/c/x", "xyz").result()));
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/c/y", "xyz").result()));
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/c/z/e", "xyz").result()));
-  EXPECT_EQ(Status(), GetStatus(store->Write("a/c/z/f", "xyz").result()));
+  TENSORSTORE_EXPECT_OK(kvstore::Write(store, "a/b", absl::Cord("xyz")));
+  TENSORSTORE_EXPECT_OK(kvstore::Write(store, "a/d", absl::Cord("xyz")));
+  TENSORSTORE_EXPECT_OK(kvstore::Write(store, "a/c/x", absl::Cord("xyz")));
+  TENSORSTORE_EXPECT_OK(kvstore::Write(store, "a/c/y", absl::Cord("xyz")));
+  TENSORSTORE_EXPECT_OK(kvstore::Write(store, "a/c/z/e", absl::Cord("xyz")));
+  TENSORSTORE_EXPECT_OK(kvstore::Write(store, "a/c/z/f", absl::Cord("xyz")));
 
   // Listing the entire stream works.
   {
     absl::Notification notification;
     std::vector<std::string> log;
     tensorstore::execution::submit(
-        store->List({""}),
+        kvstore::List(store, {}),
         CompletionNotifyingReceiver{&notification,
                                     tensorstore::LoggingReceiver{&log}});
     notification.WaitForNotification();
@@ -403,7 +418,7 @@ TEST(FileKeyValueStoreTest, List) {
     absl::Notification notification;
     std::vector<std::string> log;
     tensorstore::execution::submit(
-        store->List({"a/c/"}),
+        kvstore::List(store, {KeyRange::Prefix("a/c/")}),
         CompletionNotifyingReceiver{&notification,
                                     tensorstore::LoggingReceiver{&log}});
     notification.WaitForNotification();
@@ -426,7 +441,7 @@ TEST(FileKeyValueStoreTest, List) {
     absl::Notification notification;
     std::vector<std::string> log;
     tensorstore::execution::submit(
-        store->List({""}),
+        kvstore::List(store, {}),
         CompletionNotifyingReceiver{&notification, CancelOnStarting{{&log}}});
     notification.WaitForNotification();
 
@@ -436,7 +451,7 @@ TEST(FileKeyValueStoreTest, List) {
 
   // Cancellation in the middle of the stream stops the stream.
   struct CancelAfter2 : public tensorstore::LoggingReceiver {
-    using Key = tensorstore::KeyValueStore::Key;
+    using Key = kvstore::Key;
     tensorstore::AnyCancelReceiver cancel;
 
     void set_starting(tensorstore::AnyCancelReceiver do_cancel) {
@@ -456,7 +471,7 @@ TEST(FileKeyValueStoreTest, List) {
     absl::Notification notification;
     std::vector<std::string> log;
     tensorstore::execution::submit(
-        store->List({""}),
+        kvstore::List(store, {}),
         CompletionNotifyingReceiver{&notification, CancelAfter2{{&log}}});
     notification.WaitForNotification();
 
@@ -484,20 +499,33 @@ TEST(FileKeyValueStoreTest, InvalidSpec) {
 
   // Test with extra key.
   EXPECT_THAT(
-      KeyValueStore::Open(
-          context, {{"driver", "file"}, {"path", root}, {"extra", "key"}})
+      kvstore::Open({{"driver", "file"}, {"path", root}, {"extra", "key"}},
+                    context)
           .result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument));
 
-  // Test with missing `"path"` key.
-  EXPECT_THAT(KeyValueStore::Open(context, {{"driver", "file"}}).result(),
-              MatchesStatus(absl::StatusCode::kInvalidArgument));
-
   // Test with invalid `"path"` key.
   EXPECT_THAT(
-      KeyValueStore::Open(context, {{"driver", "file"}, {"path", 5}}, {})
-          .result(),
+      kvstore::Open({{"driver", "file"}, {"path", 5}}, context).result(),
       MatchesStatus(absl::StatusCode::kInvalidArgument));
 }
 
+TEST(FileKeyValueStoreTest, UrlRoundtrip) {
+  tensorstore::internal::TestKeyValueStoreUrlRoundtrip({{"driver", "file"}},
+                                                       "file://");
+  tensorstore::internal::TestKeyValueStoreUrlRoundtrip(
+      {{"driver", "file"}, {"path", "/abc/"}}, "file:///abc/");
+  tensorstore::internal::TestKeyValueStoreUrlRoundtrip(
+      {{"driver", "file"}, {"path", "/abc def/"}}, "file:///abc%20def/");
+}
+
+TEST(FileKeyValueStoreTest, InvalidUri) {
+  EXPECT_THAT(kvstore::Spec::FromUrl("file://abc?query"),
+              MatchesStatus(absl::StatusCode::kInvalidArgument,
+                            ".*: Query string not supported"));
+  EXPECT_THAT(kvstore::Spec::FromUrl("file://abc#fragment"),
+              MatchesStatus(absl::StatusCode::kInvalidArgument,
+                            ".*: Fragment identifier not supported"));
+}
+
 }  // namespace
