@@ -17,14 +17,13 @@
 #include <string>
 
 #include "tensorstore/data_type.h"
-#include "tensorstore/util/to_string.h"
+#include "tensorstore/util/str_cat.h"
 
 namespace tensorstore {
 namespace internal_element_pointer {
 
-std::string DescribeForCast(DataType data_type) {
-  return StrCat("pointer with ",
-                StaticCastTraits<DataType>::Describe(data_type));
+std::string DescribeForCast(DataType dtype) {
+  return StrCat("pointer with ", StaticCastTraits<DataType>::Describe(dtype));
 }
 
 }  // namespace internal_element_pointer
