@@ -18,7 +18,7 @@
 /// \file
 /// Defines the `ElementPointer` and `SharedElementPointer` classes that combine
 /// a raw pointer or shared_ptr to an `Element` with a
-/// `StaticOrDynamicDataTypeOf<Element>` value.
+/// `dtype_t<Element>` value.
 
 #include <cstddef>
 #include <memory>
@@ -26,7 +26,6 @@
 #include <type_traits>
 #include <utility>
 
-#include "absl/meta/type_traits.h"
 #include "tensorstore/data_type.h"
 #include "tensorstore/index.h"
 #include "tensorstore/internal/compressed_pair.h"
@@ -39,37 +38,41 @@
 
 namespace tensorstore {
 
-/// Tag type that represents an Element type to be held via a std::shared_ptr
-/// rather than a raw, unowned pointer.
+/// Tag type that represents an `Element` type to be held via an
+/// `std::shared_ptr` rather than a raw, unowned pointer.
+///
+/// \relates ElementPointer
 template <typename Element>
 class Shared;  // Intentionally left undefined.
 
-/// `bool`-valued metafunction that evaluates to `true` if `T` is `Shared<U>` or
-/// `const Shared<U>`.
+/// `bool`-valued metafunction that evaluates to `true` if `T` is an instance of
+/// `Shared`.
+///
+/// \relates Shared
 template <typename T>
-struct IsShared : public std::false_type {};
+constexpr inline bool IsShared = false;
+
 template <typename T>
-struct IsShared<Shared<T>> : public std::true_type {};
+constexpr inline bool IsShared<Shared<T>> = true;
 template <typename T>
-struct IsShared<const Shared<T>> : public std::true_type {};
+constexpr inline bool IsShared<const Shared<T>> = true;
 
 /// An ElementTag type is a type `T` or `Shared<T>` where `T` satisfies
 /// `IsElementType<T>`.  It specifies a pointer type of `T*` or
 /// `std::shared_ptr<T>`, respectively.
+///
+/// \relates ElementPointer
 template <typename T>
-struct IsElementTag
-    : public std::integral_constant<bool,
-                                    (IsElementType<T>::value &&
-                                     // Detect accidental `const`-qualification
-                                     // of `Shared<T>`.
-                                     !IsShared<T>::value)> {};
+constexpr inline bool IsElementTag = (IsElementType<T> &&
+                                      // Detect accidental `const`-qualification
+                                      // of `Shared<T>`.
+                                      !IsShared<T>);
 
 template <typename T>
-struct IsElementTag<Shared<T>>
-    : public std::integral_constant<bool, (IsElementType<T>::value &&
-                                           // Detect accidental nested
-                                           // `Shared<Shared<T>>`.
-                                           !IsShared<T>::value)> {};
+constexpr inline bool IsElementTag<Shared<T>> = (IsElementType<T> &&
+                                                 // Detect accidental nested
+                                                 // `Shared<Shared<T>>`.
+                                                 !IsShared<T>);
 
 /// Traits for element tag types.
 ///
@@ -78,18 +81,24 @@ struct IsElementTag<Shared<T>>
 /// element tag with the same ownership semantics for a different element type.
 ///
 /// \tparam T Type satisfying `IsElementTag<T>`.
+/// \relates ElementPointer
 template <typename T>
 struct ElementTagTraits {
-  static_assert(IsElementTag<T>::value, "T must be an ElementTag type.");
+  static_assert(IsElementTag<T>, "T must be an ElementTag type.");
+  /// Element type of `Pointer`.
   using Element = T;
+
+  /// Pointer type.
   using Pointer = T*;
+
+  /// Same kind of tag type, but for another element type `U`.
   template <typename U>
   using rebind = U;
 };
 
 template <typename T>
 struct ElementTagTraits<Shared<T>> {
-  static_assert(IsElementTag<Shared<T>>::value,
+  static_assert(IsElementTag<Shared<T>>,
                 "Shared<T> must be an ElementTag type.");
   using Element = T;
   using Pointer = std::shared_ptr<T>;
@@ -97,13 +106,7 @@ struct ElementTagTraits<Shared<T>> {
   using rebind = Shared<U>;
 };
 
-/// Metafunction with a nested `type` alias that specifies the element tag type
-/// corresponding to a given pointer type.
-///
-/// This is the inverse metafunction of `ElementTagPointerType`.
-///
-/// \tparam T Pointer type corresponding to an element tag.  Must be either `U*`
-///     or `std::shared_ptr<U>`.
+namespace internal_element_pointer {
 template <typename T>
 struct PointerElementTagType;
 
@@ -116,99 +119,113 @@ template <typename T>
 struct PointerElementTagType<std::shared_ptr<T>> {
   using type = Shared<T>;
 };
+}  // namespace internal_element_pointer
 
+/// Element tag type corresponding to a given pointer type.
+///
+/// This is the inverse metafunction of `ElementTagTraits::Pointer`.
+///
+/// \tparam T Pointer type corresponding to an element tag.  Must be either
+///     ``U*`` or ``std::shared_ptr<U>``.
+///
+/// \relates ElementPointer
 template <typename T>
-using PointerElementTag = typename PointerElementTagType<T>::type;
+using PointerElementTag =
+    typename internal_element_pointer::PointerElementTagType<T>::type;
 
 /// `bool`-valued metafunction that evaluates to `true` if `SourcePointer` may
 /// be converted to `TargetPointer` when used as an array base pointer.
 ///
 /// Specifically, the following conversions are permitted:
 ///
-///     T*                    -> U*
-///     ByteStridedPointer<T> -> U*
-///     shared_ptr<T>         -> U*
-///     shared_ptr<T>         -> shared_ptr<U>
+/// =========================    =========================
+/// `SourcePointer`              `TargetPointer`
+/// =========================    =========================
+/// ``T*``                       ``U*``
+/// ``ByteStridedPointer<T>``    ``U*``
+/// ``std::shared_ptr<T>``       ``U*``
+/// ``std::shared_ptr<T>``       ``std::shared_ptr<U>``
+/// =========================    =========================
 ///
-/// where `IsElementTypeImplicitlyConvertible<T,U>::value` is `true`.
+/// where ``IsElementTypeImplicitlyConvertible<T,U>`` is `true`.
+///
+/// \relates ElementPointer
 template <typename SourcePointer, typename TargetPointer>
-struct IsArrayBasePointerConvertible : public std::false_type {};
+constexpr inline bool IsArrayBasePointerConvertible = false;
 
 template <typename SourceElement, typename TargetElement>
-struct IsArrayBasePointerConvertible<SourceElement*, TargetElement*>
-    : public IsElementTypeImplicitlyConvertible<SourceElement, TargetElement> {
-};
+constexpr inline bool
+    IsArrayBasePointerConvertible<SourceElement*, TargetElement*> =
+        IsElementTypeImplicitlyConvertible<SourceElement, TargetElement>;
 
 template <typename SourceElement, typename TargetElement>
-struct IsArrayBasePointerConvertible<ByteStridedPointer<SourceElement>,
-                                     TargetElement*>
-    : public IsElementTypeImplicitlyConvertible<SourceElement, TargetElement> {
-};
+constexpr inline bool IsArrayBasePointerConvertible<
+    ByteStridedPointer<SourceElement>, TargetElement*> =
+    IsElementTypeImplicitlyConvertible<SourceElement, TargetElement>;
 
 template <typename SourceElement, typename TargetElement>
-struct IsArrayBasePointerConvertible<std::shared_ptr<SourceElement>,
-                                     TargetElement*>
-    : public IsElementTypeImplicitlyConvertible<SourceElement, TargetElement> {
-};
+constexpr inline bool IsArrayBasePointerConvertible<
+    std::shared_ptr<SourceElement>, TargetElement*> =
+    IsElementTypeImplicitlyConvertible<SourceElement, TargetElement>;
 
 template <typename SourceElement, typename TargetElement>
-struct IsArrayBasePointerConvertible<std::shared_ptr<SourceElement>,
-                                     std::shared_ptr<TargetElement>>
-    : public IsElementTypeImplicitlyConvertible<SourceElement, TargetElement> {
-};
+constexpr inline bool IsArrayBasePointerConvertible<
+    std::shared_ptr<SourceElement>, std::shared_ptr<TargetElement>> =
+    IsElementTypeImplicitlyConvertible<SourceElement, TargetElement>;
 
+/// Determine if `Source` and `Target` are instances of `ElementPointer` and
+/// `Source` is (explicitly) convertible to `Target`.
+///
+/// \relates ElementPointer
 template <typename Source, typename Target>
-struct IsElementPointerCastConvertible : public std::false_type {};
+constexpr inline bool IsElementPointerCastConvertible = false;
 
 template <typename ElementTag>
 class ElementPointer;
 
 template <typename SourceTag, typename TargetTag>
-struct IsElementPointerCastConvertible<ElementPointer<SourceTag>,
-                                       ElementPointer<TargetTag>>
-    : public std::integral_constant<
-          bool, ((IsShared<SourceTag>::value >= IsShared<TargetTag>::value) &&
-                 IsElementTypeExplicitlyConvertible<
-                     absl::remove_const_t<
-                         typename ElementTagTraits<SourceTag>::Element>,
-                     absl::remove_const_t<typename ElementTagTraits<
-                         TargetTag>::Element>>::value)> {};
+constexpr inline bool IsElementPointerCastConvertible<
+    ElementPointer<SourceTag>, ElementPointer<TargetTag>> =
+    ((IsShared<SourceTag> >= IsShared<TargetTag>)&&  //
+     IsElementTypeExplicitlyConvertible<
+         std::remove_const_t<typename ElementTagTraits<SourceTag>::Element>,
+         std::remove_const_t<typename ElementTagTraits<TargetTag>::Element>>);
 
-/// `bool`-valued metafunction that evaluates to `true` if `T` is `Element*` or
-/// `std::shared_ptr<Element>`, where `Element` is non-`void`.
+/// `bool`-valued metafunction that evaluates to `true` if `T` is
+/// ``Element*`` or ``std::shared_ptr<Element>``, where ``Element`` is
+/// non-`void`.
+///
+/// \relates ElementPointer
 template <typename T>
-struct IsNonVoidArrayBasePointer : public std::false_type {};
+constexpr inline bool IsNonVoidArrayBasePointer = false;
 
 template <typename T>
-struct IsNonVoidArrayBasePointer<T*>
-    : public std::integral_constant<bool, !std::is_void<T>::value> {};
+constexpr inline bool IsNonVoidArrayBasePointer<T*> = !std::is_void_v<T>;
 
 template <typename T>
-struct IsNonVoidArrayBasePointer<std::shared_ptr<T>>
-    : public std::integral_constant<bool, !std::is_void<T>::value> {};
+constexpr inline bool IsNonVoidArrayBasePointer<std::shared_ptr<T>> =
+    !std::is_void_v<T>;
 
 namespace internal_element_pointer {
 template <typename Target, typename Source>
-inline absl::enable_if_t<
-    (std::is_pointer<Target>::value ==
-     std::is_pointer<internal::remove_cvref_t<Source>>::value),
-    Target>
+inline std::enable_if_t<(std::is_pointer_v<Target> ==
+                         std::is_pointer_v<internal::remove_cvref_t<Source>>),
+                        Target>
 ConvertPointer(Source&& x) {
   return internal::StaticConstPointerCast<
       typename std::pointer_traits<Target>::element_type>(
       static_cast<Source&&>(x));
 }
 template <typename Target, typename Source>
-inline absl::enable_if_t<
-    (std::is_pointer<Target>::value >
-     std::is_pointer<internal::remove_cvref_t<Source>>::value),
-    Target>
+inline std::enable_if_t<(std::is_pointer_v<Target> >
+                         std::is_pointer_v<internal::remove_cvref_t<Source>>),
+                        Target>
 ConvertPointer(Source&& x) {
   return internal::StaticConstPointerCast<
       typename std::pointer_traits<Target>::element_type>(x.get());
 }
 
-std::string DescribeForCast(DataType data_type);
+std::string DescribeForCast(DataType dtype);
 }  // namespace internal_element_pointer
 
 template <typename ElementTagType>
@@ -216,165 +233,174 @@ class ElementPointer;
 
 /// `bool`-valued metafunction that evaluates to `true` if `T` is an instance of
 /// `ElementPointer`.
+///
+/// \relates ElementPointer
 template <typename T>
-struct IsElementPointer : public std::false_type {};
+constexpr inline bool IsElementPointer = false;
 
 template <typename PointerType>
-struct IsElementPointer<ElementPointer<PointerType>> : public std::true_type {};
+constexpr inline bool IsElementPointer<ElementPointer<PointerType>> = true;
 
 /// Pairs an array base pointer (either an `Element*` or an
-/// `std::shared_ptr<Element>`) with an `DataType` in order to support a dynamic
-/// element type determined at run time.
+/// `std::shared_ptr<Element>`) with a `DataType` in order to support a
+/// dynamic element type determined at run time.
 ///
-/// If `Element` is non-`void`, the `DataType` is not actually stored, making
-/// this object the same size as `PointerType`.
+/// If `Element` is non-`void`, the `DataType` is not actually stored,
+/// making this object the same size as `Pointer`.
 ///
-/// Instances of this type are used by the array types defined in array.h to
-/// pointer to the actual array data.
+/// Instances of this type are used by `Array` to point to the actual array
+/// data.
 ///
 /// \tparam ElementTagType An ElementTag type that is either `Element` (to
-///     represent unowned array data) or `Shared<Element>` (to represent array
-///     data with shared ownership).
+///     represent unowned array data) or `Shared<Element>` (to represent
+///     array data with shared ownership).
+/// \ingroup array
 template <typename ElementTagType>
 class ElementPointer {
   using Traits = ElementTagTraits<ElementTagType>;
 
  public:
-  static_assert(IsElementTag<ElementTagType>::value,
+  static_assert(IsElementTag<ElementTagType>,
                 "ElementTagType must be an ElementTag type.");
-  using ElementTag = ElementTagType;
-  using Pointer = typename Traits::Pointer;
-  using element_type = typename std::pointer_traits<Pointer>::element_type;
-  using Element = element_type;
-  using DataType = StaticOrDynamicDataTypeOf<Element>;
 
-  /// For compatibility with `std::pointer_traits`.
+  /// Element tag type.
+  using ElementTag = ElementTagType;
+
+  /// Underlying data pointer type, either `Element*` or
+  /// `std::shared_ptr<Element>`.
+  using Pointer = typename Traits::Pointer;
+
+  /// Underlying element type.
+  using element_type = typename std::pointer_traits<Pointer>::
+      element_type;  // NONITPICK: std::pointer_traits<Pointer>::element_type
+  using Element = element_type;
+
+  /// Static or dynamic data type representation.
+  using DataType = dtype_t<Element>;
+
+  // For compatibility with `std::pointer_traits`.
   template <typename OtherElement>
   using rebind = ElementPointer<typename Traits::template rebind<OtherElement>>;
 
   /// Initializes to a null pointer.
-  /// \post data() == nullptr
+  ///
+  /// \id default
   ElementPointer() = default;
-
-  /// Initializes to a null pointer.
-  /// \post data() == nullptr
   ElementPointer(std::nullptr_t) {}
 
-  /// Constructs from a compatible ElementPointer type.
+  /// Constructs from a compatible `ElementPointer` type.
   ///
-  /// \post this->pointer() == source.pointer()
-  /// \post this->data_type() == source.data_type()
-  template <typename Source,
-            absl::enable_if_t<
-                (IsElementPointer<internal::remove_cvref_t<Source>>::value &&
-                 IsArrayBasePointerConvertible<
-                     typename internal::remove_cvref_t<Source>::Pointer,
-                     Pointer>::value)>* = nullptr>
+  /// \id element_pointer
+  template <
+      typename Source,
+      std::enable_if_t<(IsElementPointer<internal::remove_cvref_t<Source>> &&
+                        IsArrayBasePointerConvertible<
+                            typename internal::remove_cvref_t<Source>::Pointer,
+                            Pointer>)>* = nullptr>
+  // NONITPICK: std::remove_cvref_t<Source>::Pointer
   ElementPointer(Source&& source)
-      : storage_(source.data_type(),
+      : storage_(source.dtype(),
                  internal_element_pointer::ConvertPointer<Pointer>(
                      std::forward<Source>(source).pointer())) {}
 
   /// Unchecked conversion.
-  template <
-      typename Source,
-      absl::enable_if_t<IsElementPointerCastConvertible<
-          internal::remove_cvref_t<Source>, ElementPointer>::value>* = nullptr>
+  ///
+  /// \id unchecked
+  template <typename Source,
+            std::enable_if_t<IsElementPointerCastConvertible<
+                internal::remove_cvref_t<Source>, ElementPointer>>* = nullptr>
   explicit ElementPointer(unchecked_t, Source&& source)
-      : storage_(DataType(unchecked, source.data_type()),
+      : storage_(DataType(unchecked, source.dtype()),
                  internal_element_pointer::ConvertPointer<Pointer>(
                      std::forward<Source>(source).pointer())) {}
 
-  /// Constructs from another compatible pointer with a static `Element` type.
+  /// Constructs from another compatible pointer and optional data type.
   ///
   /// \param pointer The element pointer.
-  /// \post `this->pointer() == pointer`
-  /// \post `this->data_type() == DataTypeOf<SourcePointer::element_type>()`
+  /// \param dtype The data type, must be specified if `SourcePointer` has an
+  ///     element type of `void`.
+  /// \id pointer
   template <
       typename SourcePointer,
-      absl::enable_if_t<
-          IsNonVoidArrayBasePointer<
-              internal::remove_cvref_t<SourcePointer>>::value &&
+      std::enable_if_t<
+          IsNonVoidArrayBasePointer<internal::remove_cvref_t<SourcePointer>> &&
           IsArrayBasePointerConvertible<internal::remove_cvref_t<SourcePointer>,
-                                        Pointer>::value>* = nullptr>
+                                        Pointer>>* = nullptr>
   ElementPointer(SourcePointer&& pointer)
-      : storage_(DataTypeOf<typename std::pointer_traits<
-                     internal::remove_cvref_t<SourcePointer>>::element_type>(),
+      : storage_(pointee_dtype_t<SourcePointer>(),
                  internal::static_pointer_cast<Element>(
                      internal_element_pointer::ConvertPointer<Pointer>(
                          std::forward<SourcePointer>(pointer)))) {}
+  template <typename SourcePointer,
+            std::enable_if_t<IsArrayBasePointerConvertible<
+                internal::remove_cvref_t<SourcePointer>, Pointer>>* = nullptr>
+  ElementPointer(SourcePointer&& pointer, pointee_dtype_t<SourcePointer> dtype)
+      : storage_(dtype, internal::static_pointer_cast<Element>(
+                            internal_element_pointer::ConvertPointer<Pointer>(
+                                std::forward<SourcePointer>(pointer)))) {}
 
-  /// Constructs from another compatible pointer paired with an `DataType`.
-  ///
-  /// \param pointer The element pointer.
-  /// \param data_type The element representation type.
-  /// \post this->pointer() == pointer
-  /// \post this->data_type() == data_type
-  template <
-      typename SourcePointer,
-      absl::enable_if_t<IsArrayBasePointerConvertible<
-          internal::remove_cvref_t<SourcePointer>, Pointer>::value>* = nullptr>
-  ElementPointer(SourcePointer&& pointer,
-                 StaticOrDynamicDataTypeOf<typename std::pointer_traits<
-                     internal::remove_cvref_t<SourcePointer>>::element_type>
-                     data_type)
-      : storage_(data_type,
-                 internal::static_pointer_cast<Element>(
-                     internal_element_pointer::ConvertPointer<Pointer>(
-                         std::forward<SourcePointer>(pointer)))) {}
-
-  /// Assigns from a `nullptr`, pointer type, or ElementPointer type.
+  /// Assigns from a `nullptr`, pointer type, or `ElementPointer` type.
   template <typename Source>
-  absl::enable_if_t<std::is_constructible<ElementPointer, Source&&>::value,
-                    ElementPointer&>
+  std::enable_if_t<std::is_constructible_v<ElementPointer, Source&&>,
+                   ElementPointer&>
   operator=(Source&& source) {
     return *this = ElementPointer(static_cast<Source&&>(source));
   }
 
-  constexpr DataType data_type() const { return storage_.first(); }
+  /// Returns the data type.
+  ///
+  /// \membergroup Accessors
+  constexpr DataType dtype() const { return storage_.first(); }
 
   /// Returns the raw pointer value.
+  ///
+  /// \membergroup Accessors
   Element* data() const { return internal::to_address(pointer()); }
 
-  /// Returns the raw pointer value as a ByteStridedPointer.
+  /// Returns the raw pointer value as a `ByteStridedPointer`.
+  ///
+  /// \membergroup Accessors
   ByteStridedPointer<Element> byte_strided_pointer() const { return data(); }
 
-  /// Returns a const reference to the stored pointer.
+  /// Returns a reference to the stored pointer.
+  ///
+  /// \membergroup Accessors
   const Pointer& pointer() const& { return storage_.second(); }
-
-  /// Returns a non-const reference to the stored pointer.
   Pointer& pointer() & { return storage_.second(); }
-
-  /// Returns an rvalue reference to the stored pointer.
   Pointer&& pointer() && { return static_cast<Pointer&&>(storage_.second()); }
 
   /// Returns `data() != nullptr`.
+  ///
+  /// \membergroup Accessors
   explicit operator bool() const { return data() != nullptr; }
 
-  /// Compares an element pointer against `nullptr`.
+  /// Compares the data pointers and data types.
   ///
-  /// \returns p.data() == nullptr
+  /// \id element_pointer
+  /// \membergroup Comparison
+  template <typename B>
+  friend bool operator==(const ElementPointer& a, const ElementPointer<B>& b) {
+    return a.data() == b.data() && a.dtype() == b.dtype();
+  }
+  template <typename B>
+  friend bool operator!=(const ElementPointer& a, const ElementPointer<B>& b) {
+    return !(a == b);
+  }
+
+  /// Checks if the data pointer is null.
+  ///
+  /// \id nullptr
+  /// \membergroup Comparison
   friend bool operator==(const ElementPointer& p, std::nullptr_t) {
     return p.data() == nullptr;
   }
 
-  /// Compares an element pointer against `nullptr`.
-  ///
-  /// \returns p.data() == nullptr
   friend bool operator==(std::nullptr_t, const ElementPointer& p) {
     return p.data() == nullptr;
   }
-
-  /// Compares an element pointer against `nullptr`.
-  ///
-  /// \returns p.data() != nullptr
   friend bool operator!=(const ElementPointer& p, std::nullptr_t) {
     return p.data() != nullptr;
   }
-
-  /// Compares an element pointer against `nullptr`.
-  ///
-  /// \returns p.data() != nullptr
   friend bool operator!=(std::nullptr_t, const ElementPointer& p) {
     return p.data() != nullptr;
   }
@@ -384,28 +410,14 @@ class ElementPointer {
   Storage storage_{DataType(), nullptr};
 };
 
-/// Compares two element pointers for equality.
-///
-/// \returns `a.data() == b.data() && a.data_type() == b.data_type()`
-template <typename A, typename B>
-bool operator==(const ElementPointer<A>& a, const ElementPointer<B>& b) {
-  return a.data() == b.data() && a.data_type() == b.data_type();
-}
-
-/// Compares two element pointers for inequality.
-///
-/// \returns !(a == b)
-template <typename A, typename B>
-bool operator!=(const ElementPointer<A>& a, const ElementPointer<B>& b) {
-  return !(a == b);
-}
-
 /// Represents a pointer to array data with shared ownership.
+///
+/// \relates ElementPointer
 template <typename Element>
 using SharedElementPointer = ElementPointer<Shared<Element>>;
 
-/// Specialization of `StaticCastTraits` for `ElementPointer`, which enables
-/// `StaticCast`, `StaticDataTypeCast`, and `ConstDataTypeCast`.
+// Specialization of `StaticCastTraits` for `ElementPointer`, which enables
+// `StaticCast`, `StaticDataTypeCast`, and `ConstDataTypeCast`.
 template <typename ElementTag>
 struct StaticCastTraits<ElementPointer<ElementTag>>
     : public DefaultStaticCastTraits<ElementPointer<ElementTag>> {
@@ -416,14 +428,14 @@ struct StaticCastTraits<ElementPointer<ElementTag>>
 
   template <typename OtherElementTag>
   static bool IsCompatible(const ElementPointer<OtherElementTag>& other) {
-    return IsPossiblySameDataType(typename type::DataType(), other.data_type());
+    return IsPossiblySameDataType(typename type::DataType(), other.dtype());
   }
 
   static std::string Describe() {
     return internal_element_pointer::DescribeForCast(typename type::DataType());
   }
   static std::string Describe(const type& x) {
-    return internal_element_pointer::DescribeForCast(x.data_type());
+    return internal_element_pointer::DescribeForCast(x.dtype());
   }
 };
 
@@ -437,14 +449,41 @@ struct StaticCastTraits<ElementPointer<ElementTag>>
 /// The returned SharedElementPointer can be copied more efficiently than a
 /// SharedElementPointer that does manage ownership, because it does not perform
 /// any atomic reference count operations.
+///
+/// \relates ElementPointer
+/// \membergroup Ownership conversion
+/// \id element_pointer
 template <typename Element>
-absl::enable_if_t<!IsShared<Element>::value, ElementPointer<Shared<Element>>>
+std::enable_if_t<!IsShared<Element>, ElementPointer<Shared<Element>>>
 UnownedToShared(ElementPointer<Element> element_pointer) {
   return {internal::UnownedToShared(element_pointer.pointer()),
-          element_pointer.data_type()};
+          element_pointer.dtype()};
+}
+
+/// Converts a non-`Shared` `ElementPointer` to a `SharedElementPointer` that
+/// shares ownership of the specified `owned` pointer, in the same way as the
+/// `std::shared_ptr` aliasing constructor.
+///
+/// The caller is responsible for ensuring that the returned
+/// `SharedElementPointer` is not used after the element data to which it points
+/// becomes invalid.
+///
+/// \relates ElementPointer
+/// \membergroup Ownership conversion
+/// \id owned, element_pointer
+template <typename T, typename Element>
+std::enable_if_t<!IsShared<Element>, ElementPointer<Shared<Element>>>
+UnownedToShared(const std::shared_ptr<T>& owned,
+                ElementPointer<Element> element_pointer) {
+  return {std::shared_ptr<Element>(owned, element_pointer.pointer()),
+          element_pointer.dtype()};
 }
 
 /// Adds a byte offset to a raw pointer.
+///
+/// \relates ElementPointer
+/// \membergroup Arithmetic operations
+/// \id raw
 template <typename T>
 inline T* AddByteOffset(T* x, Index byte_offset) {
   return (ByteStridedPointer<T>(x) + byte_offset).get();
@@ -453,26 +492,33 @@ inline T* AddByteOffset(T* x, Index byte_offset) {
 /// Adds a byte offset to a shared_ptr.
 ///
 /// The returned pointer shares ownership with the original.
+///
+/// \relates ElementPointer
+/// \membergroup Arithmetic operations
+/// \id shared_ptr
 template <typename T>
 inline std::shared_ptr<T> AddByteOffset(const std::shared_ptr<T>& x,
                                         Index byte_offset) {
   return std::shared_ptr<T>(x, AddByteOffset(x.get(), byte_offset));
 }
 
-/// Adds a byte offset to an ElementPointer type.
+/// Adds a byte offset to an `ElementPointer` type.
 ///
-/// If `Pointer` is an instance of `std::shared_ptr`, the returned pointer
-/// shares ownership with the original.
+/// If `IsShared<ElementTag>`, the returned pointer shares ownership with the
+/// original.
+///
+/// \relates ElementPointer
+/// \membergroup Arithmetic operations
+/// \id element_pointer
 template <typename ElementTag>
 inline ElementPointer<ElementTag> AddByteOffset(
     const ElementPointer<ElementTag>& x, Index byte_offset) {
-  return {AddByteOffset(x.pointer(), byte_offset), x.data_type()};
+  return {AddByteOffset(x.pointer(), byte_offset), x.dtype()};
 }
-
 template <typename ElementTag>
 inline ElementPointer<ElementTag> AddByteOffset(ElementPointer<ElementTag>&& x,
                                                 Index byte_offset) {
-  return {AddByteOffset(std::move(x.pointer()), byte_offset), x.data_type()};
+  return {AddByteOffset(std::move(x.pointer()), byte_offset), x.dtype()};
 }
 
 namespace internal_element_pointer {
@@ -500,7 +546,7 @@ using DeducedElementTag =
         internal::remove_cvref_t<T>>::type;
 
 template <typename Pointer>
-ElementPointer(Pointer pointer)->ElementPointer<DeducedElementTag<Pointer>>;
+ElementPointer(Pointer pointer) -> ElementPointer<DeducedElementTag<Pointer>>;
 
 }  // namespace tensorstore
 
