@@ -27,147 +27,112 @@
 
 namespace {
 
-using tensorstore::DataType;
-using tensorstore::DataTypeOf;
-using tensorstore::ElementPointer;
-using tensorstore::ElementTagTraits;
-using tensorstore::IsElementTag;
-using tensorstore::MatchesStatus;
-using tensorstore::PointerElementTag;
-using tensorstore::Result;
-using tensorstore::Shared;
-using tensorstore::SharedElementPointer;
-using tensorstore::StaticDataTypeCast;
-using tensorstore::Status;
+using ::tensorstore::DataType;
+using ::tensorstore::dtype_v;
+using ::tensorstore::ElementPointer;
+using ::tensorstore::ElementTagTraits;
+using ::tensorstore::IsElementTag;
+using ::tensorstore::MatchesStatus;
+using ::tensorstore::PointerElementTag;
+using ::tensorstore::Result;
+using ::tensorstore::Shared;
+using ::tensorstore::SharedElementPointer;
+using ::tensorstore::StaticDataTypeCast;
 
-static_assert(IsElementTag<int>::value, "");
-static_assert(IsElementTag<void>::value, "");
-static_assert(IsElementTag<const void>::value, "");
-static_assert(IsElementTag<int*>::value, "");
-static_assert(IsElementTag<const int>::value, "");
-static_assert(!IsElementTag<volatile int>::value, "");
-static_assert(!IsElementTag<int(int)>::value, "");
-static_assert(IsElementTag<int (*)(int)>::value, "");
-static_assert(IsElementTag<Shared<int>>::value, "");
-static_assert(!IsElementTag<const Shared<int>>::value, "");
-static_assert(!IsElementTag<Shared<Shared<int>>>::value, "");
-static_assert(!IsElementTag<Shared<const Shared<int>>>::value, "");
-static_assert(!IsElementTag<Shared<const Shared<Shared<int>>>>::value, "");
+static_assert(IsElementTag<int>);
+static_assert(IsElementTag<void>);
+static_assert(IsElementTag<const void>);
+static_assert(IsElementTag<int*>);
+static_assert(IsElementTag<const int>);
+static_assert(!IsElementTag<volatile int>);
+static_assert(!IsElementTag<int(int)>);
+static_assert(IsElementTag<int (*)(int)>);
+static_assert(IsElementTag<Shared<int>>);
+static_assert(!IsElementTag<const Shared<int>>);
+static_assert(!IsElementTag<Shared<Shared<int>>>);
+static_assert(!IsElementTag<Shared<const Shared<int>>>);
+static_assert(!IsElementTag<Shared<const Shared<Shared<int>>>>);
 
-static_assert(std::is_same<ElementTagTraits<int>::Pointer, int*>::value, "");
-static_assert(std::is_same<ElementTagTraits<int>::rebind<float>, float>::value,
-              "");
-static_assert(std::is_same<ElementTagTraits<Shared<int>>::Pointer,
-                           std::shared_ptr<int>>::value,
-              "");
-static_assert(std::is_same<ElementTagTraits<Shared<int>>::rebind<float>,
-                           Shared<float>>::value,
-              "");
-static_assert(std::is_same<PointerElementTag<int*>, int>::value, "");
+static_assert(std::is_same_v<ElementTagTraits<int>::Pointer, int*>);
+static_assert(std::is_same_v<ElementTagTraits<int>::rebind<float>, float>);
+static_assert(std::is_same_v<ElementTagTraits<Shared<int>>::Pointer,
+                             std::shared_ptr<int>>);
+static_assert(std::is_same_v<ElementTagTraits<Shared<int>>::rebind<float>,
+                             Shared<float>>);
+static_assert(std::is_same_v<PointerElementTag<int*>, int>);
 static_assert(
-    std::is_same<PointerElementTag<std::shared_ptr<int>>, Shared<int>>::value,
-    "");
+    std::is_same_v<PointerElementTag<std::shared_ptr<int>>, Shared<int>>);
 
 // ElementPointer conversion tests
 static_assert(
-    std::is_convertible<ElementPointer<int>, ElementPointer<const int>>::value,
-    "");
+    std::is_convertible_v<ElementPointer<int>, ElementPointer<const int>>);
 static_assert(
-    !std::is_convertible<ElementPointer<const int>, ElementPointer<int>>::value,
-    "");
+    !std::is_convertible_v<ElementPointer<const int>, ElementPointer<int>>);
+static_assert(std::is_convertible_v<ElementPointer<int>, ElementPointer<void>>);
 static_assert(
-    std::is_convertible<ElementPointer<int>, ElementPointer<void>>::value, "");
+    !std::is_convertible_v<ElementPointer<void>, ElementPointer<int>>);
 static_assert(
-    !std::is_convertible<ElementPointer<void>, ElementPointer<int>>::value, "");
+    !std::is_convertible_v<ElementPointer<const int>, ElementPointer<int>>);
 static_assert(
-    !std::is_convertible<ElementPointer<const int>, ElementPointer<int>>::value,
-    "");
-static_assert(!std::is_convertible<ElementPointer<const int>,
-                                   ElementPointer<void>>::value,
-              "");
-static_assert(std::is_convertible<ElementPointer<const int>,
-                                  ElementPointer<const void>>::value,
-              "");
-static_assert(std::is_convertible<int*, ElementPointer<int>>::value, "");
-static_assert(!std::is_convertible<const int*, ElementPointer<int>>::value, "");
-static_assert(std::is_convertible<const int*, ElementPointer<const int>>::value,
-              "");
-static_assert(std::is_convertible<int*, ElementPointer<void>>::value, "");
-static_assert(!std::is_convertible<const int*, ElementPointer<void>>::value,
-              "");
-static_assert(std::is_convertible<int*, ElementPointer<const int>>::value, "");
-static_assert(std::is_convertible<int*, ElementPointer<const void>>::value, "");
+    !std::is_convertible_v<ElementPointer<const int>, ElementPointer<void>>);
+static_assert(std::is_convertible_v<ElementPointer<const int>,
+                                    ElementPointer<const void>>);
+static_assert(std::is_convertible_v<int*, ElementPointer<int>>);
+static_assert(!std::is_convertible_v<const int*, ElementPointer<int>>);
+static_assert(std::is_convertible_v<const int*, ElementPointer<const int>>);
+static_assert(std::is_convertible_v<int*, ElementPointer<void>>);
+static_assert(!std::is_convertible_v<const int*, ElementPointer<void>>);
+static_assert(std::is_convertible_v<int*, ElementPointer<const int>>);
+static_assert(std::is_convertible_v<int*, ElementPointer<const void>>);
+static_assert(std::is_constructible_v<ElementPointer<void>, void*, DataType>);
 static_assert(
-    std::is_constructible<ElementPointer<void>, void*, DataType>::value, "");
-static_assert(
-    std::is_constructible<ElementPointer<const void>, void*, DataType>::value,
-    "");
+    std::is_constructible_v<ElementPointer<const void>, void*, DataType>);
 
-static_assert(!std::is_constructible<ElementPointer<void>, void*>::value, "");
+static_assert(!std::is_constructible_v<ElementPointer<void>, void*>);
 
 // SharedElementPointer conversion tests
-static_assert(std::is_convertible<SharedElementPointer<int>,
-                                  SharedElementPointer<const int>>::value,
-              "");
-static_assert(!std::is_convertible<SharedElementPointer<const int>,
-                                   SharedElementPointer<int>>::value,
-              "");
-static_assert(std::is_convertible<SharedElementPointer<int>,
-                                  SharedElementPointer<void>>::value,
-              "");
-static_assert(!std::is_convertible<SharedElementPointer<void>,
-                                   SharedElementPointer<int>>::value,
-              "");
-static_assert(!std::is_convertible<SharedElementPointer<const int>,
-                                   SharedElementPointer<int>>::value,
-              "");
-static_assert(!std::is_convertible<SharedElementPointer<const int>,
-                                   SharedElementPointer<void>>::value,
-              "");
-static_assert(std::is_convertible<SharedElementPointer<const int>,
-                                  SharedElementPointer<const void>>::value,
-              "");
+static_assert(std::is_convertible_v<SharedElementPointer<int>,
+                                    SharedElementPointer<const int>>);
+static_assert(!std::is_convertible_v<SharedElementPointer<const int>,
+                                     SharedElementPointer<int>>);
+static_assert(std::is_convertible_v<SharedElementPointer<int>,
+                                    SharedElementPointer<void>>);
+static_assert(!std::is_convertible_v<SharedElementPointer<void>,
+                                     SharedElementPointer<int>>);
+static_assert(!std::is_convertible_v<SharedElementPointer<const int>,
+                                     SharedElementPointer<int>>);
+static_assert(!std::is_convertible_v<SharedElementPointer<const int>,
+                                     SharedElementPointer<void>>);
+static_assert(std::is_convertible_v<SharedElementPointer<const int>,
+                                    SharedElementPointer<const void>>);
 static_assert(
-    std::is_convertible<std::shared_ptr<int>, SharedElementPointer<int>>::value,
-    "");
-static_assert(std::is_convertible<std::shared_ptr<const int>,
-                                  SharedElementPointer<const int>>::value,
-              "");
-static_assert(std::is_convertible<std::shared_ptr<int>,
-                                  SharedElementPointer<void>>::value,
-              "");
-static_assert(!std::is_convertible<std::shared_ptr<const int>,
-                                   SharedElementPointer<void>>::value,
-              "");
-static_assert(std::is_convertible<std::shared_ptr<int>,
-                                  SharedElementPointer<const int>>::value,
-              "");
-static_assert(std::is_convertible<std::shared_ptr<int>,
-                                  SharedElementPointer<const void>>::value,
-              "");
-static_assert(std::is_constructible<SharedElementPointer<void>,
-                                    std::shared_ptr<void>, DataType>::value,
-              "");
-static_assert(std::is_constructible<SharedElementPointer<const void>,
-                                    std::shared_ptr<void>, DataType>::value,
-              "");
+    std::is_convertible_v<std::shared_ptr<int>, SharedElementPointer<int>>);
+static_assert(std::is_convertible_v<std::shared_ptr<const int>,
+                                    SharedElementPointer<const int>>);
+static_assert(
+    std::is_convertible_v<std::shared_ptr<int>, SharedElementPointer<void>>);
+static_assert(!std::is_convertible_v<std::shared_ptr<const int>,
+                                     SharedElementPointer<void>>);
+static_assert(std::is_convertible_v<std::shared_ptr<int>,
+                                    SharedElementPointer<const int>>);
+static_assert(std::is_convertible_v<std::shared_ptr<int>,
+                                    SharedElementPointer<const void>>);
+static_assert(std::is_constructible_v<SharedElementPointer<void>,
+                                      std::shared_ptr<void>, DataType>);
+static_assert(std::is_constructible_v<SharedElementPointer<const void>,
+                                      std::shared_ptr<void>, DataType>);
 
 // SharedElementPointer -> ElementPointer conversion tests
 static_assert(
-    std::is_convertible<SharedElementPointer<int>, ElementPointer<int>>::value,
-    "");
-static_assert(std::is_convertible<SharedElementPointer<int>,
-                                  ElementPointer<const int>>::value,
-              "");
-static_assert(!std::is_convertible<SharedElementPointer<void>,
-                                   ElementPointer<const int>>::value,
-              "");
-static_assert(!std::is_constructible<ElementPointer<const int>,
-                                     SharedElementPointer<void>>::value,
-              "");
-static_assert(!std::is_constructible<ElementPointer<int>,
-                                     SharedElementPointer<const void>>::value,
-              "");
+    std::is_convertible_v<SharedElementPointer<int>, ElementPointer<int>>);
+static_assert(std::is_convertible_v<SharedElementPointer<int>,
+                                    ElementPointer<const int>>);
+static_assert(!std::is_convertible_v<SharedElementPointer<void>,
+                                     ElementPointer<const int>>);
+static_assert(!std::is_constructible_v<ElementPointer<const int>,
+                                       SharedElementPointer<void>>);
+static_assert(!std::is_constructible_v<ElementPointer<int>,
+                                       SharedElementPointer<const void>>);
 
 TEST(ElementPointerTest, StaticType) {
   {
@@ -184,14 +149,14 @@ TEST(ElementPointerTest, StaticType) {
   ElementPointer<float> p = &value;
   EXPECT_EQ(&value, p.data());
   EXPECT_EQ(&value, p.pointer());
-  EXPECT_EQ(DataTypeOf<float>(), p.data_type());
+  EXPECT_EQ(dtype_v<float>, p.dtype());
 
   // Test implicit construction of `ElementPointer<const float>` from
   // `ElementPointer<float>`.
   {
     ElementPointer<const float> p_const = p;
     EXPECT_EQ(&value, p_const.data());
-    EXPECT_EQ(DataTypeOf<float>(), p_const.data_type());
+    EXPECT_EQ(dtype_v<float>, p_const.dtype());
     p_const.pointer() = nullptr;
     EXPECT_EQ(nullptr, p_const.data());
   }
@@ -206,15 +171,14 @@ TEST(ElementPointerTest, StaticType) {
   // Test implicit conversion to const void.
   ElementPointer<const void> other = p;
   EXPECT_EQ(&value, other.data());
-  EXPECT_EQ(other.data_type(), p.data_type());
+  EXPECT_EQ(other.dtype(), p.dtype());
 
   // Test explicit conversion from another ElementPointer.
   {
     auto p2 = tensorstore::StaticDataTypeCast<const float>(other);
     static_assert(
-        std::is_same<decltype(p2), Result<ElementPointer<const float>>>::value,
-        "");
-    ASSERT_EQ(Status(), GetStatus(p2));
+        std::is_same_v<decltype(p2), Result<ElementPointer<const float>>>);
+    ASSERT_EQ(absl::OkStatus(), GetStatus(p2));
     EXPECT_EQ(&value, p2->data());
   }
 
@@ -228,64 +192,56 @@ TEST(ElementPointerTest, StaticType) {
     EXPECT_EQ(nullptr, p_const.data());
   }
 
-  static_assert(!std::is_assignable<ElementPointer<float>,
-                                    ElementPointer<const float>>::value,
-                "");
+  static_assert(!std::is_assignable_v<ElementPointer<float>,
+                                      ElementPointer<const float>>);
 
   static_assert(
-      !std::is_assignable<ElementPointer<int>, ElementPointer<float>>::value,
-      "");
+      !std::is_assignable_v<ElementPointer<int>, ElementPointer<float>>);
 
-  static_assert(!std::is_assignable<ElementPointer<int>, float*>::value, "");
-  static_assert(!std::is_assignable<ElementPointer<void>, void*>::value, "");
+  static_assert(!std::is_assignable_v<ElementPointer<int>, float*>);
+  static_assert(!std::is_assignable_v<ElementPointer<void>, void*>);
 
-  static_assert(!std::is_assignable<ElementPointer<const float>,
-                                    ElementPointer<const void>>::value,
-                "");
-  static_assert(!std::is_assignable<ElementPointer<float>, void*>::value, "");
+  static_assert(!std::is_assignable_v<ElementPointer<const float>,
+                                      ElementPointer<const void>>);
+  static_assert(!std::is_assignable_v<ElementPointer<float>, void*>);
 }
 
 TEST(ElementPointerTest, DynamicType) {
   {
     ElementPointer<void> p_null;
     EXPECT_EQ(nullptr, p_null.data());
-    EXPECT_EQ(DataType(), p_null.data_type());
+    EXPECT_EQ(DataType(), p_null.dtype());
   }
 
   {
     ElementPointer<void> p_null = nullptr;
     EXPECT_EQ(nullptr, p_null.data());
-    EXPECT_EQ(DataType(), p_null.data_type());
+    EXPECT_EQ(DataType(), p_null.dtype());
   }
 
   float value;
   {
     ElementPointer<void> p = &value;
     EXPECT_EQ(&value, p.data());
-    EXPECT_EQ(DataTypeOf<float>(), p.data_type());
+    EXPECT_EQ(dtype_v<float>, p.dtype());
   }
   {
-    ElementPointer<void> p = {static_cast<void*>(&value), DataTypeOf<float>()};
+    ElementPointer<void> p = {static_cast<void*>(&value), dtype_v<float>};
     EXPECT_EQ(&value, p.data());
-    EXPECT_EQ(DataTypeOf<float>(), p.data_type());
+    EXPECT_EQ(dtype_v<float>, p.dtype());
   }
 
-  static_assert(!std::is_assignable<ElementPointer<void>,
-                                    ElementPointer<const float>>::value,
-                "");
-  static_assert(!std::is_assignable<ElementPointer<void>,
-                                    SharedElementPointer<const float>>::value,
-                "");
+  static_assert(
+      !std::is_assignable_v<ElementPointer<void>, ElementPointer<const float>>);
+  static_assert(!std::is_assignable_v<ElementPointer<void>,
+                                      SharedElementPointer<const float>>);
 
-  static_assert(!std::is_assignable<ElementPointer<void>,
-                                    ElementPointer<const void>>::value,
-                "");
-  static_assert(!std::is_assignable<ElementPointer<void>,
-                                    SharedElementPointer<const void>>::value,
-                "");
-  static_assert(!std::is_assignable<ElementPointer<void>, void*>::value, "");
-  static_assert(!std::is_assignable<ElementPointer<void>, const float*>::value,
-                "");
+  static_assert(
+      !std::is_assignable_v<ElementPointer<void>, ElementPointer<const void>>);
+  static_assert(!std::is_assignable_v<ElementPointer<void>,
+                                      SharedElementPointer<const void>>);
+  static_assert(!std::is_assignable_v<ElementPointer<void>, void*>);
+  static_assert(!std::is_assignable_v<ElementPointer<void>, const float*>);
 
   // Copy construction
   {
@@ -294,7 +250,7 @@ TEST(ElementPointerTest, DynamicType) {
 
     ElementPointer<void> p_copy = p;
     EXPECT_EQ(&value, p_copy.data());
-    EXPECT_EQ(DataTypeOf<float>(), p_copy.data_type());
+    EXPECT_EQ(dtype_v<float>, p_copy.dtype());
   }
 
   // Copy assignment from `ElementPointer<void>`.
@@ -305,7 +261,7 @@ TEST(ElementPointerTest, DynamicType) {
     ElementPointer<void> p_copy;
     p_copy = p;
     EXPECT_EQ(&value, p.data());
-    EXPECT_EQ(DataTypeOf<float>(), p.data_type());
+    EXPECT_EQ(dtype_v<float>, p.dtype());
   }
 
   // Assignment from ElementPointer<float>.
@@ -313,7 +269,7 @@ TEST(ElementPointerTest, DynamicType) {
     ElementPointer<void> p;
     p = ElementPointer<float>(&value);
     EXPECT_EQ(&value, p.data());
-    EXPECT_EQ(DataTypeOf<float>(), p.data_type());
+    EXPECT_EQ(dtype_v<float>, p.dtype());
   }
 
   // Assignment from float*.
@@ -321,7 +277,7 @@ TEST(ElementPointerTest, DynamicType) {
     ElementPointer<void> p;
     p = &value;
     EXPECT_EQ(&value, p.data());
-    EXPECT_EQ(DataTypeOf<float>(), p.data_type());
+    EXPECT_EQ(dtype_v<float>, p.dtype());
   }
 
   // Assignment from SharedElementPointer<float>.
@@ -330,7 +286,7 @@ TEST(ElementPointerTest, DynamicType) {
     std::shared_ptr<float> shared_value = std::make_shared<float>();
     p = SharedElementPointer<float>(shared_value);
     EXPECT_EQ(shared_value.get(), p.data());
-    EXPECT_EQ(DataTypeOf<float>(), p.data_type());
+    EXPECT_EQ(dtype_v<float>, p.dtype());
   }
 
   // Assignment from SharedElementPointer<float>.
@@ -339,7 +295,7 @@ TEST(ElementPointerTest, DynamicType) {
     std::shared_ptr<float> shared_value = std::make_shared<float>();
     p = SharedElementPointer<void>(shared_value);
     EXPECT_EQ(shared_value.get(), p.data());
-    EXPECT_EQ(DataTypeOf<float>(), p.data_type());
+    EXPECT_EQ(dtype_v<float>, p.dtype());
   }
 }
 
@@ -361,7 +317,7 @@ TEST(SharedElementPointerTest, StaticType) {
   std::shared_ptr<float> value = std::make_shared<float>();
   SharedElementPointer<float> p = value;
   EXPECT_EQ(value.get(), p.data());
-  EXPECT_EQ(DataTypeOf<float>(), p.data_type());
+  EXPECT_EQ(dtype_v<float>, p.dtype());
 
   {
     // Test copy construction of `SharedElementPointer<float>` from
@@ -383,7 +339,7 @@ TEST(SharedElementPointerTest, StaticType) {
   {
     SharedElementPointer<const void> other = p;
     EXPECT_EQ(value.get(), other.data());
-    EXPECT_EQ(other.data_type(), p.data_type());
+    EXPECT_EQ(other.dtype(), p.dtype());
     EXPECT_EQ(3, value.use_count());
   }
   EXPECT_EQ(2, value.use_count());
@@ -405,8 +361,7 @@ TEST(SharedElementPointerTest, StaticType) {
     SharedElementPointer<void> shared_p_void = p;
     auto p_float = StaticDataTypeCast<float>(shared_p_void).value();
     static_assert(
-        std::is_same<decltype(p_float), SharedElementPointer<float>>::value,
-        "");
+        std::is_same_v<decltype(p_float), SharedElementPointer<float>>);
     EXPECT_EQ(value.get(), p_float.data());
   }
 
@@ -415,8 +370,7 @@ TEST(SharedElementPointerTest, StaticType) {
     float fvalue;
     auto f_pointer = UnownedToShared(ElementPointer<float>(&fvalue));
     static_assert(
-        std::is_same<decltype(f_pointer), SharedElementPointer<float>>::value,
-        "");
+        std::is_same_v<decltype(f_pointer), SharedElementPointer<float>>);
     EXPECT_EQ(&fvalue, f_pointer.data());
   }
 
@@ -468,19 +422,17 @@ TEST(SharedElementPointerTest, StaticType) {
     EXPECT_EQ(value.get(), p2_move.data());
   }
 
-  static_assert(!std::is_assignable<SharedElementPointer<float>,
-                                    SharedElementPointer<void>>::value,
-                "");
-  static_assert(!std::is_assignable<SharedElementPointer<float>,
-                                    std::shared_ptr<void>>::value,
-                "");
+  static_assert(!std::is_assignable_v<SharedElementPointer<float>,
+                                      SharedElementPointer<void>>);
+  static_assert(!std::is_assignable_v<SharedElementPointer<float>,
+                                      std::shared_ptr<void>>);
 }
 
 TEST(SharedElementPointerTest, DynamicType) {
   std::shared_ptr<float> value = std::make_shared<float>();
   SharedElementPointer<void> p = value;
   EXPECT_EQ(value.get(), p.data());
-  EXPECT_EQ(DataTypeOf<float>(), p.data_type());
+  EXPECT_EQ(dtype_v<float>, p.dtype());
 }
 
 TEST(ElementPointerTest, Deduction) {
