@@ -15,31 +15,62 @@
 #include "tensorstore/util/future.h"
 
 #include <atomic>
+#include <cassert>
 #include <new>
 #include <thread>  // NOLINT
 
 #include "absl/base/attributes.h"
 #include "absl/base/const_init.h"
-#include "absl/base/macros.h"
+#include "absl/base/optimization.h"
+#include "absl/hash/hash.h"
+#include "absl/status/status.h"
 #include "absl/synchronization/mutex.h"
 #include "absl/time/time.h"
 #include "tensorstore/internal/intrusive_linked_list.h"
 #include "tensorstore/internal/intrusive_ptr.h"
-#include "tensorstore/internal/void_wrapper.h"
+#include "tensorstore/internal/metrics/counter.h"
+#include "tensorstore/internal/metrics/gauge.h"
+#include "tensorstore/internal/no_destructor.h"
 #include "tensorstore/util/future_impl.h"
 
 namespace tensorstore {
-
 namespace internal_future {
+namespace {
+
+auto& live_futures = internal_metrics::Gauge<int64_t>::New(
+    "/tensorstore/futures/live", "Live futures");
+
+#if 1
+auto& future_ready_callbacks = internal_metrics::Counter<int64_t>::New(
+    "/tensorstore/futures/ready_callbacks", "Ready callbacks");
+auto& future_not_needed_callbacks = internal_metrics::Counter<int64_t>::New(
+    "/tensorstore/futures/not_needed_callbacks", "Not needed callbacks");
+auto& future_force_callbacks = internal_metrics::Counter<int64_t>::New(
+    "/tensorstore/futures/force_callbacks", "Force callbacks");
+#else
+struct MockCounter {
+  void Increment() {}
+};
+static MockCounter future_ready_callbacks;
+static MockCounter future_not_needed_callbacks;
+static MockCounter future_force_callbacks;
+#endif
+
+}  // namespace
 
 /// Special value to which CallbackListNode::next points to indicate that
 /// unregistration was requested.
 static CallbackListNode unregister_requested;
 
+struct ABSL_CACHELINE_ALIGNED CacheLineAlignedMutex {
+  absl::Mutex mutex{absl::kConstInit};
+};
+
+constexpr size_t kNumMutexes = 64;
+
 absl::Mutex* GetMutex(FutureStateBase* ptr) {
-  // FIXME: Use more than one mutex.
-  ABSL_CONST_INIT static absl::Mutex mutex(absl::kConstInit);
-  return &mutex;
+  ABSL_CONST_INIT static CacheLineAlignedMutex mutexes[kNumMutexes];
+  return &mutexes[absl::HashOf(ptr) % kNumMutexes].mutex;
 }
 
 using CallbackListAccessor =
@@ -58,7 +89,7 @@ namespace {
 /// \dchecks `callback->reference_count_.load() >= 2`.
 /// \returns A reference to `callback`.
 CallbackPointer MakeUnregisteredCallbackPointer(CallbackBase* callback) {
-  ABSL_ASSERT(callback->reference_count_.load(std::memory_order_relaxed) >= 2);
+  assert(callback->reference_count_.load(std::memory_order_relaxed) >= 2);
   callback->next = callback->prev = callback;
   callback->reference_count_.fetch_sub(1, std::memory_order_relaxed);
   return CallbackPointer(callback, internal::adopt_object_ref);
@@ -68,9 +99,10 @@ CallbackPointer MakeUnregisteredCallbackPointer(CallbackBase* callback) {
 
 CallbackPointer FutureStateBase::RegisterReadyCallback(
     ReadyCallbackBase* callback) {
-  ABSL_ASSERT(callback->reference_count_.load(std::memory_order_relaxed) >= 2);
+  assert(callback->reference_count_.load(std::memory_order_relaxed) >= 2);
   {
     absl::MutexLock lock(GetMutex(this));
+    future_ready_callbacks.Increment();
     if (!this->ready()) {
       InsertBefore(CallbackListAccessor{}, &ready_callbacks_, callback);
       return CallbackPointer(callback, internal::adopt_object_ref);
@@ -82,9 +114,10 @@ CallbackPointer FutureStateBase::RegisterReadyCallback(
 
 CallbackPointer FutureStateBase::RegisterNotNeededCallback(
     ResultNotNeededCallbackBase* callback) {
-  ABSL_ASSERT(callback->reference_count_.load(std::memory_order_relaxed) >= 2);
+  assert(callback->reference_count_.load(std::memory_order_relaxed) >= 2);
   {
     absl::MutexLock lock(GetMutex(this));
+    future_not_needed_callbacks.Increment();
     if (result_needed()) {
       InsertBefore(CallbackListAccessor{}, &promise_callbacks_, callback);
       return CallbackPointer(callback, internal::adopt_object_ref);
@@ -96,10 +129,11 @@ CallbackPointer FutureStateBase::RegisterNotNeededCallback(
 
 CallbackPointer FutureStateBase::RegisterForceCallback(
     ForceCallbackBase* callback) {
-  ABSL_ASSERT(callback->reference_count_.load(std::memory_order_relaxed) >= 2);
+  assert(callback->reference_count_.load(std::memory_order_relaxed) >= 2);
   auto* mutex = GetMutex(this);
   {
     absl::MutexLock lock(mutex);
+    future_force_callbacks.Increment();
     const auto state = state_.load(std::memory_order_acquire);
     if ((state & kResultLocked) != 0 || !has_future()) {
       // Handle result-not-needed case after unlocking the mutex.
@@ -180,13 +214,17 @@ FutureStateBase::FutureStateBase()
       future_reference_count_(1) {
   Initialize(CallbackListAccessor{}, &ready_callbacks_);
   Initialize(CallbackListAccessor{}, &promise_callbacks_);
+  live_futures.Increment();
 }
 
 namespace {
 
 void NoMorePromiseReferences(FutureStateBase* shared_state) {
-  shared_state->LockResult();
-  shared_state->CommitResult();
+  if (shared_state->LockResult()) {
+    shared_state->MarkResultWrittenAndCommitResult();
+  } else {
+    shared_state->CommitResult();
+  }
   shared_state->ReleaseCombinedReference();
 }
 
@@ -304,8 +342,7 @@ void RunForceCallbacks(FutureStateBase* shared_state) {
             InsertBefore(CallbackListAccessor{}, head, prev_node.release());
           }
         } else {
-          ABSL_ASSERT(prev_node->callback_type() ==
-                      CallbackBase::kForceCallback);
+          assert(prev_node->callback_type() == CallbackBase::kForceCallback);
           // A `kForceCallback` is always unregistered immediately after calling
           // `OnForced`.
           prev_node->next = prev_node.get();
@@ -417,14 +454,35 @@ bool FutureStateBase::LockResult() noexcept {
   return true;
 }
 
+void FutureStateBase::MarkResultWritten() noexcept {
+  const StateValue prior_state = state_.fetch_or(kResultWritten);
+  assert(prior_state & kResultLocked);
+  assert((prior_state & kResultWritten) == 0);
+  if (prior_state & kReady) {
+    RunReadyCallbacks(this);
+  }
+}
+
 bool FutureStateBase::CommitResult() noexcept {
   const StateValue prior_state = state_.fetch_or(kReady);
-  ABSL_ASSERT(prior_state & kResultLocked);
   if (prior_state & kReady) return false;
-  RunReadyCallbacks(this);
+  if (prior_state & kResultWritten) {
+    RunReadyCallbacks(this);
+  }
   return true;
 }
 
+void FutureStateBase::MarkResultWrittenAndCommitResult() noexcept {
+  [[maybe_unused]] const StateValue prior_state =
+      state_.fetch_or(kResultWrittenAndReady);
+  assert(prior_state & kResultLocked);
+  assert((prior_state & kResultWritten) == 0);
+  // MarkResultWrittenAndCommitResult must only be called after `LockResult()`
+  // returned true, so at least one of {kReady, kWrittenResult} was set by the
+  // thread and it takes ownership of RunReadyCallbacks(this).
+  RunReadyCallbacks(this);
+}
+
 bool FutureStateBase::WaitFor(absl::Duration duration) noexcept {
   if (ready()) return true;
   Force();
@@ -454,9 +512,92 @@ void FutureStateBase::Wait() noexcept {
 }
 
 FutureStateBase::~FutureStateBase() {
-  ABSL_ASSERT(promise_callbacks_.next == &promise_callbacks_);
-  ABSL_ASSERT(ready_callbacks_.next == &ready_callbacks_);
+  assert(promise_callbacks_.next == &promise_callbacks_);
+  assert(ready_callbacks_.next == &ready_callbacks_);
+  live_futures.Decrement();
 }
 
 }  // namespace internal_future
+
+ReadyFuture<const void> MakeReadyFuture() {
+  static internal::NoDestructor<ReadyFuture<const void>> future{
+      MakeReadyFuture<void>(MakeResult())};
+  return *future;
+}
+
+Future<void> WaitAllFuture(tensorstore::span<const AnyFuture> futures) {
+  auto& f = futures;
+  switch (f.size()) {
+    case 0:
+      return MakeReadyFuture<void>(absl::OkStatus());
+    case 1:
+      return PromiseFuturePair<void>::LinkError(absl::OkStatus(), f[0]).future;
+    case 2:
+      return PromiseFuturePair<void>::LinkError(absl::OkStatus(), f[0], f[1])
+          .future;
+    case 3:
+      return PromiseFuturePair<void>::LinkError(absl::OkStatus(), f[0], f[1],
+                                                f[2])
+          .future;
+    case 4:
+      return PromiseFuturePair<void>::LinkError(absl::OkStatus(), f[0], f[1],
+                                                f[2], f[3])
+          .future;
+    case 5:
+      return PromiseFuturePair<void>::LinkError(absl::OkStatus(), f[0], f[1],
+                                                f[2], f[3], f[4])
+          .future;
+    case 6:
+      return PromiseFuturePair<void>::LinkError(absl::OkStatus(), f[0], f[1],
+                                                f[2], f[3], f[4], f[5])
+          .future;
+    case 7:
+      return PromiseFuturePair<void>::LinkError(absl::OkStatus(), f[0], f[1],
+                                                f[2], f[3], f[4], f[5], f[6])
+          .future;
+    default:
+      break;
+  }
+
+  // 8 or more...
+  auto [promise, result] = PromiseFuturePair<void>::LinkError(
+      absl::OkStatus(), f[0], f[1], f[2], f[3], f[4], f[5], f[6], f[7]);
+  f = f.subspan(8);
+  while (f.size() > 8) {
+    LinkError(promise, f[0], f[1], f[2], f[3], f[4], f[5], f[6], f[7]);
+    f = f.subspan(8);
+  }
+
+  switch (f.size()) {
+    case 0:
+      return std::move(result);
+    case 1:
+      LinkError(std::move(promise), f[0]);
+      return std::move(result);
+    case 2:
+      LinkError(std::move(promise), f[0], f[1]);
+      return std::move(result);
+    case 3:
+      LinkError(std::move(promise), f[0], f[1], f[2]);
+      return std::move(result);
+    case 4:
+      LinkError(std::move(promise), f[0], f[1], f[2], f[3]);
+      return std::move(result);
+    case 5:
+      LinkError(std::move(promise), f[0], f[1], f[2], f[3], f[4]);
+      return std::move(result);
+    case 6:
+      LinkError(std::move(promise), f[0], f[1], f[2], f[3], f[4], f[5]);
+      return std::move(result);
+    case 7:
+      LinkError(std::move(promise), f[0], f[1], f[2], f[3], f[4], f[5], f[6]);
+      return std::move(result);
+    case 8:
+      LinkError(std::move(promise), f[0], f[1], f[2], f[3], f[4], f[5], f[6],
+                f[7]);
+      return std::move(result);
+  }
+  TENSORSTORE_UNREACHABLE;
+}
+
 }  // namespace tensorstore
