@@ -26,77 +26,82 @@
 
 #include <gmock/gmock.h>
 #include <gtest/gtest.h>
+#include "absl/status/status.h"
 #include "absl/time/clock.h"
 #include "absl/time/time.h"
 #include "tensorstore/internal/concurrent_testutil.h"
+#include "tensorstore/internal/metrics/collect.h"
+#include "tensorstore/internal/metrics/registry.h"
 #include "tensorstore/internal/type_traits.h"
 #include "tensorstore/util/executor.h"
 #include "tensorstore/util/future_impl.h"
 #include "tensorstore/util/result.h"
-#include "tensorstore/util/result_util.h"
 #include "tensorstore/util/status.h"
 #include "tensorstore/util/status_testutil.h"
-#include "tensorstore/util/to_string.h"
+#include "tensorstore/util/str_cat.h"
 
 namespace {
-using tensorstore::Future;
-using tensorstore::FutureCallbackRegistration;
-using tensorstore::InlineExecutor;
-using tensorstore::IsFutureConvertible;
-using tensorstore::MakeReadyFuture;
-using tensorstore::MakeResult;
-using tensorstore::MatchesStatus;
-using tensorstore::Promise;
-using tensorstore::PromiseFuturePair;
-using tensorstore::ReadyFuture;
-using tensorstore::Result;
-using tensorstore::Status;
-using tensorstore::internal::TestConcurrent;
-using tensorstore::internal_future::FutureAccess;
+using ::tensorstore::AnyFuture;
+using ::tensorstore::Future;
+using ::tensorstore::FutureCallbackRegistration;
+using ::tensorstore::InlineExecutor;
+using ::tensorstore::IsFutureConvertible;
+using ::tensorstore::MakeReadyFuture;
+using ::tensorstore::MakeResult;
+using ::tensorstore::MatchesStatus;
+using ::tensorstore::Promise;
+using ::tensorstore::PromiseFuturePair;
+using ::tensorstore::ReadyFuture;
+using ::tensorstore::Result;
+using ::tensorstore::internal::TestConcurrent;
+using ::tensorstore::internal_future::FutureAccess;
+
+static_assert(IsFutureConvertible<int, const int>);
+static_assert(!IsFutureConvertible<const int, int>);
 
-static_assert(IsFutureConvertible<int, const int>::value, "");
-static_assert(!IsFutureConvertible<const int, int>::value, "");
 static_assert(
-    std::is_same<
+    std::is_same_v<
         decltype(FutureAccess::rep_pointer(std::declval<Future<void>&>())),
-        tensorstore::internal_future::FutureStatePointer&>::value,
-    "");
+        tensorstore::internal_future::FutureStatePointer&>);
 static_assert(
-    std::is_same<
-        decltype(
-            FutureAccess::rep_pointer(std::declval<const Future<void>&>())),
-        const tensorstore::internal_future::FutureStatePointer&>::value,
-    "");
+    std::is_same_v<decltype(FutureAccess::rep_pointer(
+                       std::declval<const Future<void>&>())),
+                   const tensorstore::internal_future::FutureStatePointer&>);
 static_assert(
-    std::is_same<
+    std::is_same_v<
         decltype(FutureAccess::rep_pointer(std::declval<Future<void>&&>())),
-        tensorstore::internal_future::FutureStatePointer&&>::value,
-    "");
+        tensorstore::internal_future::FutureStatePointer&&>);
 static_assert(
-    std::is_same<
+    std::is_same_v<
         decltype(FutureAccess::rep_pointer(std::declval<Promise<void>&>())),
-        tensorstore::internal_future::PromiseStatePointer&>::value,
-    "");
+        tensorstore::internal_future::PromiseStatePointer&>);
 static_assert(
-    std::is_same<
-        decltype(
-            FutureAccess::rep_pointer(std::declval<const Promise<void>&>())),
-        const tensorstore::internal_future::PromiseStatePointer&>::value,
-    "");
+    std::is_same_v<decltype(FutureAccess::rep_pointer(
+                       std::declval<const Promise<void>&>())),
+                   const tensorstore::internal_future::PromiseStatePointer&>);
 static_assert(
-    std::is_same<
+    std::is_same_v<
         decltype(FutureAccess::rep_pointer(std::declval<Promise<void>&&>())),
-        tensorstore::internal_future::PromiseStatePointer&&>::value,
-    "");
+        tensorstore::internal_future::PromiseStatePointer&&>);
+
+// Since Future<X> can be constructed from Result<X>, ensure that the reverse
+// is not true; otherwise Result<void> may have issues.
+static_assert(!std::is_constructible_v<Result<int>, Result<Future<int>>>);
+static_assert(!std::is_convertible_v<Result<int>, Result<Future<int>>>);
+static_assert(!std::is_assignable_v<Result<int>, Result<Future<int>>>);
+
+static_assert(std::is_same_v<
+              Result<Future<void>>,
+              tensorstore::FlatResult<std::invoke_result_t<Future<void>()>>>);
 
 TEST(FutureTest, Valid) {
-  EXPECT_FALSE(Future<int>().valid());
-  EXPECT_FALSE(Promise<int>().valid());
+  EXPECT_TRUE(Future<int>().null());
+  EXPECT_TRUE(Promise<int>().null());
   auto pair = PromiseFuturePair<int>::Make();
-  EXPECT_TRUE(pair.future.valid());
-  EXPECT_TRUE(pair.promise.valid());
+  EXPECT_FALSE(pair.future.null());
+  EXPECT_FALSE(pair.promise.null());
   auto future2 = pair.promise.future();
-  EXPECT_TRUE(future2.valid());
+  EXPECT_FALSE(future2.null());
 }
 
 TEST(FutureTest, MakeReadyFuture) {
@@ -113,6 +118,12 @@ TEST(FutureTest, MakeReadyFuture) {
   EXPECT_EQ(result, future.result());
 }
 
+TEST(FutureTest, MakeInPlace) {
+  auto pair = PromiseFuturePair<int>::Make(tensorstore::in_place, 4);
+  pair.promise.reset();  // drop link.
+  EXPECT_EQ(4, pair.future.value());
+}
+
 /// Tests that a ready future can be constructed implicitly.
 TEST(FutureTest, ConstructFromValue) {
   Future<int> x = 3;
@@ -163,22 +174,105 @@ TEST(FutureTest, FlattenResultLvalue) {
 
 /// Tests that Future::SetResult works.
 TEST(FutureTest, SetResult) {
-  auto pair = PromiseFuturePair<int>::Make();
-  EXPECT_FALSE(pair.promise.ready());
-  EXPECT_TRUE(pair.promise.result_needed());
-  EXPECT_FALSE(pair.future.ready());
-  Result<int> result{tensorstore::in_place};
-  bool got_result = false;
-  pair.future.ExecuteWhenReady([&](ReadyFuture<int> r) {
-    got_result = true;
-    result = r.result();
-  });
-  EXPECT_FALSE(got_result);
-  EXPECT_TRUE(pair.promise.SetResult(5));
-  EXPECT_FALSE(pair.promise.result_needed());
-  EXPECT_TRUE(pair.future.ready());
-  EXPECT_TRUE(pair.promise.ready());
-  EXPECT_EQ(result, 5);
+  // Success
+  {
+    auto pair = PromiseFuturePair<int>::Make();
+    EXPECT_FALSE(pair.promise.ready());
+    EXPECT_TRUE(pair.promise.result_needed());
+    EXPECT_FALSE(pair.future.ready());
+    Result<int> result{tensorstore::in_place};
+    bool got_result = false;
+    pair.future.ExecuteWhenReady([&](ReadyFuture<int> r) {
+      got_result = true;
+      result = r.result();
+    });
+    EXPECT_FALSE(got_result);
+    EXPECT_TRUE(pair.promise.SetResult(5));
+    EXPECT_FALSE(pair.promise.result_needed());
+    EXPECT_TRUE(pair.future.ready());
+    EXPECT_TRUE(pair.promise.ready());
+    EXPECT_EQ(result, 5);
+  }
+  // Success
+  {
+    auto pair = PromiseFuturePair<int>::Make();
+    pair.promise.SetResult(std::in_place, 6);
+    EXPECT_TRUE(pair.future.ready());
+    EXPECT_TRUE(pair.promise.ready());
+    EXPECT_TRUE(pair.future.result().ok());
+  }
+  // Success
+  {
+    auto pair = PromiseFuturePair<int>::Make();
+    pair.promise.SetResult(MakeResult(7));
+    EXPECT_TRUE(pair.future.ready());
+    EXPECT_TRUE(pair.promise.ready());
+    EXPECT_TRUE(pair.future.result().ok());
+  }
+  // Error
+  {
+    auto pair = PromiseFuturePair<int>::Make();
+    pair.promise.SetResult(absl::InternalError("error"));
+    EXPECT_TRUE(pair.future.ready());
+    EXPECT_TRUE(pair.promise.ready());
+    EXPECT_FALSE(pair.future.result().ok());
+  }
+  // Error
+  {
+    auto pair = PromiseFuturePair<int>::Make();
+    pair.promise.SetResult(MakeResult<int>(absl::InternalError("error")));
+    EXPECT_TRUE(pair.future.ready());
+    EXPECT_TRUE(pair.promise.ready());
+    EXPECT_FALSE(pair.future.result().ok());
+  }
+}
+
+TEST(FutureTest, SetResultVoid) {
+  // Success
+  {
+    auto pair = PromiseFuturePair<void>::Make();
+    EXPECT_FALSE(pair.promise.ready());
+    EXPECT_TRUE(pair.promise.result_needed());
+    EXPECT_FALSE(pair.future.ready());
+
+    EXPECT_TRUE(pair.promise.SetResult(absl::OkStatus()));
+    EXPECT_FALSE(pair.promise.result_needed());
+    EXPECT_TRUE(pair.future.ready());
+    EXPECT_TRUE(pair.promise.ready());
+    EXPECT_TRUE(pair.future.result().ok());
+  }
+  // Success
+  {
+    auto pair = PromiseFuturePair<void>::Make();
+    pair.promise.SetResult(std::in_place);
+    EXPECT_TRUE(pair.future.ready());
+    EXPECT_TRUE(pair.promise.ready());
+    EXPECT_TRUE(pair.future.result().ok());
+  }
+  // Success
+  {
+    auto pair = PromiseFuturePair<void>::Make();
+    pair.promise.SetResult(MakeResult<void>(absl::OkStatus()));
+    EXPECT_TRUE(pair.future.ready());
+    EXPECT_TRUE(pair.promise.ready());
+    EXPECT_TRUE(pair.future.result().ok());
+  }
+  // Error
+  {
+    auto pair = PromiseFuturePair<void>::Make();
+    pair.promise.SetResult(absl::InternalError("error"));
+    EXPECT_TRUE(pair.future.ready());
+    EXPECT_TRUE(pair.promise.ready());
+    EXPECT_FALSE(pair.future.result().ok());
+  }
+  // Error
+  {
+    auto pair = PromiseFuturePair<void>::Make();
+    pair.promise.SetResult(MakeResult<void>(absl::InternalError("error")));
+    EXPECT_TRUE(pair.future.ready());
+    EXPECT_TRUE(pair.promise.ready());
+    EXPECT_FALSE(pair.future.result().ok());
+  }
 }
 
 /// Tests that Future::Wait works.
@@ -186,7 +280,7 @@ TEST(FutureTest, Wait) {
   auto pair = PromiseFuturePair<int>::Make();
   std::thread thread(
       [](Promise<int> promise) {
-        std::this_thread::sleep_for(std::chrono::milliseconds(20));
+        absl::SleepFor(absl::Milliseconds(20));
         EXPECT_TRUE(promise.SetResult(5));
       },
       std::move(pair.promise));
@@ -200,16 +294,14 @@ TEST(FutureTest, WaitForFailure) {
   // allow it to pass if it works even once.
   for (size_t i = 0; i < 100; ++i) {
     auto pair = PromiseFuturePair<int>::Make();
-    EXPECT_FALSE(
-        pair.future.WaitFor(absl::FromChrono(std::chrono::milliseconds(10))));
+    EXPECT_FALSE(pair.future.WaitFor(absl::Milliseconds(10)));
     std::thread thread(
         [](Promise<int> promise) {
-          std::this_thread::sleep_for(std::chrono::milliseconds(20));
+          absl::SleepFor(absl::Milliseconds(20));
           EXPECT_TRUE(promise.SetResult(5));
         },
         pair.promise);
-    const bool ready =
-        pair.future.WaitFor(absl::FromChrono(std::chrono::milliseconds(5)));
+    const bool ready = pair.future.WaitFor(absl::Milliseconds(5));
     thread.join();
     if (!ready) {
       // Passed
@@ -226,14 +318,12 @@ TEST(FutureTest, WaitForSuccess) {
     auto pair = PromiseFuturePair<int>::Make();
     std::thread thread(
         [](Promise<int> promise) {
-          std::this_thread::sleep_for(std::chrono::milliseconds(5));
+          absl::SleepFor(absl::Milliseconds(5));
           EXPECT_TRUE(promise.SetResult(5));
         },
         pair.promise);
-    const bool ready1 =
-        pair.future.WaitFor(absl::FromChrono(std::chrono::milliseconds(20)));
-    const bool ready2 =
-        pair.future.WaitFor(absl::FromChrono(std::chrono::milliseconds(10)));
+    const bool ready1 = pair.future.WaitFor(absl::Milliseconds(20));
+    const bool ready2 = pair.future.WaitFor(absl::Milliseconds(10));
     thread.join();
     if (ready1 && ready2) {
       // Passed
@@ -252,7 +342,7 @@ TEST(FutureTest, WaitUntilFailure) {
     EXPECT_FALSE(pair.future.WaitUntil(absl::Now() + absl::Milliseconds(10)));
     std::thread thread(
         [](Promise<int> promise) {
-          std::this_thread::sleep_for(std::chrono::milliseconds(20));
+          absl::SleepFor(absl::Milliseconds(20));
           EXPECT_TRUE(promise.SetResult(5));
         },
         pair.promise);
@@ -274,7 +364,7 @@ TEST(FutureTest, WaitUntilSuccess) {
     auto pair = PromiseFuturePair<int>::Make();
     std::thread thread(
         [](Promise<int> promise) {
-          std::this_thread::sleep_for(std::chrono::milliseconds(5));
+          absl::SleepFor(absl::Milliseconds(5));
           EXPECT_TRUE(promise.SetResult(5));
         },
         pair.promise);
@@ -773,10 +863,12 @@ TEST(FutureTest, ConvertVoid) {
 }
 
 TEST(FutureTest, ConvertVoid2) {
+  Future<const void> f;
+  Promise<const void> p;
   auto pair = PromiseFuturePair<void>::Make();
-  Future<const void> f = pair.future;
-  Promise<const void> p = pair.promise;
-  pair.promise.SetResult(absl::in_place);
+  f = pair.future;
+  p = pair.promise;
+  pair.promise.SetResult(std::in_place);
   f.value();
 }
 
@@ -844,7 +936,7 @@ TEST(AcquireFutureReferenceTest, NoExistingFutureNotReady) {
   auto pair = PromiseFuturePair<void>::Make();
   pair.future.reset();
   auto future2 = pair.promise.future();
-  EXPECT_FALSE(future2.valid());
+  EXPECT_FALSE(!future2.null());
 }
 
 TEST(AcquireFutureReferenceTest, NoExistingFutureReady) {
@@ -1179,8 +1271,8 @@ TEST(LinkTest, ConcurrentForceAndSetReady) {
         pair1 = PromiseFuturePair<int>::Make(1);
         pair2 = PromiseFuturePair<int>::Make();
         pair3 = PromiseFuturePair<int>::Make();
-        Link(pair2.promise, pair1.future);
-        Link(pair3.promise, pair2.future);
+        LinkResult(pair2.promise, pair1.future);
+        LinkResult(pair3.promise, pair2.future);
       },
       /*finalize=*/[&] {},  //
       // Concurrently:
@@ -1208,7 +1300,7 @@ TEST(LinkTest, NoCallback) {
   // Test unlinking before the future becomes ready.
   {
     auto [linked_promise, linked_future] = PromiseFuturePair<int>::Make();
-    auto link = Link(linked_promise, future);
+    auto link = LinkResult(linked_promise, future);
     EXPECT_FALSE(linked_future.ready());
     link.Unregister();
     linked_future.Force();
@@ -1219,7 +1311,7 @@ TEST(LinkTest, NoCallback) {
   // Test forcing.
   {
     auto [linked_promise, linked_future] = PromiseFuturePair<int>::Make();
-    auto link = Link(linked_promise, future);
+    auto link = LinkResult(linked_promise, future);
     EXPECT_FALSE(linked_future.ready());
     linked_future.Force();
     ASSERT_TRUE(linked_future.ready());
@@ -1230,7 +1322,7 @@ TEST(LinkTest, NoCallback) {
   // Test linking after the future is ready.
   {
     auto [linked_promise, linked_future] = PromiseFuturePair<int>::Make();
-    auto link = Link(linked_promise, future);
+    auto link = LinkResult(linked_promise, future);
     ASSERT_TRUE(linked_future.ready());
     EXPECT_THAT(linked_future.result(), ::testing::Optional(5));
   }
@@ -1317,6 +1409,24 @@ TEST(LinkErrorTest, ConcurrentForceAndSetError) {
       [&] { pairB.promise.SetResult(absl::UnknownError("")); });
 }
 
+TEST(LinkErrorTest, LinkErrorVoidImmediateSuccessFailure) {
+  auto [promise, future] = PromiseFuturePair<void>::Make();
+  LinkError(std::move(promise), MakeReadyFuture<void>(absl::OkStatus()),
+            MakeReadyFuture<void>(absl::OkStatus()));
+
+  ASSERT_TRUE(future.ready());
+  EXPECT_FALSE(future.result().ok());
+}
+
+TEST(LinkErrorTest, LinkErrorVoidImmediateSuccessOk) {
+  auto [promise, future] = PromiseFuturePair<void>::Make(absl::OkStatus());
+  LinkError(std::move(promise), MakeReadyFuture<void>(absl::OkStatus()),
+            MakeReadyFuture<void>(absl::OkStatus()));
+
+  ASSERT_TRUE(future.ready());
+  EXPECT_TRUE(future.result().ok());
+}
+
 TEST(PromiseFuturePairTest, LinkImmediateSuccess) {
   auto future = PromiseFuturePair<int>::Link(
                     [](Promise<int> p, ReadyFuture<int> f) {
@@ -1488,6 +1598,63 @@ TEST(PromiseFuturePairTest, LinkDestroyCallback) {
   EXPECT_EQ(1, sentinel.use_count());
 }
 
+// Tests that WaitAllFuture works with no futures.
+TEST(WaitAllFuture, NoFuturesSpan) {
+  std::vector<AnyFuture> futures;
+  auto future = WaitAllFuture(futures);
+  ASSERT_TRUE(future.ready());
+  ASSERT_TRUE(future.result().ok());
+}
+
+TEST(WaitAllFuture, ReadyFuture) {
+  auto future = WaitAllFuture(MakeReadyFuture<void>(absl::OkStatus()),
+                              MakeReadyFuture<void>(absl::OkStatus()));
+  ASSERT_TRUE(future.ready());
+  EXPECT_TRUE(future.result().ok());
+
+  future = WaitAllFuture(MakeReadyFuture<void>(absl::OkStatus()),
+                         MakeReadyFuture<void>(absl::OkStatus()),
+                         MakeReadyFuture<void>(absl::InternalError("")));
+  ASSERT_TRUE(future.ready());
+  ASSERT_FALSE(future.result().ok());
+}
+
+TEST(WaitAllFuture, ReadyFutureSpanError) {
+  std::vector<AnyFuture> futures{MakeReadyFuture<void>(absl::OkStatus()),
+                                 MakeReadyFuture<void>(absl::OkStatus())};
+  auto future = WaitAllFuture(futures);
+  ASSERT_TRUE(future.ready());
+  EXPECT_TRUE(future.result().ok());
+
+  futures.push_back(MakeReadyFuture<void>(absl::InternalError("")));
+  future = WaitAllFuture(futures);
+  ASSERT_TRUE(future.ready());
+  ASSERT_FALSE(future.result().ok());
+}
+
+TEST(WaitAllFuture, ReadyFutureSpan) {
+  std::vector<AnyFuture> futures;
+  for (int i = 0; i < 17; i++) {
+    auto future = WaitAllFuture(futures);
+    ASSERT_TRUE(future.ready());
+    EXPECT_TRUE(future.result().ok());
+    futures.emplace_back(MakeReadyFuture<void>(absl::OkStatus()));
+  }
+}
+
+TEST(WaitAllFuture, NonVoidFuture) {
+  auto a = PromiseFuturePair<int>::Make();
+  auto b = PromiseFuturePair<int>::Make();
+
+  auto future = WaitAllFuture(a.future, b.future);
+  ASSERT_FALSE(future.ready());
+  a.promise.SetResult(2);
+  ASSERT_FALSE(future.ready());
+  b.promise.SetResult(absl::InternalError(""));
+  ASSERT_TRUE(future.ready());
+  EXPECT_FALSE(future.result().ok());
+}
+
 // Tests that MapFuture works with no futures.
 TEST(MapFutureTest, NoFutures) {
   auto future = MapFuture(InlineExecutor{}, [] { return 3; });
@@ -1531,6 +1698,14 @@ TEST(MapFutureTest, LValueReference) {
   EXPECT_EQ(10, a.value());
 }
 
+TEST(MapFutureTest, ReturnFuture) {
+  Future<int> a = 5;
+  auto f = MapFuture(
+      InlineExecutor{},
+      [](Result<int> a) -> Future<int> { return a.value() + 3; }, a);
+  EXPECT_THAT(f.result(), ::testing::Optional(8));
+}
+
 TEST(MapFutureValueTest, BothReady) {
   auto a = MakeReadyFuture<int>(3);
   auto b = MakeReadyFuture<int>(5);
@@ -1564,10 +1739,17 @@ TEST(MapFutureValueTest, ValueToError) {
               MatchesStatus(absl::StatusCode::kUnknown, "Got value: 3"));
 }
 
+TEST(MapFutureValueTest, ReturnFuture) {
+  Future<int> a = 5;
+  auto f = MapFutureValue(
+      InlineExecutor{}, [](int a) -> Future<int> { return a + 3; }, a);
+  EXPECT_THAT(f.result(), ::testing::Optional(8));
+}
+
 TEST(MapFutureErrorTest, Success) {
   auto pair = PromiseFuturePair<int>::Make();
   auto mapped = MapFutureError(
-      InlineExecutor{}, [](Status status) { return 5; }, pair.future);
+      InlineExecutor{}, [](absl::Status status) { return 5; }, pair.future);
   EXPECT_FALSE(mapped.ready());
   pair.promise.SetResult(7);
   EXPECT_EQ(7, mapped.result());
@@ -1577,7 +1759,7 @@ TEST(MapFutureErrorTest, ErrorMappedToSuccess) {
   auto pair = PromiseFuturePair<int>::Make();
   auto mapped = MapFutureError(
       InlineExecutor{},
-      [](Status status) {
+      [](absl::Status status) {
         EXPECT_EQ(absl::UnknownError("message"), status);
         return 5;
       },
@@ -1591,7 +1773,7 @@ TEST(MapFutureErrorTest, ErrorMappedToError) {
   auto pair = PromiseFuturePair<int>::Make();
   auto mapped = MapFutureError(
       InlineExecutor{},
-      [](Status status) {
+      [](absl::Status status) {
         return tensorstore::MaybeAnnotateStatus(status, "Mapped");
       },
       pair.future);
@@ -1601,4 +1783,98 @@ TEST(MapFutureErrorTest, ErrorMappedToError) {
               MatchesStatus(absl::StatusCode::kUnknown, "Mapped: message"));
 }
 
+TEST(MakeReadyFutureTest, Basic) {
+  auto future = MakeReadyFuture();
+  static_assert(std::is_same_v<ReadyFuture<const void>, decltype(future)>);
+  EXPECT_TRUE(future.ready());
+  EXPECT_EQ(MakeResult(), future.result());
+}
+
+TEST(FutureTest, SetDeferredResult) {
+  auto [promise, future] = PromiseFuturePair<int>::Make();
+  SetDeferredResult(promise, 2);
+  EXPECT_FALSE(future.ready());
+  SetDeferredResult(promise, 3);
+  EXPECT_FALSE(future.ready());
+  promise = Promise<int>();
+  ASSERT_TRUE(future.ready());
+  EXPECT_THAT(future.result(), ::testing::Optional(2));
+}
+
+TEST(FutureTest, SetDeferredResultAfterReady) {
+  auto [promise, future] = PromiseFuturePair<int>::Make();
+  promise.SetResult(1);
+  ASSERT_TRUE(future.ready());
+  SetDeferredResult(promise, 2);
+  ASSERT_TRUE(future.ready());
+  EXPECT_THAT(future.result(), ::testing::Optional(1));
+}
+
+TEST(FutureTest, SetDeferredResultSetReady) {
+  auto [promise, future] = PromiseFuturePair<int>::Make();
+  int value = 1;
+  future.ExecuteWhenReady(
+      [&](ReadyFuture<int> r) { value *= (r.result().ok()) ? 2 : 3; });
+
+  SetDeferredResult(promise, absl::InternalError("1"));
+  SetDeferredResult(promise, absl::InternalError("2"));
+  promise.SetReady();
+  future.Wait();
+
+  // promise & future refs still exist.
+  EXPECT_EQ(3, value);
+}
+// Tests that `RETURN_IF_ERROR` saves the status before the expression lifetime
+// ends.
+TEST(FutureTest, ReturnIfError) {
+  auto do_test = [] {
+    TENSORSTORE_RETURN_IF_ERROR(MakeReadyFuture<int>(42).result(), false);
+    return true;
+  };
+  EXPECT_EQ(true, do_test());
+}
+
+TEST(FutureTest, UntypedExecuteWhenReadyAlreadyDone) {
+  Future<int> f(3);
+  bool ran = false;
+  f.UntypedExecuteWhenReady([&](AnyFuture f) { ran = true; });
+  EXPECT_TRUE(ran);
+}
+
+TEST(FutureTest, UntypedExecuteWhenReadyNotAlreadyDone) {
+  auto [promise, future] = PromiseFuturePair<int>::Make();
+  bool ran = false;
+  future.UntypedExecuteWhenReady([&](AnyFuture f) { ran = true; });
+  EXPECT_FALSE(ran);
+  promise.SetResult(5);
+  EXPECT_TRUE(ran);
+}
+
+TEST(FutureTest, FutureResultFuture) {
+  auto [promise, future] = PromiseFuturePair<int>::Make();
+  Result<Future<int>> rf(future);
+  Future<int> f(rf);
+  promise.SetResult(5);
+  EXPECT_EQ(5, f.value());
+}
+
+TEST(FutureTest, Live) {
+  auto& registry = tensorstore::internal_metrics::GetMetricRegistry();
+  EXPECT_EQ(
+      0, std::get<int64_t>(
+             registry.Collect("/tensorstore/futures/live")->gauges[0].value));
+
+  /// While the future is active we expect that the live metric has a value.
+  {
+    auto [promise, future] = PromiseFuturePair<int>::Make();
+    EXPECT_NE(
+        0, std::get<int64_t>(
+               registry.Collect("/tensorstore/futures/live")->gauges[0].value));
+  }
+
+  EXPECT_EQ(
+      0, std::get<int64_t>(
+             registry.Collect("/tensorstore/futures/live")->gauges[0].value));
+}
+
 }  // namespace
