@@ -27,27 +27,31 @@
 #include "tensorstore/util/internal/iterate_impl.h"
 #include "tensorstore/util/iterate.h"
 #include "tensorstore/util/span.h"
-#include "tensorstore/util/to_string.h"
+#include "tensorstore/util/str_cat.h"
 
 namespace {
 
-using tensorstore::ArrayIterateResult;
-using tensorstore::ContiguousLayoutOrder;
-using tensorstore::DimensionIndex;
-using tensorstore::include_repeated_elements;
-using tensorstore::Index;
-using tensorstore::IterationConstraints;
-using tensorstore::LayoutOrderConstraint;
-using tensorstore::skip_repeated_elements;
-using tensorstore::span;
-using tensorstore::internal_iterate::
+using ::tensorstore::ArrayIterateResult;
+using ::tensorstore::c_order;
+using ::tensorstore::ContiguousLayoutOrder;
+using ::tensorstore::DimensionIndex;
+using ::tensorstore::fortran_order;
+using ::tensorstore::include_repeated_elements;
+using ::tensorstore::Index;
+using ::tensorstore::IterationConstraints;
+using ::tensorstore::LayoutOrderConstraint;
+using ::tensorstore::skip_repeated_elements;
+using ::tensorstore::span;
+using ::tensorstore::internal::AdvanceIndices;
+using ::tensorstore::internal::DefaultIterationResult;
+using ::tensorstore::internal_iterate::
     ComputeStridedLayoutDimensionIterationOrder;
-using tensorstore::internal_iterate::ExtractInnerShapeAndStrides;
-using tensorstore::internal_iterate::InnerShapeAndStrides;
-using tensorstore::internal_iterate::PermuteAndSimplifyStridedIterationLayout;
-using tensorstore::internal_iterate::SimplifyStridedIterationLayout;
-using tensorstore::internal_iterate::StridedIterationLayout;
-using testing::ElementsAre;
+using ::tensorstore::internal_iterate::ExtractInnerShapeAndStrides;
+using ::tensorstore::internal_iterate::InnerShapeAndStrides;
+using ::tensorstore::internal_iterate::PermuteAndSimplifyStridedIterationLayout;
+using ::tensorstore::internal_iterate::SimplifyStridedIterationLayout;
+using ::tensorstore::internal_iterate::StridedIterationLayout;
+using ::testing::ElementsAre;
 
 TEST(LayoutOrderConstraint, Basic) {
   static_assert(!LayoutOrderConstraint{}, "");
@@ -453,6 +457,21 @@ TEST(IterateOverStridedLayoutsTest, InnerRank0ContiguousC) {
   EXPECT_EQ(expected_result, result);
 }
 
+TEST(IterateOverStridedLayoutsTest, EmptyDomain) {
+  const Index shape[] = {0, 3};
+  const Index strides[] = {0, 1};
+
+  std::vector<int> result;
+  auto func = [&](int a) {
+    result.emplace_back(a);
+    return true;
+  };
+  IterateOverStridedLayouts(shape, {{strides}}, func,
+                            {ContiguousLayoutOrder::c, skip_repeated_elements},
+                            0);
+  EXPECT_THAT(result, ::testing::ElementsAre());
+}
+
 TEST(IterateOverStridedLayoutsTest, InnerRank0ContiguousCStop) {
   const Index shape[] = {2, 3};
   const Index strides0[] = {3, 1};
@@ -523,4 +542,70 @@ TEST(ArrayIterateResultTest, PrintToOstream) {
             tensorstore::StrCat(ArrayIterateResult{true, 3}));
 }
 
+template <ContiguousLayoutOrder Order>
+std::vector<std::vector<int>> GetIndexVectors(std::vector<int> shape) {
+  std::vector<std::vector<int>> result;
+
+  std::vector<int> indices(shape.size());
+  do {
+    result.push_back(indices);
+  } while (AdvanceIndices<Order>(indices.size(), indices.data(), shape.data()));
+  return result;
+}
+
+template <ContiguousLayoutOrder Order>
+std::vector<std::vector<int>> GetIndexVectors(std::vector<int> inclusive_min,
+                                              std::vector<int> exclusive_max) {
+  std::vector<std::vector<int>> result;
+
+  std::vector<int> indices = inclusive_min;
+  do {
+    result.push_back(indices);
+  } while (AdvanceIndices<Order>(indices.size(), indices.data(),
+                                 inclusive_min.data(), exclusive_max.data()));
+  return result;
+}
+
+TEST(AdvanceIndicesTest, COrderRank0) {
+  EXPECT_THAT(GetIndexVectors<c_order>({}), ElementsAre(ElementsAre()));
+}
+
+TEST(AdvanceIndicesTest, FortranOrderRank0) {
+  EXPECT_THAT(GetIndexVectors<fortran_order>({}), ElementsAre(ElementsAre()));
+}
+
+TEST(AdvanceIndicesTest, COrderShape) {
+  EXPECT_THAT(GetIndexVectors<c_order>({2, 3}),
+              ElementsAre(  //
+                  ElementsAre(0, 0), ElementsAre(0, 1), ElementsAre(0, 2),
+                  ElementsAre(1, 0), ElementsAre(1, 1), ElementsAre(1, 2)));
+}
+
+TEST(AdvanceIndicesTest, FortranOrderShape) {
+  EXPECT_THAT(GetIndexVectors<fortran_order>({2, 3}),
+              ElementsAre(ElementsAre(0, 0), ElementsAre(1, 0),  //
+                          ElementsAre(0, 1), ElementsAre(1, 1),  //
+                          ElementsAre(0, 2), ElementsAre(1, 2)));
+}
+
+TEST(AdvanceIndicesTest, COrderInclusiveMinExclusiveMax) {
+  EXPECT_THAT(
+      GetIndexVectors<c_order>({1, 2}, {3, 5}),
+      ElementsAre(ElementsAre(1, 2), ElementsAre(1, 3), ElementsAre(1, 4),
+                  ElementsAre(2, 2), ElementsAre(2, 3), ElementsAre(2, 4)));
+}
+
+TEST(AdvanceIndicesTest, FortranOrderInclusiveMinExclusiveMax) {
+  EXPECT_THAT(GetIndexVectors<fortran_order>({1, 2}, {3, 5}),
+              ElementsAre(ElementsAre(1, 2), ElementsAre(2, 2),  //
+                          ElementsAre(1, 3), ElementsAre(2, 3),  //
+                          ElementsAre(1, 4), ElementsAre(2, 4)));
+}
+
+// Test bool specialization.
+static_assert(DefaultIterationResult<bool>::value() == true, "");
+
+// Test default definition.
+static_assert(DefaultIterationResult<int>::value() == 0, "");
+
 }  // namespace
