@@ -21,23 +21,33 @@
 #include <utility>
 #include <vector>
 
+#include <gmock/gmock.h>
 #include <gtest/gtest.h>
 #include "tensorstore/util/status.h"
+#include "tensorstore/util/status_testutil.h"
 
 namespace {
 
-using tensorstore::Result;
-using tensorstore::Status;
+using ::tensorstore::ChainResult;
+using ::tensorstore::FlatMapResultType;
+using ::tensorstore::FlatResult;
+using ::tensorstore::Result;
+using ::tensorstore::UnwrapQualifiedResultType;
+using ::tensorstore::UnwrapResultType;
 
-static_assert(std::is_convertible<Result<int>, Result<float>>::value, "");
+static_assert(std::is_convertible_v<Result<int>, Result<float>>, "");
+static_assert(!std::is_convertible_v<Result<int>, Result<std::string>>, "");
+static_assert(std::is_same_v<int, Result<int>::value_type>, "");
 
-static_assert(!std::is_convertible<Result<int>, Result<std::string>>::value,
+/// validate some internal helper methods.
+static_assert(!tensorstore::internal_result::is_result_status_or_inplace<int>,
               "");
-
-static_assert(std::is_same<int, Result<int>::value_type>::value, "");
+static_assert(
+    tensorstore::internal_result::is_result_status_or_inplace<Result<void>>,
+    "");
 
 TEST(ResultTest, ConstructDefault) {
-  Result<int> result{absl::in_place};
+  Result<int> result{std::in_place};
   EXPECT_TRUE(result.has_value());
   EXPECT_TRUE(result.ok());
   EXPECT_TRUE(result);
@@ -53,30 +63,29 @@ TEST(ResultTest, ConstructValue) {
 }
 
 TEST(ResultDeathTest, ConstructSuccess) {
-  tensorstore::Status status;
+  absl::Status status;
   ASSERT_DEATH(Result<int> result(status), "status");
 }
 
 TEST(ResultDeathTest, AssignSuccess) {
   // Cannot get blood from a stone.
-  tensorstore::Status status;
-  Result<int> result{absl::in_place};
+  absl::Status status;
+  Result<int> result{std::in_place};
   ASSERT_DEATH(result = status, "status");
 }
 
 TEST(ResultDeathTest, ConstructDefaultHasNoStatus) {
-  Result<int> result{absl::in_place};
+  Result<int> result{std::in_place};
   ASSERT_DEATH(result.status(), "has_value");
 }
 
 TEST(ResultDeathTest, ConstructStatusHasNoValue) {
   // Cannot get blood from a stone.
-  tensorstore::Status status(absl::StatusCode::kUnknown,
-                             "My custom error message");
+  absl::Status status = absl::UnknownError("My custom error message");
   Result<int> result(status);
 
   ASSERT_DEATH(result.value(), "");
-  ASSERT_DEATH(static_cast<const Result<int>&>(result).value(), "");
+  ASSERT_DEATH(static_cast<const Result<int> &>(result).value(), "");
   ASSERT_DEATH(std::move(result).value(), "My custom error message");
 }
 
@@ -92,7 +101,8 @@ TEST(ResultDeathTest, ValidResultChecksOnStatus) {
 }
 
 TEST(ResultTest, ConstructStatus) {
-  tensorstore::Status status(absl::StatusCode::kUnknown, "Message");
+  absl::Status status = absl::UnknownError("Message");
+
   Result<int> result(status);
   EXPECT_FALSE(result);
   EXPECT_FALSE(result.ok());
@@ -101,12 +111,42 @@ TEST(ResultTest, ConstructStatus) {
 }
 
 TEST(ResultTest, ConstructStatusMove) {
-  tensorstore::Status status(absl::StatusCode::kUnknown, "Message");
+  absl::Status status = absl::UnknownError("Message");
   auto message = status.message();  // string_view
   Result<int> result(std::move(status));
   EXPECT_EQ(message.data(), result.status().message().data());
 }
 
+TEST(ResultTest, ConstructVoidStatus) {
+  absl::Status status = absl::UnknownError("Message");
+  Result<void> result(status);
+  EXPECT_FALSE(result);
+  EXPECT_FALSE(result.ok());
+
+  EXPECT_EQ(status, result.status());
+}
+
+TEST(ResultTest, ConstructVoidStatusMove) {
+  absl::Status status = absl::UnknownError("Message");
+  auto message = status.message();  // string_view
+  Result<void> result(std::move(status));
+  EXPECT_EQ(message.data(), result.status().message().data());
+}
+
+TEST(ResultTest, ConstructVoidOkStatus) {
+  absl::Status status = absl::OkStatus();
+  Result<void> result(status);
+  EXPECT_TRUE(result);
+  EXPECT_TRUE(result.ok());
+}
+
+TEST(ResultTest, ConstructVoidOkStatusMove) {
+  absl::Status status = absl::OkStatus();
+  Result<void> result(std::move(status));
+  EXPECT_TRUE(result);
+  EXPECT_TRUE(result.ok());
+}
+
 TEST(ResultTest, ConstructValueMove) {
   std::unique_ptr<int> value(new int(3));
   Result<std::unique_ptr<int>> result(std::move(value));
@@ -153,15 +193,19 @@ TEST(ResultTest, ConstructCopyFailure) {
 }
 
 TEST(ResultTest, Comparison) {
-  // Compare with value.
+  // Compare with same type.
   const Result<int> r(1);
+  EXPECT_EQ(true, r == Result<int>(1));
+  EXPECT_EQ(false, r != Result<int>(1));
+  EXPECT_EQ(true, Result<int>{std::in_place} == Result<int>{std::in_place});
+  EXPECT_EQ(false, Result<int>{std::in_place} == r);
+  EXPECT_EQ(false, Result<int>{std::in_place} == r);
+
+  // Compare with different type
   EXPECT_EQ(true, r == Result<float>(1));
   EXPECT_EQ(false, r != Result<float>(1));
   EXPECT_EQ(false, Result<int>(2) == Result<float>(1));
   EXPECT_EQ(true, Result<int>(2) != Result<float>(1));
-  EXPECT_EQ(true, Result<int>{absl::in_place} == Result<int>{absl::in_place});
-  EXPECT_EQ(false, Result<int>{absl::in_place} == r);
-  EXPECT_EQ(false, Result<int>{absl::in_place} == r);
 
   Result<int> err{absl::UnknownError("Message")};
   Result<int> err2 = err;
@@ -171,25 +215,38 @@ TEST(ResultTest, Comparison) {
   EXPECT_EQ(false, err != err2);
   EXPECT_EQ(false, err == Result<int>(1));
   EXPECT_EQ(true, err != Result<int>(1));
-  EXPECT_EQ(false, err == Result<int>(absl::in_place));
-  EXPECT_EQ(true, err != Result<int>(absl::in_place));
+  EXPECT_EQ(false, err == Result<int>(std::in_place));
+  EXPECT_EQ(true, err != Result<int>(std::in_place));
 
   // Compare with values.
   EXPECT_EQ(true, r == 1);
   EXPECT_EQ(false, r == 2);
   EXPECT_EQ(false, r != 1);
   EXPECT_EQ(true, r != 2);
+
+  // Compare Result<void>
+  const Result<void> rv = tensorstore::MakeResult();
+  Result<void> err3 = absl::UnknownError("Message");
+  EXPECT_TRUE(rv == rv);
+  EXPECT_FALSE(rv != rv);
+  EXPECT_TRUE(rv == Result<void>(std::in_place));
+  EXPECT_TRUE(err3 == err3);
+  EXPECT_FALSE(err3 != err3);
+  EXPECT_FALSE(rv == err3);
+  EXPECT_FALSE(err3 == rv);
+  EXPECT_TRUE(rv != err3);
+  EXPECT_TRUE(err3 != rv);
 }
 
 TEST(ResultTest, AssignMoveFailure) {
   Result<int> result(absl::UnknownError("Hello"));
   auto message = result.status().message();
-  Result<int> result2{absl::in_place};
+  Result<int> result2{std::in_place};
   result2 = std::move(result);
   EXPECT_FALSE(result2);
   EXPECT_EQ(message.data(), result2.status().message().data());
 
-  Result<int> other(absl::in_place);
+  Result<int> other(std::in_place);
   result2 = std::move(other);
   EXPECT_TRUE(result2.has_value());
   EXPECT_EQ(0, *result2);
@@ -206,20 +263,20 @@ TEST(ResultTest, AssignMoveSuccess) {
 
 TEST(ResultTest, AssignCopyFailure) {
   Result<int> result(absl::UnknownError("Hello"));
-  Result<int> result2{absl::in_place};
+  Result<int> result2{std::in_place};
   result2 = result;
   EXPECT_EQ(result2, result);
   EXPECT_FALSE(result2);
   EXPECT_EQ(result2.status(), result.status());
 
-  result2 = Result<int>(absl::in_place);
+  result2 = Result<int>(std::in_place);
   EXPECT_TRUE(result2.has_value());
   EXPECT_EQ(0, *result2);
 }
 
 TEST(ResultTest, AssignCopySuccess) {
   Result<int> result(3);
-  Result<int> result2{absl::in_place};
+  Result<int> result2{std::in_place};
   result2 = result;
   EXPECT_EQ(result2, result);
   EXPECT_TRUE(result2);
@@ -228,7 +285,7 @@ TEST(ResultTest, AssignCopySuccess) {
 
 TEST(ResultTest, ConvertingValueConstructor) {
   float a = 3.0f;
-  const tensorstore::Status b(absl::StatusCode::kUnknown, "Hello");
+  const absl::Status b(absl::StatusCode::kUnknown, "Hello");
 
   Result<double> c(a);
   EXPECT_TRUE(c);
@@ -275,7 +332,7 @@ TEST(ResultTest, ConvertingConstructor) {
 TEST(ResultTest, Swap) {
   using std::swap;
 
-  tensorstore::Status status(absl::StatusCode::kUnknown, "Message");
+  absl::Status status(absl::StatusCode::kUnknown, "Message");
 
   Result<int> result(status);
   Result<int> result2(3);
@@ -424,43 +481,43 @@ TEST(ResultTest, Message) {
 
 TEST(ResultTest, Value) {
   static_assert(
-      std::is_same<decltype(std::declval<const Result<int>&>().value()),
-                   const int&>::value,
+      std::is_same_v<decltype(std::declval<const Result<int> &>().value()),
+                     const int &>,
       "");
   static_assert(
-      std::is_same<decltype(std::declval<Result<int>&>().value()), int&>::value,
+      std::is_same_v<decltype(std::declval<Result<int> &>().value()), int &>,
       "");
   static_assert(
-      std::is_same<decltype(std::declval<Result<int>&&>().value()), int>::value,
+      std::is_same_v<decltype(std::declval<Result<int> &&>().value()), int>,
       "");
 
-  static_assert(std::is_same<decltype(*std::declval<const Result<int>&>()),
-                             const int&>::value,
+  static_assert(std::is_same_v<decltype(*std::declval<const Result<int> &>()),
+                               const int &>,
+                "");
+  static_assert(std::is_same_v<decltype(*std::declval<Result<int> &>()), int &>,
                 "");
-  static_assert(
-      std::is_same<decltype(*std::declval<Result<int>&>()), int&>::value, "");
 
   static_assert(
-      std::is_same<decltype(std::declval<const Result<int>&>().operator->()),
-                   const int*>::value,
+      std::is_same_v<decltype(std::declval<const Result<int> &>().operator->()),
+                     const int *>,
       "");
   static_assert(
-      std::is_same<decltype(std::declval<Result<int>&>().operator->()),
-                   int*>::value,
+      std::is_same_v<decltype(std::declval<Result<int> &>().operator->()),
+                     int *>,
       "");
 
   Result<int> result = 3;
   EXPECT_EQ(3, result.value());
   EXPECT_EQ(3, *result);
   EXPECT_EQ(3, result.value_or(4));
-  EXPECT_EQ(3, static_cast<const Result<int>&>(result).value());
+  EXPECT_EQ(3, static_cast<const Result<int> &>(result).value());
   EXPECT_EQ(3, std::move(result).value());
 
   std::vector<int> vec{1, 2, 3};
-  const int* data = vec.data();
+  const int *data = vec.data();
   Result<std::vector<int>> result2(std::move(vec));
   EXPECT_EQ(3, result2->size());
-  EXPECT_EQ(3, static_cast<const Result<std::vector<int>>&>(result2)->size());
+  EXPECT_EQ(3, static_cast<const Result<std::vector<int>> &>(result2)->size());
   std::vector<int> vec2 = std::move(result2).value();
   EXPECT_EQ(data, vec2.data());
 }
@@ -566,15 +623,15 @@ TEST(ResultVoidTest, MoveConstructFromStatus) {
 }
 
 TEST(ResultVoidTest, CopyConstructFromStatus) {
-  Status s = absl::UnknownError("C");
+  absl::Status s = absl::UnknownError("C");
   Result<void> r = s;
   EXPECT_EQ(absl::UnknownError("C"), r.status());
 }
 
 TEST(ResultVoidTest, CopyAssignFromStatus) {
-  Status s = absl::UnknownError("C");
-  Result<void> t(absl::in_place);
-  Result<void> r{absl::in_place};
+  absl::Status s = absl::UnknownError("C");
+  Result<void> t(std::in_place);
+  Result<void> r{std::in_place};
   r = s;
   EXPECT_EQ(absl::UnknownError("C"), r.status());
   r = t;
@@ -584,12 +641,12 @@ TEST(ResultVoidTest, CopyAssignFromStatus) {
 }
 
 TEST(ResultVoidTest, MoveAssignFromStatus) {
-  Result<void> r{absl::in_place};
+  Result<void> r{std::in_place};
   r = absl::UnknownError("C");
   EXPECT_EQ(absl::UnknownError("C"), r.status());
-  r = Result<void>(absl::in_place);
+  r = Result<void>(std::in_place);
   EXPECT_TRUE(r.has_value());
-  r = Result<void>(absl::in_place);
+  r = Result<void>(std::in_place);
   EXPECT_TRUE(r.has_value());
 }
 
@@ -608,7 +665,7 @@ TEST(ResultVoidTest, MakeResultVoid) {
 }
 
 TEST(ResultVoidTest, ReturnVoid) {
-  auto fn = []() -> Result<void> { return {absl::in_place}; };
+  auto fn = []() -> Result<void> { return {std::in_place}; };
 
   Result<void> r = fn();
   EXPECT_TRUE(r.has_value());
@@ -617,10 +674,10 @@ TEST(ResultVoidTest, ReturnVoid) {
 struct MoveOnly {
   MoveOnly(int value) : value(value) {}
 
-  MoveOnly(MoveOnly const&) = delete;
-  MoveOnly& operator=(const MoveOnly&) = delete;
-  MoveOnly(MoveOnly&&) = default;
-  MoveOnly& operator=(MoveOnly&&) = default;
+  MoveOnly(MoveOnly const &) = delete;
+  MoveOnly &operator=(const MoveOnly &) = delete;
+  MoveOnly(MoveOnly &&) = default;
+  MoveOnly &operator=(MoveOnly &&) = default;
 
   int value;
 };
@@ -643,7 +700,7 @@ TEST(ResultTest, MoveOnlyConstruct) {
     EXPECT_EQ(3, x.value().value);
   }
   {
-    Result<MoveOnly> x(absl::in_place, 3);
+    Result<MoveOnly> x(std::in_place, 3);
     EXPECT_EQ(3, x.value().value);
   }
   {
@@ -658,7 +715,7 @@ TEST(ResultTest, MoveOnlyConstruct) {
   }
   {
     MoveOnly w(3);
-    Result<MoveOnly> y(absl::in_place, 4);
+    Result<MoveOnly> y(std::in_place, 4);
     y = std::move(w);
     EXPECT_EQ(3, y.value().value);
   }
@@ -684,7 +741,7 @@ TEST(ResultTest, MoveOnlyConstruct) {
     EXPECT_EQ(3, y.value().value);
   }
   {
-    Result<MoveOnly> y{absl::in_place, 4};
+    Result<MoveOnly> y{std::in_place, 4};
     y = absl::UnknownError("D");
     EXPECT_FALSE(y);
   }
@@ -711,10 +768,10 @@ TEST(ResultTest, MoveOnlyFunc) {
 struct CopyOnly {
   CopyOnly(int value) : value(value) {}
 
-  CopyOnly(CopyOnly const&) = default;
-  CopyOnly& operator=(const CopyOnly&) = default;
-  CopyOnly(CopyOnly&&) = delete;
-  CopyOnly& operator=(CopyOnly&&) = delete;
+  CopyOnly(CopyOnly const &) = default;
+  CopyOnly &operator=(const CopyOnly &) = default;
+  CopyOnly(CopyOnly &&) = delete;
+  CopyOnly &operator=(CopyOnly &&) = delete;
 
   int value;
 };
@@ -737,7 +794,7 @@ TEST(ResultTest, CopyOnlyConstruct) {
     EXPECT_EQ(3, x.value().value);
   }
   {
-    Result<CopyOnly> x(absl::in_place, 3);
+    Result<CopyOnly> x(std::in_place, 3);
     EXPECT_EQ(3, x.value().value);
   }
   {
@@ -752,7 +809,7 @@ TEST(ResultTest, CopyOnlyConstruct) {
   }
   {
     CopyOnly w(3);
-    Result<CopyOnly> y(absl::in_place, 4);
+    Result<CopyOnly> y(std::in_place, 4);
     y = w;
     EXPECT_EQ(3, y.value().value);
   }
@@ -769,7 +826,7 @@ TEST(ResultTest, CopyOnlyConstruct) {
   }
   // No overload for y = CopyOnly(3)
   {
-    Result<CopyOnly> y{absl::in_place, 4};
+    Result<CopyOnly> y{std::in_place, 4};
     y = absl::UnknownError("D");
     EXPECT_FALSE(y);
   }
@@ -786,13 +843,13 @@ struct Explicit {
   explicit Explicit(MoveOnly v) : value(v.value) {}
   explicit Explicit(CopyOnly v) : value(v.value) {}
 
-  Explicit(const Explicit& x) : value(x.value) {}
-  Explicit& operator=(const Explicit& x) {
+  Explicit(const Explicit &x) : value(x.value) {}
+  Explicit &operator=(const Explicit &x) {
     value = x.value;
     return *this;
   }
-  Explicit(Explicit&& x) : value(x.value) {}
-  Explicit& operator=(Explicit&& x) {
+  Explicit(Explicit &&x) : value(x.value) {}
+  Explicit &operator=(Explicit &&x) {
     value = std::move(x).value;
     return *this;
   }
@@ -807,7 +864,7 @@ TEST(ResultTest, ExplicitConstruct) {
     EXPECT_EQ(3, x.value().value);
   }
   {
-    Result<Explicit> x(absl::in_place, 3);
+    Result<Explicit> x(std::in_place, 3);
     EXPECT_EQ(3, x.value().value);
   }
   {
@@ -822,7 +879,7 @@ TEST(ResultTest, ExplicitConstruct) {
   }
   {
     Explicit w(3);
-    Result<Explicit> y(absl::in_place, 4);
+    Result<Explicit> y(std::in_place, 4);
     y = w;
     EXPECT_EQ(3, y.value().value);
   }
@@ -839,7 +896,7 @@ TEST(ResultTest, ExplicitConstruct) {
   }
   {
     Explicit w(3);
-    Result<Explicit> y(absl::in_place, 4);
+    Result<Explicit> y(std::in_place, 4);
     y = std::move(w);
     EXPECT_EQ(3, y.value().value);
   }
@@ -864,7 +921,7 @@ TEST(ResultTest, ExplicitConstruct) {
     EXPECT_EQ(3, y.value().value);
   }
   {
-    Result<Explicit> y{absl::in_place, 4};
+    Result<Explicit> y{std::in_place, 4};
     y = absl::UnknownError("D");
     EXPECT_FALSE(y);
   }
@@ -872,13 +929,13 @@ TEST(ResultTest, ExplicitConstruct) {
 
 TEST(ResultTest, ExplicitConvertingConstructor) {
   {
-    Result<CopyOnly> w{absl::in_place, 4};
+    Result<CopyOnly> w{std::in_place, 4};
     Result<Explicit> y{w};
     EXPECT_TRUE(y);
     EXPECT_EQ(4, y.value().value);
   }
   {
-    Result<MoveOnly> w{absl::in_place, 4};
+    Result<MoveOnly> w{std::in_place, 4};
     Result<Explicit> y{std::move(w)};
     EXPECT_TRUE(y);
     EXPECT_EQ(4, y.value().value);
@@ -892,7 +949,7 @@ struct X {
 TEST(ResultTest, AdvancedConversion) {
   // See result_nc_test.cc as well.
   Result<int> int_result(3);
-  Result<X> x_result(absl::in_place, int_result);
+  Result<X> x_result(std::in_place, int_result);
 }
 
 TEST(ResultTest, MakeResult) {
@@ -912,7 +969,7 @@ TEST(ResultTest, MakeResult) {
     EXPECT_EQ(0, *result);
   }
 
-  const Status err(absl::StatusCode::kUnknown, "C");
+  const absl::Status err(absl::StatusCode::kUnknown, "C");
   {
     auto result = tensorstore::MakeResult(err);
     EXPECT_FALSE(result.has_value());
@@ -923,59 +980,6 @@ TEST(ResultTest, MakeResult) {
   }
 }
 
-// Tests of Result::Construct.
-
-TEST(ResultConstructTest, AssignStatusCopy) {
-  Result<int> res = absl::UnknownError("");
-  const Status status(absl::StatusCode::kUnknown, "new");
-  res.Construct(status);
-  EXPECT_FALSE(res);
-  EXPECT_EQ(status, res.status());
-}
-
-TEST(ResultConstructTest, AssignStatusMove) {
-  Result<int> res = absl::UnknownError("");
-  Status status(absl::StatusCode::kUnknown, "new");
-  res.Construct(absl::UnknownError("new"));
-  EXPECT_FALSE(res);
-  EXPECT_EQ(status, res.status());
-}
-
-TEST(ResultConstructTest, AssignResultCopy) {
-  Result<int> res = absl::UnknownError("");
-  Result<int> other(3);
-  res.Construct(other);
-  EXPECT_EQ(res, other);
-}
-
-TEST(ResultConstructTest, AssignResultMove) {
-  Result<int> res = absl::UnknownError("");
-  Result<int> other(3);
-  res.Construct(Result<int>(3));
-  EXPECT_EQ(res, other);
-}
-
-TEST(ResultConstructTest, AssignValue) {
-  Result<int> res = absl::UnknownError("");
-  Result<int> other(3);
-  res.Construct(3);
-  EXPECT_EQ(res, other);
-}
-
-TEST(ResultConstructTest, AssignInPlace) {
-  Result<int> res = absl::UnknownError("");
-  Result<int> other(3);
-  res.Construct(tensorstore::in_place, 3);
-  EXPECT_EQ(res, other);
-}
-
-TEST(ResultConstructTest, AssignInPlaceInitializerList) {
-  Result<std::vector<int>> res = absl::UnknownError("");
-  Result<std::vector<int>> other(tensorstore::in_place, {1, 2, 3});
-  res.Construct(tensorstore::in_place, {1, 2, 3});
-  EXPECT_EQ(res, other);
-}
-
 TEST(ResultTest, AssignOrReturn) {
   const auto Helper = [](Result<int> r) {
     TENSORSTORE_ASSIGN_OR_RETURN(auto x, r);
@@ -1001,4 +1005,215 @@ TEST(ResultTest, AssignOrReturnAnnotate) {
             Helper(absl::UnknownError("Got error")));
 }
 
+/// FIXME: Is FlatMapResultType pulling it's weight?
+
+static_assert(std::is_same_v<UnwrapResultType<int>, int>);
+static_assert(std::is_same_v<UnwrapResultType<Result<int>>, int>);
+static_assert(std::is_same_v<UnwrapResultType<absl::Status>, void>);
+static_assert(std::is_same_v<UnwrapQualifiedResultType<absl::Status>, void>);
+static_assert(std::is_same_v<UnwrapQualifiedResultType<Result<int>>, int>);
+static_assert(std::is_same_v<UnwrapQualifiedResultType<Result<int> &>, int &>);
+static_assert(std::is_same_v<UnwrapQualifiedResultType<const Result<int> &>,
+                             const int &>);
+
+static_assert(
+    std::is_same_v<UnwrapQualifiedResultType<Result<int> &&>, int &&>);
+
+/// FIXME: Typically a meta-function like FlatResult would be named MakeResult<>
+/// or similar.
+
+static_assert(std::is_same_v<int, typename tensorstore::internal_result::
+                                      UnwrapResultHelper<int>::type>,
+              "");
+
+static_assert(
+    std::is_same_v<Result<int>, typename tensorstore::internal_result::
+                                    UnwrapResultHelper<int>::result_type>,
+    "");
+
+static_assert(std::is_same_v<Result<int>, FlatResult<Result<int>>>);
+static_assert(std::is_same_v<Result<int>, FlatResult<int>>);
+static_assert(
+    std::is_same_v<Result<int>,
+                   tensorstore::FlatResult<std::invoke_result_t<int()>>>,
+    "");
+
+TEST(ChainResultTest, Example) {
+  auto func1 = [](int x) -> float { return 1.0f + x; };
+  auto func2 = [](float x) -> Result<std::string> {
+    return absl::StrCat("fn.", x);
+  };
+  auto func3 = [](std::string_view x) -> bool { return x.length() > 4; };
+
+  Result<bool> y1 = ChainResult(Result<int>(3), func1, func2, func3);
+  Result<bool> y2 = ChainResult(3, func1, func2, func3);
+
+  EXPECT_TRUE(y1.has_value());
+  EXPECT_TRUE(y2.has_value());
+
+  EXPECT_EQ(y1.value(), y2.value());
+}
+
+TEST(ChainResultTest, Basic) {
+  EXPECT_EQ(Result<int>(2), ChainResult(2));
+  EXPECT_EQ(Result<int>(2), ChainResult(Result<int>(2)));
+
+  EXPECT_EQ(Result<int>(3),
+            ChainResult(Result<int>(2), [](int x) { return x + 1; }));
+
+  EXPECT_EQ(Result<float>(1.5), ChainResult(
+                                    Result<int>(2), [](int x) { return x + 1; },
+                                    [](int x) { return x / 2.0f; }));
+
+  EXPECT_EQ(Result<int>(absl::UnknownError("A")),
+            ChainResult(Result<int>(absl::UnknownError("A")),
+                        [](int x) { return x + 1; }));
+}
+
+TEST(MapResultTest, Basic) {
+  absl::Status status;
+
+  EXPECT_EQ(Result<int>(absl::UnknownError("A")),
+            tensorstore::MapResult(std::plus<int>(),
+                                   Result<int>(absl::UnknownError("A")),
+                                   Result<int>(absl::UnknownError("B"))));
+  EXPECT_EQ(Result<int>(absl::UnknownError("B")),
+            tensorstore::MapResult(std::plus<int>(), 1,
+                                   Result<int>(absl::UnknownError("B"))));
+  EXPECT_EQ(Result<int>(3), tensorstore::MapResult(std::plus<int>(), 1, 2));
+  EXPECT_EQ(
+      Result<int>(absl::UnknownError("C")),
+      tensorstore::MapResult(
+          [](int a, int b) { return Result<int>(absl::UnknownError("C")); }, 1,
+          2));
+}
+
+TEST(PipelineOperator, Basic) {
+  auto func1 = [](int x) -> float { return 1.0f + x; };
+  auto func2 = [](float x) -> Result<std::string> {
+    return absl::StrCat("fn.", x);
+  };
+  auto func3 = [](std::string_view x) -> bool { return x.length() > 4; };
+
+  auto y1 = Result<int>(3) | func1 | func2 | func3;
+  static_assert(std::is_same_v<decltype(y1), Result<bool>>);
+  EXPECT_THAT(y1, ::testing::Optional(false));
+
+  auto y2 = Result<float>(2.5) | func2;
+  static_assert(std::is_same_v<decltype(y2), Result<std::string>>);
+  EXPECT_THAT(y2, ::testing::Optional(std::string("fn.2.5")));
+}
+
+TEST(ResultOfStatusTest, Value) {
+  Result<absl::Status> r(std::in_place, absl::InvalidArgumentError("abc"));
+  ASSERT_TRUE(r.has_value());
+  EXPECT_EQ(absl::InvalidArgumentError("abc"), *r);
+}
+
+TEST(ResultOfStatusTest, Error) {
+  Result<absl::Status> r(absl::InvalidArgumentError("abc"));
+  ASSERT_FALSE(r.has_value());
+  EXPECT_EQ(r.status(), absl::InvalidArgumentError("abc"));
+}
+
+TEST(ResultOfResultTest, ValueValue) {
+  Result<Result<int>> r(std::in_place, Result<int>(3));
+  ASSERT_TRUE(r.has_value());
+  EXPECT_EQ(Result<int>(3), *r);
+}
+
+TEST(ResultOfResultTest, ValueError) {
+  Result<Result<int>> r(std::in_place, absl::InvalidArgumentError("abc"));
+  ASSERT_TRUE(r.has_value());
+  EXPECT_EQ(Result<int>(absl::InvalidArgumentError("abc")), *r);
+}
+
+TEST(ResultOfResultTest, Error) {
+  Result<Result<int>> r(absl::InvalidArgumentError("abc"));
+  ASSERT_FALSE(r.has_value());
+  EXPECT_EQ(absl::InvalidArgumentError("abc"), r.status());
+}
+
+TEST(ResultTest, DefaultConstruct) {
+  Result<int> r;
+  ASSERT_FALSE(r.has_value());
+  EXPECT_EQ(absl::UnknownError(""), r.status());
+}
+
+TEST(ResultTest, ConstructVoidOk) {
+  Result<int> r(3);
+  Result<void> s(r);
+  ASSERT_TRUE(s.has_value());
+}
+
+TEST(ResultTest, MoveConstructVoidOk) {
+  Result<int> r(3);
+  Result<void> s(std::move(r));
+  ASSERT_TRUE(s.has_value());
+}
+
+TEST(ResultTest, ConstructVoid) {
+  Result<int> r(absl::InvalidArgumentError("abc"));
+  Result<void> s(r);
+  ASSERT_FALSE(s.has_value());
+  EXPECT_EQ(absl::InvalidArgumentError("abc"), s.status());
+}
+
+TEST(ResultTest, MoveConstructVoid) {
+  Result<int> r;
+  Result<void> s(std::move(r));
+  ASSERT_FALSE(s.has_value());
+  EXPECT_EQ(absl::UnknownError(""), s.status());
+}
+
+TEST(ResultTest, AssignmentConstructVoidOk) {
+  Result<int> r(3);
+  Result<void> s = r;
+  ASSERT_TRUE(s.has_value());
+}
+
+TEST(ResultTest, AssignmentConstructVoid) {
+  Result<int> r;
+  Result<void> s = r;
+  ASSERT_FALSE(s.has_value());
+}
+
+TEST(ResultTest, AssignmentVoidOk) {
+  Result<int> r(3);
+  Result<void> s;
+  ASSERT_FALSE(s.has_value());
+  s = r;
+  ASSERT_TRUE(s.has_value());
+}
+
+TEST(ResultTest, AssignmentVoid) {
+  Result<int> r(absl::InvalidArgumentError("abc"));
+  Result<void> s;
+  ASSERT_FALSE(s.has_value());
+  s = r;
+  ASSERT_FALSE(s.has_value());
+  EXPECT_EQ(absl::InvalidArgumentError("abc"), s.status());
+}
+
+TEST(ResultTest, MoveAssignmentVoidOk) {
+  Result<int> r(3);
+  Result<void> s;
+  ASSERT_FALSE(s.has_value());
+  s = std::move(r);
+  ASSERT_TRUE(s.has_value());
+}
+
+TEST(ResultTest, MoveAssignmentVoid) {
+  Result<int> r(absl::InvalidArgumentError("abc"));
+  Result<void> s;
+  ASSERT_FALSE(s.has_value());
+  s = std::move(r);
+  ASSERT_FALSE(s.has_value());
+  EXPECT_EQ(absl::InvalidArgumentError("abc"), s.status());
+}
+
+struct NotEqualityComparable {};
+static_assert(!tensorstore::internal::IsEqualityComparable<
+              Result<NotEqualityComparable>, Result<NotEqualityComparable>>);
+
 }  // namespace
